# Generated from language/FandangoParser.g4 by ANTLR 4.13.2
# encoding: utf-8
from antlr4 import *
from io import StringIO
import sys

if sys.version_info[1] > 5:
    from typing import TextIO
else:
    from typing.io import TextIO


def serializedATN():
    return [
        4,
        1,
        119,
        2707,
        2,
        0,
        7,
        0,
        2,
        1,
        7,
        1,
        2,
        2,
        7,
        2,
        2,
        3,
        7,
        3,
        2,
        4,
        7,
        4,
        2,
        5,
        7,
        5,
        2,
        6,
        7,
        6,
        2,
        7,
        7,
        7,
        2,
        8,
        7,
        8,
        2,
        9,
        7,
        9,
        2,
        10,
        7,
        10,
        2,
        11,
        7,
        11,
        2,
        12,
        7,
        12,
        2,
        13,
        7,
        13,
        2,
        14,
        7,
        14,
        2,
        15,
        7,
        15,
        2,
        16,
        7,
        16,
        2,
        17,
        7,
        17,
        2,
        18,
        7,
        18,
        2,
        19,
        7,
        19,
        2,
        20,
        7,
        20,
        2,
        21,
        7,
        21,
        2,
        22,
        7,
        22,
        2,
        23,
        7,
        23,
        2,
        24,
        7,
        24,
        2,
        25,
        7,
        25,
        2,
        26,
        7,
        26,
        2,
        27,
        7,
        27,
        2,
        28,
        7,
        28,
        2,
        29,
        7,
        29,
        2,
        30,
        7,
        30,
        2,
        31,
        7,
        31,
        2,
        32,
        7,
        32,
        2,
        33,
        7,
        33,
        2,
        34,
        7,
        34,
        2,
        35,
        7,
        35,
        2,
        36,
        7,
        36,
        2,
        37,
        7,
        37,
        2,
        38,
        7,
        38,
        2,
        39,
        7,
        39,
        2,
        40,
        7,
        40,
        2,
        41,
        7,
        41,
        2,
        42,
        7,
        42,
        2,
        43,
        7,
        43,
        2,
        44,
        7,
        44,
        2,
        45,
        7,
        45,
        2,
        46,
        7,
        46,
        2,
        47,
        7,
        47,
        2,
        48,
        7,
        48,
        2,
        49,
        7,
        49,
        2,
        50,
        7,
        50,
        2,
        51,
        7,
        51,
        2,
        52,
        7,
        52,
        2,
        53,
        7,
        53,
        2,
        54,
        7,
        54,
        2,
        55,
        7,
        55,
        2,
        56,
        7,
        56,
        2,
        57,
        7,
        57,
        2,
        58,
        7,
        58,
        2,
        59,
        7,
        59,
        2,
        60,
        7,
        60,
        2,
        61,
        7,
        61,
        2,
        62,
        7,
        62,
        2,
        63,
        7,
        63,
        2,
        64,
        7,
        64,
        2,
        65,
        7,
        65,
        2,
        66,
        7,
        66,
        2,
        67,
        7,
        67,
        2,
        68,
        7,
        68,
        2,
        69,
        7,
        69,
        2,
        70,
        7,
        70,
        2,
        71,
        7,
        71,
        2,
        72,
        7,
        72,
        2,
        73,
        7,
        73,
        2,
        74,
        7,
        74,
        2,
        75,
        7,
        75,
        2,
        76,
        7,
        76,
        2,
        77,
        7,
        77,
        2,
        78,
        7,
        78,
        2,
        79,
        7,
        79,
        2,
        80,
        7,
        80,
        2,
        81,
        7,
        81,
        2,
        82,
        7,
        82,
        2,
        83,
        7,
        83,
        2,
        84,
        7,
        84,
        2,
        85,
        7,
        85,
        2,
        86,
        7,
        86,
        2,
        87,
        7,
        87,
        2,
        88,
        7,
        88,
        2,
        89,
        7,
        89,
        2,
        90,
        7,
        90,
        2,
        91,
        7,
        91,
        2,
        92,
        7,
        92,
        2,
        93,
        7,
        93,
        2,
        94,
        7,
        94,
        2,
        95,
        7,
        95,
        2,
        96,
        7,
        96,
        2,
        97,
        7,
        97,
        2,
        98,
        7,
        98,
        2,
        99,
        7,
        99,
        2,
        100,
        7,
        100,
        2,
        101,
        7,
        101,
        2,
        102,
        7,
        102,
        2,
        103,
        7,
        103,
        2,
        104,
        7,
        104,
        2,
        105,
        7,
        105,
        2,
        106,
        7,
        106,
        2,
        107,
        7,
        107,
        2,
        108,
        7,
        108,
        2,
        109,
        7,
        109,
        2,
        110,
        7,
        110,
        2,
        111,
        7,
        111,
        2,
        112,
        7,
        112,
        2,
        113,
        7,
        113,
        2,
        114,
        7,
        114,
        2,
        115,
        7,
        115,
        2,
        116,
        7,
        116,
        2,
        117,
        7,
        117,
        2,
        118,
        7,
        118,
        2,
        119,
        7,
        119,
        2,
        120,
        7,
        120,
        2,
        121,
        7,
        121,
        2,
        122,
        7,
        122,
        2,
        123,
        7,
        123,
        2,
        124,
        7,
        124,
        2,
        125,
        7,
        125,
        2,
        126,
        7,
        126,
        2,
        127,
        7,
        127,
        2,
        128,
        7,
        128,
        2,
        129,
        7,
        129,
        2,
        130,
        7,
        130,
        2,
        131,
        7,
        131,
        2,
        132,
        7,
        132,
        2,
        133,
        7,
        133,
        2,
        134,
        7,
        134,
        2,
        135,
        7,
        135,
        2,
        136,
        7,
        136,
        2,
        137,
        7,
        137,
        2,
        138,
        7,
        138,
        2,
        139,
        7,
        139,
        2,
        140,
        7,
        140,
        2,
        141,
        7,
        141,
        2,
        142,
        7,
        142,
        2,
        143,
        7,
        143,
        2,
        144,
        7,
        144,
        2,
        145,
        7,
        145,
        2,
        146,
        7,
        146,
        2,
        147,
        7,
        147,
        2,
        148,
        7,
        148,
        2,
        149,
        7,
        149,
        2,
        150,
        7,
        150,
        2,
        151,
        7,
        151,
        2,
        152,
        7,
        152,
        2,
        153,
        7,
        153,
        2,
        154,
        7,
        154,
        2,
        155,
        7,
        155,
        2,
        156,
        7,
        156,
        2,
        157,
        7,
        157,
        2,
        158,
        7,
        158,
        2,
        159,
        7,
        159,
        2,
        160,
        7,
        160,
        2,
        161,
        7,
        161,
        2,
        162,
        7,
        162,
        2,
        163,
        7,
        163,
        2,
        164,
        7,
        164,
        2,
        165,
        7,
        165,
        2,
        166,
        7,
        166,
        2,
        167,
        7,
        167,
        2,
        168,
        7,
        168,
        2,
        169,
        7,
        169,
        2,
        170,
        7,
        170,
        2,
        171,
        7,
        171,
        2,
        172,
        7,
        172,
        2,
        173,
        7,
        173,
        2,
        174,
        7,
        174,
        2,
        175,
        7,
        175,
        2,
        176,
        7,
        176,
        2,
        177,
        7,
        177,
        2,
        178,
        7,
        178,
        2,
        179,
        7,
        179,
        2,
        180,
        7,
        180,
        2,
        181,
        7,
        181,
        2,
        182,
        7,
        182,
        2,
        183,
        7,
        183,
        2,
        184,
        7,
        184,
        2,
        185,
        7,
        185,
        2,
        186,
        7,
        186,
        2,
        187,
        7,
        187,
        2,
        188,
        7,
        188,
        2,
        189,
        7,
        189,
        2,
        190,
        7,
        190,
        2,
        191,
        7,
        191,
        2,
        192,
        7,
        192,
        2,
        193,
        7,
        193,
        2,
        194,
        7,
        194,
        2,
        195,
        7,
        195,
        2,
        196,
        7,
        196,
        2,
        197,
        7,
        197,
        2,
        198,
        7,
        198,
        2,
        199,
        7,
        199,
        2,
        200,
        7,
        200,
        2,
        201,
        7,
        201,
        2,
        202,
        7,
        202,
        2,
        203,
        7,
        203,
        2,
        204,
        7,
        204,
        2,
        205,
        7,
        205,
        2,
        206,
        7,
        206,
        2,
        207,
        7,
        207,
        2,
        208,
        7,
        208,
        2,
        209,
        7,
        209,
        2,
        210,
        7,
        210,
        2,
        211,
        7,
        211,
        2,
        212,
        7,
        212,
        2,
        213,
        7,
        213,
        2,
        214,
        7,
        214,
        2,
        215,
        7,
        215,
        2,
        216,
        7,
        216,
        2,
        217,
        7,
        217,
        2,
        218,
        7,
        218,
        2,
        219,
        7,
        219,
        2,
        220,
        7,
        220,
        2,
        221,
        7,
        221,
        2,
        222,
        7,
        222,
        2,
        223,
        7,
        223,
        1,
        0,
        1,
        0,
        1,
        0,
        1,
        1,
        5,
        1,
        453,
        8,
        1,
        10,
        1,
        12,
        1,
        456,
        9,
        1,
        1,
        1,
        1,
        1,
        5,
        1,
        460,
        8,
        1,
        10,
        1,
        12,
        1,
        463,
        9,
        1,
        5,
        1,
        465,
        8,
        1,
        10,
        1,
        12,
        1,
        468,
        9,
        1,
        1,
        2,
        1,
        2,
        1,
        2,
        3,
        2,
        473,
        8,
        2,
        1,
        3,
        1,
        3,
        1,
        3,
        1,
        3,
        1,
        3,
        3,
        3,
        480,
        8,
        3,
        1,
        3,
        1,
        3,
        1,
        3,
        1,
        3,
        1,
        3,
        1,
        3,
        1,
        3,
        3,
        3,
        489,
        8,
        3,
        1,
        3,
        1,
        3,
        1,
        3,
        1,
        3,
        1,
        3,
        1,
        3,
        1,
        3,
        1,
        3,
        3,
        3,
        499,
        8,
        3,
        1,
        3,
        1,
        3,
        3,
        3,
        503,
        8,
        3,
        1,
        4,
        1,
        4,
        1,
        4,
        5,
        4,
        508,
        8,
        4,
        10,
        4,
        12,
        4,
        511,
        9,
        4,
        1,
        5,
        1,
        5,
        5,
        5,
        515,
        8,
        5,
        10,
        5,
        12,
        5,
        518,
        9,
        5,
        1,
        6,
        1,
        6,
        1,
        6,
        1,
        6,
        1,
        6,
        3,
        6,
        525,
        8,
        6,
        1,
        7,
        1,
        7,
        1,
        7,
        1,
        8,
        1,
        8,
        1,
        8,
        1,
        9,
        1,
        9,
        1,
        9,
        1,
        10,
        1,
        10,
        1,
        10,
        1,
        10,
        1,
        10,
        1,
        10,
        1,
        10,
        1,
        10,
        3,
        10,
        544,
        8,
        10,
        1,
        10,
        1,
        10,
        3,
        10,
        548,
        8,
        10,
        1,
        10,
        1,
        10,
        3,
        10,
        552,
        8,
        10,
        1,
        11,
        5,
        11,
        555,
        8,
        11,
        10,
        11,
        12,
        11,
        558,
        9,
        11,
        1,
        11,
        1,
        11,
        1,
        11,
        1,
        11,
        1,
        11,
        1,
        11,
        1,
        11,
        1,
        11,
        3,
        11,
        568,
        8,
        11,
        1,
        11,
        5,
        11,
        571,
        8,
        11,
        10,
        11,
        12,
        11,
        574,
        9,
        11,
        1,
        12,
        1,
        12,
        3,
        12,
        578,
        8,
        12,
        1,
        12,
        1,
        12,
        1,
        12,
        1,
        13,
        1,
        13,
        1,
        13,
        1,
        13,
        1,
        13,
        1,
        13,
        1,
        13,
        1,
        13,
        1,
        13,
        1,
        13,
        1,
        13,
        1,
        13,
        1,
        13,
        1,
        13,
        1,
        13,
        1,
        13,
        3,
        13,
        599,
        8,
        13,
        1,
        14,
        5,
        14,
        602,
        8,
        14,
        10,
        14,
        12,
        14,
        605,
        9,
        14,
        1,
        14,
        1,
        14,
        1,
        14,
        1,
        14,
        1,
        14,
        3,
        14,
        612,
        8,
        14,
        1,
        14,
        5,
        14,
        615,
        8,
        14,
        10,
        14,
        12,
        14,
        618,
        9,
        14,
        1,
        15,
        5,
        15,
        621,
        8,
        15,
        10,
        15,
        12,
        15,
        624,
        9,
        15,
        1,
        15,
        1,
        15,
        1,
        15,
        1,
        15,
        1,
        15,
        1,
        15,
        1,
        15,
        1,
        15,
        1,
        15,
        1,
        15,
        1,
        15,
        1,
        15,
        1,
        15,
        1,
        15,
        1,
        15,
        3,
        15,
        641,
        8,
        15,
        1,
        15,
        5,
        15,
        644,
        8,
        15,
        10,
        15,
        12,
        15,
        647,
        9,
        15,
        1,
        16,
        5,
        16,
        650,
        8,
        16,
        10,
        16,
        12,
        16,
        653,
        9,
        16,
        1,
        16,
        1,
        16,
        1,
        16,
        5,
        16,
        658,
        8,
        16,
        10,
        16,
        12,
        16,
        661,
        9,
        16,
        1,
        16,
        5,
        16,
        664,
        8,
        16,
        10,
        16,
        12,
        16,
        667,
        9,
        16,
        1,
        17,
        5,
        17,
        670,
        8,
        17,
        10,
        17,
        12,
        17,
        673,
        9,
        17,
        1,
        17,
        1,
        17,
        1,
        17,
        5,
        17,
        678,
        8,
        17,
        10,
        17,
        12,
        17,
        681,
        9,
        17,
        1,
        17,
        5,
        17,
        684,
        8,
        17,
        10,
        17,
        12,
        17,
        687,
        9,
        17,
        1,
        18,
        5,
        18,
        690,
        8,
        18,
        10,
        18,
        12,
        18,
        693,
        9,
        18,
        1,
        18,
        1,
        18,
        1,
        18,
        1,
        18,
        1,
        18,
        1,
        18,
        3,
        18,
        701,
        8,
        18,
        1,
        18,
        5,
        18,
        704,
        8,
        18,
        10,
        18,
        12,
        18,
        707,
        9,
        18,
        1,
        19,
        1,
        19,
        1,
        19,
        1,
        19,
        1,
        20,
        1,
        20,
        1,
        20,
        1,
        20,
        1,
        20,
        1,
        20,
        1,
        20,
        1,
        20,
        3,
        20,
        721,
        8,
        20,
        1,
        21,
        1,
        21,
        1,
        21,
        1,
        21,
        1,
        21,
        3,
        21,
        728,
        8,
        21,
        1,
        22,
        1,
        22,
        1,
        22,
        1,
        22,
        1,
        22,
        1,
        22,
        1,
        22,
        1,
        22,
        1,
        22,
        5,
        22,
        739,
        8,
        22,
        10,
        22,
        12,
        22,
        742,
        9,
        22,
        1,
        23,
        1,
        23,
        1,
        23,
        1,
        23,
        1,
        23,
        1,
        23,
        1,
        23,
        1,
        23,
        1,
        23,
        1,
        23,
        1,
        23,
        3,
        23,
        755,
        8,
        23,
        1,
        24,
        1,
        24,
        1,
        24,
        1,
        24,
        1,
        24,
        3,
        24,
        762,
        8,
        24,
        1,
        25,
        1,
        25,
        1,
        25,
        5,
        25,
        767,
        8,
        25,
        10,
        25,
        12,
        25,
        770,
        9,
        25,
        1,
        25,
        3,
        25,
        773,
        8,
        25,
        1,
        26,
        1,
        26,
        1,
        26,
        1,
        26,
        3,
        26,
        779,
        8,
        26,
        1,
        27,
        1,
        27,
        1,
        27,
        5,
        27,
        784,
        8,
        27,
        10,
        27,
        12,
        27,
        787,
        9,
        27,
        1,
        27,
        3,
        27,
        790,
        8,
        27,
        1,
        28,
        1,
        28,
        3,
        28,
        794,
        8,
        28,
        1,
        28,
        1,
        28,
        3,
        28,
        798,
        8,
        28,
        1,
        28,
        3,
        28,
        801,
        8,
        28,
        1,
        28,
        1,
        28,
        3,
        28,
        805,
        8,
        28,
        1,
        28,
        1,
        28,
        3,
        28,
        809,
        8,
        28,
        3,
        28,
        811,
        8,
        28,
        1,
        29,
        1,
        29,
        3,
        29,
        815,
        8,
        29,
        1,
        30,
        5,
        30,
        818,
        8,
        30,
        10,
        30,
        12,
        30,
        821,
        9,
        30,
        1,
        30,
        1,
        30,
        5,
        30,
        825,
        8,
        30,
        10,
        30,
        12,
        30,
        828,
        9,
        30,
        1,
        31,
        3,
        31,
        831,
        8,
        31,
        1,
        31,
        3,
        31,
        834,
        8,
        31,
        1,
        32,
        1,
        32,
        1,
        33,
        1,
        33,
        5,
        33,
        840,
        8,
        33,
        10,
        33,
        12,
        33,
        843,
        9,
        33,
        1,
        33,
        3,
        33,
        846,
        8,
        33,
        1,
        34,
        1,
        34,
        3,
        34,
        850,
        8,
        34,
        1,
        34,
        1,
        34,
        1,
        34,
        1,
        34,
        5,
        34,
        856,
        8,
        34,
        10,
        34,
        12,
        34,
        859,
        9,
        34,
        1,
        34,
        3,
        34,
        862,
        8,
        34,
        1,
        35,
        1,
        35,
        4,
        35,
        866,
        8,
        35,
        11,
        35,
        12,
        35,
        867,
        1,
        36,
        1,
        36,
        3,
        36,
        872,
        8,
        36,
        1,
        37,
        1,
        37,
        1,
        37,
        1,
        37,
        1,
        37,
        3,
        37,
        879,
        8,
        37,
        1,
        38,
        1,
        38,
        1,
        38,
        5,
        38,
        884,
        8,
        38,
        10,
        38,
        12,
        38,
        887,
        9,
        38,
        1,
        38,
        3,
        38,
        890,
        8,
        38,
        1,
        38,
        4,
        38,
        893,
        8,
        38,
        11,
        38,
        12,
        38,
        894,
        1,
        38,
        5,
        38,
        898,
        8,
        38,
        10,
        38,
        12,
        38,
        901,
        9,
        38,
        1,
        38,
        3,
        38,
        904,
        8,
        38,
        1,
        39,
        1,
        39,
        1,
        39,
        1,
        39,
        1,
        39,
        1,
        39,
        1,
        39,
        1,
        39,
        1,
        39,
        1,
        39,
        1,
        39,
        1,
        39,
        1,
        39,
        1,
        39,
        3,
        39,
        920,
        8,
        39,
        1,
        40,
        1,
        40,
        1,
        40,
        1,
        40,
        1,
        40,
        1,
        40,
        1,
        40,
        1,
        40,
        3,
        40,
        930,
        8,
        40,
        1,
        41,
        1,
        41,
        1,
        41,
        1,
        41,
        1,
        41,
        3,
        41,
        937,
        8,
        41,
        1,
        41,
        1,
        41,
        1,
        41,
        1,
        41,
        1,
        41,
        3,
        41,
        944,
        8,
        41,
        1,
        41,
        1,
        41,
        1,
        41,
        1,
        41,
        3,
        41,
        950,
        8,
        41,
        1,
        41,
        1,
        41,
        1,
        41,
        4,
        41,
        955,
        8,
        41,
        11,
        41,
        12,
        41,
        956,
        1,
        41,
        1,
        41,
        3,
        41,
        961,
        8,
        41,
        1,
        41,
        1,
        41,
        1,
        41,
        1,
        41,
        3,
        41,
        967,
        8,
        41,
        3,
        41,
        969,
        8,
        41,
        1,
        42,
        1,
        42,
        3,
        42,
        973,
        8,
        42,
        1,
        43,
        1,
        43,
        1,
        44,
        1,
        44,
        3,
        44,
        979,
        8,
        44,
        1,
        45,
        1,
        45,
        1,
        45,
        1,
        45,
        3,
        45,
        985,
        8,
        45,
        1,
        45,
        3,
        45,
        988,
        8,
        45,
        1,
        46,
        1,
        46,
        1,
        46,
        1,
        46,
        5,
        46,
        994,
        8,
        46,
        10,
        46,
        12,
        46,
        997,
        9,
        46,
        1,
        47,
        1,
        47,
        1,
        47,
        1,
        47,
        5,
        47,
        1003,
        8,
        47,
        10,
        47,
        12,
        47,
        1006,
        9,
        47,
        1,
        48,
        1,
        48,
        1,
        48,
        1,
        49,
        1,
        49,
        1,
        50,
        1,
        50,
        1,
        50,
        1,
        50,
        3,
        50,
        1017,
        8,
        50,
        1,
        51,
        1,
        51,
        3,
        51,
        1021,
        8,
        51,
        1,
        52,
        1,
        52,
        1,
        52,
        1,
        53,
        1,
        53,
        5,
        53,
        1028,
        8,
        53,
        10,
        53,
        12,
        53,
        1031,
        9,
        53,
        1,
        53,
        1,
        53,
        1,
        53,
        1,
        53,
        1,
        53,
        1,
        53,
        4,
        53,
        1039,
        8,
        53,
        11,
        53,
        12,
        53,
        1040,
        1,
        53,
        1,
        53,
        3,
        53,
        1045,
        8,
        53,
        1,
        54,
        1,
        54,
        1,
        54,
        3,
        54,
        1050,
        8,
        54,
        1,
        54,
        1,
        54,
        1,
        54,
        1,
        54,
        3,
        54,
        1056,
        8,
        54,
        1,
        55,
        1,
        55,
        1,
        55,
        5,
        55,
        1061,
        8,
        55,
        10,
        55,
        12,
        55,
        1064,
        9,
        55,
        1,
        56,
        1,
        56,
        1,
        56,
        3,
        56,
        1069,
        8,
        56,
        1,
        57,
        1,
        57,
        1,
        57,
        5,
        57,
        1074,
        8,
        57,
        10,
        57,
        12,
        57,
        1077,
        9,
        57,
        1,
        58,
        1,
        58,
        1,
        58,
        3,
        58,
        1082,
        8,
        58,
        1,
        59,
        1,
        59,
        1,
        59,
        1,
        59,
        1,
        59,
        1,
        59,
        5,
        59,
        1090,
        8,
        59,
        10,
        59,
        12,
        59,
        1093,
        9,
        59,
        1,
        60,
        1,
        60,
        1,
        60,
        1,
        60,
        1,
        60,
        1,
        60,
        3,
        60,
        1101,
        8,
        60,
        1,
        61,
        1,
        61,
        1,
        61,
        1,
        61,
        4,
        61,
        1107,
        8,
        61,
        11,
        61,
        12,
        61,
        1108,
        1,
        62,
        3,
        62,
        1112,
        8,
        62,
        1,
        62,
        1,
        62,
        1,
        63,
        1,
        63,
        1,
        63,
        3,
        63,
        1119,
        8,
        63,
        1,
        63,
        1,
        63,
        3,
        63,
        1123,
        8,
        63,
        1,
        63,
        3,
        63,
        1126,
        8,
        63,
        1,
        63,
        1,
        63,
        1,
        63,
        1,
        64,
        3,
        64,
        1132,
        8,
        64,
        1,
        64,
        1,
        64,
        1,
        65,
        3,
        65,
        1137,
        8,
        65,
        1,
        65,
        1,
        65,
        1,
        65,
        3,
        65,
        1142,
        8,
        65,
        1,
        65,
        1,
        65,
        3,
        65,
        1146,
        8,
        65,
        1,
        65,
        1,
        65,
        1,
        65,
        3,
        65,
        1151,
        8,
        65,
        1,
        65,
        1,
        65,
        3,
        65,
        1155,
        8,
        65,
        1,
        65,
        1,
        65,
        1,
        66,
        1,
        66,
        1,
        67,
        1,
        67,
        5,
        67,
        1163,
        8,
        67,
        10,
        67,
        12,
        67,
        1166,
        9,
        67,
        1,
        67,
        5,
        67,
        1169,
        8,
        67,
        10,
        67,
        12,
        67,
        1172,
        9,
        67,
        1,
        67,
        3,
        67,
        1175,
        8,
        67,
        1,
        67,
        1,
        67,
        5,
        67,
        1179,
        8,
        67,
        10,
        67,
        12,
        67,
        1182,
        9,
        67,
        1,
        67,
        3,
        67,
        1185,
        8,
        67,
        1,
        67,
        4,
        67,
        1188,
        8,
        67,
        11,
        67,
        12,
        67,
        1189,
        1,
        67,
        5,
        67,
        1193,
        8,
        67,
        10,
        67,
        12,
        67,
        1196,
        9,
        67,
        1,
        67,
        3,
        67,
        1199,
        8,
        67,
        1,
        67,
        4,
        67,
        1202,
        8,
        67,
        11,
        67,
        12,
        67,
        1203,
        1,
        67,
        3,
        67,
        1207,
        8,
        67,
        1,
        67,
        3,
        67,
        1210,
        8,
        67,
        1,
        68,
        4,
        68,
        1213,
        8,
        68,
        11,
        68,
        12,
        68,
        1214,
        1,
        68,
        1,
        68,
        3,
        68,
        1219,
        8,
        68,
        1,
        69,
        5,
        69,
        1222,
        8,
        69,
        10,
        69,
        12,
        69,
        1225,
        9,
        69,
        1,
        69,
        4,
        69,
        1228,
        8,
        69,
        11,
        69,
        12,
        69,
        1229,
        1,
        69,
        1,
        69,
        3,
        69,
        1234,
        8,
        69,
        1,
        70,
        1,
        70,
        1,
        70,
        5,
        70,
        1239,
        8,
        70,
        10,
        70,
        12,
        70,
        1242,
        9,
        70,
        1,
        70,
        3,
        70,
        1245,
        8,
        70,
        1,
        70,
        1,
        70,
        1,
        70,
        5,
        70,
        1250,
        8,
        70,
        10,
        70,
        12,
        70,
        1253,
        9,
        70,
        1,
        70,
        3,
        70,
        1256,
        8,
        70,
        1,
        70,
        1,
        70,
        1,
        70,
        4,
        70,
        1261,
        8,
        70,
        11,
        70,
        12,
        70,
        1262,
        1,
        70,
        3,
        70,
        1266,
        8,
        70,
        1,
        70,
        3,
        70,
        1269,
        8,
        70,
        1,
        71,
        1,
        71,
        1,
        71,
        1,
        72,
        1,
        72,
        1,
        72,
        1,
        72,
        3,
        72,
        1278,
        8,
        72,
        1,
        73,
        1,
        73,
        1,
        73,
        1,
        73,
        3,
        73,
        1284,
        8,
        73,
        1,
        74,
        1,
        74,
        1,
        74,
        1,
        74,
        1,
        74,
        1,
        74,
        1,
        74,
        3,
        74,
        1293,
        8,
        74,
        1,
        75,
        1,
        75,
        3,
        75,
        1297,
        8,
        75,
        1,
        75,
        1,
        75,
        1,
        75,
        1,
        75,
        3,
        75,
        1303,
        8,
        75,
        3,
        75,
        1305,
        8,
        75,
        1,
        76,
        1,
        76,
        3,
        76,
        1309,
        8,
        76,
        1,
        77,
        1,
        77,
        1,
        77,
        1,
        78,
        1,
        78,
        1,
        78,
        1,
        79,
        1,
        79,
        1,
        79,
        1,
        80,
        1,
        80,
        1,
        80,
        1,
        81,
        1,
        81,
        1,
        81,
        1,
        81,
        1,
        81,
        1,
        81,
        1,
        81,
        1,
        81,
        1,
        81,
        1,
        81,
        1,
        81,
        3,
        81,
        1334,
        8,
        81,
        3,
        81,
        1336,
        8,
        81,
        1,
        82,
        1,
        82,
        1,
        82,
        1,
        82,
        1,
        82,
        1,
        82,
        1,
        82,
        1,
        82,
        1,
        82,
        1,
        82,
        1,
        82,
        3,
        82,
        1349,
        8,
        82,
        3,
        82,
        1351,
        8,
        82,
        1,
        83,
        1,
        83,
        1,
        83,
        1,
        83,
        1,
        84,
        1,
        84,
        1,
        84,
        1,
        84,
        1,
        84,
        3,
        84,
        1362,
        8,
        84,
        1,
        85,
        1,
        85,
        1,
        85,
        1,
        85,
        1,
        85,
        1,
        85,
        1,
        85,
        3,
        85,
        1371,
        8,
        85,
        1,
        85,
        1,
        85,
        1,
        85,
        1,
        85,
        1,
        85,
        1,
        85,
        1,
        85,
        1,
        85,
        3,
        85,
        1381,
        8,
        85,
        3,
        85,
        1383,
        8,
        85,
        1,
        86,
        1,
        86,
        1,
        86,
        1,
        86,
        1,
        86,
        5,
        86,
        1390,
        8,
        86,
        10,
        86,
        12,
        86,
        1393,
        9,
        86,
        1,
        86,
        3,
        86,
        1396,
        8,
        86,
        1,
        86,
        1,
        86,
        1,
        86,
        1,
        86,
        1,
        86,
        1,
        86,
        1,
        86,
        1,
        86,
        5,
        86,
        1406,
        8,
        86,
        10,
        86,
        12,
        86,
        1409,
        9,
        86,
        1,
        86,
        1,
        86,
        1,
        86,
        1,
        86,
        1,
        86,
        1,
        86,
        1,
        86,
        1,
        86,
        1,
        86,
        5,
        86,
        1420,
        8,
        86,
        10,
        86,
        12,
        86,
        1423,
        9,
        86,
        1,
        86,
        3,
        86,
        1426,
        8,
        86,
        1,
        86,
        1,
        86,
        1,
        86,
        1,
        86,
        1,
        86,
        1,
        86,
        1,
        86,
        1,
        86,
        1,
        86,
        5,
        86,
        1437,
        8,
        86,
        10,
        86,
        12,
        86,
        1440,
        9,
        86,
        1,
        86,
        1,
        86,
        1,
        86,
        3,
        86,
        1445,
        8,
        86,
        1,
        87,
        1,
        87,
        1,
        87,
        1,
        87,
        1,
        87,
        3,
        87,
        1452,
        8,
        87,
        1,
        88,
        1,
        88,
        1,
        88,
        1,
        88,
        1,
        88,
        1,
        88,
        1,
        88,
        1,
        88,
        1,
        88,
        4,
        88,
        1463,
        8,
        88,
        11,
        88,
        12,
        88,
        1464,
        1,
        88,
        3,
        88,
        1468,
        8,
        88,
        1,
        88,
        3,
        88,
        1471,
        8,
        88,
        1,
        88,
        1,
        88,
        1,
        88,
        1,
        88,
        4,
        88,
        1477,
        8,
        88,
        11,
        88,
        12,
        88,
        1478,
        1,
        88,
        3,
        88,
        1482,
        8,
        88,
        1,
        88,
        3,
        88,
        1485,
        8,
        88,
        3,
        88,
        1487,
        8,
        88,
        1,
        89,
        1,
        89,
        1,
        89,
        1,
        89,
        3,
        89,
        1493,
        8,
        89,
        1,
        89,
        1,
        89,
        1,
        89,
        1,
        89,
        1,
        89,
        1,
        89,
        3,
        89,
        1501,
        8,
        89,
        1,
        90,
        1,
        90,
        1,
        90,
        1,
        90,
        1,
        90,
        3,
        90,
        1508,
        8,
        90,
        1,
        90,
        1,
        90,
        1,
        90,
        1,
        91,
        1,
        91,
        1,
        91,
        1,
        91,
        1,
        92,
        1,
        92,
        1,
        92,
        1,
        92,
        1,
        92,
        1,
        92,
        4,
        92,
        1523,
        8,
        92,
        11,
        92,
        12,
        92,
        1524,
        1,
        92,
        1,
        92,
        1,
        93,
        1,
        93,
        1,
        93,
        3,
        93,
        1532,
        8,
        93,
        1,
        93,
        3,
        93,
        1535,
        8,
        93,
        1,
        94,
        1,
        94,
        1,
        94,
        3,
        94,
        1540,
        8,
        94,
        1,
        94,
        1,
        94,
        1,
        94,
        1,
        95,
        1,
        95,
        1,
        95,
        1,
        96,
        1,
        96,
        3,
        96,
        1550,
        8,
        96,
        1,
        97,
        1,
        97,
        3,
        97,
        1554,
        8,
        97,
        1,
        98,
        1,
        98,
        1,
        98,
        1,
        98,
        1,
        99,
        1,
        99,
        1,
        99,
        5,
        99,
        1563,
        8,
        99,
        10,
        99,
        12,
        99,
        1566,
        9,
        99,
        1,
        100,
        1,
        100,
        1,
        100,
        1,
        100,
        1,
        100,
        1,
        100,
        1,
        100,
        1,
        100,
        3,
        100,
        1576,
        8,
        100,
        1,
        101,
        1,
        101,
        1,
        101,
        1,
        101,
        1,
        101,
        1,
        101,
        3,
        101,
        1584,
        8,
        101,
        1,
        102,
        1,
        102,
        1,
        102,
        1,
        102,
        1,
        102,
        1,
        102,
        3,
        102,
        1592,
        8,
        102,
        1,
        103,
        1,
        103,
        1,
        103,
        1,
        103,
        1,
        103,
        1,
        103,
        1,
        103,
        1,
        103,
        3,
        103,
        1602,
        8,
        103,
        1,
        104,
        1,
        104,
        1,
        104,
        3,
        104,
        1607,
        8,
        104,
        1,
        105,
        1,
        105,
        1,
        105,
        3,
        105,
        1612,
        8,
        105,
        1,
        106,
        1,
        106,
        1,
        107,
        1,
        107,
        1,
        108,
        1,
        108,
        1,
        109,
        1,
        109,
        1,
        110,
        1,
        110,
        1,
        111,
        1,
        111,
        1,
        112,
        1,
        112,
        1,
        112,
        1,
        112,
        1,
        113,
        1,
        113,
        1,
        113,
        1,
        113,
        1,
        113,
        1,
        113,
        5,
        113,
        1636,
        8,
        113,
        10,
        113,
        12,
        113,
        1639,
        9,
        113,
        1,
        114,
        1,
        114,
        1,
        114,
        1,
        114,
        1,
        115,
        1,
        115,
        3,
        115,
        1647,
        8,
        115,
        1,
        115,
        1,
        115,
        1,
        115,
        3,
        115,
        1652,
        8,
        115,
        1,
        115,
        3,
        115,
        1655,
        8,
        115,
        1,
        116,
        1,
        116,
        1,
        116,
        3,
        116,
        1660,
        8,
        116,
        1,
        117,
        1,
        117,
        1,
        117,
        5,
        117,
        1665,
        8,
        117,
        10,
        117,
        12,
        117,
        1668,
        9,
        117,
        1,
        117,
        3,
        117,
        1671,
        8,
        117,
        1,
        118,
        1,
        118,
        3,
        118,
        1675,
        8,
        118,
        1,
        119,
        1,
        119,
        1,
        119,
        1,
        119,
        3,
        119,
        1681,
        8,
        119,
        1,
        120,
        1,
        120,
        1,
        120,
        1,
        120,
        1,
        120,
        3,
        120,
        1688,
        8,
        120,
        1,
        120,
        1,
        120,
        1,
        120,
        1,
        120,
        1,
        120,
        1,
        120,
        1,
        120,
        3,
        120,
        1697,
        8,
        120,
        1,
        120,
        1,
        120,
        1,
        120,
        1,
        120,
        1,
        120,
        3,
        120,
        1704,
        8,
        120,
        1,
        120,
        1,
        120,
        3,
        120,
        1708,
        8,
        120,
        1,
        121,
        1,
        121,
        1,
        121,
        5,
        121,
        1713,
        8,
        121,
        10,
        121,
        12,
        121,
        1716,
        9,
        121,
        1,
        122,
        1,
        122,
        3,
        122,
        1720,
        8,
        122,
        1,
        122,
        1,
        122,
        1,
        122,
        1,
        123,
        1,
        123,
        1,
        123,
        1,
        124,
        1,
        124,
        1,
        124,
        1,
        124,
        1,
        124,
        1,
        124,
        1,
        124,
        1,
        124,
        3,
        124,
        1736,
        8,
        124,
        1,
        124,
        1,
        124,
        1,
        124,
        1,
        124,
        1,
        124,
        1,
        124,
        3,
        124,
        1744,
        8,
        124,
        1,
        124,
        1,
        124,
        1,
        124,
        1,
        124,
        1,
        124,
        1,
        124,
        1,
        124,
        1,
        124,
        3,
        124,
        1754,
        8,
        124,
        1,
        124,
        1,
        124,
        3,
        124,
        1758,
        8,
        124,
        1,
        125,
        1,
        125,
        1,
        125,
        5,
        125,
        1763,
        8,
        125,
        10,
        125,
        12,
        125,
        1766,
        9,
        125,
        1,
        126,
        1,
        126,
        1,
        126,
        5,
        126,
        1771,
        8,
        126,
        10,
        126,
        12,
        126,
        1774,
        9,
        126,
        1,
        127,
        1,
        127,
        1,
        127,
        1,
        127,
        1,
        128,
        1,
        128,
        1,
        128,
        3,
        128,
        1783,
        8,
        128,
        1,
        128,
        1,
        128,
        1,
        128,
        1,
        129,
        1,
        129,
        1,
        129,
        1,
        129,
        1,
        130,
        1,
        130,
        1,
        130,
        5,
        130,
        1795,
        8,
        130,
        10,
        130,
        12,
        130,
        1798,
        9,
        130,
        1,
        130,
        3,
        130,
        1801,
        8,
        130,
        1,
        131,
        1,
        131,
        3,
        131,
        1805,
        8,
        131,
        1,
        131,
        1,
        131,
        1,
        131,
        1,
        131,
        3,
        131,
        1811,
        8,
        131,
        1,
        132,
        1,
        132,
        1,
        132,
        1,
        133,
        1,
        133,
        1,
        133,
        5,
        133,
        1819,
        8,
        133,
        10,
        133,
        12,
        133,
        1822,
        9,
        133,
        1,
        133,
        3,
        133,
        1825,
        8,
        133,
        1,
        134,
        1,
        134,
        1,
        134,
        1,
        134,
        1,
        134,
        1,
        134,
        1,
        134,
        1,
        134,
        3,
        134,
        1835,
        8,
        134,
        1,
        135,
        1,
        135,
        1,
        135,
        1,
        135,
        1,
        135,
        3,
        135,
        1842,
        8,
        135,
        3,
        135,
        1844,
        8,
        135,
        1,
        136,
        1,
        136,
        1,
        136,
        5,
        136,
        1849,
        8,
        136,
        10,
        136,
        12,
        136,
        1852,
        9,
        136,
        1,
        136,
        3,
        136,
        1855,
        8,
        136,
        1,
        137,
        1,
        137,
        1,
        137,
        3,
        137,
        1860,
        8,
        137,
        1,
        138,
        1,
        138,
        1,
        138,
        5,
        138,
        1865,
        8,
        138,
        10,
        138,
        12,
        138,
        1868,
        9,
        138,
        1,
        138,
        3,
        138,
        1871,
        8,
        138,
        1,
        139,
        1,
        139,
        1,
        139,
        3,
        139,
        1876,
        8,
        139,
        1,
        140,
        1,
        140,
        1,
        140,
        1,
        140,
        1,
        141,
        1,
        141,
        3,
        141,
        1884,
        8,
        141,
        1,
        142,
        1,
        142,
        1,
        142,
        5,
        142,
        1889,
        8,
        142,
        10,
        142,
        12,
        142,
        1892,
        9,
        142,
        1,
        143,
        1,
        143,
        1,
        143,
        5,
        143,
        1897,
        8,
        143,
        10,
        143,
        12,
        143,
        1900,
        9,
        143,
        1,
        144,
        1,
        144,
        1,
        144,
        3,
        144,
        1905,
        8,
        144,
        1,
        145,
        1,
        145,
        5,
        145,
        1909,
        8,
        145,
        10,
        145,
        12,
        145,
        1912,
        9,
        145,
        1,
        146,
        1,
        146,
        1,
        146,
        1,
        146,
        1,
        146,
        1,
        146,
        1,
        146,
        1,
        146,
        1,
        146,
        1,
        146,
        3,
        146,
        1924,
        8,
        146,
        1,
        147,
        1,
        147,
        1,
        147,
        1,
        148,
        1,
        148,
        1,
        148,
        1,
        149,
        1,
        149,
        1,
        149,
        1,
        150,
        1,
        150,
        1,
        150,
        1,
        151,
        1,
        151,
        1,
        151,
        1,
        152,
        1,
        152,
        1,
        152,
        1,
        153,
        1,
        153,
        1,
        153,
        1,
        153,
        1,
        154,
        1,
        154,
        1,
        154,
        1,
        155,
        1,
        155,
        1,
        155,
        1,
        155,
        1,
        156,
        1,
        156,
        1,
        156,
        1,
        157,
        1,
        157,
        1,
        157,
        1,
        157,
        1,
        157,
        1,
        157,
        5,
        157,
        1964,
        8,
        157,
        10,
        157,
        12,
        157,
        1967,
        9,
        157,
        1,
        158,
        1,
        158,
        1,
        158,
        1,
        158,
        1,
        158,
        1,
        158,
        5,
        158,
        1975,
        8,
        158,
        10,
        158,
        12,
        158,
        1978,
        9,
        158,
        1,
        159,
        1,
        159,
        1,
        159,
        1,
        159,
        1,
        159,
        1,
        159,
        5,
        159,
        1986,
        8,
        159,
        10,
        159,
        12,
        159,
        1989,
        9,
        159,
        1,
        160,
        1,
        160,
        1,
        160,
        1,
        160,
        1,
        160,
        1,
        160,
        1,
        160,
        1,
        160,
        1,
        160,
        5,
        160,
        2000,
        8,
        160,
        10,
        160,
        12,
        160,
        2003,
        9,
        160,
        1,
        161,
        1,
        161,
        1,
        161,
        1,
        161,
        1,
        161,
        1,
        161,
        1,
        161,
        1,
        161,
        1,
        161,
        5,
        161,
        2014,
        8,
        161,
        10,
        161,
        12,
        161,
        2017,
        9,
        161,
        1,
        162,
        1,
        162,
        1,
        162,
        1,
        162,
        1,
        162,
        1,
        162,
        1,
        162,
        1,
        162,
        1,
        162,
        1,
        162,
        1,
        162,
        1,
        162,
        1,
        162,
        1,
        162,
        1,
        162,
        1,
        162,
        1,
        162,
        1,
        162,
        5,
        162,
        2037,
        8,
        162,
        10,
        162,
        12,
        162,
        2040,
        9,
        162,
        1,
        163,
        1,
        163,
        1,
        163,
        1,
        163,
        1,
        163,
        1,
        163,
        1,
        163,
        3,
        163,
        2049,
        8,
        163,
        1,
        164,
        1,
        164,
        1,
        164,
        1,
        164,
        1,
        164,
        3,
        164,
        2056,
        8,
        164,
        1,
        165,
        1,
        165,
        1,
        165,
        3,
        165,
        2061,
        8,
        165,
        1,
        166,
        1,
        166,
        1,
        166,
        1,
        166,
        1,
        166,
        1,
        166,
        1,
        166,
        1,
        166,
        1,
        166,
        1,
        166,
        1,
        166,
        3,
        166,
        2074,
        8,
        166,
        1,
        166,
        1,
        166,
        1,
        166,
        1,
        166,
        1,
        166,
        1,
        166,
        5,
        166,
        2082,
        8,
        166,
        10,
        166,
        12,
        166,
        2085,
        9,
        166,
        1,
        167,
        1,
        167,
        3,
        167,
        2089,
        8,
        167,
        1,
        167,
        1,
        167,
        1,
        167,
        3,
        167,
        2094,
        8,
        167,
        5,
        167,
        2096,
        8,
        167,
        10,
        167,
        12,
        167,
        2099,
        9,
        167,
        1,
        167,
        3,
        167,
        2102,
        8,
        167,
        1,
        168,
        3,
        168,
        2105,
        8,
        168,
        1,
        168,
        1,
        168,
        3,
        168,
        2109,
        8,
        168,
        1,
        168,
        1,
        168,
        3,
        168,
        2113,
        8,
        168,
        3,
        168,
        2115,
        8,
        168,
        1,
        168,
        3,
        168,
        2118,
        8,
        168,
        1,
        169,
        1,
        169,
        1,
        169,
        1,
        169,
        1,
        169,
        1,
        169,
        1,
        169,
        1,
        169,
        1,
        169,
        1,
        169,
        3,
        169,
        2130,
        8,
        169,
        1,
        169,
        1,
        169,
        3,
        169,
        2134,
        8,
        169,
        1,
        169,
        1,
        169,
        1,
        169,
        1,
        169,
        3,
        169,
        2140,
        8,
        169,
        1,
        169,
        3,
        169,
        2143,
        8,
        169,
        1,
        170,
        1,
        170,
        1,
        170,
        3,
        170,
        2148,
        8,
        170,
        1,
        170,
        1,
        170,
        1,
        171,
        1,
        171,
        3,
        171,
        2154,
        8,
        171,
        1,
        171,
        1,
        171,
        1,
        171,
        1,
        172,
        1,
        172,
        1,
        173,
        1,
        173,
        5,
        173,
        2163,
        8,
        173,
        10,
        173,
        12,
        173,
        2166,
        9,
        173,
        1,
        173,
        5,
        173,
        2169,
        8,
        173,
        10,
        173,
        12,
        173,
        2172,
        9,
        173,
        1,
        173,
        3,
        173,
        2175,
        8,
        173,
        1,
        173,
        1,
        173,
        5,
        173,
        2179,
        8,
        173,
        10,
        173,
        12,
        173,
        2182,
        9,
        173,
        1,
        173,
        3,
        173,
        2185,
        8,
        173,
        1,
        173,
        4,
        173,
        2188,
        8,
        173,
        11,
        173,
        12,
        173,
        2189,
        1,
        173,
        5,
        173,
        2193,
        8,
        173,
        10,
        173,
        12,
        173,
        2196,
        9,
        173,
        1,
        173,
        3,
        173,
        2199,
        8,
        173,
        1,
        173,
        4,
        173,
        2202,
        8,
        173,
        11,
        173,
        12,
        173,
        2203,
        1,
        173,
        3,
        173,
        2207,
        8,
        173,
        1,
        173,
        3,
        173,
        2210,
        8,
        173,
        1,
        174,
        4,
        174,
        2213,
        8,
        174,
        11,
        174,
        12,
        174,
        2214,
        1,
        174,
        1,
        174,
        3,
        174,
        2219,
        8,
        174,
        1,
        175,
        5,
        175,
        2222,
        8,
        175,
        10,
        175,
        12,
        175,
        2225,
        9,
        175,
        1,
        175,
        4,
        175,
        2228,
        8,
        175,
        11,
        175,
        12,
        175,
        2229,
        1,
        175,
        1,
        175,
        3,
        175,
        2234,
        8,
        175,
        1,
        176,
        1,
        176,
        1,
        176,
        5,
        176,
        2239,
        8,
        176,
        10,
        176,
        12,
        176,
        2242,
        9,
        176,
        1,
        176,
        3,
        176,
        2245,
        8,
        176,
        1,
        176,
        1,
        176,
        1,
        176,
        4,
        176,
        2250,
        8,
        176,
        11,
        176,
        12,
        176,
        2251,
        1,
        176,
        3,
        176,
        2255,
        8,
        176,
        1,
        176,
        3,
        176,
        2258,
        8,
        176,
        1,
        177,
        1,
        177,
        1,
        177,
        1,
        178,
        1,
        178,
        3,
        178,
        2265,
        8,
        178,
        1,
        179,
        1,
        179,
        1,
        179,
        3,
        179,
        2270,
        8,
        179,
        1,
        180,
        1,
        180,
        3,
        180,
        2274,
        8,
        180,
        1,
        180,
        3,
        180,
        2277,
        8,
        180,
        1,
        181,
        1,
        181,
        1,
        182,
        1,
        182,
        3,
        182,
        2283,
        8,
        182,
        1,
        183,
        1,
        183,
        1,
        183,
        3,
        183,
        2288,
        8,
        183,
        1,
        183,
        3,
        183,
        2291,
        8,
        183,
        1,
        183,
        3,
        183,
        2294,
        8,
        183,
        1,
        183,
        3,
        183,
        2297,
        8,
        183,
        1,
        183,
        1,
        183,
        1,
        184,
        1,
        184,
        1,
        184,
        3,
        184,
        2304,
        8,
        184,
        1,
        185,
        1,
        185,
        5,
        185,
        2308,
        8,
        185,
        10,
        185,
        12,
        185,
        2311,
        9,
        185,
        1,
        186,
        1,
        186,
        3,
        186,
        2315,
        8,
        186,
        1,
        187,
        1,
        187,
        5,
        187,
        2319,
        8,
        187,
        10,
        187,
        12,
        187,
        2322,
        9,
        187,
        1,
        187,
        1,
        187,
        1,
        188,
        1,
        188,
        1,
        189,
        1,
        189,
        4,
        189,
        2330,
        8,
        189,
        11,
        189,
        12,
        189,
        2331,
        1,
        190,
        1,
        190,
        3,
        190,
        2336,
        8,
        190,
        1,
        190,
        1,
        190,
        1,
        191,
        1,
        191,
        1,
        191,
        1,
        191,
        1,
        191,
        3,
        191,
        2345,
        8,
        191,
        1,
        191,
        1,
        191,
        1,
        192,
        1,
        192,
        1,
        192,
        1,
        192,
        1,
        193,
        1,
        193,
        3,
        193,
        2355,
        8,
        193,
        1,
        193,
        1,
        193,
        1,
        194,
        1,
        194,
        1,
        194,
        5,
        194,
        2362,
        8,
        194,
        10,
        194,
        12,
        194,
        2365,
        9,
        194,
        1,
        194,
        3,
        194,
        2368,
        8,
        194,
        1,
        195,
        1,
        195,
        1,
        195,
        3,
        195,
        2373,
        8,
        195,
        1,
        196,
        1,
        196,
        1,
        196,
        1,
        196,
        1,
        197,
        4,
        197,
        2380,
        8,
        197,
        11,
        197,
        12,
        197,
        2381,
        1,
        198,
        3,
        198,
        2385,
        8,
        198,
        1,
        198,
        1,
        198,
        1,
        198,
        1,
        198,
        1,
        198,
        1,
        198,
        5,
        198,
        2393,
        8,
        198,
        10,
        198,
        12,
        198,
        2396,
        9,
        198,
        1,
        199,
        1,
        199,
        1,
        199,
        1,
        199,
        1,
        199,
        1,
        200,
        1,
        200,
        1,
        200,
        1,
        200,
        1,
        200,
        1,
        201,
        1,
        201,
        1,
        201,
        3,
        201,
        2411,
        8,
        201,
        1,
        201,
        1,
        201,
        1,
        201,
        1,
        202,
        1,
        202,
        1,
        202,
        1,
        202,
        1,
        202,
        1,
        203,
        1,
        203,
        3,
        203,
        2423,
        8,
        203,
        1,
        204,
        1,
        204,
        1,
        204,
        5,
        204,
        2428,
        8,
        204,
        10,
        204,
        12,
        204,
        2431,
        9,
        204,
        1,
        204,
        1,
        204,
        3,
        204,
        2435,
        8,
        204,
        1,
        204,
        3,
        204,
        2438,
        8,
        204,
        1,
        205,
        1,
        205,
        1,
        205,
        3,
        205,
        2443,
        8,
        205,
        1,
        206,
        1,
        206,
        1,
        206,
        5,
        206,
        2448,
        8,
        206,
        10,
        206,
        12,
        206,
        2451,
        9,
        206,
        1,
        206,
        1,
        206,
        1,
        206,
        1,
        206,
        5,
        206,
        2457,
        8,
        206,
        10,
        206,
        12,
        206,
        2460,
        9,
        206,
        1,
        206,
        1,
        206,
        1,
        206,
        5,
        206,
        2465,
        8,
        206,
        10,
        206,
        12,
        206,
        2468,
        9,
        206,
        1,
        206,
        1,
        206,
        1,
        206,
        5,
        206,
        2473,
        8,
        206,
        10,
        206,
        12,
        206,
        2476,
        9,
        206,
        3,
        206,
        2478,
        8,
        206,
        1,
        207,
        1,
        207,
        1,
        207,
        1,
        208,
        1,
        208,
        1,
        208,
        1,
        208,
        3,
        208,
        2487,
        8,
        208,
        1,
        209,
        1,
        209,
        1,
        209,
        1,
        209,
        1,
        209,
        3,
        209,
        2494,
        8,
        209,
        1,
        210,
        1,
        210,
        1,
        210,
        5,
        210,
        2499,
        8,
        210,
        10,
        210,
        12,
        210,
        2502,
        9,
        210,
        1,
        210,
        3,
        210,
        2505,
        8,
        210,
        1,
        211,
        1,
        211,
        1,
        211,
        5,
        211,
        2510,
        8,
        211,
        10,
        211,
        12,
        211,
        2513,
        9,
        211,
        1,
        211,
        3,
        211,
        2516,
        8,
        211,
        1,
        212,
        1,
        212,
        1,
        212,
        4,
        212,
        2521,
        8,
        212,
        11,
        212,
        12,
        212,
        2522,
        1,
        212,
        3,
        212,
        2526,
        8,
        212,
        1,
        212,
        1,
        212,
        1,
        212,
        3,
        212,
        2531,
        8,
        212,
        1,
        213,
        1,
        213,
        1,
        213,
        3,
        213,
        2536,
        8,
        213,
        1,
        214,
        1,
        214,
        1,
        214,
        1,
        214,
        1,
        214,
        1,
        214,
        1,
        214,
        1,
        214,
        1,
        214,
        1,
        214,
        3,
        214,
        2548,
        8,
        214,
        1,
        215,
        1,
        215,
        1,
        215,
        1,
        215,
        1,
        215,
        1,
        215,
        1,
        215,
        3,
        215,
        2557,
        8,
        215,
        1,
        215,
        1,
        215,
        1,
        215,
        3,
        215,
        2562,
        8,
        215,
        1,
        215,
        3,
        215,
        2565,
        8,
        215,
        1,
        216,
        1,
        216,
        1,
        216,
        1,
        216,
        1,
        216,
        1,
        216,
        3,
        216,
        2573,
        8,
        216,
        1,
        217,
        1,
        217,
        1,
        217,
        1,
        217,
        1,
        217,
        1,
        217,
        1,
        217,
        1,
        217,
        1,
        217,
        3,
        217,
        2584,
        8,
        217,
        1,
        218,
        1,
        218,
        1,
        218,
        1,
        218,
        1,
        218,
        1,
        218,
        1,
        218,
        1,
        218,
        1,
        218,
        1,
        218,
        1,
        218,
        1,
        218,
        1,
        218,
        1,
        218,
        1,
        218,
        1,
        218,
        3,
        218,
        2602,
        8,
        218,
        1,
        218,
        5,
        218,
        2605,
        8,
        218,
        10,
        218,
        12,
        218,
        2608,
        9,
        218,
        1,
        219,
        1,
        219,
        1,
        219,
        5,
        219,
        2613,
        8,
        219,
        10,
        219,
        12,
        219,
        2616,
        9,
        219,
        1,
        219,
        3,
        219,
        2619,
        8,
        219,
        1,
        220,
        1,
        220,
        1,
        220,
        1,
        220,
        1,
        220,
        1,
        220,
        1,
        220,
        1,
        220,
        1,
        220,
        1,
        220,
        3,
        220,
        2631,
        8,
        220,
        1,
        221,
        1,
        221,
        1,
        221,
        3,
        221,
        2636,
        8,
        221,
        1,
        221,
        1,
        221,
        1,
        221,
        3,
        221,
        2641,
        8,
        221,
        1,
        221,
        3,
        221,
        2644,
        8,
        221,
        1,
        222,
        1,
        222,
        1,
        222,
        5,
        222,
        2649,
        8,
        222,
        10,
        222,
        12,
        222,
        2652,
        9,
        222,
        1,
        222,
        1,
        222,
        1,
        222,
        1,
        222,
        1,
        222,
        1,
        222,
        1,
        222,
        1,
        222,
        1,
        222,
        1,
        222,
        5,
        222,
        2664,
        8,
        222,
        10,
        222,
        12,
        222,
        2667,
        9,
        222,
        1,
        222,
        1,
        222,
        1,
        222,
        1,
        222,
        1,
        222,
        1,
        222,
        1,
        222,
        5,
        222,
        2676,
        8,
        222,
        10,
        222,
        12,
        222,
        2679,
        9,
        222,
        1,
        222,
        1,
        222,
        1,
        222,
        1,
        222,
        1,
        222,
        1,
        222,
        1,
        222,
        1,
        222,
        1,
        222,
        1,
        222,
        1,
        222,
        1,
        222,
        1,
        222,
        1,
        222,
        1,
        222,
        1,
        222,
        1,
        222,
        5,
        222,
        2698,
        8,
        222,
        10,
        222,
        12,
        222,
        2701,
        9,
        222,
        3,
        222,
        2703,
        8,
        222,
        1,
        223,
        1,
        223,
        1,
        223,
        0,
        11,
        44,
        118,
        226,
        314,
        316,
        318,
        320,
        322,
        324,
        332,
        436,
        224,
        0,
        2,
        4,
        6,
        8,
        10,
        12,
        14,
        16,
        18,
        20,
        22,
        24,
        26,
        28,
        30,
        32,
        34,
        36,
        38,
        40,
        42,
        44,
        46,
        48,
        50,
        52,
        54,
        56,
        58,
        60,
        62,
        64,
        66,
        68,
        70,
        72,
        74,
        76,
        78,
        80,
        82,
        84,
        86,
        88,
        90,
        92,
        94,
        96,
        98,
        100,
        102,
        104,
        106,
        108,
        110,
        112,
        114,
        116,
        118,
        120,
        122,
        124,
        126,
        128,
        130,
        132,
        134,
        136,
        138,
        140,
        142,
        144,
        146,
        148,
        150,
        152,
        154,
        156,
        158,
        160,
        162,
        164,
        166,
        168,
        170,
        172,
        174,
        176,
        178,
        180,
        182,
        184,
        186,
        188,
        190,
        192,
        194,
        196,
        198,
        200,
        202,
        204,
        206,
        208,
        210,
        212,
        214,
        216,
        218,
        220,
        222,
        224,
        226,
        228,
        230,
        232,
        234,
        236,
        238,
        240,
        242,
        244,
        246,
        248,
        250,
        252,
        254,
        256,
        258,
        260,
        262,
        264,
        266,
        268,
        270,
        272,
        274,
        276,
        278,
        280,
        282,
        284,
        286,
        288,
        290,
        292,
        294,
        296,
        298,
        300,
        302,
        304,
        306,
        308,
        310,
        312,
        314,
        316,
        318,
        320,
        322,
        324,
        326,
        328,
        330,
        332,
        334,
        336,
        338,
        340,
        342,
        344,
        346,
        348,
        350,
        352,
        354,
        356,
        358,
        360,
        362,
        364,
        366,
        368,
        370,
        372,
        374,
        376,
        378,
        380,
        382,
        384,
        386,
        388,
        390,
        392,
        394,
        396,
        398,
        400,
        402,
        404,
        406,
        408,
        410,
        412,
        414,
        416,
        418,
        420,
        422,
        424,
        426,
        428,
        430,
        432,
        434,
        436,
        438,
        440,
        442,
        444,
        446,
        0,
        4,
        2,
        1,
        72,
        72,
        114,
        114,
        1,
        0,
        90,
        96,
        1,
        0,
        99,
        111,
        2,
        0,
        64,
        64,
        66,
        66,
        2930,
        0,
        448,
        1,
        0,
        0,
        0,
        2,
        454,
        1,
        0,
        0,
        0,
        4,
        472,
        1,
        0,
        0,
        0,
        6,
        502,
        1,
        0,
        0,
        0,
        8,
        504,
        1,
        0,
        0,
        0,
        10,
        512,
        1,
        0,
        0,
        0,
        12,
        524,
        1,
        0,
        0,
        0,
        14,
        526,
        1,
        0,
        0,
        0,
        16,
        529,
        1,
        0,
        0,
        0,
        18,
        532,
        1,
        0,
        0,
        0,
        20,
        551,
        1,
        0,
        0,
        0,
        22,
        556,
        1,
        0,
        0,
        0,
        24,
        575,
        1,
        0,
        0,
        0,
        26,
        598,
        1,
        0,
        0,
        0,
        28,
        603,
        1,
        0,
        0,
        0,
        30,
        622,
        1,
        0,
        0,
        0,
        32,
        651,
        1,
        0,
        0,
        0,
        34,
        671,
        1,
        0,
        0,
        0,
        36,
        691,
        1,
        0,
        0,
        0,
        38,
        708,
        1,
        0,
        0,
        0,
        40,
        720,
        1,
        0,
        0,
        0,
        42,
        727,
        1,
        0,
        0,
        0,
        44,
        729,
        1,
        0,
        0,
        0,
        46,
        754,
        1,
        0,
        0,
        0,
        48,
        761,
        1,
        0,
        0,
        0,
        50,
        763,
        1,
        0,
        0,
        0,
        52,
        774,
        1,
        0,
        0,
        0,
        54,
        780,
        1,
        0,
        0,
        0,
        56,
        810,
        1,
        0,
        0,
        0,
        58,
        814,
        1,
        0,
        0,
        0,
        60,
        819,
        1,
        0,
        0,
        0,
        62,
        830,
        1,
        0,
        0,
        0,
        64,
        835,
        1,
        0,
        0,
        0,
        66,
        837,
        1,
        0,
        0,
        0,
        68,
        847,
        1,
        0,
        0,
        0,
        70,
        865,
        1,
        0,
        0,
        0,
        72,
        871,
        1,
        0,
        0,
        0,
        74,
        878,
        1,
        0,
        0,
        0,
        76,
        880,
        1,
        0,
        0,
        0,
        78,
        919,
        1,
        0,
        0,
        0,
        80,
        929,
        1,
        0,
        0,
        0,
        82,
        968,
        1,
        0,
        0,
        0,
        84,
        972,
        1,
        0,
        0,
        0,
        86,
        974,
        1,
        0,
        0,
        0,
        88,
        976,
        1,
        0,
        0,
        0,
        90,
        987,
        1,
        0,
        0,
        0,
        92,
        989,
        1,
        0,
        0,
        0,
        94,
        998,
        1,
        0,
        0,
        0,
        96,
        1007,
        1,
        0,
        0,
        0,
        98,
        1010,
        1,
        0,
        0,
        0,
        100,
        1012,
        1,
        0,
        0,
        0,
        102,
        1020,
        1,
        0,
        0,
        0,
        104,
        1022,
        1,
        0,
        0,
        0,
        106,
        1044,
        1,
        0,
        0,
        0,
        108,
        1055,
        1,
        0,
        0,
        0,
        110,
        1057,
        1,
        0,
        0,
        0,
        112,
        1065,
        1,
        0,
        0,
        0,
        114,
        1070,
        1,
        0,
        0,
        0,
        116,
        1078,
        1,
        0,
        0,
        0,
        118,
        1083,
        1,
        0,
        0,
        0,
        120,
        1100,
        1,
        0,
        0,
        0,
        122,
        1106,
        1,
        0,
        0,
        0,
        124,
        1111,
        1,
        0,
        0,
        0,
        126,
        1115,
        1,
        0,
        0,
        0,
        128,
        1131,
        1,
        0,
        0,
        0,
        130,
        1136,
        1,
        0,
        0,
        0,
        132,
        1158,
        1,
        0,
        0,
        0,
        134,
        1209,
        1,
        0,
        0,
        0,
        136,
        1212,
        1,
        0,
        0,
        0,
        138,
        1223,
        1,
        0,
        0,
        0,
        140,
        1268,
        1,
        0,
        0,
        0,
        142,
        1270,
        1,
        0,
        0,
        0,
        144,
        1277,
        1,
        0,
        0,
        0,
        146,
        1283,
        1,
        0,
        0,
        0,
        148,
        1292,
        1,
        0,
        0,
        0,
        150,
        1304,
        1,
        0,
        0,
        0,
        152,
        1306,
        1,
        0,
        0,
        0,
        154,
        1310,
        1,
        0,
        0,
        0,
        156,
        1313,
        1,
        0,
        0,
        0,
        158,
        1316,
        1,
        0,
        0,
        0,
        160,
        1319,
        1,
        0,
        0,
        0,
        162,
        1335,
        1,
        0,
        0,
        0,
        164,
        1350,
        1,
        0,
        0,
        0,
        166,
        1352,
        1,
        0,
        0,
        0,
        168,
        1356,
        1,
        0,
        0,
        0,
        170,
        1382,
        1,
        0,
        0,
        0,
        172,
        1444,
        1,
        0,
        0,
        0,
        174,
        1451,
        1,
        0,
        0,
        0,
        176,
        1486,
        1,
        0,
        0,
        0,
        178,
        1500,
        1,
        0,
        0,
        0,
        180,
        1502,
        1,
        0,
        0,
        0,
        182,
        1512,
        1,
        0,
        0,
        0,
        184,
        1516,
        1,
        0,
        0,
        0,
        186,
        1534,
        1,
        0,
        0,
        0,
        188,
        1536,
        1,
        0,
        0,
        0,
        190,
        1544,
        1,
        0,
        0,
        0,
        192,
        1549,
        1,
        0,
        0,
        0,
        194,
        1553,
        1,
        0,
        0,
        0,
        196,
        1555,
        1,
        0,
        0,
        0,
        198,
        1559,
        1,
        0,
        0,
        0,
        200,
        1575,
        1,
        0,
        0,
        0,
        202,
        1583,
        1,
        0,
        0,
        0,
        204,
        1591,
        1,
        0,
        0,
        0,
        206,
        1601,
        1,
        0,
        0,
        0,
        208,
        1606,
        1,
        0,
        0,
        0,
        210,
        1611,
        1,
        0,
        0,
        0,
        212,
        1613,
        1,
        0,
        0,
        0,
        214,
        1615,
        1,
        0,
        0,
        0,
        216,
        1617,
        1,
        0,
        0,
        0,
        218,
        1619,
        1,
        0,
        0,
        0,
        220,
        1621,
        1,
        0,
        0,
        0,
        222,
        1623,
        1,
        0,
        0,
        0,
        224,
        1625,
        1,
        0,
        0,
        0,
        226,
        1629,
        1,
        0,
        0,
        0,
        228,
        1640,
        1,
        0,
        0,
        0,
        230,
        1654,
        1,
        0,
        0,
        0,
        232,
        1656,
        1,
        0,
        0,
        0,
        234,
        1661,
        1,
        0,
        0,
        0,
        236,
        1674,
        1,
        0,
        0,
        0,
        238,
        1680,
        1,
        0,
        0,
        0,
        240,
        1707,
        1,
        0,
        0,
        0,
        242,
        1709,
        1,
        0,
        0,
        0,
        244,
        1719,
        1,
        0,
        0,
        0,
        246,
        1724,
        1,
        0,
        0,
        0,
        248,
        1757,
        1,
        0,
        0,
        0,
        250,
        1759,
        1,
        0,
        0,
        0,
        252,
        1767,
        1,
        0,
        0,
        0,
        254,
        1775,
        1,
        0,
        0,
        0,
        256,
        1779,
        1,
        0,
        0,
        0,
        258,
        1787,
        1,
        0,
        0,
        0,
        260,
        1791,
        1,
        0,
        0,
        0,
        262,
        1810,
        1,
        0,
        0,
        0,
        264,
        1812,
        1,
        0,
        0,
        0,
        266,
        1815,
        1,
        0,
        0,
        0,
        268,
        1834,
        1,
        0,
        0,
        0,
        270,
        1843,
        1,
        0,
        0,
        0,
        272,
        1845,
        1,
        0,
        0,
        0,
        274,
        1859,
        1,
        0,
        0,
        0,
        276,
        1861,
        1,
        0,
        0,
        0,
        278,
        1875,
        1,
        0,
        0,
        0,
        280,
        1877,
        1,
        0,
        0,
        0,
        282,
        1883,
        1,
        0,
        0,
        0,
        284,
        1885,
        1,
        0,
        0,
        0,
        286,
        1893,
        1,
        0,
        0,
        0,
        288,
        1904,
        1,
        0,
        0,
        0,
        290,
        1906,
        1,
        0,
        0,
        0,
        292,
        1923,
        1,
        0,
        0,
        0,
        294,
        1925,
        1,
        0,
        0,
        0,
        296,
        1928,
        1,
        0,
        0,
        0,
        298,
        1931,
        1,
        0,
        0,
        0,
        300,
        1934,
        1,
        0,
        0,
        0,
        302,
        1937,
        1,
        0,
        0,
        0,
        304,
        1940,
        1,
        0,
        0,
        0,
        306,
        1943,
        1,
        0,
        0,
        0,
        308,
        1947,
        1,
        0,
        0,
        0,
        310,
        1950,
        1,
        0,
        0,
        0,
        312,
        1954,
        1,
        0,
        0,
        0,
        314,
        1957,
        1,
        0,
        0,
        0,
        316,
        1968,
        1,
        0,
        0,
        0,
        318,
        1979,
        1,
        0,
        0,
        0,
        320,
        1990,
        1,
        0,
        0,
        0,
        322,
        2004,
        1,
        0,
        0,
        0,
        324,
        2018,
        1,
        0,
        0,
        0,
        326,
        2048,
        1,
        0,
        0,
        0,
        328,
        2055,
        1,
        0,
        0,
        0,
        330,
        2060,
        1,
        0,
        0,
        0,
        332,
        2062,
        1,
        0,
        0,
        0,
        334,
        2088,
        1,
        0,
        0,
        0,
        336,
        2117,
        1,
        0,
        0,
        0,
        338,
        2142,
        1,
        0,
        0,
        0,
        340,
        2144,
        1,
        0,
        0,
        0,
        342,
        2151,
        1,
        0,
        0,
        0,
        344,
        2158,
        1,
        0,
        0,
        0,
        346,
        2209,
        1,
        0,
        0,
        0,
        348,
        2212,
        1,
        0,
        0,
        0,
        350,
        2223,
        1,
        0,
        0,
        0,
        352,
        2257,
        1,
        0,
        0,
        0,
        354,
        2259,
        1,
        0,
        0,
        0,
        356,
        2262,
        1,
        0,
        0,
        0,
        358,
        2266,
        1,
        0,
        0,
        0,
        360,
        2271,
        1,
        0,
        0,
        0,
        362,
        2278,
        1,
        0,
        0,
        0,
        364,
        2282,
        1,
        0,
        0,
        0,
        366,
        2284,
        1,
        0,
        0,
        0,
        368,
        2303,
        1,
        0,
        0,
        0,
        370,
        2305,
        1,
        0,
        0,
        0,
        372,
        2314,
        1,
        0,
        0,
        0,
        374,
        2316,
        1,
        0,
        0,
        0,
        376,
        2325,
        1,
        0,
        0,
        0,
        378,
        2329,
        1,
        0,
        0,
        0,
        380,
        2333,
        1,
        0,
        0,
        0,
        382,
        2339,
        1,
        0,
        0,
        0,
        384,
        2348,
        1,
        0,
        0,
        0,
        386,
        2352,
        1,
        0,
        0,
        0,
        388,
        2358,
        1,
        0,
        0,
        0,
        390,
        2372,
        1,
        0,
        0,
        0,
        392,
        2374,
        1,
        0,
        0,
        0,
        394,
        2379,
        1,
        0,
        0,
        0,
        396,
        2384,
        1,
        0,
        0,
        0,
        398,
        2397,
        1,
        0,
        0,
        0,
        400,
        2402,
        1,
        0,
        0,
        0,
        402,
        2407,
        1,
        0,
        0,
        0,
        404,
        2415,
        1,
        0,
        0,
        0,
        406,
        2420,
        1,
        0,
        0,
        0,
        408,
        2437,
        1,
        0,
        0,
        0,
        410,
        2442,
        1,
        0,
        0,
        0,
        412,
        2477,
        1,
        0,
        0,
        0,
        414,
        2479,
        1,
        0,
        0,
        0,
        416,
        2486,
        1,
        0,
        0,
        0,
        418,
        2493,
        1,
        0,
        0,
        0,
        420,
        2495,
        1,
        0,
        0,
        0,
        422,
        2506,
        1,
        0,
        0,
        0,
        424,
        2530,
        1,
        0,
        0,
        0,
        426,
        2535,
        1,
        0,
        0,
        0,
        428,
        2547,
        1,
        0,
        0,
        0,
        430,
        2564,
        1,
        0,
        0,
        0,
        432,
        2572,
        1,
        0,
        0,
        0,
        434,
        2583,
        1,
        0,
        0,
        0,
        436,
        2585,
        1,
        0,
        0,
        0,
        438,
        2609,
        1,
        0,
        0,
        0,
        440,
        2630,
        1,
        0,
        0,
        0,
        442,
        2643,
        1,
        0,
        0,
        0,
        444,
        2702,
        1,
        0,
        0,
        0,
        446,
        2704,
        1,
        0,
        0,
        0,
        448,
        449,
        3,
        2,
        1,
        0,
        449,
        450,
        5,
        0,
        0,
        1,
        450,
        1,
        1,
        0,
        0,
        0,
        451,
        453,
        5,
        114,
        0,
        0,
        452,
        451,
        1,
        0,
        0,
        0,
        453,
        456,
        1,
        0,
        0,
        0,
        454,
        452,
        1,
        0,
        0,
        0,
        454,
        455,
        1,
        0,
        0,
        0,
        455,
        466,
        1,
        0,
        0,
        0,
        456,
        454,
        1,
        0,
        0,
        0,
        457,
        461,
        3,
        4,
        2,
        0,
        458,
        460,
        5,
        114,
        0,
        0,
        459,
        458,
        1,
        0,
        0,
        0,
        460,
        463,
        1,
        0,
        0,
        0,
        461,
        459,
        1,
        0,
        0,
        0,
        461,
        462,
        1,
        0,
        0,
        0,
        462,
        465,
        1,
        0,
        0,
        0,
        463,
        461,
        1,
        0,
        0,
        0,
        464,
        457,
        1,
        0,
        0,
        0,
        465,
        468,
        1,
        0,
        0,
        0,
        466,
        464,
        1,
        0,
        0,
        0,
        466,
        467,
        1,
        0,
        0,
        0,
        467,
        3,
        1,
        0,
        0,
        0,
        468,
        466,
        1,
        0,
        0,
        0,
        469,
        473,
        3,
        6,
        3,
        0,
        470,
        473,
        3,
        26,
        13,
        0,
        471,
        473,
        3,
        58,
        29,
        0,
        472,
        469,
        1,
        0,
        0,
        0,
        472,
        470,
        1,
        0,
        0,
        0,
        472,
        471,
        1,
        0,
        0,
        0,
        473,
        5,
        1,
        0,
        0,
        0,
        474,
        475,
        5,
        8,
        0,
        0,
        475,
        476,
        5,
        61,
        0,
        0,
        476,
        479,
        3,
        8,
        4,
        0,
        477,
        478,
        5,
        112,
        0,
        0,
        478,
        480,
        3,
        268,
        134,
        0,
        479,
        477,
        1,
        0,
        0,
        0,
        479,
        480,
        1,
        0,
        0,
        0,
        480,
        481,
        1,
        0,
        0,
        0,
        481,
        482,
        7,
        0,
        0,
        0,
        482,
        503,
        1,
        0,
        0,
        0,
        483,
        484,
        5,
        8,
        0,
        0,
        484,
        485,
        5,
        61,
        0,
        0,
        485,
        488,
        3,
        8,
        4,
        0,
        486,
        487,
        5,
        74,
        0,
        0,
        487,
        489,
        3,
        268,
        134,
        0,
        488,
        486,
        1,
        0,
        0,
        0,
        488,
        489,
        1,
        0,
        0,
        0,
        489,
        490,
        1,
        0,
        0,
        0,
        490,
        491,
        7,
        0,
        0,
        0,
        491,
        503,
        1,
        0,
        0,
        0,
        492,
        493,
        5,
        8,
        0,
        0,
        493,
        494,
        5,
        61,
        0,
        0,
        494,
        498,
        3,
        8,
        4,
        0,
        495,
        496,
        5,
        71,
        0,
        0,
        496,
        497,
        5,
        71,
        0,
        0,
        497,
        499,
        3,
        268,
        134,
        0,
        498,
        495,
        1,
        0,
        0,
        0,
        498,
        499,
        1,
        0,
        0,
        0,
        499,
        500,
        1,
        0,
        0,
        0,
        500,
        501,
        7,
        0,
        0,
        0,
        501,
        503,
        1,
        0,
        0,
        0,
        502,
        474,
        1,
        0,
        0,
        0,
        502,
        483,
        1,
        0,
        0,
        0,
        502,
        492,
        1,
        0,
        0,
        0,
        503,
        7,
        1,
        0,
        0,
        0,
        504,
        509,
        3,
        10,
        5,
        0,
        505,
        506,
        5,
        77,
        0,
        0,
        506,
        508,
        3,
        10,
        5,
        0,
        507,
        505,
        1,
        0,
        0,
        0,
        508,
        511,
        1,
        0,
        0,
        0,
        509,
        507,
        1,
        0,
        0,
        0,
        509,
        510,
        1,
        0,
        0,
        0,
        510,
        9,
        1,
        0,
        0,
        0,
        511,
        509,
        1,
        0,
        0,
        0,
        512,
        516,
        3,
        12,
        6,
        0,
        513,
        515,
        3,
        12,
        6,
        0,
        514,
        513,
        1,
        0,
        0,
        0,
        515,
        518,
        1,
        0,
        0,
        0,
        516,
        514,
        1,
        0,
        0,
        0,
        516,
        517,
        1,
        0,
        0,
        0,
        517,
        11,
        1,
        0,
        0,
        0,
        518,
        516,
        1,
        0,
        0,
        0,
        519,
        525,
        3,
        22,
        11,
        0,
        520,
        525,
        3,
        14,
        7,
        0,
        521,
        525,
        3,
        16,
        8,
        0,
        522,
        525,
        3,
        18,
        9,
        0,
        523,
        525,
        3,
        20,
        10,
        0,
        524,
        519,
        1,
        0,
        0,
        0,
        524,
        520,
        1,
        0,
        0,
        0,
        524,
        521,
        1,
        0,
        0,
        0,
        524,
        522,
        1,
        0,
        0,
        0,
        524,
        523,
        1,
        0,
        0,
        0,
        525,
        13,
        1,
        0,
        0,
        0,
        526,
        527,
        3,
        22,
        11,
        0,
        527,
        528,
        5,
        67,
        0,
        0,
        528,
        15,
        1,
        0,
        0,
        0,
        529,
        530,
        3,
        22,
        11,
        0,
        530,
        531,
        5,
        82,
        0,
        0,
        531,
        17,
        1,
        0,
        0,
        0,
        532,
        533,
        3,
        22,
        11,
        0,
        533,
        534,
        5,
        62,
        0,
        0,
        534,
        19,
        1,
        0,
        0,
        0,
        535,
        536,
        3,
        22,
        11,
        0,
        536,
        537,
        5,
        88,
        0,
        0,
        537,
        538,
        5,
        4,
        0,
        0,
        538,
        539,
        5,
        89,
        0,
        0,
        539,
        552,
        1,
        0,
        0,
        0,
        540,
        541,
        3,
        22,
        11,
        0,
        541,
        543,
        5,
        88,
        0,
        0,
        542,
        544,
        5,
        4,
        0,
        0,
        543,
        542,
        1,
        0,
        0,
        0,
        543,
        544,
        1,
        0,
        0,
        0,
        544,
        545,
        1,
        0,
        0,
        0,
        545,
        547,
        5,
        70,
        0,
        0,
        546,
        548,
        5,
        4,
        0,
        0,
        547,
        546,
        1,
        0,
        0,
        0,
        547,
        548,
        1,
        0,
        0,
        0,
        548,
        549,
        1,
        0,
        0,
        0,
        549,
        550,
        5,
        89,
        0,
        0,
        550,
        552,
        1,
        0,
        0,
        0,
        551,
        535,
        1,
        0,
        0,
        0,
        551,
        540,
        1,
        0,
        0,
        0,
        552,
        21,
        1,
        0,
        0,
        0,
        553,
        555,
        5,
        114,
        0,
        0,
        554,
        553,
        1,
        0,
        0,
        0,
        555,
        558,
        1,
        0,
        0,
        0,
        556,
        554,
        1,
        0,
        0,
        0,
        556,
        557,
        1,
        0,
        0,
        0,
        557,
        567,
        1,
        0,
        0,
        0,
        558,
        556,
        1,
        0,
        0,
        0,
        559,
        568,
        5,
        8,
        0,
        0,
        560,
        568,
        5,
        3,
        0,
        0,
        561,
        568,
        5,
        4,
        0,
        0,
        562,
        563,
        5,
        68,
        0,
        0,
        563,
        564,
        3,
        8,
        4,
        0,
        564,
        565,
        5,
        69,
        0,
        0,
        565,
        568,
        1,
        0,
        0,
        0,
        566,
        568,
        3,
        24,
        12,
        0,
        567,
        559,
        1,
        0,
        0,
        0,
        567,
        560,
        1,
        0,
        0,
        0,
        567,
        561,
        1,
        0,
        0,
        0,
        567,
        562,
        1,
        0,
        0,
        0,
        567,
        566,
        1,
        0,
        0,
        0,
        568,
        572,
        1,
        0,
        0,
        0,
        569,
        571,
        5,
        114,
        0,
        0,
        570,
        569,
        1,
        0,
        0,
        0,
        571,
        574,
        1,
        0,
        0,
        0,
        572,
        570,
        1,
        0,
        0,
        0,
        572,
        573,
        1,
        0,
        0,
        0,
        573,
        23,
        1,
        0,
        0,
        0,
        574,
        572,
        1,
        0,
        0,
        0,
        575,
        577,
        5,
        75,
        0,
        0,
        576,
        578,
        5,
        78,
        0,
        0,
        577,
        576,
        1,
        0,
        0,
        0,
        577,
        578,
        1,
        0,
        0,
        0,
        578,
        579,
        1,
        0,
        0,
        0,
        579,
        580,
        5,
        3,
        0,
        0,
        580,
        581,
        5,
        76,
        0,
        0,
        581,
        25,
        1,
        0,
        0,
        0,
        582,
        583,
        5,
        46,
        0,
        0,
        583,
        584,
        3,
        28,
        14,
        0,
        584,
        585,
        7,
        0,
        0,
        0,
        585,
        599,
        1,
        0,
        0,
        0,
        586,
        587,
        5,
        46,
        0,
        0,
        587,
        588,
        5,
        51,
        0,
        0,
        588,
        589,
        3,
        40,
        20,
        0,
        589,
        590,
        7,
        0,
        0,
        0,
        590,
        599,
        1,
        0,
        0,
        0,
        591,
        592,
        3,
        28,
        14,
        0,
        592,
        593,
        5,
        72,
        0,
        0,
        593,
        599,
        1,
        0,
        0,
        0,
        594,
        595,
        5,
        51,
        0,
        0,
        595,
        596,
        3,
        40,
        20,
        0,
        596,
        597,
        5,
        72,
        0,
        0,
        597,
        599,
        1,
        0,
        0,
        0,
        598,
        582,
        1,
        0,
        0,
        0,
        598,
        586,
        1,
        0,
        0,
        0,
        598,
        591,
        1,
        0,
        0,
        0,
        598,
        594,
        1,
        0,
        0,
        0,
        599,
        27,
        1,
        0,
        0,
        0,
        600,
        602,
        5,
        114,
        0,
        0,
        601,
        600,
        1,
        0,
        0,
        0,
        602,
        605,
        1,
        0,
        0,
        0,
        603,
        601,
        1,
        0,
        0,
        0,
        603,
        604,
        1,
        0,
        0,
        0,
        604,
        611,
        1,
        0,
        0,
        0,
        605,
        603,
        1,
        0,
        0,
        0,
        606,
        607,
        3,
        30,
        15,
        0,
        607,
        608,
        5,
        98,
        0,
        0,
        608,
        609,
        3,
        28,
        14,
        0,
        609,
        612,
        1,
        0,
        0,
        0,
        610,
        612,
        3,
        30,
        15,
        0,
        611,
        606,
        1,
        0,
        0,
        0,
        611,
        610,
        1,
        0,
        0,
        0,
        612,
        616,
        1,
        0,
        0,
        0,
        613,
        615,
        5,
        114,
        0,
        0,
        614,
        613,
        1,
        0,
        0,
        0,
        615,
        618,
        1,
        0,
        0,
        0,
        616,
        614,
        1,
        0,
        0,
        0,
        616,
        617,
        1,
        0,
        0,
        0,
        617,
        29,
        1,
        0,
        0,
        0,
        618,
        616,
        1,
        0,
        0,
        0,
        619,
        621,
        5,
        114,
        0,
        0,
        620,
        619,
        1,
        0,
        0,
        0,
        621,
        624,
        1,
        0,
        0,
        0,
        622,
        620,
        1,
        0,
        0,
        0,
        622,
        623,
        1,
        0,
        0,
        0,
        623,
        640,
        1,
        0,
        0,
        0,
        624,
        622,
        1,
        0,
        0,
        0,
        625,
        626,
        5,
        49,
        0,
        0,
        626,
        627,
        5,
        8,
        0,
        0,
        627,
        628,
        5,
        30,
        0,
        0,
        628,
        629,
        3,
        44,
        22,
        0,
        629,
        630,
        5,
        71,
        0,
        0,
        630,
        631,
        3,
        30,
        15,
        0,
        631,
        641,
        1,
        0,
        0,
        0,
        632,
        633,
        5,
        50,
        0,
        0,
        633,
        634,
        5,
        8,
        0,
        0,
        634,
        635,
        5,
        30,
        0,
        0,
        635,
        636,
        3,
        44,
        22,
        0,
        636,
        637,
        5,
        71,
        0,
        0,
        637,
        638,
        3,
        30,
        15,
        0,
        638,
        641,
        1,
        0,
        0,
        0,
        639,
        641,
        3,
        32,
        16,
        0,
        640,
        625,
        1,
        0,
        0,
        0,
        640,
        632,
        1,
        0,
        0,
        0,
        640,
        639,
        1,
        0,
        0,
        0,
        641,
        645,
        1,
        0,
        0,
        0,
        642,
        644,
        5,
        114,
        0,
        0,
        643,
        642,
        1,
        0,
        0,
        0,
        644,
        647,
        1,
        0,
        0,
        0,
        645,
        643,
        1,
        0,
        0,
        0,
        645,
        646,
        1,
        0,
        0,
        0,
        646,
        31,
        1,
        0,
        0,
        0,
        647,
        645,
        1,
        0,
        0,
        0,
        648,
        650,
        5,
        114,
        0,
        0,
        649,
        648,
        1,
        0,
        0,
        0,
        650,
        653,
        1,
        0,
        0,
        0,
        651,
        649,
        1,
        0,
        0,
        0,
        651,
        652,
        1,
        0,
        0,
        0,
        652,
        654,
        1,
        0,
        0,
        0,
        653,
        651,
        1,
        0,
        0,
        0,
        654,
        659,
        3,
        34,
        17,
        0,
        655,
        656,
        5,
        37,
        0,
        0,
        656,
        658,
        3,
        34,
        17,
        0,
        657,
        655,
        1,
        0,
        0,
        0,
        658,
        661,
        1,
        0,
        0,
        0,
        659,
        657,
        1,
        0,
        0,
        0,
        659,
        660,
        1,
        0,
        0,
        0,
        660,
        665,
        1,
        0,
        0,
        0,
        661,
        659,
        1,
        0,
        0,
        0,
        662,
        664,
        5,
        114,
        0,
        0,
        663,
        662,
        1,
        0,
        0,
        0,
        664,
        667,
        1,
        0,
        0,
        0,
        665,
        663,
        1,
        0,
        0,
        0,
        665,
        666,
        1,
        0,
        0,
        0,
        666,
        33,
        1,
        0,
        0,
        0,
        667,
        665,
        1,
        0,
        0,
        0,
        668,
        670,
        5,
        114,
        0,
        0,
        669,
        668,
        1,
        0,
        0,
        0,
        670,
        673,
        1,
        0,
        0,
        0,
        671,
        669,
        1,
        0,
        0,
        0,
        671,
        672,
        1,
        0,
        0,
        0,
        672,
        674,
        1,
        0,
        0,
        0,
        673,
        671,
        1,
        0,
        0,
        0,
        674,
        679,
        3,
        36,
        18,
        0,
        675,
        676,
        5,
        9,
        0,
        0,
        676,
        678,
        3,
        36,
        18,
        0,
        677,
        675,
        1,
        0,
        0,
        0,
        678,
        681,
        1,
        0,
        0,
        0,
        679,
        677,
        1,
        0,
        0,
        0,
        679,
        680,
        1,
        0,
        0,
        0,
        680,
        685,
        1,
        0,
        0,
        0,
        681,
        679,
        1,
        0,
        0,
        0,
        682,
        684,
        5,
        114,
        0,
        0,
        683,
        682,
        1,
        0,
        0,
        0,
        684,
        687,
        1,
        0,
        0,
        0,
        685,
        683,
        1,
        0,
        0,
        0,
        685,
        686,
        1,
        0,
        0,
        0,
        686,
        35,
        1,
        0,
        0,
        0,
        687,
        685,
        1,
        0,
        0,
        0,
        688,
        690,
        5,
        114,
        0,
        0,
        689,
        688,
        1,
        0,
        0,
        0,
        690,
        693,
        1,
        0,
        0,
        0,
        691,
        689,
        1,
        0,
        0,
        0,
        691,
        692,
        1,
        0,
        0,
        0,
        692,
        700,
        1,
        0,
        0,
        0,
        693,
        691,
        1,
        0,
        0,
        0,
        694,
        701,
        3,
        38,
        19,
        0,
        695,
        696,
        5,
        68,
        0,
        0,
        696,
        697,
        3,
        28,
        14,
        0,
        697,
        698,
        5,
        69,
        0,
        0,
        698,
        701,
        1,
        0,
        0,
        0,
        699,
        701,
        3,
        40,
        20,
        0,
        700,
        694,
        1,
        0,
        0,
        0,
        700,
        695,
        1,
        0,
        0,
        0,
        700,
        699,
        1,
        0,
        0,
        0,
        701,
        705,
        1,
        0,
        0,
        0,
        702,
        704,
        5,
        114,
        0,
        0,
        703,
        702,
        1,
        0,
        0,
        0,
        704,
        707,
        1,
        0,
        0,
        0,
        705,
        703,
        1,
        0,
        0,
        0,
        705,
        706,
        1,
        0,
        0,
        0,
        706,
        37,
        1,
        0,
        0,
        0,
        707,
        705,
        1,
        0,
        0,
        0,
        708,
        709,
        3,
        40,
        20,
        0,
        709,
        710,
        7,
        1,
        0,
        0,
        710,
        711,
        3,
        40,
        20,
        0,
        711,
        39,
        1,
        0,
        0,
        0,
        712,
        721,
        3,
        42,
        21,
        0,
        713,
        721,
        3,
        288,
        144,
        0,
        714,
        715,
        3,
        288,
        144,
        0,
        715,
        716,
        5,
        28,
        0,
        0,
        716,
        717,
        3,
        288,
        144,
        0,
        717,
        718,
        5,
        21,
        0,
        0,
        718,
        719,
        3,
        288,
        144,
        0,
        719,
        721,
        1,
        0,
        0,
        0,
        720,
        712,
        1,
        0,
        0,
        0,
        720,
        713,
        1,
        0,
        0,
        0,
        720,
        714,
        1,
        0,
        0,
        0,
        721,
        41,
        1,
        0,
        0,
        0,
        722,
        723,
        5,
        77,
        0,
        0,
        723,
        724,
        3,
        44,
        22,
        0,
        724,
        725,
        5,
        77,
        0,
        0,
        725,
        728,
        1,
        0,
        0,
        0,
        726,
        728,
        3,
        44,
        22,
        0,
        727,
        722,
        1,
        0,
        0,
        0,
        727,
        726,
        1,
        0,
        0,
        0,
        728,
        43,
        1,
        0,
        0,
        0,
        729,
        730,
        6,
        22,
        -1,
        0,
        730,
        731,
        3,
        46,
        23,
        0,
        731,
        740,
        1,
        0,
        0,
        0,
        732,
        733,
        10,
        2,
        0,
        0,
        733,
        734,
        5,
        64,
        0,
        0,
        734,
        739,
        3,
        46,
        23,
        0,
        735,
        736,
        10,
        1,
        0,
        0,
        736,
        737,
        5,
        65,
        0,
        0,
        737,
        739,
        3,
        46,
        23,
        0,
        738,
        732,
        1,
        0,
        0,
        0,
        738,
        735,
        1,
        0,
        0,
        0,
        739,
        742,
        1,
        0,
        0,
        0,
        740,
        738,
        1,
        0,
        0,
        0,
        740,
        741,
        1,
        0,
        0,
        0,
        741,
        45,
        1,
        0,
        0,
        0,
        742,
        740,
        1,
        0,
        0,
        0,
        743,
        755,
        3,
        48,
        24,
        0,
        744,
        745,
        3,
        48,
        24,
        0,
        745,
        746,
        5,
        75,
        0,
        0,
        746,
        747,
        3,
        54,
        27,
        0,
        747,
        748,
        5,
        76,
        0,
        0,
        748,
        755,
        1,
        0,
        0,
        0,
        749,
        750,
        3,
        48,
        24,
        0,
        750,
        751,
        5,
        88,
        0,
        0,
        751,
        752,
        3,
        50,
        25,
        0,
        752,
        753,
        5,
        89,
        0,
        0,
        753,
        755,
        1,
        0,
        0,
        0,
        754,
        743,
        1,
        0,
        0,
        0,
        754,
        744,
        1,
        0,
        0,
        0,
        754,
        749,
        1,
        0,
        0,
        0,
        755,
        47,
        1,
        0,
        0,
        0,
        756,
        762,
        5,
        8,
        0,
        0,
        757,
        758,
        5,
        68,
        0,
        0,
        758,
        759,
        3,
        44,
        22,
        0,
        759,
        760,
        5,
        69,
        0,
        0,
        760,
        762,
        1,
        0,
        0,
        0,
        761,
        756,
        1,
        0,
        0,
        0,
        761,
        757,
        1,
        0,
        0,
        0,
        762,
        49,
        1,
        0,
        0,
        0,
        763,
        768,
        3,
        52,
        26,
        0,
        764,
        765,
        5,
        70,
        0,
        0,
        765,
        767,
        3,
        52,
        26,
        0,
        766,
        764,
        1,
        0,
        0,
        0,
        767,
        770,
        1,
        0,
        0,
        0,
        768,
        766,
        1,
        0,
        0,
        0,
        768,
        769,
        1,
        0,
        0,
        0,
        769,
        772,
        1,
        0,
        0,
        0,
        770,
        768,
        1,
        0,
        0,
        0,
        771,
        773,
        5,
        70,
        0,
        0,
        772,
        771,
        1,
        0,
        0,
        0,
        772,
        773,
        1,
        0,
        0,
        0,
        773,
        51,
        1,
        0,
        0,
        0,
        774,
        775,
        5,
        67,
        0,
        0,
        775,
        778,
        5,
        8,
        0,
        0,
        776,
        777,
        5,
        71,
        0,
        0,
        777,
        779,
        3,
        56,
        28,
        0,
        778,
        776,
        1,
        0,
        0,
        0,
        778,
        779,
        1,
        0,
        0,
        0,
        779,
        53,
        1,
        0,
        0,
        0,
        780,
        785,
        3,
        56,
        28,
        0,
        781,
        782,
        5,
        70,
        0,
        0,
        782,
        784,
        3,
        56,
        28,
        0,
        783,
        781,
        1,
        0,
        0,
        0,
        784,
        787,
        1,
        0,
        0,
        0,
        785,
        783,
        1,
        0,
        0,
        0,
        785,
        786,
        1,
        0,
        0,
        0,
        786,
        789,
        1,
        0,
        0,
        0,
        787,
        785,
        1,
        0,
        0,
        0,
        788,
        790,
        5,
        70,
        0,
        0,
        789,
        788,
        1,
        0,
        0,
        0,
        789,
        790,
        1,
        0,
        0,
        0,
        790,
        55,
        1,
        0,
        0,
        0,
        791,
        811,
        5,
        4,
        0,
        0,
        792,
        794,
        5,
        4,
        0,
        0,
        793,
        792,
        1,
        0,
        0,
        0,
        793,
        794,
        1,
        0,
        0,
        0,
        794,
        795,
        1,
        0,
        0,
        0,
        795,
        797,
        5,
        71,
        0,
        0,
        796,
        798,
        5,
        4,
        0,
        0,
        797,
        796,
        1,
        0,
        0,
        0,
        797,
        798,
        1,
        0,
        0,
        0,
        798,
        811,
        1,
        0,
        0,
        0,
        799,
        801,
        5,
        4,
        0,
        0,
        800,
        799,
        1,
        0,
        0,
        0,
        800,
        801,
        1,
        0,
        0,
        0,
        801,
        802,
        1,
        0,
        0,
        0,
        802,
        804,
        5,
        71,
        0,
        0,
        803,
        805,
        5,
        4,
        0,
        0,
        804,
        803,
        1,
        0,
        0,
        0,
        804,
        805,
        1,
        0,
        0,
        0,
        805,
        806,
        1,
        0,
        0,
        0,
        806,
        808,
        5,
        71,
        0,
        0,
        807,
        809,
        5,
        4,
        0,
        0,
        808,
        807,
        1,
        0,
        0,
        0,
        808,
        809,
        1,
        0,
        0,
        0,
        809,
        811,
        1,
        0,
        0,
        0,
        810,
        791,
        1,
        0,
        0,
        0,
        810,
        793,
        1,
        0,
        0,
        0,
        810,
        800,
        1,
        0,
        0,
        0,
        811,
        57,
        1,
        0,
        0,
        0,
        812,
        815,
        3,
        80,
        40,
        0,
        813,
        815,
        3,
        78,
        39,
        0,
        814,
        812,
        1,
        0,
        0,
        0,
        814,
        813,
        1,
        0,
        0,
        0,
        815,
        59,
        1,
        0,
        0,
        0,
        816,
        818,
        5,
        114,
        0,
        0,
        817,
        816,
        1,
        0,
        0,
        0,
        818,
        821,
        1,
        0,
        0,
        0,
        819,
        817,
        1,
        0,
        0,
        0,
        819,
        820,
        1,
        0,
        0,
        0,
        820,
        822,
        1,
        0,
        0,
        0,
        821,
        819,
        1,
        0,
        0,
        0,
        822,
        826,
        3,
        72,
        36,
        0,
        823,
        825,
        5,
        114,
        0,
        0,
        824,
        823,
        1,
        0,
        0,
        0,
        825,
        828,
        1,
        0,
        0,
        0,
        826,
        824,
        1,
        0,
        0,
        0,
        826,
        827,
        1,
        0,
        0,
        0,
        827,
        61,
        1,
        0,
        0,
        0,
        828,
        826,
        1,
        0,
        0,
        0,
        829,
        831,
        3,
        70,
        35,
        0,
        830,
        829,
        1,
        0,
        0,
        0,
        830,
        831,
        1,
        0,
        0,
        0,
        831,
        833,
        1,
        0,
        0,
        0,
        832,
        834,
        5,
        0,
        0,
        1,
        833,
        832,
        1,
        0,
        0,
        0,
        833,
        834,
        1,
        0,
        0,
        0,
        834,
        63,
        1,
        0,
        0,
        0,
        835,
        836,
        3,
        74,
        37,
        0,
        836,
        65,
        1,
        0,
        0,
        0,
        837,
        841,
        3,
        266,
        133,
        0,
        838,
        840,
        5,
        114,
        0,
        0,
        839,
        838,
        1,
        0,
        0,
        0,
        840,
        843,
        1,
        0,
        0,
        0,
        841,
        839,
        1,
        0,
        0,
        0,
        841,
        842,
        1,
        0,
        0,
        0,
        842,
        845,
        1,
        0,
        0,
        0,
        843,
        841,
        1,
        0,
        0,
        0,
        844,
        846,
        5,
        0,
        0,
        1,
        845,
        844,
        1,
        0,
        0,
        0,
        845,
        846,
        1,
        0,
        0,
        0,
        846,
        67,
        1,
        0,
        0,
        0,
        847,
        849,
        5,
        68,
        0,
        0,
        848,
        850,
        3,
        444,
        222,
        0,
        849,
        848,
        1,
        0,
        0,
        0,
        849,
        850,
        1,
        0,
        0,
        0,
        850,
        851,
        1,
        0,
        0,
        0,
        851,
        852,
        5,
        69,
        0,
        0,
        852,
        853,
        5,
        98,
        0,
        0,
        853,
        857,
        3,
        268,
        134,
        0,
        854,
        856,
        5,
        114,
        0,
        0,
        855,
        854,
        1,
        0,
        0,
        0,
        856,
        859,
        1,
        0,
        0,
        0,
        857,
        855,
        1,
        0,
        0,
        0,
        857,
        858,
        1,
        0,
        0,
        0,
        858,
        861,
        1,
        0,
        0,
        0,
        859,
        857,
        1,
        0,
        0,
        0,
        860,
        862,
        5,
        0,
        0,
        1,
        861,
        860,
        1,
        0,
        0,
        0,
        861,
        862,
        1,
        0,
        0,
        0,
        862,
        69,
        1,
        0,
        0,
        0,
        863,
        866,
        3,
        72,
        36,
        0,
        864,
        866,
        5,
        114,
        0,
        0,
        865,
        863,
        1,
        0,
        0,
        0,
        865,
        864,
        1,
        0,
        0,
        0,
        866,
        867,
        1,
        0,
        0,
        0,
        867,
        865,
        1,
        0,
        0,
        0,
        867,
        868,
        1,
        0,
        0,
        0,
        868,
        71,
        1,
        0,
        0,
        0,
        869,
        872,
        3,
        80,
        40,
        0,
        870,
        872,
        3,
        76,
        38,
        0,
        871,
        869,
        1,
        0,
        0,
        0,
        871,
        870,
        1,
        0,
        0,
        0,
        872,
        73,
        1,
        0,
        0,
        0,
        873,
        874,
        3,
        80,
        40,
        0,
        874,
        875,
        5,
        114,
        0,
        0,
        875,
        879,
        1,
        0,
        0,
        0,
        876,
        879,
        3,
        76,
        38,
        0,
        877,
        879,
        5,
        114,
        0,
        0,
        878,
        873,
        1,
        0,
        0,
        0,
        878,
        876,
        1,
        0,
        0,
        0,
        878,
        877,
        1,
        0,
        0,
        0,
        879,
        75,
        1,
        0,
        0,
        0,
        880,
        885,
        3,
        78,
        39,
        0,
        881,
        882,
        5,
        72,
        0,
        0,
        882,
        884,
        3,
        78,
        39,
        0,
        883,
        881,
        1,
        0,
        0,
        0,
        884,
        887,
        1,
        0,
        0,
        0,
        885,
        883,
        1,
        0,
        0,
        0,
        885,
        886,
        1,
        0,
        0,
        0,
        886,
        889,
        1,
        0,
        0,
        0,
        887,
        885,
        1,
        0,
        0,
        0,
        888,
        890,
        5,
        72,
        0,
        0,
        889,
        888,
        1,
        0,
        0,
        0,
        889,
        890,
        1,
        0,
        0,
        0,
        890,
        903,
        1,
        0,
        0,
        0,
        891,
        893,
        5,
        114,
        0,
        0,
        892,
        891,
        1,
        0,
        0,
        0,
        893,
        894,
        1,
        0,
        0,
        0,
        894,
        892,
        1,
        0,
        0,
        0,
        894,
        895,
        1,
        0,
        0,
        0,
        895,
        904,
        1,
        0,
        0,
        0,
        896,
        898,
        5,
        114,
        0,
        0,
        897,
        896,
        1,
        0,
        0,
        0,
        898,
        901,
        1,
        0,
        0,
        0,
        899,
        897,
        1,
        0,
        0,
        0,
        899,
        900,
        1,
        0,
        0,
        0,
        900,
        902,
        1,
        0,
        0,
        0,
        901,
        899,
        1,
        0,
        0,
        0,
        902,
        904,
        5,
        0,
        0,
        1,
        903,
        892,
        1,
        0,
        0,
        0,
        903,
        899,
        1,
        0,
        0,
        0,
        904,
        77,
        1,
        0,
        0,
        0,
        905,
        920,
        3,
        82,
        41,
        0,
        906,
        920,
        3,
        256,
        128,
        0,
        907,
        920,
        3,
        272,
        136,
        0,
        908,
        920,
        3,
        88,
        44,
        0,
        909,
        920,
        3,
        102,
        51,
        0,
        910,
        920,
        3,
        90,
        45,
        0,
        911,
        920,
        5,
        38,
        0,
        0,
        912,
        920,
        3,
        96,
        48,
        0,
        913,
        920,
        3,
        98,
        49,
        0,
        914,
        920,
        3,
        100,
        50,
        0,
        915,
        920,
        5,
        14,
        0,
        0,
        916,
        920,
        5,
        17,
        0,
        0,
        917,
        920,
        3,
        92,
        46,
        0,
        918,
        920,
        3,
        94,
        47,
        0,
        919,
        905,
        1,
        0,
        0,
        0,
        919,
        906,
        1,
        0,
        0,
        0,
        919,
        907,
        1,
        0,
        0,
        0,
        919,
        908,
        1,
        0,
        0,
        0,
        919,
        909,
        1,
        0,
        0,
        0,
        919,
        910,
        1,
        0,
        0,
        0,
        919,
        911,
        1,
        0,
        0,
        0,
        919,
        912,
        1,
        0,
        0,
        0,
        919,
        913,
        1,
        0,
        0,
        0,
        919,
        914,
        1,
        0,
        0,
        0,
        919,
        915,
        1,
        0,
        0,
        0,
        919,
        916,
        1,
        0,
        0,
        0,
        919,
        917,
        1,
        0,
        0,
        0,
        919,
        918,
        1,
        0,
        0,
        0,
        920,
        79,
        1,
        0,
        0,
        0,
        921,
        930,
        3,
        128,
        64,
        0,
        922,
        930,
        3,
        162,
        81,
        0,
        923,
        930,
        3,
        124,
        62,
        0,
        924,
        930,
        3,
        172,
        86,
        0,
        925,
        930,
        3,
        170,
        85,
        0,
        926,
        930,
        3,
        176,
        88,
        0,
        927,
        930,
        3,
        168,
        84,
        0,
        928,
        930,
        3,
        184,
        92,
        0,
        929,
        921,
        1,
        0,
        0,
        0,
        929,
        922,
        1,
        0,
        0,
        0,
        929,
        923,
        1,
        0,
        0,
        0,
        929,
        924,
        1,
        0,
        0,
        0,
        929,
        925,
        1,
        0,
        0,
        0,
        929,
        926,
        1,
        0,
        0,
        0,
        929,
        927,
        1,
        0,
        0,
        0,
        929,
        928,
        1,
        0,
        0,
        0,
        930,
        81,
        1,
        0,
        0,
        0,
        931,
        932,
        5,
        52,
        0,
        0,
        932,
        933,
        5,
        71,
        0,
        0,
        933,
        936,
        3,
        268,
        134,
        0,
        934,
        935,
        5,
        74,
        0,
        0,
        935,
        937,
        3,
        84,
        42,
        0,
        936,
        934,
        1,
        0,
        0,
        0,
        936,
        937,
        1,
        0,
        0,
        0,
        937,
        969,
        1,
        0,
        0,
        0,
        938,
        939,
        5,
        68,
        0,
        0,
        939,
        940,
        3,
        432,
        216,
        0,
        940,
        941,
        5,
        69,
        0,
        0,
        941,
        944,
        1,
        0,
        0,
        0,
        942,
        944,
        3,
        434,
        217,
        0,
        943,
        938,
        1,
        0,
        0,
        0,
        943,
        942,
        1,
        0,
        0,
        0,
        944,
        945,
        1,
        0,
        0,
        0,
        945,
        946,
        5,
        71,
        0,
        0,
        946,
        949,
        3,
        268,
        134,
        0,
        947,
        948,
        5,
        74,
        0,
        0,
        948,
        950,
        3,
        84,
        42,
        0,
        949,
        947,
        1,
        0,
        0,
        0,
        949,
        950,
        1,
        0,
        0,
        0,
        950,
        969,
        1,
        0,
        0,
        0,
        951,
        952,
        3,
        420,
        210,
        0,
        952,
        953,
        5,
        74,
        0,
        0,
        953,
        955,
        1,
        0,
        0,
        0,
        954,
        951,
        1,
        0,
        0,
        0,
        955,
        956,
        1,
        0,
        0,
        0,
        956,
        954,
        1,
        0,
        0,
        0,
        956,
        957,
        1,
        0,
        0,
        0,
        957,
        960,
        1,
        0,
        0,
        0,
        958,
        961,
        3,
        270,
        135,
        0,
        959,
        961,
        3,
        272,
        136,
        0,
        960,
        958,
        1,
        0,
        0,
        0,
        960,
        959,
        1,
        0,
        0,
        0,
        961,
        969,
        1,
        0,
        0,
        0,
        962,
        963,
        3,
        432,
        216,
        0,
        963,
        966,
        3,
        86,
        43,
        0,
        964,
        967,
        3,
        270,
        135,
        0,
        965,
        967,
        3,
        272,
        136,
        0,
        966,
        964,
        1,
        0,
        0,
        0,
        966,
        965,
        1,
        0,
        0,
        0,
        967,
        969,
        1,
        0,
        0,
        0,
        968,
        931,
        1,
        0,
        0,
        0,
        968,
        943,
        1,
        0,
        0,
        0,
        968,
        954,
        1,
        0,
        0,
        0,
        968,
        962,
        1,
        0,
        0,
        0,
        969,
        83,
        1,
        0,
        0,
        0,
        970,
        973,
        3,
        270,
        135,
        0,
        971,
        973,
        3,
        272,
        136,
        0,
        972,
        970,
        1,
        0,
        0,
        0,
        972,
        971,
        1,
        0,
        0,
        0,
        973,
        85,
        1,
        0,
        0,
        0,
        974,
        975,
        7,
        2,
        0,
        0,
        975,
        87,
        1,
        0,
        0,
        0,
        976,
        978,
        5,
        40,
        0,
        0,
        977,
        979,
        3,
        272,
        136,
        0,
        978,
        977,
        1,
        0,
        0,
        0,
        978,
        979,
        1,
        0,
        0,
        0,
        979,
        89,
        1,
        0,
        0,
        0,
        980,
        981,
        5,
        39,
        0,
        0,
        981,
        984,
        3,
        268,
        134,
        0,
        982,
        983,
        5,
        26,
        0,
        0,
        983,
        985,
        3,
        268,
        134,
        0,
        984,
        982,
        1,
        0,
        0,
        0,
        984,
        985,
        1,
        0,
        0,
        0,
        985,
        988,
        1,
        0,
        0,
        0,
        986,
        988,
        5,
        39,
        0,
        0,
        987,
        980,
        1,
        0,
        0,
        0,
        987,
        986,
        1,
        0,
        0,
        0,
        988,
        91,
        1,
        0,
        0,
        0,
        989,
        990,
        5,
        27,
        0,
        0,
        990,
        995,
        5,
        52,
        0,
        0,
        991,
        992,
        5,
        70,
        0,
        0,
        992,
        994,
        5,
        52,
        0,
        0,
        993,
        991,
        1,
        0,
        0,
        0,
        994,
        997,
        1,
        0,
        0,
        0,
        995,
        993,
        1,
        0,
        0,
        0,
        995,
        996,
        1,
        0,
        0,
        0,
        996,
        93,
        1,
        0,
        0,
        0,
        997,
        995,
        1,
        0,
        0,
        0,
        998,
        999,
        5,
        35,
        0,
        0,
        999,
        1004,
        5,
        52,
        0,
        0,
        1000,
        1001,
        5,
        70,
        0,
        0,
        1001,
        1003,
        5,
        52,
        0,
        0,
        1002,
        1000,
        1,
        0,
        0,
        0,
        1003,
        1006,
        1,
        0,
        0,
        0,
        1004,
        1002,
        1,
        0,
        0,
        0,
        1004,
        1005,
        1,
        0,
        0,
        0,
        1005,
        95,
        1,
        0,
        0,
        0,
        1006,
        1004,
        1,
        0,
        0,
        0,
        1007,
        1008,
        5,
        19,
        0,
        0,
        1008,
        1009,
        3,
        438,
        219,
        0,
        1009,
        97,
        1,
        0,
        0,
        0,
        1010,
        1011,
        3,
        270,
        135,
        0,
        1011,
        99,
        1,
        0,
        0,
        0,
        1012,
        1013,
        5,
        11,
        0,
        0,
        1013,
        1016,
        3,
        268,
        134,
        0,
        1014,
        1015,
        5,
        70,
        0,
        0,
        1015,
        1017,
        3,
        268,
        134,
        0,
        1016,
        1014,
        1,
        0,
        0,
        0,
        1016,
        1017,
        1,
        0,
        0,
        0,
        1017,
        101,
        1,
        0,
        0,
        0,
        1018,
        1021,
        3,
        104,
        52,
        0,
        1019,
        1021,
        3,
        106,
        53,
        0,
        1020,
        1018,
        1,
        0,
        0,
        0,
        1020,
        1019,
        1,
        0,
        0,
        0,
        1021,
        103,
        1,
        0,
        0,
        0,
        1022,
        1023,
        5,
        29,
        0,
        0,
        1023,
        1024,
        3,
        114,
        57,
        0,
        1024,
        105,
        1,
        0,
        0,
        0,
        1025,
        1029,
        5,
        26,
        0,
        0,
        1026,
        1028,
        7,
        3,
        0,
        0,
        1027,
        1026,
        1,
        0,
        0,
        0,
        1028,
        1031,
        1,
        0,
        0,
        0,
        1029,
        1027,
        1,
        0,
        0,
        0,
        1029,
        1030,
        1,
        0,
        0,
        0,
        1030,
        1032,
        1,
        0,
        0,
        0,
        1031,
        1029,
        1,
        0,
        0,
        0,
        1032,
        1033,
        3,
        118,
        59,
        0,
        1033,
        1034,
        5,
        29,
        0,
        0,
        1034,
        1035,
        3,
        108,
        54,
        0,
        1035,
        1045,
        1,
        0,
        0,
        0,
        1036,
        1038,
        5,
        26,
        0,
        0,
        1037,
        1039,
        7,
        3,
        0,
        0,
        1038,
        1037,
        1,
        0,
        0,
        0,
        1039,
        1040,
        1,
        0,
        0,
        0,
        1040,
        1038,
        1,
        0,
        0,
        0,
        1040,
        1041,
        1,
        0,
        0,
        0,
        1041,
        1042,
        1,
        0,
        0,
        0,
        1042,
        1043,
        5,
        29,
        0,
        0,
        1043,
        1045,
        3,
        108,
        54,
        0,
        1044,
        1025,
        1,
        0,
        0,
        0,
        1044,
        1036,
        1,
        0,
        0,
        0,
        1045,
        107,
        1,
        0,
        0,
        0,
        1046,
        1047,
        5,
        68,
        0,
        0,
        1047,
        1049,
        3,
        110,
        55,
        0,
        1048,
        1050,
        5,
        70,
        0,
        0,
        1049,
        1048,
        1,
        0,
        0,
        0,
        1049,
        1050,
        1,
        0,
        0,
        0,
        1050,
        1051,
        1,
        0,
        0,
        0,
        1051,
        1052,
        5,
        69,
        0,
        0,
        1052,
        1056,
        1,
        0,
        0,
        0,
        1053,
        1056,
        3,
        110,
        55,
        0,
        1054,
        1056,
        5,
        67,
        0,
        0,
        1055,
        1046,
        1,
        0,
        0,
        0,
        1055,
        1053,
        1,
        0,
        0,
        0,
        1055,
        1054,
        1,
        0,
        0,
        0,
        1056,
        109,
        1,
        0,
        0,
        0,
        1057,
        1062,
        3,
        112,
        56,
        0,
        1058,
        1059,
        5,
        70,
        0,
        0,
        1059,
        1061,
        3,
        112,
        56,
        0,
        1060,
        1058,
        1,
        0,
        0,
        0,
        1061,
        1064,
        1,
        0,
        0,
        0,
        1062,
        1060,
        1,
        0,
        0,
        0,
        1062,
        1063,
        1,
        0,
        0,
        0,
        1063,
        111,
        1,
        0,
        0,
        0,
        1064,
        1062,
        1,
        0,
        0,
        0,
        1065,
        1068,
        5,
        52,
        0,
        0,
        1066,
        1067,
        5,
        10,
        0,
        0,
        1067,
        1069,
        5,
        52,
        0,
        0,
        1068,
        1066,
        1,
        0,
        0,
        0,
        1068,
        1069,
        1,
        0,
        0,
        0,
        1069,
        113,
        1,
        0,
        0,
        0,
        1070,
        1075,
        3,
        116,
        58,
        0,
        1071,
        1072,
        5,
        70,
        0,
        0,
        1072,
        1074,
        3,
        116,
        58,
        0,
        1073,
        1071,
        1,
        0,
        0,
        0,
        1074,
        1077,
        1,
        0,
        0,
        0,
        1075,
        1073,
        1,
        0,
        0,
        0,
        1075,
        1076,
        1,
        0,
        0,
        0,
        1076,
        115,
        1,
        0,
        0,
        0,
        1077,
        1075,
        1,
        0,
        0,
        0,
        1078,
        1081,
        3,
        118,
        59,
        0,
        1079,
        1080,
        5,
        10,
        0,
        0,
        1080,
        1082,
        5,
        52,
        0,
        0,
        1081,
        1079,
        1,
        0,
        0,
        0,
        1081,
        1082,
        1,
        0,
        0,
        0,
        1082,
        117,
        1,
        0,
        0,
        0,
        1083,
        1084,
        6,
        59,
        -1,
        0,
        1084,
        1085,
        5,
        52,
        0,
        0,
        1085,
        1091,
        1,
        0,
        0,
        0,
        1086,
        1087,
        10,
        2,
        0,
        0,
        1087,
        1088,
        5,
        64,
        0,
        0,
        1088,
        1090,
        5,
        52,
        0,
        0,
        1089,
        1086,
        1,
        0,
        0,
        0,
        1090,
        1093,
        1,
        0,
        0,
        0,
        1091,
        1089,
        1,
        0,
        0,
        0,
        1091,
        1092,
        1,
        0,
        0,
        0,
        1092,
        119,
        1,
        0,
        0,
        0,
        1093,
        1091,
        1,
        0,
        0,
        0,
        1094,
        1095,
        5,
        114,
        0,
        0,
        1095,
        1096,
        5,
        1,
        0,
        0,
        1096,
        1097,
        3,
        70,
        35,
        0,
        1097,
        1098,
        5,
        2,
        0,
        0,
        1098,
        1101,
        1,
        0,
        0,
        0,
        1099,
        1101,
        3,
        76,
        38,
        0,
        1100,
        1094,
        1,
        0,
        0,
        0,
        1100,
        1099,
        1,
        0,
        0,
        0,
        1101,
        121,
        1,
        0,
        0,
        0,
        1102,
        1103,
        5,
        97,
        0,
        0,
        1103,
        1104,
        3,
        282,
        141,
        0,
        1104,
        1105,
        5,
        114,
        0,
        0,
        1105,
        1107,
        1,
        0,
        0,
        0,
        1106,
        1102,
        1,
        0,
        0,
        0,
        1107,
        1108,
        1,
        0,
        0,
        0,
        1108,
        1106,
        1,
        0,
        0,
        0,
        1108,
        1109,
        1,
        0,
        0,
        0,
        1109,
        123,
        1,
        0,
        0,
        0,
        1110,
        1112,
        3,
        122,
        61,
        0,
        1111,
        1110,
        1,
        0,
        0,
        0,
        1111,
        1112,
        1,
        0,
        0,
        0,
        1112,
        1113,
        1,
        0,
        0,
        0,
        1113,
        1114,
        3,
        126,
        63,
        0,
        1114,
        125,
        1,
        0,
        0,
        0,
        1115,
        1116,
        5,
        16,
        0,
        0,
        1116,
        1118,
        5,
        52,
        0,
        0,
        1117,
        1119,
        3,
        258,
        129,
        0,
        1118,
        1117,
        1,
        0,
        0,
        0,
        1118,
        1119,
        1,
        0,
        0,
        0,
        1119,
        1125,
        1,
        0,
        0,
        0,
        1120,
        1122,
        5,
        68,
        0,
        0,
        1121,
        1123,
        3,
        406,
        203,
        0,
        1122,
        1121,
        1,
        0,
        0,
        0,
        1122,
        1123,
        1,
        0,
        0,
        0,
        1123,
        1124,
        1,
        0,
        0,
        0,
        1124,
        1126,
        5,
        69,
        0,
        0,
        1125,
        1120,
        1,
        0,
        0,
        0,
        1125,
        1126,
        1,
        0,
        0,
        0,
        1126,
        1127,
        1,
        0,
        0,
        0,
        1127,
        1128,
        5,
        71,
        0,
        0,
        1128,
        1129,
        3,
        120,
        60,
        0,
        1129,
        127,
        1,
        0,
        0,
        0,
        1130,
        1132,
        3,
        122,
        61,
        0,
        1131,
        1130,
        1,
        0,
        0,
        0,
        1131,
        1132,
        1,
        0,
        0,
        0,
        1132,
        1133,
        1,
        0,
        0,
        0,
        1133,
        1134,
        3,
        130,
        65,
        0,
        1134,
        129,
        1,
        0,
        0,
        0,
        1135,
        1137,
        5,
        12,
        0,
        0,
        1136,
        1135,
        1,
        0,
        0,
        0,
        1136,
        1137,
        1,
        0,
        0,
        0,
        1137,
        1138,
        1,
        0,
        0,
        0,
        1138,
        1139,
        5,
        18,
        0,
        0,
        1139,
        1141,
        5,
        52,
        0,
        0,
        1140,
        1142,
        3,
        258,
        129,
        0,
        1141,
        1140,
        1,
        0,
        0,
        0,
        1141,
        1142,
        1,
        0,
        0,
        0,
        1142,
        1143,
        1,
        0,
        0,
        0,
        1143,
        1145,
        5,
        68,
        0,
        0,
        1144,
        1146,
        3,
        132,
        66,
        0,
        1145,
        1144,
        1,
        0,
        0,
        0,
        1145,
        1146,
        1,
        0,
        0,
        0,
        1146,
        1147,
        1,
        0,
        0,
        0,
        1147,
        1150,
        5,
        69,
        0,
        0,
        1148,
        1149,
        5,
        98,
        0,
        0,
        1149,
        1151,
        3,
        268,
        134,
        0,
        1150,
        1148,
        1,
        0,
        0,
        0,
        1150,
        1151,
        1,
        0,
        0,
        0,
        1151,
        1152,
        1,
        0,
        0,
        0,
        1152,
        1154,
        5,
        71,
        0,
        0,
        1153,
        1155,
        3,
        446,
        223,
        0,
        1154,
        1153,
        1,
        0,
        0,
        0,
        1154,
        1155,
        1,
        0,
        0,
        0,
        1155,
        1156,
        1,
        0,
        0,
        0,
        1156,
        1157,
        3,
        120,
        60,
        0,
        1157,
        131,
        1,
        0,
        0,
        0,
        1158,
        1159,
        3,
        134,
        67,
        0,
        1159,
        133,
        1,
        0,
        0,
        0,
        1160,
        1164,
        3,
        136,
        68,
        0,
        1161,
        1163,
        3,
        144,
        72,
        0,
        1162,
        1161,
        1,
        0,
        0,
        0,
        1163,
        1166,
        1,
        0,
        0,
        0,
        1164,
        1162,
        1,
        0,
        0,
        0,
        1164,
        1165,
        1,
        0,
        0,
        0,
        1165,
        1170,
        1,
        0,
        0,
        0,
        1166,
        1164,
        1,
        0,
        0,
        0,
        1167,
        1169,
        3,
        148,
        74,
        0,
        1168,
        1167,
        1,
        0,
        0,
        0,
        1169,
        1172,
        1,
        0,
        0,
        0,
        1170,
        1168,
        1,
        0,
        0,
        0,
        1170,
        1171,
        1,
        0,
        0,
        0,
        1171,
        1174,
        1,
        0,
        0,
        0,
        1172,
        1170,
        1,
        0,
        0,
        0,
        1173,
        1175,
        3,
        140,
        70,
        0,
        1174,
        1173,
        1,
        0,
        0,
        0,
        1174,
        1175,
        1,
        0,
        0,
        0,
        1175,
        1210,
        1,
        0,
        0,
        0,
        1176,
        1180,
        3,
        138,
        69,
        0,
        1177,
        1179,
        3,
        148,
        74,
        0,
        1178,
        1177,
        1,
        0,
        0,
        0,
        1179,
        1182,
        1,
        0,
        0,
        0,
        1180,
        1178,
        1,
        0,
        0,
        0,
        1180,
        1181,
        1,
        0,
        0,
        0,
        1181,
        1184,
        1,
        0,
        0,
        0,
        1182,
        1180,
        1,
        0,
        0,
        0,
        1183,
        1185,
        3,
        140,
        70,
        0,
        1184,
        1183,
        1,
        0,
        0,
        0,
        1184,
        1185,
        1,
        0,
        0,
        0,
        1185,
        1210,
        1,
        0,
        0,
        0,
        1186,
        1188,
        3,
        144,
        72,
        0,
        1187,
        1186,
        1,
        0,
        0,
        0,
        1188,
        1189,
        1,
        0,
        0,
        0,
        1189,
        1187,
        1,
        0,
        0,
        0,
        1189,
        1190,
        1,
        0,
        0,
        0,
        1190,
        1194,
        1,
        0,
        0,
        0,
        1191,
        1193,
        3,
        148,
        74,
        0,
        1192,
        1191,
        1,
        0,
        0,
        0,
        1193,
        1196,
        1,
        0,
        0,
        0,
        1194,
        1192,
        1,
        0,
        0,
        0,
        1194,
        1195,
        1,
        0,
        0,
        0,
        1195,
        1198,
        1,
        0,
        0,
        0,
        1196,
        1194,
        1,
        0,
        0,
        0,
        1197,
        1199,
        3,
        140,
        70,
        0,
        1198,
        1197,
        1,
        0,
        0,
        0,
        1198,
        1199,
        1,
        0,
        0,
        0,
        1199,
        1210,
        1,
        0,
        0,
        0,
        1200,
        1202,
        3,
        148,
        74,
        0,
        1201,
        1200,
        1,
        0,
        0,
        0,
        1202,
        1203,
        1,
        0,
        0,
        0,
        1203,
        1201,
        1,
        0,
        0,
        0,
        1203,
        1204,
        1,
        0,
        0,
        0,
        1204,
        1206,
        1,
        0,
        0,
        0,
        1205,
        1207,
        3,
        140,
        70,
        0,
        1206,
        1205,
        1,
        0,
        0,
        0,
        1206,
        1207,
        1,
        0,
        0,
        0,
        1207,
        1210,
        1,
        0,
        0,
        0,
        1208,
        1210,
        3,
        140,
        70,
        0,
        1209,
        1160,
        1,
        0,
        0,
        0,
        1209,
        1176,
        1,
        0,
        0,
        0,
        1209,
        1187,
        1,
        0,
        0,
        0,
        1209,
        1201,
        1,
        0,
        0,
        0,
        1209,
        1208,
        1,
        0,
        0,
        0,
        1210,
        135,
        1,
        0,
        0,
        0,
        1211,
        1213,
        3,
        144,
        72,
        0,
        1212,
        1211,
        1,
        0,
        0,
        0,
        1213,
        1214,
        1,
        0,
        0,
        0,
        1214,
        1212,
        1,
        0,
        0,
        0,
        1214,
        1215,
        1,
        0,
        0,
        0,
        1215,
        1216,
        1,
        0,
        0,
        0,
        1216,
        1218,
        5,
        84,
        0,
        0,
        1217,
        1219,
        5,
        70,
        0,
        0,
        1218,
        1217,
        1,
        0,
        0,
        0,
        1218,
        1219,
        1,
        0,
        0,
        0,
        1219,
        137,
        1,
        0,
        0,
        0,
        1220,
        1222,
        3,
        144,
        72,
        0,
        1221,
        1220,
        1,
        0,
        0,
        0,
        1222,
        1225,
        1,
        0,
        0,
        0,
        1223,
        1221,
        1,
        0,
        0,
        0,
        1223,
        1224,
        1,
        0,
        0,
        0,
        1224,
        1227,
        1,
        0,
        0,
        0,
        1225,
        1223,
        1,
        0,
        0,
        0,
        1226,
        1228,
        3,
        148,
        74,
        0,
        1227,
        1226,
        1,
        0,
        0,
        0,
        1228,
        1229,
        1,
        0,
        0,
        0,
        1229,
        1227,
        1,
        0,
        0,
        0,
        1229,
        1230,
        1,
        0,
        0,
        0,
        1230,
        1231,
        1,
        0,
        0,
        0,
        1231,
        1233,
        5,
        84,
        0,
        0,
        1232,
        1234,
        5,
        70,
        0,
        0,
        1233,
        1232,
        1,
        0,
        0,
        0,
        1233,
        1234,
        1,
        0,
        0,
        0,
        1234,
        139,
        1,
        0,
        0,
        0,
        1235,
        1236,
        5,
        67,
        0,
        0,
        1236,
        1240,
        3,
        144,
        72,
        0,
        1237,
        1239,
        3,
        150,
        75,
        0,
        1238,
        1237,
        1,
        0,
        0,
        0,
        1239,
        1242,
        1,
        0,
        0,
        0,
        1240,
        1238,
        1,
        0,
        0,
        0,
        1240,
        1241,
        1,
        0,
        0,
        0,
        1241,
        1244,
        1,
        0,
        0,
        0,
        1242,
        1240,
        1,
        0,
        0,
        0,
        1243,
        1245,
        3,
        142,
        71,
        0,
        1244,
        1243,
        1,
        0,
        0,
        0,
        1244,
        1245,
        1,
        0,
        0,
        0,
        1245,
        1269,
        1,
        0,
        0,
        0,
        1246,
        1247,
        5,
        67,
        0,
        0,
        1247,
        1251,
        3,
        146,
        73,
        0,
        1248,
        1250,
        3,
        150,
        75,
        0,
        1249,
        1248,
        1,
        0,
        0,
        0,
        1250,
        1253,
        1,
        0,
        0,
        0,
        1251,
        1249,
        1,
        0,
        0,
        0,
        1251,
        1252,
        1,
        0,
        0,
        0,
        1252,
        1255,
        1,
        0,
        0,
        0,
        1253,
        1251,
        1,
        0,
        0,
        0,
        1254,
        1256,
        3,
        142,
        71,
        0,
        1255,
        1254,
        1,
        0,
        0,
        0,
        1255,
        1256,
        1,
        0,
        0,
        0,
        1256,
        1269,
        1,
        0,
        0,
        0,
        1257,
        1258,
        5,
        67,
        0,
        0,
        1258,
        1260,
        5,
        70,
        0,
        0,
        1259,
        1261,
        3,
        150,
        75,
        0,
        1260,
        1259,
        1,
        0,
        0,
        0,
        1261,
        1262,
        1,
        0,
        0,
        0,
        1262,
        1260,
        1,
        0,
        0,
        0,
        1262,
        1263,
        1,
        0,
        0,
        0,
        1263,
        1265,
        1,
        0,
        0,
        0,
        1264,
        1266,
        3,
        142,
        71,
        0,
        1265,
        1264,
        1,
        0,
        0,
        0,
        1265,
        1266,
        1,
        0,
        0,
        0,
        1266,
        1269,
        1,
        0,
        0,
        0,
        1267,
        1269,
        3,
        142,
        71,
        0,
        1268,
        1235,
        1,
        0,
        0,
        0,
        1268,
        1246,
        1,
        0,
        0,
        0,
        1268,
        1257,
        1,
        0,
        0,
        0,
        1268,
        1267,
        1,
        0,
        0,
        0,
        1269,
        141,
        1,
        0,
        0,
        0,
        1270,
        1271,
        5,
        73,
        0,
        0,
        1271,
        1272,
        3,
        144,
        72,
        0,
        1272,
        143,
        1,
        0,
        0,
        0,
        1273,
        1274,
        3,
        152,
        76,
        0,
        1274,
        1275,
        5,
        70,
        0,
        0,
        1275,
        1278,
        1,
        0,
        0,
        0,
        1276,
        1278,
        3,
        152,
        76,
        0,
        1277,
        1273,
        1,
        0,
        0,
        0,
        1277,
        1276,
        1,
        0,
        0,
        0,
        1278,
        145,
        1,
        0,
        0,
        0,
        1279,
        1280,
        3,
        154,
        77,
        0,
        1280,
        1281,
        5,
        70,
        0,
        0,
        1281,
        1284,
        1,
        0,
        0,
        0,
        1282,
        1284,
        3,
        154,
        77,
        0,
        1283,
        1279,
        1,
        0,
        0,
        0,
        1283,
        1282,
        1,
        0,
        0,
        0,
        1284,
        147,
        1,
        0,
        0,
        0,
        1285,
        1286,
        3,
        152,
        76,
        0,
        1286,
        1287,
        3,
        160,
        80,
        0,
        1287,
        1288,
        5,
        70,
        0,
        0,
        1288,
        1293,
        1,
        0,
        0,
        0,
        1289,
        1290,
        3,
        152,
        76,
        0,
        1290,
        1291,
        3,
        160,
        80,
        0,
        1291,
        1293,
        1,
        0,
        0,
        0,
        1292,
        1285,
        1,
        0,
        0,
        0,
        1292,
        1289,
        1,
        0,
        0,
        0,
        1293,
        149,
        1,
        0,
        0,
        0,
        1294,
        1296,
        3,
        152,
        76,
        0,
        1295,
        1297,
        3,
        160,
        80,
        0,
        1296,
        1295,
        1,
        0,
        0,
        0,
        1296,
        1297,
        1,
        0,
        0,
        0,
        1297,
        1298,
        1,
        0,
        0,
        0,
        1298,
        1299,
        5,
        70,
        0,
        0,
        1299,
        1305,
        1,
        0,
        0,
        0,
        1300,
        1302,
        3,
        152,
        76,
        0,
        1301,
        1303,
        3,
        160,
        80,
        0,
        1302,
        1301,
        1,
        0,
        0,
        0,
        1302,
        1303,
        1,
        0,
        0,
        0,
        1303,
        1305,
        1,
        0,
        0,
        0,
        1304,
        1294,
        1,
        0,
        0,
        0,
        1304,
        1300,
        1,
        0,
        0,
        0,
        1305,
        151,
        1,
        0,
        0,
        0,
        1306,
        1308,
        5,
        52,
        0,
        0,
        1307,
        1309,
        3,
        156,
        78,
        0,
        1308,
        1307,
        1,
        0,
        0,
        0,
        1308,
        1309,
        1,
        0,
        0,
        0,
        1309,
        153,
        1,
        0,
        0,
        0,
        1310,
        1311,
        5,
        52,
        0,
        0,
        1311,
        1312,
        3,
        158,
        79,
        0,
        1312,
        155,
        1,
        0,
        0,
        0,
        1313,
        1314,
        5,
        71,
        0,
        0,
        1314,
        1315,
        3,
        268,
        134,
        0,
        1315,
        157,
        1,
        0,
        0,
        0,
        1316,
        1317,
        5,
        71,
        0,
        0,
        1317,
        1318,
        3,
        274,
        137,
        0,
        1318,
        159,
        1,
        0,
        0,
        0,
        1319,
        1320,
        5,
        74,
        0,
        0,
        1320,
        1321,
        3,
        268,
        134,
        0,
        1321,
        161,
        1,
        0,
        0,
        0,
        1322,
        1323,
        5,
        28,
        0,
        0,
        1323,
        1324,
        3,
        282,
        141,
        0,
        1324,
        1325,
        5,
        71,
        0,
        0,
        1325,
        1326,
        3,
        120,
        60,
        0,
        1326,
        1327,
        3,
        164,
        82,
        0,
        1327,
        1336,
        1,
        0,
        0,
        0,
        1328,
        1329,
        5,
        28,
        0,
        0,
        1329,
        1330,
        3,
        282,
        141,
        0,
        1330,
        1331,
        5,
        71,
        0,
        0,
        1331,
        1333,
        3,
        120,
        60,
        0,
        1332,
        1334,
        3,
        166,
        83,
        0,
        1333,
        1332,
        1,
        0,
        0,
        0,
        1333,
        1334,
        1,
        0,
        0,
        0,
        1334,
        1336,
        1,
        0,
        0,
        0,
        1335,
        1322,
        1,
        0,
        0,
        0,
        1335,
        1328,
        1,
        0,
        0,
        0,
        1336,
        163,
        1,
        0,
        0,
        0,
        1337,
        1338,
        5,
        20,
        0,
        0,
        1338,
        1339,
        3,
        282,
        141,
        0,
        1339,
        1340,
        5,
        71,
        0,
        0,
        1340,
        1341,
        3,
        120,
        60,
        0,
        1341,
        1342,
        3,
        164,
        82,
        0,
        1342,
        1351,
        1,
        0,
        0,
        0,
        1343,
        1344,
        5,
        20,
        0,
        0,
        1344,
        1345,
        3,
        282,
        141,
        0,
        1345,
        1346,
        5,
        71,
        0,
        0,
        1346,
        1348,
        3,
        120,
        60,
        0,
        1347,
        1349,
        3,
        166,
        83,
        0,
        1348,
        1347,
        1,
        0,
        0,
        0,
        1348,
        1349,
        1,
        0,
        0,
        0,
        1349,
        1351,
        1,
        0,
        0,
        0,
        1350,
        1337,
        1,
        0,
        0,
        0,
        1350,
        1343,
        1,
        0,
        0,
        0,
        1351,
        165,
        1,
        0,
        0,
        0,
        1352,
        1353,
        5,
        21,
        0,
        0,
        1353,
        1354,
        5,
        71,
        0,
        0,
        1354,
        1355,
        3,
        120,
        60,
        0,
        1355,
        167,
        1,
        0,
        0,
        0,
        1356,
        1357,
        5,
        45,
        0,
        0,
        1357,
        1358,
        3,
        282,
        141,
        0,
        1358,
        1359,
        5,
        71,
        0,
        0,
        1359,
        1361,
        3,
        120,
        60,
        0,
        1360,
        1362,
        3,
        166,
        83,
        0,
        1361,
        1360,
        1,
        0,
        0,
        0,
        1361,
        1362,
        1,
        0,
        0,
        0,
        1362,
        169,
        1,
        0,
        0,
        0,
        1363,
        1364,
        5,
        25,
        0,
        0,
        1364,
        1365,
        3,
        420,
        210,
        0,
        1365,
        1366,
        5,
        30,
        0,
        0,
        1366,
        1367,
        3,
        272,
        136,
        0,
        1367,
        1368,
        5,
        71,
        0,
        0,
        1368,
        1370,
        3,
        120,
        60,
        0,
        1369,
        1371,
        3,
        166,
        83,
        0,
        1370,
        1369,
        1,
        0,
        0,
        0,
        1370,
        1371,
        1,
        0,
        0,
        0,
        1371,
        1383,
        1,
        0,
        0,
        0,
        1372,
        1373,
        5,
        12,
        0,
        0,
        1373,
        1374,
        5,
        25,
        0,
        0,
        1374,
        1375,
        3,
        420,
        210,
        0,
        1375,
        1376,
        5,
        30,
        0,
        0,
        1376,
        1377,
        3,
        272,
        136,
        0,
        1377,
        1378,
        5,
        71,
        0,
        0,
        1378,
        1380,
        3,
        120,
        60,
        0,
        1379,
        1381,
        3,
        166,
        83,
        0,
        1380,
        1379,
        1,
        0,
        0,
        0,
        1380,
        1381,
        1,
        0,
        0,
        0,
        1381,
        1383,
        1,
        0,
        0,
        0,
        1382,
        1363,
        1,
        0,
        0,
        0,
        1382,
        1372,
        1,
        0,
        0,
        0,
        1383,
        171,
        1,
        0,
        0,
        0,
        1384,
        1385,
        5,
        47,
        0,
        0,
        1385,
        1386,
        5,
        68,
        0,
        0,
        1386,
        1391,
        3,
        174,
        87,
        0,
        1387,
        1388,
        5,
        70,
        0,
        0,
        1388,
        1390,
        3,
        174,
        87,
        0,
        1389,
        1387,
        1,
        0,
        0,
        0,
        1390,
        1393,
        1,
        0,
        0,
        0,
        1391,
        1389,
        1,
        0,
        0,
        0,
        1391,
        1392,
        1,
        0,
        0,
        0,
        1392,
        1395,
        1,
        0,
        0,
        0,
        1393,
        1391,
        1,
        0,
        0,
        0,
        1394,
        1396,
        5,
        70,
        0,
        0,
        1395,
        1394,
        1,
        0,
        0,
        0,
        1395,
        1396,
        1,
        0,
        0,
        0,
        1396,
        1397,
        1,
        0,
        0,
        0,
        1397,
        1398,
        5,
        69,
        0,
        0,
        1398,
        1399,
        5,
        71,
        0,
        0,
        1399,
        1400,
        3,
        120,
        60,
        0,
        1400,
        1445,
        1,
        0,
        0,
        0,
        1401,
        1402,
        5,
        47,
        0,
        0,
        1402,
        1407,
        3,
        174,
        87,
        0,
        1403,
        1404,
        5,
        70,
        0,
        0,
        1404,
        1406,
        3,
        174,
        87,
        0,
        1405,
        1403,
        1,
        0,
        0,
        0,
        1406,
        1409,
        1,
        0,
        0,
        0,
        1407,
        1405,
        1,
        0,
        0,
        0,
        1407,
        1408,
        1,
        0,
        0,
        0,
        1408,
        1410,
        1,
        0,
        0,
        0,
        1409,
        1407,
        1,
        0,
        0,
        0,
        1410,
        1411,
        5,
        71,
        0,
        0,
        1411,
        1412,
        3,
        120,
        60,
        0,
        1412,
        1445,
        1,
        0,
        0,
        0,
        1413,
        1414,
        5,
        12,
        0,
        0,
        1414,
        1415,
        5,
        47,
        0,
        0,
        1415,
        1416,
        5,
        68,
        0,
        0,
        1416,
        1421,
        3,
        174,
        87,
        0,
        1417,
        1418,
        5,
        70,
        0,
        0,
        1418,
        1420,
        3,
        174,
        87,
        0,
        1419,
        1417,
        1,
        0,
        0,
        0,
        1420,
        1423,
        1,
        0,
        0,
        0,
        1421,
        1419,
        1,
        0,
        0,
        0,
        1421,
        1422,
        1,
        0,
        0,
        0,
        1422,
        1425,
        1,
        0,
        0,
        0,
        1423,
        1421,
        1,
        0,
        0,
        0,
        1424,
        1426,
        5,
        70,
        0,
        0,
        1425,
        1424,
        1,
        0,
        0,
        0,
        1425,
        1426,
        1,
        0,
        0,
        0,
        1426,
        1427,
        1,
        0,
        0,
        0,
        1427,
        1428,
        5,
        69,
        0,
        0,
        1428,
        1429,
        5,
        71,
        0,
        0,
        1429,
        1430,
        3,
        120,
        60,
        0,
        1430,
        1445,
        1,
        0,
        0,
        0,
        1431,
        1432,
        5,
        12,
        0,
        0,
        1432,
        1433,
        5,
        47,
        0,
        0,
        1433,
        1438,
        3,
        174,
        87,
        0,
        1434,
        1435,
        5,
        70,
        0,
        0,
        1435,
        1437,
        3,
        174,
        87,
        0,
        1436,
        1434,
        1,
        0,
        0,
        0,
        1437,
        1440,
        1,
        0,
        0,
        0,
        1438,
        1436,
        1,
        0,
        0,
        0,
        1438,
        1439,
        1,
        0,
        0,
        0,
        1439,
        1441,
        1,
        0,
        0,
        0,
        1440,
        1438,
        1,
        0,
        0,
        0,
        1441,
        1442,
        5,
        71,
        0,
        0,
        1442,
        1443,
        3,
        120,
        60,
        0,
        1443,
        1445,
        1,
        0,
        0,
        0,
        1444,
        1384,
        1,
        0,
        0,
        0,
        1444,
        1401,
        1,
        0,
        0,
        0,
        1444,
        1413,
        1,
        0,
        0,
        0,
        1444,
        1431,
        1,
        0,
        0,
        0,
        1445,
        173,
        1,
        0,
        0,
        0,
        1446,
        1447,
        3,
        268,
        134,
        0,
        1447,
        1448,
        5,
        10,
        0,
        0,
        1448,
        1449,
        3,
        426,
        213,
        0,
        1449,
        1452,
        1,
        0,
        0,
        0,
        1450,
        1452,
        3,
        268,
        134,
        0,
        1451,
        1446,
        1,
        0,
        0,
        0,
        1451,
        1450,
        1,
        0,
        0,
        0,
        1452,
        175,
        1,
        0,
        0,
        0,
        1453,
        1454,
        5,
        42,
        0,
        0,
        1454,
        1455,
        5,
        71,
        0,
        0,
        1455,
        1456,
        3,
        120,
        60,
        0,
        1456,
        1457,
        3,
        182,
        91,
        0,
        1457,
        1487,
        1,
        0,
        0,
        0,
        1458,
        1459,
        5,
        42,
        0,
        0,
        1459,
        1460,
        5,
        71,
        0,
        0,
        1460,
        1462,
        3,
        120,
        60,
        0,
        1461,
        1463,
        3,
        178,
        89,
        0,
        1462,
        1461,
        1,
        0,
        0,
        0,
        1463,
        1464,
        1,
        0,
        0,
        0,
        1464,
        1462,
        1,
        0,
        0,
        0,
        1464,
        1465,
        1,
        0,
        0,
        0,
        1465,
        1467,
        1,
        0,
        0,
        0,
        1466,
        1468,
        3,
        166,
        83,
        0,
        1467,
        1466,
        1,
        0,
        0,
        0,
        1467,
        1468,
        1,
        0,
        0,
        0,
        1468,
        1470,
        1,
        0,
        0,
        0,
        1469,
        1471,
        3,
        182,
        91,
        0,
        1470,
        1469,
        1,
        0,
        0,
        0,
        1470,
        1471,
        1,
        0,
        0,
        0,
        1471,
        1487,
        1,
        0,
        0,
        0,
        1472,
        1473,
        5,
        42,
        0,
        0,
        1473,
        1474,
        5,
        71,
        0,
        0,
        1474,
        1476,
        3,
        120,
        60,
        0,
        1475,
        1477,
        3,
        180,
        90,
        0,
        1476,
        1475,
        1,
        0,
        0,
        0,
        1477,
        1478,
        1,
        0,
        0,
        0,
        1478,
        1476,
        1,
        0,
        0,
        0,
        1478,
        1479,
        1,
        0,
        0,
        0,
        1479,
        1481,
        1,
        0,
        0,
        0,
        1480,
        1482,
        3,
        166,
        83,
        0,
        1481,
        1480,
        1,
        0,
        0,
        0,
        1481,
        1482,
        1,
        0,
        0,
        0,
        1482,
        1484,
        1,
        0,
        0,
        0,
        1483,
        1485,
        3,
        182,
        91,
        0,
        1484,
        1483,
        1,
        0,
        0,
        0,
        1484,
        1485,
        1,
        0,
        0,
        0,
        1485,
        1487,
        1,
        0,
        0,
        0,
        1486,
        1453,
        1,
        0,
        0,
        0,
        1486,
        1458,
        1,
        0,
        0,
        0,
        1486,
        1472,
        1,
        0,
        0,
        0,
        1487,
        177,
        1,
        0,
        0,
        0,
        1488,
        1489,
        5,
        22,
        0,
        0,
        1489,
        1492,
        3,
        268,
        134,
        0,
        1490,
        1491,
        5,
        10,
        0,
        0,
        1491,
        1493,
        5,
        52,
        0,
        0,
        1492,
        1490,
        1,
        0,
        0,
        0,
        1492,
        1493,
        1,
        0,
        0,
        0,
        1493,
        1494,
        1,
        0,
        0,
        0,
        1494,
        1495,
        5,
        71,
        0,
        0,
        1495,
        1496,
        3,
        120,
        60,
        0,
        1496,
        1501,
        1,
        0,
        0,
        0,
        1497,
        1498,
        5,
        22,
        0,
        0,
        1498,
        1499,
        5,
        71,
        0,
        0,
        1499,
        1501,
        3,
        120,
        60,
        0,
        1500,
        1488,
        1,
        0,
        0,
        0,
        1500,
        1497,
        1,
        0,
        0,
        0,
        1501,
        179,
        1,
        0,
        0,
        0,
        1502,
        1503,
        5,
        22,
        0,
        0,
        1503,
        1504,
        5,
        67,
        0,
        0,
        1504,
        1507,
        3,
        268,
        134,
        0,
        1505,
        1506,
        5,
        10,
        0,
        0,
        1506,
        1508,
        5,
        52,
        0,
        0,
        1507,
        1505,
        1,
        0,
        0,
        0,
        1507,
        1508,
        1,
        0,
        0,
        0,
        1508,
        1509,
        1,
        0,
        0,
        0,
        1509,
        1510,
        5,
        71,
        0,
        0,
        1510,
        1511,
        3,
        120,
        60,
        0,
        1511,
        181,
        1,
        0,
        0,
        0,
        1512,
        1513,
        5,
        24,
        0,
        0,
        1513,
        1514,
        5,
        71,
        0,
        0,
        1514,
        1515,
        3,
        120,
        60,
        0,
        1515,
        183,
        1,
        0,
        0,
        0,
        1516,
        1517,
        5,
        33,
        0,
        0,
        1517,
        1518,
        3,
        186,
        93,
        0,
        1518,
        1519,
        5,
        71,
        0,
        0,
        1519,
        1520,
        5,
        114,
        0,
        0,
        1520,
        1522,
        5,
        1,
        0,
        0,
        1521,
        1523,
        3,
        188,
        94,
        0,
        1522,
        1521,
        1,
        0,
        0,
        0,
        1523,
        1524,
        1,
        0,
        0,
        0,
        1524,
        1522,
        1,
        0,
        0,
        0,
        1524,
        1525,
        1,
        0,
        0,
        0,
        1525,
        1526,
        1,
        0,
        0,
        0,
        1526,
        1527,
        5,
        2,
        0,
        0,
        1527,
        185,
        1,
        0,
        0,
        0,
        1528,
        1529,
        3,
        278,
        139,
        0,
        1529,
        1531,
        5,
        70,
        0,
        0,
        1530,
        1532,
        3,
        276,
        138,
        0,
        1531,
        1530,
        1,
        0,
        0,
        0,
        1531,
        1532,
        1,
        0,
        0,
        0,
        1532,
        1535,
        1,
        0,
        0,
        0,
        1533,
        1535,
        3,
        282,
        141,
        0,
        1534,
        1528,
        1,
        0,
        0,
        0,
        1534,
        1533,
        1,
        0,
        0,
        0,
        1535,
        187,
        1,
        0,
        0,
        0,
        1536,
        1537,
        5,
        15,
        0,
        0,
        1537,
        1539,
        3,
        192,
        96,
        0,
        1538,
        1540,
        3,
        190,
        95,
        0,
        1539,
        1538,
        1,
        0,
        0,
        0,
        1539,
        1540,
        1,
        0,
        0,
        0,
        1540,
        1541,
        1,
        0,
        0,
        0,
        1541,
        1542,
        5,
        71,
        0,
        0,
        1542,
        1543,
        3,
        120,
        60,
        0,
        1543,
        189,
        1,
        0,
        0,
        0,
        1544,
        1545,
        5,
        28,
        0,
        0,
        1545,
        1546,
        3,
        282,
        141,
        0,
        1546,
        191,
        1,
        0,
        0,
        0,
        1547,
        1550,
        3,
        232,
        116,
        0,
        1548,
        1550,
        3,
        194,
        97,
        0,
        1549,
        1547,
        1,
        0,
        0,
        0,
        1549,
        1548,
        1,
        0,
        0,
        0,
        1550,
        193,
        1,
        0,
        0,
        0,
        1551,
        1554,
        3,
        196,
        98,
        0,
        1552,
        1554,
        3,
        198,
        99,
        0,
        1553,
        1551,
        1,
        0,
        0,
        0,
        1553,
        1552,
        1,
        0,
        0,
        0,
        1554,
        195,
        1,
        0,
        0,
        0,
        1555,
        1556,
        3,
        198,
        99,
        0,
        1556,
        1557,
        5,
        10,
        0,
        0,
        1557,
        1558,
        3,
        218,
        109,
        0,
        1558,
        197,
        1,
        0,
        0,
        0,
        1559,
        1564,
        3,
        200,
        100,
        0,
        1560,
        1561,
        5,
        77,
        0,
        0,
        1561,
        1563,
        3,
        200,
        100,
        0,
        1562,
        1560,
        1,
        0,
        0,
        0,
        1563,
        1566,
        1,
        0,
        0,
        0,
        1564,
        1562,
        1,
        0,
        0,
        0,
        1564,
        1565,
        1,
        0,
        0,
        0,
        1565,
        199,
        1,
        0,
        0,
        0,
        1566,
        1564,
        1,
        0,
        0,
        0,
        1567,
        1576,
        3,
        202,
        101,
        0,
        1568,
        1576,
        3,
        216,
        108,
        0,
        1569,
        1576,
        3,
        220,
        110,
        0,
        1570,
        1576,
        3,
        222,
        111,
        0,
        1571,
        1576,
        3,
        228,
        114,
        0,
        1572,
        1576,
        3,
        230,
        115,
        0,
        1573,
        1576,
        3,
        240,
        120,
        0,
        1574,
        1576,
        3,
        248,
        124,
        0,
        1575,
        1567,
        1,
        0,
        0,
        0,
        1575,
        1568,
        1,
        0,
        0,
        0,
        1575,
        1569,
        1,
        0,
        0,
        0,
        1575,
        1570,
        1,
        0,
        0,
        0,
        1575,
        1571,
        1,
        0,
        0,
        0,
        1575,
        1572,
        1,
        0,
        0,
        0,
        1575,
        1573,
        1,
        0,
        0,
        0,
        1575,
        1574,
        1,
        0,
        0,
        0,
        1576,
        201,
        1,
        0,
        0,
        0,
        1577,
        1584,
        3,
        208,
        104,
        0,
        1578,
        1584,
        3,
        206,
        103,
        0,
        1579,
        1584,
        3,
        378,
        189,
        0,
        1580,
        1584,
        5,
        34,
        0,
        0,
        1581,
        1584,
        5,
        41,
        0,
        0,
        1582,
        1584,
        5,
        23,
        0,
        0,
        1583,
        1577,
        1,
        0,
        0,
        0,
        1583,
        1578,
        1,
        0,
        0,
        0,
        1583,
        1579,
        1,
        0,
        0,
        0,
        1583,
        1580,
        1,
        0,
        0,
        0,
        1583,
        1581,
        1,
        0,
        0,
        0,
        1583,
        1582,
        1,
        0,
        0,
        0,
        1584,
        203,
        1,
        0,
        0,
        0,
        1585,
        1592,
        3,
        208,
        104,
        0,
        1586,
        1592,
        3,
        206,
        103,
        0,
        1587,
        1592,
        3,
        378,
        189,
        0,
        1588,
        1592,
        5,
        34,
        0,
        0,
        1589,
        1592,
        5,
        41,
        0,
        0,
        1590,
        1592,
        5,
        23,
        0,
        0,
        1591,
        1585,
        1,
        0,
        0,
        0,
        1591,
        1586,
        1,
        0,
        0,
        0,
        1591,
        1587,
        1,
        0,
        0,
        0,
        1591,
        1588,
        1,
        0,
        0,
        0,
        1591,
        1589,
        1,
        0,
        0,
        0,
        1591,
        1590,
        1,
        0,
        0,
        0,
        1592,
        205,
        1,
        0,
        0,
        0,
        1593,
        1594,
        3,
        210,
        105,
        0,
        1594,
        1595,
        5,
        82,
        0,
        0,
        1595,
        1596,
        3,
        214,
        107,
        0,
        1596,
        1602,
        1,
        0,
        0,
        0,
        1597,
        1598,
        3,
        210,
        105,
        0,
        1598,
        1599,
        5,
        83,
        0,
        0,
        1599,
        1600,
        3,
        214,
        107,
        0,
        1600,
        1602,
        1,
        0,
        0,
        0,
        1601,
        1593,
        1,
        0,
        0,
        0,
        1601,
        1597,
        1,
        0,
        0,
        0,
        1602,
        207,
        1,
        0,
        0,
        0,
        1603,
        1607,
        5,
        4,
        0,
        0,
        1604,
        1605,
        5,
        83,
        0,
        0,
        1605,
        1607,
        5,
        4,
        0,
        0,
        1606,
        1603,
        1,
        0,
        0,
        0,
        1606,
        1604,
        1,
        0,
        0,
        0,
        1607,
        209,
        1,
        0,
        0,
        0,
        1608,
        1612,
        3,
        212,
        106,
        0,
        1609,
        1610,
        5,
        83,
        0,
        0,
        1610,
        1612,
        3,
        212,
        106,
        0,
        1611,
        1608,
        1,
        0,
        0,
        0,
        1611,
        1609,
        1,
        0,
        0,
        0,
        1612,
        211,
        1,
        0,
        0,
        0,
        1613,
        1614,
        5,
        4,
        0,
        0,
        1614,
        213,
        1,
        0,
        0,
        0,
        1615,
        1616,
        5,
        4,
        0,
        0,
        1616,
        215,
        1,
        0,
        0,
        0,
        1617,
        1618,
        3,
        218,
        109,
        0,
        1618,
        217,
        1,
        0,
        0,
        0,
        1619,
        1620,
        5,
        52,
        0,
        0,
        1620,
        219,
        1,
        0,
        0,
        0,
        1621,
        1622,
        5,
        44,
        0,
        0,
        1622,
        221,
        1,
        0,
        0,
        0,
        1623,
        1624,
        3,
        224,
        112,
        0,
        1624,
        223,
        1,
        0,
        0,
        0,
        1625,
        1626,
        3,
        226,
        113,
        0,
        1626,
        1627,
        5,
        64,
        0,
        0,
        1627,
        1628,
        5,
        52,
        0,
        0,
        1628,
        225,
        1,
        0,
        0,
        0,
        1629,
        1630,
        6,
        113,
        -1,
        0,
        1630,
        1631,
        5,
        52,
        0,
        0,
        1631,
        1637,
        1,
        0,
        0,
        0,
        1632,
        1633,
        10,
        2,
        0,
        0,
        1633,
        1634,
        5,
        64,
        0,
        0,
        1634,
        1636,
        5,
        52,
        0,
        0,
        1635,
        1632,
        1,
        0,
        0,
        0,
        1636,
        1639,
        1,
        0,
        0,
        0,
        1637,
        1635,
        1,
        0,
        0,
        0,
        1637,
        1638,
        1,
        0,
        0,
        0,
        1638,
        227,
        1,
        0,
        0,
        0,
        1639,
        1637,
        1,
        0,
        0,
        0,
        1640,
        1641,
        5,
        68,
        0,
        0,
        1641,
        1642,
        3,
        194,
        97,
        0,
        1642,
        1643,
        5,
        69,
        0,
        0,
        1643,
        229,
        1,
        0,
        0,
        0,
        1644,
        1646,
        5,
        75,
        0,
        0,
        1645,
        1647,
        3,
        234,
        117,
        0,
        1646,
        1645,
        1,
        0,
        0,
        0,
        1646,
        1647,
        1,
        0,
        0,
        0,
        1647,
        1648,
        1,
        0,
        0,
        0,
        1648,
        1655,
        5,
        76,
        0,
        0,
        1649,
        1651,
        5,
        68,
        0,
        0,
        1650,
        1652,
        3,
        232,
        116,
        0,
        1651,
        1650,
        1,
        0,
        0,
        0,
        1651,
        1652,
        1,
        0,
        0,
        0,
        1652,
        1653,
        1,
        0,
        0,
        0,
        1653,
        1655,
        5,
        69,
        0,
        0,
        1654,
        1644,
        1,
        0,
        0,
        0,
        1654,
        1649,
        1,
        0,
        0,
        0,
        1655,
        231,
        1,
        0,
        0,
        0,
        1656,
        1657,
        3,
        236,
        118,
        0,
        1657,
        1659,
        5,
        70,
        0,
        0,
        1658,
        1660,
        3,
        234,
        117,
        0,
        1659,
        1658,
        1,
        0,
        0,
        0,
        1659,
        1660,
        1,
        0,
        0,
        0,
        1660,
        233,
        1,
        0,
        0,
        0,
        1661,
        1666,
        3,
        236,
        118,
        0,
        1662,
        1663,
        5,
        70,
        0,
        0,
        1663,
        1665,
        3,
        236,
        118,
        0,
        1664,
        1662,
        1,
        0,
        0,
        0,
        1665,
        1668,
        1,
        0,
        0,
        0,
        1666,
        1664,
        1,
        0,
        0,
        0,
        1666,
        1667,
        1,
        0,
        0,
        0,
        1667,
        1670,
        1,
        0,
        0,
        0,
        1668,
        1666,
        1,
        0,
        0,
        0,
        1669,
        1671,
        5,
        70,
        0,
        0,
        1670,
        1669,
        1,
        0,
        0,
        0,
        1670,
        1671,
        1,
        0,
        0,
        0,
        1671,
        235,
        1,
        0,
        0,
        0,
        1672,
        1675,
        3,
        238,
        119,
        0,
        1673,
        1675,
        3,
        194,
        97,
        0,
        1674,
        1672,
        1,
        0,
        0,
        0,
        1674,
        1673,
        1,
        0,
        0,
        0,
        1675,
        237,
        1,
        0,
        0,
        0,
        1676,
        1677,
        5,
        67,
        0,
        0,
        1677,
        1681,
        3,
        218,
        109,
        0,
        1678,
        1679,
        5,
        67,
        0,
        0,
        1679,
        1681,
        3,
        220,
        110,
        0,
        1680,
        1676,
        1,
        0,
        0,
        0,
        1680,
        1678,
        1,
        0,
        0,
        0,
        1681,
        239,
        1,
        0,
        0,
        0,
        1682,
        1683,
        5,
        88,
        0,
        0,
        1683,
        1708,
        5,
        89,
        0,
        0,
        1684,
        1685,
        5,
        88,
        0,
        0,
        1685,
        1687,
        3,
        246,
        123,
        0,
        1686,
        1688,
        5,
        70,
        0,
        0,
        1687,
        1686,
        1,
        0,
        0,
        0,
        1687,
        1688,
        1,
        0,
        0,
        0,
        1688,
        1689,
        1,
        0,
        0,
        0,
        1689,
        1690,
        5,
        89,
        0,
        0,
        1690,
        1708,
        1,
        0,
        0,
        0,
        1691,
        1692,
        5,
        88,
        0,
        0,
        1692,
        1693,
        3,
        242,
        121,
        0,
        1693,
        1694,
        5,
        70,
        0,
        0,
        1694,
        1696,
        3,
        246,
        123,
        0,
        1695,
        1697,
        5,
        70,
        0,
        0,
        1696,
        1695,
        1,
        0,
        0,
        0,
        1696,
        1697,
        1,
        0,
        0,
        0,
        1697,
        1698,
        1,
        0,
        0,
        0,
        1698,
        1699,
        5,
        89,
        0,
        0,
        1699,
        1708,
        1,
        0,
        0,
        0,
        1700,
        1701,
        5,
        88,
        0,
        0,
        1701,
        1703,
        3,
        242,
        121,
        0,
        1702,
        1704,
        5,
        70,
        0,
        0,
        1703,
        1702,
        1,
        0,
        0,
        0,
        1703,
        1704,
        1,
        0,
        0,
        0,
        1704,
        1705,
        1,
        0,
        0,
        0,
        1705,
        1706,
        5,
        89,
        0,
        0,
        1706,
        1708,
        1,
        0,
        0,
        0,
        1707,
        1682,
        1,
        0,
        0,
        0,
        1707,
        1684,
        1,
        0,
        0,
        0,
        1707,
        1691,
        1,
        0,
        0,
        0,
        1707,
        1700,
        1,
        0,
        0,
        0,
        1708,
        241,
        1,
        0,
        0,
        0,
        1709,
        1714,
        3,
        244,
        122,
        0,
        1710,
        1711,
        5,
        70,
        0,
        0,
        1711,
        1713,
        3,
        244,
        122,
        0,
        1712,
        1710,
        1,
        0,
        0,
        0,
        1713,
        1716,
        1,
        0,
        0,
        0,
        1714,
        1712,
        1,
        0,
        0,
        0,
        1714,
        1715,
        1,
        0,
        0,
        0,
        1715,
        243,
        1,
        0,
        0,
        0,
        1716,
        1714,
        1,
        0,
        0,
        0,
        1717,
        1720,
        3,
        204,
        102,
        0,
        1718,
        1720,
        3,
        224,
        112,
        0,
        1719,
        1717,
        1,
        0,
        0,
        0,
        1719,
        1718,
        1,
        0,
        0,
        0,
        1720,
        1721,
        1,
        0,
        0,
        0,
        1721,
        1722,
        5,
        71,
        0,
        0,
        1722,
        1723,
        3,
        194,
        97,
        0,
        1723,
        245,
        1,
        0,
        0,
        0,
        1724,
        1725,
        5,
        73,
        0,
        0,
        1725,
        1726,
        3,
        218,
        109,
        0,
        1726,
        247,
        1,
        0,
        0,
        0,
        1727,
        1728,
        3,
        226,
        113,
        0,
        1728,
        1729,
        5,
        68,
        0,
        0,
        1729,
        1730,
        5,
        69,
        0,
        0,
        1730,
        1758,
        1,
        0,
        0,
        0,
        1731,
        1732,
        3,
        226,
        113,
        0,
        1732,
        1733,
        5,
        68,
        0,
        0,
        1733,
        1735,
        3,
        250,
        125,
        0,
        1734,
        1736,
        5,
        70,
        0,
        0,
        1735,
        1734,
        1,
        0,
        0,
        0,
        1735,
        1736,
        1,
        0,
        0,
        0,
        1736,
        1737,
        1,
        0,
        0,
        0,
        1737,
        1738,
        5,
        69,
        0,
        0,
        1738,
        1758,
        1,
        0,
        0,
        0,
        1739,
        1740,
        3,
        226,
        113,
        0,
        1740,
        1741,
        5,
        68,
        0,
        0,
        1741,
        1743,
        3,
        252,
        126,
        0,
        1742,
        1744,
        5,
        70,
        0,
        0,
        1743,
        1742,
        1,
        0,
        0,
        0,
        1743,
        1744,
        1,
        0,
        0,
        0,
        1744,
        1745,
        1,
        0,
        0,
        0,
        1745,
        1746,
        5,
        69,
        0,
        0,
        1746,
        1758,
        1,
        0,
        0,
        0,
        1747,
        1748,
        3,
        226,
        113,
        0,
        1748,
        1749,
        5,
        68,
        0,
        0,
        1749,
        1750,
        3,
        250,
        125,
        0,
        1750,
        1751,
        5,
        70,
        0,
        0,
        1751,
        1753,
        3,
        252,
        126,
        0,
        1752,
        1754,
        5,
        70,
        0,
        0,
        1753,
        1752,
        1,
        0,
        0,
        0,
        1753,
        1754,
        1,
        0,
        0,
        0,
        1754,
        1755,
        1,
        0,
        0,
        0,
        1755,
        1756,
        5,
        69,
        0,
        0,
        1756,
        1758,
        1,
        0,
        0,
        0,
        1757,
        1727,
        1,
        0,
        0,
        0,
        1757,
        1731,
        1,
        0,
        0,
        0,
        1757,
        1739,
        1,
        0,
        0,
        0,
        1757,
        1747,
        1,
        0,
        0,
        0,
        1758,
        249,
        1,
        0,
        0,
        0,
        1759,
        1764,
        3,
        194,
        97,
        0,
        1760,
        1761,
        5,
        70,
        0,
        0,
        1761,
        1763,
        3,
        194,
        97,
        0,
        1762,
        1760,
        1,
        0,
        0,
        0,
        1763,
        1766,
        1,
        0,
        0,
        0,
        1764,
        1762,
        1,
        0,
        0,
        0,
        1764,
        1765,
        1,
        0,
        0,
        0,
        1765,
        251,
        1,
        0,
        0,
        0,
        1766,
        1764,
        1,
        0,
        0,
        0,
        1767,
        1772,
        3,
        254,
        127,
        0,
        1768,
        1769,
        5,
        70,
        0,
        0,
        1769,
        1771,
        3,
        254,
        127,
        0,
        1770,
        1768,
        1,
        0,
        0,
        0,
        1771,
        1774,
        1,
        0,
        0,
        0,
        1772,
        1770,
        1,
        0,
        0,
        0,
        1772,
        1773,
        1,
        0,
        0,
        0,
        1773,
        253,
        1,
        0,
        0,
        0,
        1774,
        1772,
        1,
        0,
        0,
        0,
        1775,
        1776,
        5,
        52,
        0,
        0,
        1776,
        1777,
        5,
        74,
        0,
        0,
        1777,
        1778,
        3,
        194,
        97,
        0,
        1778,
        255,
        1,
        0,
        0,
        0,
        1779,
        1780,
        5,
        43,
        0,
        0,
        1780,
        1782,
        5,
        52,
        0,
        0,
        1781,
        1783,
        3,
        258,
        129,
        0,
        1782,
        1781,
        1,
        0,
        0,
        0,
        1782,
        1783,
        1,
        0,
        0,
        0,
        1783,
        1784,
        1,
        0,
        0,
        0,
        1784,
        1785,
        5,
        74,
        0,
        0,
        1785,
        1786,
        3,
        268,
        134,
        0,
        1786,
        257,
        1,
        0,
        0,
        0,
        1787,
        1788,
        5,
        75,
        0,
        0,
        1788,
        1789,
        3,
        260,
        130,
        0,
        1789,
        1790,
        5,
        76,
        0,
        0,
        1790,
        259,
        1,
        0,
        0,
        0,
        1791,
        1796,
        3,
        262,
        131,
        0,
        1792,
        1793,
        5,
        70,
        0,
        0,
        1793,
        1795,
        3,
        262,
        131,
        0,
        1794,
        1792,
        1,
        0,
        0,
        0,
        1795,
        1798,
        1,
        0,
        0,
        0,
        1796,
        1794,
        1,
        0,
        0,
        0,
        1796,
        1797,
        1,
        0,
        0,
        0,
        1797,
        1800,
        1,
        0,
        0,
        0,
        1798,
        1796,
        1,
        0,
        0,
        0,
        1799,
        1801,
        5,
        70,
        0,
        0,
        1800,
        1799,
        1,
        0,
        0,
        0,
        1800,
        1801,
        1,
        0,
        0,
        0,
        1801,
        261,
        1,
        0,
        0,
        0,
        1802,
        1804,
        5,
        52,
        0,
        0,
        1803,
        1805,
        3,
        264,
        132,
        0,
        1804,
        1803,
        1,
        0,
        0,
        0,
        1804,
        1805,
        1,
        0,
        0,
        0,
        1805,
        1811,
        1,
        0,
        0,
        0,
        1806,
        1807,
        5,
        67,
        0,
        0,
        1807,
        1811,
        5,
        52,
        0,
        0,
        1808,
        1809,
        5,
        73,
        0,
        0,
        1809,
        1811,
        5,
        52,
        0,
        0,
        1810,
        1802,
        1,
        0,
        0,
        0,
        1810,
        1806,
        1,
        0,
        0,
        0,
        1810,
        1808,
        1,
        0,
        0,
        0,
        1811,
        263,
        1,
        0,
        0,
        0,
        1812,
        1813,
        5,
        71,
        0,
        0,
        1813,
        1814,
        3,
        268,
        134,
        0,
        1814,
        265,
        1,
        0,
        0,
        0,
        1815,
        1820,
        3,
        268,
        134,
        0,
        1816,
        1817,
        5,
        70,
        0,
        0,
        1817,
        1819,
        3,
        268,
        134,
        0,
        1818,
        1816,
        1,
        0,
        0,
        0,
        1819,
        1822,
        1,
        0,
        0,
        0,
        1820,
        1818,
        1,
        0,
        0,
        0,
        1820,
        1821,
        1,
        0,
        0,
        0,
        1821,
        1824,
        1,
        0,
        0,
        0,
        1822,
        1820,
        1,
        0,
        0,
        0,
        1823,
        1825,
        5,
        70,
        0,
        0,
        1824,
        1823,
        1,
        0,
        0,
        0,
        1824,
        1825,
        1,
        0,
        0,
        0,
        1825,
        267,
        1,
        0,
        0,
        0,
        1826,
        1827,
        3,
        284,
        142,
        0,
        1827,
        1828,
        5,
        28,
        0,
        0,
        1828,
        1829,
        3,
        284,
        142,
        0,
        1829,
        1830,
        5,
        21,
        0,
        0,
        1830,
        1831,
        3,
        268,
        134,
        0,
        1831,
        1835,
        1,
        0,
        0,
        0,
        1832,
        1835,
        3,
        284,
        142,
        0,
        1833,
        1835,
        3,
        342,
        171,
        0,
        1834,
        1826,
        1,
        0,
        0,
        0,
        1834,
        1832,
        1,
        0,
        0,
        0,
        1834,
        1833,
        1,
        0,
        0,
        0,
        1835,
        269,
        1,
        0,
        0,
        0,
        1836,
        1837,
        5,
        48,
        0,
        0,
        1837,
        1838,
        5,
        26,
        0,
        0,
        1838,
        1844,
        3,
        268,
        134,
        0,
        1839,
        1841,
        5,
        48,
        0,
        0,
        1840,
        1842,
        3,
        272,
        136,
        0,
        1841,
        1840,
        1,
        0,
        0,
        0,
        1841,
        1842,
        1,
        0,
        0,
        0,
        1842,
        1844,
        1,
        0,
        0,
        0,
        1843,
        1836,
        1,
        0,
        0,
        0,
        1843,
        1839,
        1,
        0,
        0,
        0,
        1844,
        271,
        1,
        0,
        0,
        0,
        1845,
        1850,
        3,
        274,
        137,
        0,
        1846,
        1847,
        5,
        70,
        0,
        0,
        1847,
        1849,
        3,
        274,
        137,
        0,
        1848,
        1846,
        1,
        0,
        0,
        0,
        1849,
        1852,
        1,
        0,
        0,
        0,
        1850,
        1848,
        1,
        0,
        0,
        0,
        1850,
        1851,
        1,
        0,
        0,
        0,
        1851,
        1854,
        1,
        0,
        0,
        0,
        1852,
        1850,
        1,
        0,
        0,
        0,
        1853,
        1855,
        5,
        70,
        0,
        0,
        1854,
        1853,
        1,
        0,
        0,
        0,
        1854,
        1855,
        1,
        0,
        0,
        0,
        1855,
        273,
        1,
        0,
        0,
        0,
        1856,
        1857,
        5,
        67,
        0,
        0,
        1857,
        1860,
        3,
        314,
        157,
        0,
        1858,
        1860,
        3,
        268,
        134,
        0,
        1859,
        1856,
        1,
        0,
        0,
        0,
        1859,
        1858,
        1,
        0,
        0,
        0,
        1860,
        275,
        1,
        0,
        0,
        0,
        1861,
        1866,
        3,
        278,
        139,
        0,
        1862,
        1863,
        5,
        70,
        0,
        0,
        1863,
        1865,
        3,
        278,
        139,
        0,
        1864,
        1862,
        1,
        0,
        0,
        0,
        1865,
        1868,
        1,
        0,
        0,
        0,
        1866,
        1864,
        1,
        0,
        0,
        0,
        1866,
        1867,
        1,
        0,
        0,
        0,
        1867,
        1870,
        1,
        0,
        0,
        0,
        1868,
        1866,
        1,
        0,
        0,
        0,
        1869,
        1871,
        5,
        70,
        0,
        0,
        1870,
        1869,
        1,
        0,
        0,
        0,
        1870,
        1871,
        1,
        0,
        0,
        0,
        1871,
        277,
        1,
        0,
        0,
        0,
        1872,
        1873,
        5,
        67,
        0,
        0,
        1873,
        1876,
        3,
        314,
        157,
        0,
        1874,
        1876,
        3,
        282,
        141,
        0,
        1875,
        1872,
        1,
        0,
        0,
        0,
        1875,
        1874,
        1,
        0,
        0,
        0,
        1876,
        279,
        1,
        0,
        0,
        0,
        1877,
        1878,
        5,
        52,
        0,
        0,
        1878,
        1879,
        5,
        112,
        0,
        0,
        1879,
        1880,
        3,
        268,
        134,
        0,
        1880,
        281,
        1,
        0,
        0,
        0,
        1881,
        1884,
        3,
        280,
        140,
        0,
        1882,
        1884,
        3,
        268,
        134,
        0,
        1883,
        1881,
        1,
        0,
        0,
        0,
        1883,
        1882,
        1,
        0,
        0,
        0,
        1884,
        283,
        1,
        0,
        0,
        0,
        1885,
        1890,
        3,
        286,
        143,
        0,
        1886,
        1887,
        5,
        37,
        0,
        0,
        1887,
        1889,
        3,
        286,
        143,
        0,
        1888,
        1886,
        1,
        0,
        0,
        0,
        1889,
        1892,
        1,
        0,
        0,
        0,
        1890,
        1888,
        1,
        0,
        0,
        0,
        1890,
        1891,
        1,
        0,
        0,
        0,
        1891,
        285,
        1,
        0,
        0,
        0,
        1892,
        1890,
        1,
        0,
        0,
        0,
        1893,
        1898,
        3,
        288,
        144,
        0,
        1894,
        1895,
        5,
        9,
        0,
        0,
        1895,
        1897,
        3,
        288,
        144,
        0,
        1896,
        1894,
        1,
        0,
        0,
        0,
        1897,
        1900,
        1,
        0,
        0,
        0,
        1898,
        1896,
        1,
        0,
        0,
        0,
        1898,
        1899,
        1,
        0,
        0,
        0,
        1899,
        287,
        1,
        0,
        0,
        0,
        1900,
        1898,
        1,
        0,
        0,
        0,
        1901,
        1902,
        5,
        36,
        0,
        0,
        1902,
        1905,
        3,
        288,
        144,
        0,
        1903,
        1905,
        3,
        290,
        145,
        0,
        1904,
        1901,
        1,
        0,
        0,
        0,
        1904,
        1903,
        1,
        0,
        0,
        0,
        1905,
        289,
        1,
        0,
        0,
        0,
        1906,
        1910,
        3,
        314,
        157,
        0,
        1907,
        1909,
        3,
        292,
        146,
        0,
        1908,
        1907,
        1,
        0,
        0,
        0,
        1909,
        1912,
        1,
        0,
        0,
        0,
        1910,
        1908,
        1,
        0,
        0,
        0,
        1910,
        1911,
        1,
        0,
        0,
        0,
        1911,
        291,
        1,
        0,
        0,
        0,
        1912,
        1910,
        1,
        0,
        0,
        0,
        1913,
        1924,
        3,
        294,
        147,
        0,
        1914,
        1924,
        3,
        296,
        148,
        0,
        1915,
        1924,
        3,
        298,
        149,
        0,
        1916,
        1924,
        3,
        300,
        150,
        0,
        1917,
        1924,
        3,
        302,
        151,
        0,
        1918,
        1924,
        3,
        304,
        152,
        0,
        1919,
        1924,
        3,
        306,
        153,
        0,
        1920,
        1924,
        3,
        308,
        154,
        0,
        1921,
        1924,
        3,
        310,
        155,
        0,
        1922,
        1924,
        3,
        312,
        156,
        0,
        1923,
        1913,
        1,
        0,
        0,
        0,
        1923,
        1914,
        1,
        0,
        0,
        0,
        1923,
        1915,
        1,
        0,
        0,
        0,
        1923,
        1916,
        1,
        0,
        0,
        0,
        1923,
        1917,
        1,
        0,
        0,
        0,
        1923,
        1918,
        1,
        0,
        0,
        0,
        1923,
        1919,
        1,
        0,
        0,
        0,
        1923,
        1920,
        1,
        0,
        0,
        0,
        1923,
        1921,
        1,
        0,
        0,
        0,
        1923,
        1922,
        1,
        0,
        0,
        0,
        1924,
        293,
        1,
        0,
        0,
        0,
        1925,
        1926,
        5,
        92,
        0,
        0,
        1926,
        1927,
        3,
        314,
        157,
        0,
        1927,
        295,
        1,
        0,
        0,
        0,
        1928,
        1929,
        5,
        96,
        0,
        0,
        1929,
        1930,
        3,
        314,
        157,
        0,
        1930,
        297,
        1,
        0,
        0,
        0,
        1931,
        1932,
        5,
        94,
        0,
        0,
        1932,
        1933,
        3,
        314,
        157,
        0,
        1933,
        299,
        1,
        0,
        0,
        0,
        1934,
        1935,
        5,
        90,
        0,
        0,
        1935,
        1936,
        3,
        314,
        157,
        0,
        1936,
        301,
        1,
        0,
        0,
        0,
        1937,
        1938,
        5,
        93,
        0,
        0,
        1938,
        1939,
        3,
        314,
        157,
        0,
        1939,
        303,
        1,
        0,
        0,
        0,
        1940,
        1941,
        5,
        91,
        0,
        0,
        1941,
        1942,
        3,
        314,
        157,
        0,
        1942,
        305,
        1,
        0,
        0,
        0,
        1943,
        1944,
        5,
        36,
        0,
        0,
        1944,
        1945,
        5,
        30,
        0,
        0,
        1945,
        1946,
        3,
        314,
        157,
        0,
        1946,
        307,
        1,
        0,
        0,
        0,
        1947,
        1948,
        5,
        30,
        0,
        0,
        1948,
        1949,
        3,
        314,
        157,
        0,
        1949,
        309,
        1,
        0,
        0,
        0,
        1950,
        1951,
        5,
        31,
        0,
        0,
        1951,
        1952,
        5,
        36,
        0,
        0,
        1952,
        1953,
        3,
        314,
        157,
        0,
        1953,
        311,
        1,
        0,
        0,
        0,
        1954,
        1955,
        5,
        31,
        0,
        0,
        1955,
        1956,
        3,
        314,
        157,
        0,
        1956,
        313,
        1,
        0,
        0,
        0,
        1957,
        1958,
        6,
        157,
        -1,
        0,
        1958,
        1959,
        3,
        316,
        158,
        0,
        1959,
        1965,
        1,
        0,
        0,
        0,
        1960,
        1961,
        10,
        2,
        0,
        0,
        1961,
        1962,
        5,
        77,
        0,
        0,
        1962,
        1964,
        3,
        316,
        158,
        0,
        1963,
        1960,
        1,
        0,
        0,
        0,
        1964,
        1967,
        1,
        0,
        0,
        0,
        1965,
        1963,
        1,
        0,
        0,
        0,
        1965,
        1966,
        1,
        0,
        0,
        0,
        1966,
        315,
        1,
        0,
        0,
        0,
        1967,
        1965,
        1,
        0,
        0,
        0,
        1968,
        1969,
        6,
        158,
        -1,
        0,
        1969,
        1970,
        3,
        318,
        159,
        0,
        1970,
        1976,
        1,
        0,
        0,
        0,
        1971,
        1972,
        10,
        2,
        0,
        0,
        1972,
        1973,
        5,
        78,
        0,
        0,
        1973,
        1975,
        3,
        318,
        159,
        0,
        1974,
        1971,
        1,
        0,
        0,
        0,
        1975,
        1978,
        1,
        0,
        0,
        0,
        1976,
        1974,
        1,
        0,
        0,
        0,
        1976,
        1977,
        1,
        0,
        0,
        0,
        1977,
        317,
        1,
        0,
        0,
        0,
        1978,
        1976,
        1,
        0,
        0,
        0,
        1979,
        1980,
        6,
        159,
        -1,
        0,
        1980,
        1981,
        3,
        320,
        160,
        0,
        1981,
        1987,
        1,
        0,
        0,
        0,
        1982,
        1983,
        10,
        2,
        0,
        0,
        1983,
        1984,
        5,
        79,
        0,
        0,
        1984,
        1986,
        3,
        320,
        160,
        0,
        1985,
        1982,
        1,
        0,
        0,
        0,
        1986,
        1989,
        1,
        0,
        0,
        0,
        1987,
        1985,
        1,
        0,
        0,
        0,
        1987,
        1988,
        1,
        0,
        0,
        0,
        1988,
        319,
        1,
        0,
        0,
        0,
        1989,
        1987,
        1,
        0,
        0,
        0,
        1990,
        1991,
        6,
        160,
        -1,
        0,
        1991,
        1992,
        3,
        322,
        161,
        0,
        1992,
        2001,
        1,
        0,
        0,
        0,
        1993,
        1994,
        10,
        3,
        0,
        0,
        1994,
        1995,
        5,
        80,
        0,
        0,
        1995,
        2000,
        3,
        322,
        161,
        0,
        1996,
        1997,
        10,
        2,
        0,
        0,
        1997,
        1998,
        5,
        81,
        0,
        0,
        1998,
        2000,
        3,
        322,
        161,
        0,
        1999,
        1993,
        1,
        0,
        0,
        0,
        1999,
        1996,
        1,
        0,
        0,
        0,
        2000,
        2003,
        1,
        0,
        0,
        0,
        2001,
        1999,
        1,
        0,
        0,
        0,
        2001,
        2002,
        1,
        0,
        0,
        0,
        2002,
        321,
        1,
        0,
        0,
        0,
        2003,
        2001,
        1,
        0,
        0,
        0,
        2004,
        2005,
        6,
        161,
        -1,
        0,
        2005,
        2006,
        3,
        324,
        162,
        0,
        2006,
        2015,
        1,
        0,
        0,
        0,
        2007,
        2008,
        10,
        3,
        0,
        0,
        2008,
        2009,
        5,
        82,
        0,
        0,
        2009,
        2014,
        3,
        324,
        162,
        0,
        2010,
        2011,
        10,
        2,
        0,
        0,
        2011,
        2012,
        5,
        83,
        0,
        0,
        2012,
        2014,
        3,
        324,
        162,
        0,
        2013,
        2007,
        1,
        0,
        0,
        0,
        2013,
        2010,
        1,
        0,
        0,
        0,
        2014,
        2017,
        1,
        0,
        0,
        0,
        2015,
        2013,
        1,
        0,
        0,
        0,
        2015,
        2016,
        1,
        0,
        0,
        0,
        2016,
        323,
        1,
        0,
        0,
        0,
        2017,
        2015,
        1,
        0,
        0,
        0,
        2018,
        2019,
        6,
        162,
        -1,
        0,
        2019,
        2020,
        3,
        326,
        163,
        0,
        2020,
        2038,
        1,
        0,
        0,
        0,
        2021,
        2022,
        10,
        6,
        0,
        0,
        2022,
        2023,
        5,
        67,
        0,
        0,
        2023,
        2037,
        3,
        326,
        163,
        0,
        2024,
        2025,
        10,
        5,
        0,
        0,
        2025,
        2026,
        5,
        84,
        0,
        0,
        2026,
        2037,
        3,
        326,
        163,
        0,
        2027,
        2028,
        10,
        4,
        0,
        0,
        2028,
        2029,
        5,
        86,
        0,
        0,
        2029,
        2037,
        3,
        326,
        163,
        0,
        2030,
        2031,
        10,
        3,
        0,
        0,
        2031,
        2032,
        5,
        85,
        0,
        0,
        2032,
        2037,
        3,
        326,
        163,
        0,
        2033,
        2034,
        10,
        2,
        0,
        0,
        2034,
        2035,
        5,
        97,
        0,
        0,
        2035,
        2037,
        3,
        326,
        163,
        0,
        2036,
        2021,
        1,
        0,
        0,
        0,
        2036,
        2024,
        1,
        0,
        0,
        0,
        2036,
        2027,
        1,
        0,
        0,
        0,
        2036,
        2030,
        1,
        0,
        0,
        0,
        2036,
        2033,
        1,
        0,
        0,
        0,
        2037,
        2040,
        1,
        0,
        0,
        0,
        2038,
        2036,
        1,
        0,
        0,
        0,
        2038,
        2039,
        1,
        0,
        0,
        0,
        2039,
        325,
        1,
        0,
        0,
        0,
        2040,
        2038,
        1,
        0,
        0,
        0,
        2041,
        2042,
        5,
        82,
        0,
        0,
        2042,
        2049,
        3,
        326,
        163,
        0,
        2043,
        2044,
        5,
        83,
        0,
        0,
        2044,
        2049,
        3,
        326,
        163,
        0,
        2045,
        2046,
        5,
        87,
        0,
        0,
        2046,
        2049,
        3,
        326,
        163,
        0,
        2047,
        2049,
        3,
        328,
        164,
        0,
        2048,
        2041,
        1,
        0,
        0,
        0,
        2048,
        2043,
        1,
        0,
        0,
        0,
        2048,
        2045,
        1,
        0,
        0,
        0,
        2048,
        2047,
        1,
        0,
        0,
        0,
        2049,
        327,
        1,
        0,
        0,
        0,
        2050,
        2051,
        3,
        330,
        165,
        0,
        2051,
        2052,
        5,
        73,
        0,
        0,
        2052,
        2053,
        3,
        326,
        163,
        0,
        2053,
        2056,
        1,
        0,
        0,
        0,
        2054,
        2056,
        3,
        330,
        165,
        0,
        2055,
        2050,
        1,
        0,
        0,
        0,
        2055,
        2054,
        1,
        0,
        0,
        0,
        2056,
        329,
        1,
        0,
        0,
        0,
        2057,
        2058,
        5,
        13,
        0,
        0,
        2058,
        2061,
        3,
        332,
        166,
        0,
        2059,
        2061,
        3,
        332,
        166,
        0,
        2060,
        2057,
        1,
        0,
        0,
        0,
        2060,
        2059,
        1,
        0,
        0,
        0,
        2061,
        331,
        1,
        0,
        0,
        0,
        2062,
        2063,
        6,
        166,
        -1,
        0,
        2063,
        2064,
        3,
        338,
        169,
        0,
        2064,
        2083,
        1,
        0,
        0,
        0,
        2065,
        2066,
        10,
        5,
        0,
        0,
        2066,
        2067,
        5,
        64,
        0,
        0,
        2067,
        2082,
        5,
        52,
        0,
        0,
        2068,
        2069,
        10,
        4,
        0,
        0,
        2069,
        2082,
        3,
        402,
        201,
        0,
        2070,
        2071,
        10,
        3,
        0,
        0,
        2071,
        2073,
        5,
        68,
        0,
        0,
        2072,
        2074,
        3,
        406,
        203,
        0,
        2073,
        2072,
        1,
        0,
        0,
        0,
        2073,
        2074,
        1,
        0,
        0,
        0,
        2074,
        2075,
        1,
        0,
        0,
        0,
        2075,
        2082,
        5,
        69,
        0,
        0,
        2076,
        2077,
        10,
        2,
        0,
        0,
        2077,
        2078,
        5,
        75,
        0,
        0,
        2078,
        2079,
        3,
        334,
        167,
        0,
        2079,
        2080,
        5,
        76,
        0,
        0,
        2080,
        2082,
        1,
        0,
        0,
        0,
        2081,
        2065,
        1,
        0,
        0,
        0,
        2081,
        2068,
        1,
        0,
        0,
        0,
        2081,
        2070,
        1,
        0,
        0,
        0,
        2081,
        2076,
        1,
        0,
        0,
        0,
        2082,
        2085,
        1,
        0,
        0,
        0,
        2083,
        2081,
        1,
        0,
        0,
        0,
        2083,
        2084,
        1,
        0,
        0,
        0,
        2084,
        333,
        1,
        0,
        0,
        0,
        2085,
        2083,
        1,
        0,
        0,
        0,
        2086,
        2089,
        3,
        336,
        168,
        0,
        2087,
        2089,
        3,
        414,
        207,
        0,
        2088,
        2086,
        1,
        0,
        0,
        0,
        2088,
        2087,
        1,
        0,
        0,
        0,
        2089,
        2097,
        1,
        0,
        0,
        0,
        2090,
        2093,
        5,
        70,
        0,
        0,
        2091,
        2094,
        3,
        336,
        168,
        0,
        2092,
        2094,
        3,
        414,
        207,
        0,
        2093,
        2091,
        1,
        0,
        0,
        0,
        2093,
        2092,
        1,
        0,
        0,
        0,
        2094,
        2096,
        1,
        0,
        0,
        0,
        2095,
        2090,
        1,
        0,
        0,
        0,
        2096,
        2099,
        1,
        0,
        0,
        0,
        2097,
        2095,
        1,
        0,
        0,
        0,
        2097,
        2098,
        1,
        0,
        0,
        0,
        2098,
        2101,
        1,
        0,
        0,
        0,
        2099,
        2097,
        1,
        0,
        0,
        0,
        2100,
        2102,
        5,
        70,
        0,
        0,
        2101,
        2100,
        1,
        0,
        0,
        0,
        2101,
        2102,
        1,
        0,
        0,
        0,
        2102,
        335,
        1,
        0,
        0,
        0,
        2103,
        2105,
        3,
        268,
        134,
        0,
        2104,
        2103,
        1,
        0,
        0,
        0,
        2104,
        2105,
        1,
        0,
        0,
        0,
        2105,
        2106,
        1,
        0,
        0,
        0,
        2106,
        2108,
        5,
        71,
        0,
        0,
        2107,
        2109,
        3,
        268,
        134,
        0,
        2108,
        2107,
        1,
        0,
        0,
        0,
        2108,
        2109,
        1,
        0,
        0,
        0,
        2109,
        2114,
        1,
        0,
        0,
        0,
        2110,
        2112,
        5,
        71,
        0,
        0,
        2111,
        2113,
        3,
        268,
        134,
        0,
        2112,
        2111,
        1,
        0,
        0,
        0,
        2112,
        2113,
        1,
        0,
        0,
        0,
        2113,
        2115,
        1,
        0,
        0,
        0,
        2114,
        2110,
        1,
        0,
        0,
        0,
        2114,
        2115,
        1,
        0,
        0,
        0,
        2115,
        2118,
        1,
        0,
        0,
        0,
        2116,
        2118,
        3,
        282,
        141,
        0,
        2117,
        2104,
        1,
        0,
        0,
        0,
        2117,
        2116,
        1,
        0,
        0,
        0,
        2118,
        337,
        1,
        0,
        0,
        0,
        2119,
        2143,
        3,
        42,
        21,
        0,
        2120,
        2143,
        5,
        52,
        0,
        0,
        2121,
        2143,
        5,
        41,
        0,
        0,
        2122,
        2143,
        5,
        23,
        0,
        0,
        2123,
        2143,
        5,
        34,
        0,
        0,
        2124,
        2143,
        3,
        378,
        189,
        0,
        2125,
        2143,
        5,
        4,
        0,
        0,
        2126,
        2130,
        3,
        382,
        191,
        0,
        2127,
        2130,
        3,
        340,
        170,
        0,
        2128,
        2130,
        3,
        402,
        201,
        0,
        2129,
        2126,
        1,
        0,
        0,
        0,
        2129,
        2127,
        1,
        0,
        0,
        0,
        2129,
        2128,
        1,
        0,
        0,
        0,
        2130,
        2143,
        1,
        0,
        0,
        0,
        2131,
        2134,
        3,
        380,
        190,
        0,
        2132,
        2134,
        3,
        398,
        199,
        0,
        2133,
        2131,
        1,
        0,
        0,
        0,
        2133,
        2132,
        1,
        0,
        0,
        0,
        2134,
        2143,
        1,
        0,
        0,
        0,
        2135,
        2140,
        3,
        386,
        193,
        0,
        2136,
        2140,
        3,
        384,
        192,
        0,
        2137,
        2140,
        3,
        404,
        202,
        0,
        2138,
        2140,
        3,
        400,
        200,
        0,
        2139,
        2135,
        1,
        0,
        0,
        0,
        2139,
        2136,
        1,
        0,
        0,
        0,
        2139,
        2137,
        1,
        0,
        0,
        0,
        2139,
        2138,
        1,
        0,
        0,
        0,
        2140,
        2143,
        1,
        0,
        0,
        0,
        2141,
        2143,
        5,
        66,
        0,
        0,
        2142,
        2119,
        1,
        0,
        0,
        0,
        2142,
        2120,
        1,
        0,
        0,
        0,
        2142,
        2121,
        1,
        0,
        0,
        0,
        2142,
        2122,
        1,
        0,
        0,
        0,
        2142,
        2123,
        1,
        0,
        0,
        0,
        2142,
        2124,
        1,
        0,
        0,
        0,
        2142,
        2125,
        1,
        0,
        0,
        0,
        2142,
        2129,
        1,
        0,
        0,
        0,
        2142,
        2133,
        1,
        0,
        0,
        0,
        2142,
        2139,
        1,
        0,
        0,
        0,
        2142,
        2141,
        1,
        0,
        0,
        0,
        2143,
        339,
        1,
        0,
        0,
        0,
        2144,
        2147,
        5,
        68,
        0,
        0,
        2145,
        2148,
        3,
        270,
        135,
        0,
        2146,
        2148,
        3,
        282,
        141,
        0,
        2147,
        2145,
        1,
        0,
        0,
        0,
        2147,
        2146,
        1,
        0,
        0,
        0,
        2148,
        2149,
        1,
        0,
        0,
        0,
        2149,
        2150,
        5,
        69,
        0,
        0,
        2150,
        341,
        1,
        0,
        0,
        0,
        2151,
        2153,
        5,
        32,
        0,
        0,
        2152,
        2154,
        3,
        344,
        172,
        0,
        2153,
        2152,
        1,
        0,
        0,
        0,
        2153,
        2154,
        1,
        0,
        0,
        0,
        2154,
        2155,
        1,
        0,
        0,
        0,
        2155,
        2156,
        5,
        71,
        0,
        0,
        2156,
        2157,
        3,
        268,
        134,
        0,
        2157,
        343,
        1,
        0,
        0,
        0,
        2158,
        2159,
        3,
        346,
        173,
        0,
        2159,
        345,
        1,
        0,
        0,
        0,
        2160,
        2164,
        3,
        348,
        174,
        0,
        2161,
        2163,
        3,
        356,
        178,
        0,
        2162,
        2161,
        1,
        0,
        0,
        0,
        2163,
        2166,
        1,
        0,
        0,
        0,
        2164,
        2162,
        1,
        0,
        0,
        0,
        2164,
        2165,
        1,
        0,
        0,
        0,
        2165,
        2170,
        1,
        0,
        0,
        0,
        2166,
        2164,
        1,
        0,
        0,
        0,
        2167,
        2169,
        3,
        358,
        179,
        0,
        2168,
        2167,
        1,
        0,
        0,
        0,
        2169,
        2172,
        1,
        0,
        0,
        0,
        2170,
        2168,
        1,
        0,
        0,
        0,
        2170,
        2171,
        1,
        0,
        0,
        0,
        2171,
        2174,
        1,
        0,
        0,
        0,
        2172,
        2170,
        1,
        0,
        0,
        0,
        2173,
        2175,
        3,
        352,
        176,
        0,
        2174,
        2173,
        1,
        0,
        0,
        0,
        2174,
        2175,
        1,
        0,
        0,
        0,
        2175,
        2210,
        1,
        0,
        0,
        0,
        2176,
        2180,
        3,
        350,
        175,
        0,
        2177,
        2179,
        3,
        358,
        179,
        0,
        2178,
        2177,
        1,
        0,
        0,
        0,
        2179,
        2182,
        1,
        0,
        0,
        0,
        2180,
        2178,
        1,
        0,
        0,
        0,
        2180,
        2181,
        1,
        0,
        0,
        0,
        2181,
        2184,
        1,
        0,
        0,
        0,
        2182,
        2180,
        1,
        0,
        0,
        0,
        2183,
        2185,
        3,
        352,
        176,
        0,
        2184,
        2183,
        1,
        0,
        0,
        0,
        2184,
        2185,
        1,
        0,
        0,
        0,
        2185,
        2210,
        1,
        0,
        0,
        0,
        2186,
        2188,
        3,
        356,
        178,
        0,
        2187,
        2186,
        1,
        0,
        0,
        0,
        2188,
        2189,
        1,
        0,
        0,
        0,
        2189,
        2187,
        1,
        0,
        0,
        0,
        2189,
        2190,
        1,
        0,
        0,
        0,
        2190,
        2194,
        1,
        0,
        0,
        0,
        2191,
        2193,
        3,
        358,
        179,
        0,
        2192,
        2191,
        1,
        0,
        0,
        0,
        2193,
        2196,
        1,
        0,
        0,
        0,
        2194,
        2192,
        1,
        0,
        0,
        0,
        2194,
        2195,
        1,
        0,
        0,
        0,
        2195,
        2198,
        1,
        0,
        0,
        0,
        2196,
        2194,
        1,
        0,
        0,
        0,
        2197,
        2199,
        3,
        352,
        176,
        0,
        2198,
        2197,
        1,
        0,
        0,
        0,
        2198,
        2199,
        1,
        0,
        0,
        0,
        2199,
        2210,
        1,
        0,
        0,
        0,
        2200,
        2202,
        3,
        358,
        179,
        0,
        2201,
        2200,
        1,
        0,
        0,
        0,
        2202,
        2203,
        1,
        0,
        0,
        0,
        2203,
        2201,
        1,
        0,
        0,
        0,
        2203,
        2204,
        1,
        0,
        0,
        0,
        2204,
        2206,
        1,
        0,
        0,
        0,
        2205,
        2207,
        3,
        352,
        176,
        0,
        2206,
        2205,
        1,
        0,
        0,
        0,
        2206,
        2207,
        1,
        0,
        0,
        0,
        2207,
        2210,
        1,
        0,
        0,
        0,
        2208,
        2210,
        3,
        352,
        176,
        0,
        2209,
        2160,
        1,
        0,
        0,
        0,
        2209,
        2176,
        1,
        0,
        0,
        0,
        2209,
        2187,
        1,
        0,
        0,
        0,
        2209,
        2201,
        1,
        0,
        0,
        0,
        2209,
        2208,
        1,
        0,
        0,
        0,
        2210,
        347,
        1,
        0,
        0,
        0,
        2211,
        2213,
        3,
        356,
        178,
        0,
        2212,
        2211,
        1,
        0,
        0,
        0,
        2213,
        2214,
        1,
        0,
        0,
        0,
        2214,
        2212,
        1,
        0,
        0,
        0,
        2214,
        2215,
        1,
        0,
        0,
        0,
        2215,
        2216,
        1,
        0,
        0,
        0,
        2216,
        2218,
        5,
        84,
        0,
        0,
        2217,
        2219,
        5,
        70,
        0,
        0,
        2218,
        2217,
        1,
        0,
        0,
        0,
        2218,
        2219,
        1,
        0,
        0,
        0,
        2219,
        349,
        1,
        0,
        0,
        0,
        2220,
        2222,
        3,
        356,
        178,
        0,
        2221,
        2220,
        1,
        0,
        0,
        0,
        2222,
        2225,
        1,
        0,
        0,
        0,
        2223,
        2221,
        1,
        0,
        0,
        0,
        2223,
        2224,
        1,
        0,
        0,
        0,
        2224,
        2227,
        1,
        0,
        0,
        0,
        2225,
        2223,
        1,
        0,
        0,
        0,
        2226,
        2228,
        3,
        358,
        179,
        0,
        2227,
        2226,
        1,
        0,
        0,
        0,
        2228,
        2229,
        1,
        0,
        0,
        0,
        2229,
        2227,
        1,
        0,
        0,
        0,
        2229,
        2230,
        1,
        0,
        0,
        0,
        2230,
        2231,
        1,
        0,
        0,
        0,
        2231,
        2233,
        5,
        84,
        0,
        0,
        2232,
        2234,
        5,
        70,
        0,
        0,
        2233,
        2232,
        1,
        0,
        0,
        0,
        2233,
        2234,
        1,
        0,
        0,
        0,
        2234,
        351,
        1,
        0,
        0,
        0,
        2235,
        2236,
        5,
        67,
        0,
        0,
        2236,
        2240,
        3,
        356,
        178,
        0,
        2237,
        2239,
        3,
        360,
        180,
        0,
        2238,
        2237,
        1,
        0,
        0,
        0,
        2239,
        2242,
        1,
        0,
        0,
        0,
        2240,
        2238,
        1,
        0,
        0,
        0,
        2240,
        2241,
        1,
        0,
        0,
        0,
        2241,
        2244,
        1,
        0,
        0,
        0,
        2242,
        2240,
        1,
        0,
        0,
        0,
        2243,
        2245,
        3,
        354,
        177,
        0,
        2244,
        2243,
        1,
        0,
        0,
        0,
        2244,
        2245,
        1,
        0,
        0,
        0,
        2245,
        2258,
        1,
        0,
        0,
        0,
        2246,
        2247,
        5,
        67,
        0,
        0,
        2247,
        2249,
        5,
        70,
        0,
        0,
        2248,
        2250,
        3,
        360,
        180,
        0,
        2249,
        2248,
        1,
        0,
        0,
        0,
        2250,
        2251,
        1,
        0,
        0,
        0,
        2251,
        2249,
        1,
        0,
        0,
        0,
        2251,
        2252,
        1,
        0,
        0,
        0,
        2252,
        2254,
        1,
        0,
        0,
        0,
        2253,
        2255,
        3,
        354,
        177,
        0,
        2254,
        2253,
        1,
        0,
        0,
        0,
        2254,
        2255,
        1,
        0,
        0,
        0,
        2255,
        2258,
        1,
        0,
        0,
        0,
        2256,
        2258,
        3,
        354,
        177,
        0,
        2257,
        2235,
        1,
        0,
        0,
        0,
        2257,
        2246,
        1,
        0,
        0,
        0,
        2257,
        2256,
        1,
        0,
        0,
        0,
        2258,
        353,
        1,
        0,
        0,
        0,
        2259,
        2260,
        5,
        73,
        0,
        0,
        2260,
        2261,
        3,
        356,
        178,
        0,
        2261,
        355,
        1,
        0,
        0,
        0,
        2262,
        2264,
        3,
        362,
        181,
        0,
        2263,
        2265,
        5,
        70,
        0,
        0,
        2264,
        2263,
        1,
        0,
        0,
        0,
        2264,
        2265,
        1,
        0,
        0,
        0,
        2265,
        357,
        1,
        0,
        0,
        0,
        2266,
        2267,
        3,
        362,
        181,
        0,
        2267,
        2269,
        3,
        160,
        80,
        0,
        2268,
        2270,
        5,
        70,
        0,
        0,
        2269,
        2268,
        1,
        0,
        0,
        0,
        2269,
        2270,
        1,
        0,
        0,
        0,
        2270,
        359,
        1,
        0,
        0,
        0,
        2271,
        2273,
        3,
        362,
        181,
        0,
        2272,
        2274,
        3,
        160,
        80,
        0,
        2273,
        2272,
        1,
        0,
        0,
        0,
        2273,
        2274,
        1,
        0,
        0,
        0,
        2274,
        2276,
        1,
        0,
        0,
        0,
        2275,
        2277,
        5,
        70,
        0,
        0,
        2276,
        2275,
        1,
        0,
        0,
        0,
        2276,
        2277,
        1,
        0,
        0,
        0,
        2277,
        361,
        1,
        0,
        0,
        0,
        2278,
        2279,
        5,
        52,
        0,
        0,
        2279,
        363,
        1,
        0,
        0,
        0,
        2280,
        2283,
        3,
        366,
        183,
        0,
        2281,
        2283,
        5,
        117,
        0,
        0,
        2282,
        2280,
        1,
        0,
        0,
        0,
        2282,
        2281,
        1,
        0,
        0,
        0,
        2283,
        365,
        1,
        0,
        0,
        0,
        2284,
        2287,
        5,
        88,
        0,
        0,
        2285,
        2288,
        3,
        270,
        135,
        0,
        2286,
        2288,
        3,
        272,
        136,
        0,
        2287,
        2285,
        1,
        0,
        0,
        0,
        2287,
        2286,
        1,
        0,
        0,
        0,
        2288,
        2290,
        1,
        0,
        0,
        0,
        2289,
        2291,
        5,
        74,
        0,
        0,
        2290,
        2289,
        1,
        0,
        0,
        0,
        2290,
        2291,
        1,
        0,
        0,
        0,
        2291,
        2293,
        1,
        0,
        0,
        0,
        2292,
        2294,
        3,
        368,
        184,
        0,
        2293,
        2292,
        1,
        0,
        0,
        0,
        2293,
        2294,
        1,
        0,
        0,
        0,
        2294,
        2296,
        1,
        0,
        0,
        0,
        2295,
        2297,
        3,
        370,
        185,
        0,
        2296,
        2295,
        1,
        0,
        0,
        0,
        2296,
        2297,
        1,
        0,
        0,
        0,
        2297,
        2298,
        1,
        0,
        0,
        0,
        2298,
        2299,
        5,
        89,
        0,
        0,
        2299,
        367,
        1,
        0,
        0,
        0,
        2300,
        2304,
        1,
        0,
        0,
        0,
        2301,
        2302,
        5,
        113,
        0,
        0,
        2302,
        2304,
        5,
        52,
        0,
        0,
        2303,
        2300,
        1,
        0,
        0,
        0,
        2303,
        2301,
        1,
        0,
        0,
        0,
        2304,
        369,
        1,
        0,
        0,
        0,
        2305,
        2309,
        5,
        71,
        0,
        0,
        2306,
        2308,
        3,
        372,
        186,
        0,
        2307,
        2306,
        1,
        0,
        0,
        0,
        2308,
        2311,
        1,
        0,
        0,
        0,
        2309,
        2307,
        1,
        0,
        0,
        0,
        2309,
        2310,
        1,
        0,
        0,
        0,
        2310,
        371,
        1,
        0,
        0,
        0,
        2311,
        2309,
        1,
        0,
        0,
        0,
        2312,
        2315,
        5,
        117,
        0,
        0,
        2313,
        2315,
        3,
        366,
        183,
        0,
        2314,
        2312,
        1,
        0,
        0,
        0,
        2314,
        2313,
        1,
        0,
        0,
        0,
        2315,
        373,
        1,
        0,
        0,
        0,
        2316,
        2320,
        5,
        118,
        0,
        0,
        2317,
        2319,
        3,
        364,
        182,
        0,
        2318,
        2317,
        1,
        0,
        0,
        0,
        2319,
        2322,
        1,
        0,
        0,
        0,
        2320,
        2318,
        1,
        0,
        0,
        0,
        2320,
        2321,
        1,
        0,
        0,
        0,
        2321,
        2323,
        1,
        0,
        0,
        0,
        2322,
        2320,
        1,
        0,
        0,
        0,
        2323,
        2324,
        5,
        119,
        0,
        0,
        2324,
        375,
        1,
        0,
        0,
        0,
        2325,
        2326,
        5,
        3,
        0,
        0,
        2326,
        377,
        1,
        0,
        0,
        0,
        2327,
        2330,
        3,
        374,
        187,
        0,
        2328,
        2330,
        3,
        376,
        188,
        0,
        2329,
        2327,
        1,
        0,
        0,
        0,
        2329,
        2328,
        1,
        0,
        0,
        0,
        2330,
        2331,
        1,
        0,
        0,
        0,
        2331,
        2329,
        1,
        0,
        0,
        0,
        2331,
        2332,
        1,
        0,
        0,
        0,
        2332,
        379,
        1,
        0,
        0,
        0,
        2333,
        2335,
        5,
        75,
        0,
        0,
        2334,
        2336,
        3,
        276,
        138,
        0,
        2335,
        2334,
        1,
        0,
        0,
        0,
        2335,
        2336,
        1,
        0,
        0,
        0,
        2336,
        2337,
        1,
        0,
        0,
        0,
        2337,
        2338,
        5,
        76,
        0,
        0,
        2338,
        381,
        1,
        0,
        0,
        0,
        2339,
        2344,
        5,
        68,
        0,
        0,
        2340,
        2341,
        3,
        278,
        139,
        0,
        2341,
        2342,
        5,
        70,
        0,
        0,
        2342,
        2343,
        3,
        276,
        138,
        0,
        2343,
        2345,
        1,
        0,
        0,
        0,
        2344,
        2340,
        1,
        0,
        0,
        0,
        2344,
        2345,
        1,
        0,
        0,
        0,
        2345,
        2346,
        1,
        0,
        0,
        0,
        2346,
        2347,
        5,
        69,
        0,
        0,
        2347,
        383,
        1,
        0,
        0,
        0,
        2348,
        2349,
        5,
        88,
        0,
        0,
        2349,
        2350,
        3,
        276,
        138,
        0,
        2350,
        2351,
        5,
        89,
        0,
        0,
        2351,
        385,
        1,
        0,
        0,
        0,
        2352,
        2354,
        5,
        88,
        0,
        0,
        2353,
        2355,
        3,
        388,
        194,
        0,
        2354,
        2353,
        1,
        0,
        0,
        0,
        2354,
        2355,
        1,
        0,
        0,
        0,
        2355,
        2356,
        1,
        0,
        0,
        0,
        2356,
        2357,
        5,
        89,
        0,
        0,
        2357,
        387,
        1,
        0,
        0,
        0,
        2358,
        2363,
        3,
        390,
        195,
        0,
        2359,
        2360,
        5,
        70,
        0,
        0,
        2360,
        2362,
        3,
        390,
        195,
        0,
        2361,
        2359,
        1,
        0,
        0,
        0,
        2362,
        2365,
        1,
        0,
        0,
        0,
        2363,
        2361,
        1,
        0,
        0,
        0,
        2363,
        2364,
        1,
        0,
        0,
        0,
        2364,
        2367,
        1,
        0,
        0,
        0,
        2365,
        2363,
        1,
        0,
        0,
        0,
        2366,
        2368,
        5,
        70,
        0,
        0,
        2367,
        2366,
        1,
        0,
        0,
        0,
        2367,
        2368,
        1,
        0,
        0,
        0,
        2368,
        389,
        1,
        0,
        0,
        0,
        2369,
        2370,
        5,
        73,
        0,
        0,
        2370,
        2373,
        3,
        314,
        157,
        0,
        2371,
        2373,
        3,
        392,
        196,
        0,
        2372,
        2369,
        1,
        0,
        0,
        0,
        2372,
        2371,
        1,
        0,
        0,
        0,
        2373,
        391,
        1,
        0,
        0,
        0,
        2374,
        2375,
        3,
        268,
        134,
        0,
        2375,
        2376,
        5,
        71,
        0,
        0,
        2376,
        2377,
        3,
        268,
        134,
        0,
        2377,
        393,
        1,
        0,
        0,
        0,
        2378,
        2380,
        3,
        396,
        198,
        0,
        2379,
        2378,
        1,
        0,
        0,
        0,
        2380,
        2381,
        1,
        0,
        0,
        0,
        2381,
        2379,
        1,
        0,
        0,
        0,
        2381,
        2382,
        1,
        0,
        0,
        0,
        2382,
        395,
        1,
        0,
        0,
        0,
        2383,
        2385,
        5,
        12,
        0,
        0,
        2384,
        2383,
        1,
        0,
        0,
        0,
        2384,
        2385,
        1,
        0,
        0,
        0,
        2385,
        2386,
        1,
        0,
        0,
        0,
        2386,
        2387,
        5,
        25,
        0,
        0,
        2387,
        2388,
        3,
        420,
        210,
        0,
        2388,
        2389,
        5,
        30,
        0,
        0,
        2389,
        2394,
        3,
        284,
        142,
        0,
        2390,
        2391,
        5,
        28,
        0,
        0,
        2391,
        2393,
        3,
        284,
        142,
        0,
        2392,
        2390,
        1,
        0,
        0,
        0,
        2393,
        2396,
        1,
        0,
        0,
        0,
        2394,
        2392,
        1,
        0,
        0,
        0,
        2394,
        2395,
        1,
        0,
        0,
        0,
        2395,
        397,
        1,
        0,
        0,
        0,
        2396,
        2394,
        1,
        0,
        0,
        0,
        2397,
        2398,
        5,
        75,
        0,
        0,
        2398,
        2399,
        3,
        282,
        141,
        0,
        2399,
        2400,
        3,
        394,
        197,
        0,
        2400,
        2401,
        5,
        76,
        0,
        0,
        2401,
        399,
        1,
        0,
        0,
        0,
        2402,
        2403,
        5,
        88,
        0,
        0,
        2403,
        2404,
        3,
        282,
        141,
        0,
        2404,
        2405,
        3,
        394,
        197,
        0,
        2405,
        2406,
        5,
        89,
        0,
        0,
        2406,
        401,
        1,
        0,
        0,
        0,
        2407,
        2410,
        5,
        68,
        0,
        0,
        2408,
        2411,
        3,
        280,
        140,
        0,
        2409,
        2411,
        3,
        268,
        134,
        0,
        2410,
        2408,
        1,
        0,
        0,
        0,
        2410,
        2409,
        1,
        0,
        0,
        0,
        2411,
        2412,
        1,
        0,
        0,
        0,
        2412,
        2413,
        3,
        394,
        197,
        0,
        2413,
        2414,
        5,
        69,
        0,
        0,
        2414,
        403,
        1,
        0,
        0,
        0,
        2415,
        2416,
        5,
        88,
        0,
        0,
        2416,
        2417,
        3,
        392,
        196,
        0,
        2417,
        2418,
        3,
        394,
        197,
        0,
        2418,
        2419,
        5,
        89,
        0,
        0,
        2419,
        405,
        1,
        0,
        0,
        0,
        2420,
        2422,
        3,
        408,
        204,
        0,
        2421,
        2423,
        5,
        70,
        0,
        0,
        2422,
        2421,
        1,
        0,
        0,
        0,
        2422,
        2423,
        1,
        0,
        0,
        0,
        2423,
        407,
        1,
        0,
        0,
        0,
        2424,
        2429,
        3,
        410,
        205,
        0,
        2425,
        2426,
        5,
        70,
        0,
        0,
        2426,
        2428,
        3,
        410,
        205,
        0,
        2427,
        2425,
        1,
        0,
        0,
        0,
        2428,
        2431,
        1,
        0,
        0,
        0,
        2429,
        2427,
        1,
        0,
        0,
        0,
        2429,
        2430,
        1,
        0,
        0,
        0,
        2430,
        2434,
        1,
        0,
        0,
        0,
        2431,
        2429,
        1,
        0,
        0,
        0,
        2432,
        2433,
        5,
        70,
        0,
        0,
        2433,
        2435,
        3,
        412,
        206,
        0,
        2434,
        2432,
        1,
        0,
        0,
        0,
        2434,
        2435,
        1,
        0,
        0,
        0,
        2435,
        2438,
        1,
        0,
        0,
        0,
        2436,
        2438,
        3,
        412,
        206,
        0,
        2437,
        2424,
        1,
        0,
        0,
        0,
        2437,
        2436,
        1,
        0,
        0,
        0,
        2438,
        409,
        1,
        0,
        0,
        0,
        2439,
        2443,
        3,
        414,
        207,
        0,
        2440,
        2443,
        3,
        280,
        140,
        0,
        2441,
        2443,
        3,
        268,
        134,
        0,
        2442,
        2439,
        1,
        0,
        0,
        0,
        2442,
        2440,
        1,
        0,
        0,
        0,
        2442,
        2441,
        1,
        0,
        0,
        0,
        2443,
        411,
        1,
        0,
        0,
        0,
        2444,
        2449,
        3,
        416,
        208,
        0,
        2445,
        2446,
        5,
        70,
        0,
        0,
        2446,
        2448,
        3,
        416,
        208,
        0,
        2447,
        2445,
        1,
        0,
        0,
        0,
        2448,
        2451,
        1,
        0,
        0,
        0,
        2449,
        2447,
        1,
        0,
        0,
        0,
        2449,
        2450,
        1,
        0,
        0,
        0,
        2450,
        2452,
        1,
        0,
        0,
        0,
        2451,
        2449,
        1,
        0,
        0,
        0,
        2452,
        2453,
        5,
        70,
        0,
        0,
        2453,
        2458,
        3,
        418,
        209,
        0,
        2454,
        2455,
        5,
        70,
        0,
        0,
        2455,
        2457,
        3,
        418,
        209,
        0,
        2456,
        2454,
        1,
        0,
        0,
        0,
        2457,
        2460,
        1,
        0,
        0,
        0,
        2458,
        2456,
        1,
        0,
        0,
        0,
        2458,
        2459,
        1,
        0,
        0,
        0,
        2459,
        2478,
        1,
        0,
        0,
        0,
        2460,
        2458,
        1,
        0,
        0,
        0,
        2461,
        2466,
        3,
        416,
        208,
        0,
        2462,
        2463,
        5,
        70,
        0,
        0,
        2463,
        2465,
        3,
        416,
        208,
        0,
        2464,
        2462,
        1,
        0,
        0,
        0,
        2465,
        2468,
        1,
        0,
        0,
        0,
        2466,
        2464,
        1,
        0,
        0,
        0,
        2466,
        2467,
        1,
        0,
        0,
        0,
        2467,
        2478,
        1,
        0,
        0,
        0,
        2468,
        2466,
        1,
        0,
        0,
        0,
        2469,
        2474,
        3,
        418,
        209,
        0,
        2470,
        2471,
        5,
        70,
        0,
        0,
        2471,
        2473,
        3,
        418,
        209,
        0,
        2472,
        2470,
        1,
        0,
        0,
        0,
        2473,
        2476,
        1,
        0,
        0,
        0,
        2474,
        2472,
        1,
        0,
        0,
        0,
        2474,
        2475,
        1,
        0,
        0,
        0,
        2475,
        2478,
        1,
        0,
        0,
        0,
        2476,
        2474,
        1,
        0,
        0,
        0,
        2477,
        2444,
        1,
        0,
        0,
        0,
        2477,
        2461,
        1,
        0,
        0,
        0,
        2477,
        2469,
        1,
        0,
        0,
        0,
        2478,
        413,
        1,
        0,
        0,
        0,
        2479,
        2480,
        5,
        67,
        0,
        0,
        2480,
        2481,
        3,
        268,
        134,
        0,
        2481,
        415,
        1,
        0,
        0,
        0,
        2482,
        2483,
        5,
        52,
        0,
        0,
        2483,
        2484,
        5,
        74,
        0,
        0,
        2484,
        2487,
        3,
        268,
        134,
        0,
        2485,
        2487,
        3,
        414,
        207,
        0,
        2486,
        2482,
        1,
        0,
        0,
        0,
        2486,
        2485,
        1,
        0,
        0,
        0,
        2487,
        417,
        1,
        0,
        0,
        0,
        2488,
        2489,
        5,
        52,
        0,
        0,
        2489,
        2490,
        5,
        74,
        0,
        0,
        2490,
        2494,
        3,
        268,
        134,
        0,
        2491,
        2492,
        5,
        73,
        0,
        0,
        2492,
        2494,
        3,
        268,
        134,
        0,
        2493,
        2488,
        1,
        0,
        0,
        0,
        2493,
        2491,
        1,
        0,
        0,
        0,
        2494,
        419,
        1,
        0,
        0,
        0,
        2495,
        2500,
        3,
        426,
        213,
        0,
        2496,
        2497,
        5,
        70,
        0,
        0,
        2497,
        2499,
        3,
        426,
        213,
        0,
        2498,
        2496,
        1,
        0,
        0,
        0,
        2499,
        2502,
        1,
        0,
        0,
        0,
        2500,
        2498,
        1,
        0,
        0,
        0,
        2500,
        2501,
        1,
        0,
        0,
        0,
        2501,
        2504,
        1,
        0,
        0,
        0,
        2502,
        2500,
        1,
        0,
        0,
        0,
        2503,
        2505,
        5,
        70,
        0,
        0,
        2504,
        2503,
        1,
        0,
        0,
        0,
        2504,
        2505,
        1,
        0,
        0,
        0,
        2505,
        421,
        1,
        0,
        0,
        0,
        2506,
        2511,
        3,
        426,
        213,
        0,
        2507,
        2508,
        5,
        70,
        0,
        0,
        2508,
        2510,
        3,
        426,
        213,
        0,
        2509,
        2507,
        1,
        0,
        0,
        0,
        2510,
        2513,
        1,
        0,
        0,
        0,
        2511,
        2509,
        1,
        0,
        0,
        0,
        2511,
        2512,
        1,
        0,
        0,
        0,
        2512,
        2515,
        1,
        0,
        0,
        0,
        2513,
        2511,
        1,
        0,
        0,
        0,
        2514,
        2516,
        5,
        70,
        0,
        0,
        2515,
        2514,
        1,
        0,
        0,
        0,
        2515,
        2516,
        1,
        0,
        0,
        0,
        2516,
        423,
        1,
        0,
        0,
        0,
        2517,
        2520,
        3,
        426,
        213,
        0,
        2518,
        2519,
        5,
        70,
        0,
        0,
        2519,
        2521,
        3,
        426,
        213,
        0,
        2520,
        2518,
        1,
        0,
        0,
        0,
        2521,
        2522,
        1,
        0,
        0,
        0,
        2522,
        2520,
        1,
        0,
        0,
        0,
        2522,
        2523,
        1,
        0,
        0,
        0,
        2523,
        2525,
        1,
        0,
        0,
        0,
        2524,
        2526,
        5,
        70,
        0,
        0,
        2525,
        2524,
        1,
        0,
        0,
        0,
        2525,
        2526,
        1,
        0,
        0,
        0,
        2526,
        2531,
        1,
        0,
        0,
        0,
        2527,
        2528,
        3,
        426,
        213,
        0,
        2528,
        2529,
        5,
        70,
        0,
        0,
        2529,
        2531,
        1,
        0,
        0,
        0,
        2530,
        2517,
        1,
        0,
        0,
        0,
        2530,
        2527,
        1,
        0,
        0,
        0,
        2531,
        425,
        1,
        0,
        0,
        0,
        2532,
        2533,
        5,
        67,
        0,
        0,
        2533,
        2536,
        3,
        426,
        213,
        0,
        2534,
        2536,
        3,
        428,
        214,
        0,
        2535,
        2532,
        1,
        0,
        0,
        0,
        2535,
        2534,
        1,
        0,
        0,
        0,
        2536,
        427,
        1,
        0,
        0,
        0,
        2537,
        2538,
        3,
        436,
        218,
        0,
        2538,
        2539,
        5,
        64,
        0,
        0,
        2539,
        2540,
        5,
        52,
        0,
        0,
        2540,
        2548,
        1,
        0,
        0,
        0,
        2541,
        2542,
        3,
        436,
        218,
        0,
        2542,
        2543,
        5,
        75,
        0,
        0,
        2543,
        2544,
        3,
        334,
        167,
        0,
        2544,
        2545,
        5,
        76,
        0,
        0,
        2545,
        2548,
        1,
        0,
        0,
        0,
        2546,
        2548,
        3,
        430,
        215,
        0,
        2547,
        2537,
        1,
        0,
        0,
        0,
        2547,
        2541,
        1,
        0,
        0,
        0,
        2547,
        2546,
        1,
        0,
        0,
        0,
        2548,
        429,
        1,
        0,
        0,
        0,
        2549,
        2565,
        5,
        52,
        0,
        0,
        2550,
        2551,
        5,
        68,
        0,
        0,
        2551,
        2552,
        3,
        428,
        214,
        0,
        2552,
        2553,
        5,
        69,
        0,
        0,
        2553,
        2565,
        1,
        0,
        0,
        0,
        2554,
        2556,
        5,
        68,
        0,
        0,
        2555,
        2557,
        3,
        424,
        212,
        0,
        2556,
        2555,
        1,
        0,
        0,
        0,
        2556,
        2557,
        1,
        0,
        0,
        0,
        2557,
        2558,
        1,
        0,
        0,
        0,
        2558,
        2565,
        5,
        69,
        0,
        0,
        2559,
        2561,
        5,
        75,
        0,
        0,
        2560,
        2562,
        3,
        422,
        211,
        0,
        2561,
        2560,
        1,
        0,
        0,
        0,
        2561,
        2562,
        1,
        0,
        0,
        0,
        2562,
        2563,
        1,
        0,
        0,
        0,
        2563,
        2565,
        5,
        76,
        0,
        0,
        2564,
        2549,
        1,
        0,
        0,
        0,
        2564,
        2550,
        1,
        0,
        0,
        0,
        2564,
        2554,
        1,
        0,
        0,
        0,
        2564,
        2559,
        1,
        0,
        0,
        0,
        2565,
        431,
        1,
        0,
        0,
        0,
        2566,
        2573,
        3,
        434,
        217,
        0,
        2567,
        2573,
        5,
        52,
        0,
        0,
        2568,
        2569,
        5,
        68,
        0,
        0,
        2569,
        2570,
        3,
        432,
        216,
        0,
        2570,
        2571,
        5,
        69,
        0,
        0,
        2571,
        2573,
        1,
        0,
        0,
        0,
        2572,
        2566,
        1,
        0,
        0,
        0,
        2572,
        2567,
        1,
        0,
        0,
        0,
        2572,
        2568,
        1,
        0,
        0,
        0,
        2573,
        433,
        1,
        0,
        0,
        0,
        2574,
        2575,
        3,
        436,
        218,
        0,
        2575,
        2576,
        5,
        64,
        0,
        0,
        2576,
        2577,
        5,
        52,
        0,
        0,
        2577,
        2584,
        1,
        0,
        0,
        0,
        2578,
        2579,
        3,
        436,
        218,
        0,
        2579,
        2580,
        5,
        75,
        0,
        0,
        2580,
        2581,
        3,
        334,
        167,
        0,
        2581,
        2582,
        5,
        76,
        0,
        0,
        2582,
        2584,
        1,
        0,
        0,
        0,
        2583,
        2574,
        1,
        0,
        0,
        0,
        2583,
        2578,
        1,
        0,
        0,
        0,
        2584,
        435,
        1,
        0,
        0,
        0,
        2585,
        2586,
        6,
        218,
        -1,
        0,
        2586,
        2587,
        3,
        338,
        169,
        0,
        2587,
        2606,
        1,
        0,
        0,
        0,
        2588,
        2589,
        10,
        5,
        0,
        0,
        2589,
        2590,
        5,
        64,
        0,
        0,
        2590,
        2605,
        5,
        52,
        0,
        0,
        2591,
        2592,
        10,
        4,
        0,
        0,
        2592,
        2593,
        5,
        75,
        0,
        0,
        2593,
        2594,
        3,
        334,
        167,
        0,
        2594,
        2595,
        5,
        76,
        0,
        0,
        2595,
        2605,
        1,
        0,
        0,
        0,
        2596,
        2597,
        10,
        3,
        0,
        0,
        2597,
        2605,
        3,
        402,
        201,
        0,
        2598,
        2599,
        10,
        2,
        0,
        0,
        2599,
        2601,
        5,
        68,
        0,
        0,
        2600,
        2602,
        3,
        406,
        203,
        0,
        2601,
        2600,
        1,
        0,
        0,
        0,
        2601,
        2602,
        1,
        0,
        0,
        0,
        2602,
        2603,
        1,
        0,
        0,
        0,
        2603,
        2605,
        5,
        69,
        0,
        0,
        2604,
        2588,
        1,
        0,
        0,
        0,
        2604,
        2591,
        1,
        0,
        0,
        0,
        2604,
        2596,
        1,
        0,
        0,
        0,
        2604,
        2598,
        1,
        0,
        0,
        0,
        2605,
        2608,
        1,
        0,
        0,
        0,
        2606,
        2604,
        1,
        0,
        0,
        0,
        2606,
        2607,
        1,
        0,
        0,
        0,
        2607,
        437,
        1,
        0,
        0,
        0,
        2608,
        2606,
        1,
        0,
        0,
        0,
        2609,
        2614,
        3,
        440,
        220,
        0,
        2610,
        2611,
        5,
        70,
        0,
        0,
        2611,
        2613,
        3,
        440,
        220,
        0,
        2612,
        2610,
        1,
        0,
        0,
        0,
        2613,
        2616,
        1,
        0,
        0,
        0,
        2614,
        2612,
        1,
        0,
        0,
        0,
        2614,
        2615,
        1,
        0,
        0,
        0,
        2615,
        2618,
        1,
        0,
        0,
        0,
        2616,
        2614,
        1,
        0,
        0,
        0,
        2617,
        2619,
        5,
        70,
        0,
        0,
        2618,
        2617,
        1,
        0,
        0,
        0,
        2618,
        2619,
        1,
        0,
        0,
        0,
        2619,
        439,
        1,
        0,
        0,
        0,
        2620,
        2621,
        3,
        436,
        218,
        0,
        2621,
        2622,
        5,
        64,
        0,
        0,
        2622,
        2623,
        5,
        52,
        0,
        0,
        2623,
        2631,
        1,
        0,
        0,
        0,
        2624,
        2625,
        3,
        436,
        218,
        0,
        2625,
        2626,
        5,
        75,
        0,
        0,
        2626,
        2627,
        3,
        334,
        167,
        0,
        2627,
        2628,
        5,
        76,
        0,
        0,
        2628,
        2631,
        1,
        0,
        0,
        0,
        2629,
        2631,
        3,
        442,
        221,
        0,
        2630,
        2620,
        1,
        0,
        0,
        0,
        2630,
        2624,
        1,
        0,
        0,
        0,
        2630,
        2629,
        1,
        0,
        0,
        0,
        2631,
        441,
        1,
        0,
        0,
        0,
        2632,
        2644,
        5,
        52,
        0,
        0,
        2633,
        2635,
        5,
        68,
        0,
        0,
        2634,
        2636,
        3,
        438,
        219,
        0,
        2635,
        2634,
        1,
        0,
        0,
        0,
        2635,
        2636,
        1,
        0,
        0,
        0,
        2636,
        2637,
        1,
        0,
        0,
        0,
        2637,
        2644,
        5,
        69,
        0,
        0,
        2638,
        2640,
        5,
        75,
        0,
        0,
        2639,
        2641,
        3,
        438,
        219,
        0,
        2640,
        2639,
        1,
        0,
        0,
        0,
        2640,
        2641,
        1,
        0,
        0,
        0,
        2641,
        2642,
        1,
        0,
        0,
        0,
        2642,
        2644,
        5,
        76,
        0,
        0,
        2643,
        2632,
        1,
        0,
        0,
        0,
        2643,
        2633,
        1,
        0,
        0,
        0,
        2643,
        2638,
        1,
        0,
        0,
        0,
        2644,
        443,
        1,
        0,
        0,
        0,
        2645,
        2650,
        3,
        268,
        134,
        0,
        2646,
        2647,
        5,
        70,
        0,
        0,
        2647,
        2649,
        3,
        268,
        134,
        0,
        2648,
        2646,
        1,
        0,
        0,
        0,
        2649,
        2652,
        1,
        0,
        0,
        0,
        2650,
        2648,
        1,
        0,
        0,
        0,
        2650,
        2651,
        1,
        0,
        0,
        0,
        2651,
        2653,
        1,
        0,
        0,
        0,
        2652,
        2650,
        1,
        0,
        0,
        0,
        2653,
        2654,
        5,
        70,
        0,
        0,
        2654,
        2655,
        5,
        67,
        0,
        0,
        2655,
        2656,
        3,
        268,
        134,
        0,
        2656,
        2657,
        5,
        70,
        0,
        0,
        2657,
        2658,
        5,
        73,
        0,
        0,
        2658,
        2659,
        3,
        268,
        134,
        0,
        2659,
        2703,
        1,
        0,
        0,
        0,
        2660,
        2665,
        3,
        268,
        134,
        0,
        2661,
        2662,
        5,
        70,
        0,
        0,
        2662,
        2664,
        3,
        268,
        134,
        0,
        2663,
        2661,
        1,
        0,
        0,
        0,
        2664,
        2667,
        1,
        0,
        0,
        0,
        2665,
        2663,
        1,
        0,
        0,
        0,
        2665,
        2666,
        1,
        0,
        0,
        0,
        2666,
        2668,
        1,
        0,
        0,
        0,
        2667,
        2665,
        1,
        0,
        0,
        0,
        2668,
        2669,
        5,
        70,
        0,
        0,
        2669,
        2670,
        5,
        67,
        0,
        0,
        2670,
        2671,
        3,
        268,
        134,
        0,
        2671,
        2703,
        1,
        0,
        0,
        0,
        2672,
        2677,
        3,
        268,
        134,
        0,
        2673,
        2674,
        5,
        70,
        0,
        0,
        2674,
        2676,
        3,
        268,
        134,
        0,
        2675,
        2673,
        1,
        0,
        0,
        0,
        2676,
        2679,
        1,
        0,
        0,
        0,
        2677,
        2675,
        1,
        0,
        0,
        0,
        2677,
        2678,
        1,
        0,
        0,
        0,
        2678,
        2680,
        1,
        0,
        0,
        0,
        2679,
        2677,
        1,
        0,
        0,
        0,
        2680,
        2681,
        5,
        70,
        0,
        0,
        2681,
        2682,
        5,
        73,
        0,
        0,
        2682,
        2683,
        3,
        268,
        134,
        0,
        2683,
        2703,
        1,
        0,
        0,
        0,
        2684,
        2685,
        5,
        67,
        0,
        0,
        2685,
        2686,
        3,
        268,
        134,
        0,
        2686,
        2687,
        5,
        70,
        0,
        0,
        2687,
        2688,
        5,
        73,
        0,
        0,
        2688,
        2689,
        3,
        268,
        134,
        0,
        2689,
        2703,
        1,
        0,
        0,
        0,
        2690,
        2691,
        5,
        67,
        0,
        0,
        2691,
        2703,
        3,
        268,
        134,
        0,
        2692,
        2693,
        5,
        73,
        0,
        0,
        2693,
        2703,
        3,
        268,
        134,
        0,
        2694,
        2699,
        3,
        268,
        134,
        0,
        2695,
        2696,
        5,
        70,
        0,
        0,
        2696,
        2698,
        3,
        268,
        134,
        0,
        2697,
        2695,
        1,
        0,
        0,
        0,
        2698,
        2701,
        1,
        0,
        0,
        0,
        2699,
        2697,
        1,
        0,
        0,
        0,
        2699,
        2700,
        1,
        0,
        0,
        0,
        2700,
        2703,
        1,
        0,
        0,
        0,
        2701,
        2699,
        1,
        0,
        0,
        0,
        2702,
        2645,
        1,
        0,
        0,
        0,
        2702,
        2660,
        1,
        0,
        0,
        0,
        2702,
        2672,
        1,
        0,
        0,
        0,
        2702,
        2684,
        1,
        0,
        0,
        0,
        2702,
        2690,
        1,
        0,
        0,
        0,
        2702,
        2692,
        1,
        0,
        0,
        0,
        2702,
        2694,
        1,
        0,
        0,
        0,
        2703,
        445,
        1,
        0,
        0,
        0,
        2704,
        2705,
        5,
        114,
        0,
        0,
        2705,
        447,
        1,
        0,
        0,
        0,
        337,
        454,
        461,
        466,
        472,
        479,
        488,
        498,
        502,
        509,
        516,
        524,
        543,
        547,
        551,
        556,
        567,
        572,
        577,
        598,
        603,
        611,
        616,
        622,
        640,
        645,
        651,
        659,
        665,
        671,
        679,
        685,
        691,
        700,
        705,
        720,
        727,
        738,
        740,
        754,
        761,
        768,
        772,
        778,
        785,
        789,
        793,
        797,
        800,
        804,
        808,
        810,
        814,
        819,
        826,
        830,
        833,
        841,
        845,
        849,
        857,
        861,
        865,
        867,
        871,
        878,
        885,
        889,
        894,
        899,
        903,
        919,
        929,
        936,
        943,
        949,
        956,
        960,
        966,
        968,
        972,
        978,
        984,
        987,
        995,
        1004,
        1016,
        1020,
        1029,
        1040,
        1044,
        1049,
        1055,
        1062,
        1068,
        1075,
        1081,
        1091,
        1100,
        1108,
        1111,
        1118,
        1122,
        1125,
        1131,
        1136,
        1141,
        1145,
        1150,
        1154,
        1164,
        1170,
        1174,
        1180,
        1184,
        1189,
        1194,
        1198,
        1203,
        1206,
        1209,
        1214,
        1218,
        1223,
        1229,
        1233,
        1240,
        1244,
        1251,
        1255,
        1262,
        1265,
        1268,
        1277,
        1283,
        1292,
        1296,
        1302,
        1304,
        1308,
        1333,
        1335,
        1348,
        1350,
        1361,
        1370,
        1380,
        1382,
        1391,
        1395,
        1407,
        1421,
        1425,
        1438,
        1444,
        1451,
        1464,
        1467,
        1470,
        1478,
        1481,
        1484,
        1486,
        1492,
        1500,
        1507,
        1524,
        1531,
        1534,
        1539,
        1549,
        1553,
        1564,
        1575,
        1583,
        1591,
        1601,
        1606,
        1611,
        1637,
        1646,
        1651,
        1654,
        1659,
        1666,
        1670,
        1674,
        1680,
        1687,
        1696,
        1703,
        1707,
        1714,
        1719,
        1735,
        1743,
        1753,
        1757,
        1764,
        1772,
        1782,
        1796,
        1800,
        1804,
        1810,
        1820,
        1824,
        1834,
        1841,
        1843,
        1850,
        1854,
        1859,
        1866,
        1870,
        1875,
        1883,
        1890,
        1898,
        1904,
        1910,
        1923,
        1965,
        1976,
        1987,
        1999,
        2001,
        2013,
        2015,
        2036,
        2038,
        2048,
        2055,
        2060,
        2073,
        2081,
        2083,
        2088,
        2093,
        2097,
        2101,
        2104,
        2108,
        2112,
        2114,
        2117,
        2129,
        2133,
        2139,
        2142,
        2147,
        2153,
        2164,
        2170,
        2174,
        2180,
        2184,
        2189,
        2194,
        2198,
        2203,
        2206,
        2209,
        2214,
        2218,
        2223,
        2229,
        2233,
        2240,
        2244,
        2251,
        2254,
        2257,
        2264,
        2269,
        2273,
        2276,
        2282,
        2287,
        2290,
        2293,
        2296,
        2303,
        2309,
        2314,
        2320,
        2329,
        2331,
        2335,
        2344,
        2354,
        2363,
        2367,
        2372,
        2381,
        2384,
        2394,
        2410,
        2422,
        2429,
        2434,
        2437,
        2442,
        2449,
        2458,
        2466,
        2474,
        2477,
        2486,
        2493,
        2500,
        2504,
        2511,
        2515,
        2522,
        2525,
        2530,
        2535,
        2547,
        2556,
        2561,
        2564,
        2572,
        2583,
        2601,
        2604,
        2606,
        2614,
        2618,
        2630,
        2635,
        2640,
        2643,
        2650,
        2665,
        2677,
        2699,
        2702,
    ]


class FandangoParser(Parser):

    grammarFileName = "FandangoParser.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [DFA(ds, i) for i, ds in enumerate(atn.decisionToState)]

    sharedContextCache = PredictionContextCache()

    literalNames = [
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "'<py>'",
        "'</py>'",
        "<INVALID>",
        "'and'",
        "'as'",
        "'assert'",
        "'async'",
        "'await'",
        "'break'",
        "'case'",
        "'class'",
        "'continue'",
        "'def'",
        "'del'",
        "'elif'",
        "'else'",
        "'except'",
        "'False'",
        "'finally'",
        "'for'",
        "'from'",
        "'global'",
        "'if'",
        "'import'",
        "'in'",
        "'is'",
        "'lambda'",
        "'match'",
        "'None'",
        "'nonlocal'",
        "'not'",
        "'or'",
        "'pass'",
        "'raise'",
        "'return'",
        "'True'",
        "'try'",
        "'type'",
        "'_'",
        "'while'",
        "'where'",
        "'with'",
        "'yield'",
        "'forall'",
        "'exists'",
        "'fitness'",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "'::='",
        "'?'",
        "'\\'",
        "'.'",
        "'..'",
        "'...'",
        "'*'",
        "'('",
        "')'",
        "','",
        "':'",
        "';'",
        "'**'",
        "'='",
        "'['",
        "']'",
        "'|'",
        "'^'",
        "'&'",
        "'<<'",
        "'>>'",
        "'+'",
        "'-'",
        "'/'",
        "'%'",
        "'//'",
        "'~'",
        "'{'",
        "'}'",
        "'<'",
        "'>'",
        "'=='",
        "'>='",
        "'<='",
        "'<>'",
        "'!='",
        "'@'",
        "'->'",
        "'+='",
        "'-='",
        "'*='",
        "'@='",
        "'/='",
        "'%='",
        "'&='",
        "'|='",
        "'^='",
        "'<<='",
        "'>>='",
        "'**='",
        "'//='",
        "':='",
        "'!'",
    ]

    symbolicNames = [
        "<INVALID>",
        "INDENT",
        "DEDENT",
        "STRING",
        "NUMBER",
        "INTEGER",
        "PYTHON_START",
        "PYTHON_END",
        "NONTERMINAL",
        "AND",
        "AS",
        "ASSERT",
        "ASYNC",
        "AWAIT",
        "BREAK",
        "CASE",
        "CLASS",
        "CONTINUE",
        "DEF",
        "DEL",
        "ELIF",
        "ELSE",
        "EXCEPT",
        "FALSE",
        "FINALLY",
        "FOR",
        "FROM",
        "GLOBAL",
        "IF",
        "IMPORT",
        "IN",
        "IS",
        "LAMBDA",
        "MATCH",
        "NONE",
        "NONLOCAL",
        "NOT",
        "OR",
        "PASS",
        "RAISE",
        "RETURN",
        "TRUE",
        "TRY",
        "TYPE",
        "UNDERSCORE",
        "WHILE",
        "WHERE",
        "WITH",
        "YIELD",
        "FORALL",
        "EXISTS",
        "FITNESS",
        "NAME",
        "STRING_LITERAL",
        "BYTES_LITERAL",
        "DECIMAL_INTEGER",
        "OCT_INTEGER",
        "HEX_INTEGER",
        "BIN_INTEGER",
        "FLOAT_NUMBER",
        "IMAG_NUMBER",
        "GRAMMAR_ASSIGN",
        "QUESTION",
        "BACKSLASH",
        "DOT",
        "DOTDOT",
        "ELLIPSIS",
        "STAR",
        "OPEN_PAREN",
        "CLOSE_PAREN",
        "COMMA",
        "COLON",
        "SEMI_COLON",
        "POWER",
        "ASSIGN",
        "OPEN_BRACK",
        "CLOSE_BRACK",
        "OR_OP",
        "XOR",
        "AND_OP",
        "LEFT_SHIFT",
        "RIGHT_SHIFT",
        "ADD",
        "MINUS",
        "DIV",
        "MOD",
        "IDIV",
        "NOT_OP",
        "OPEN_BRACE",
        "CLOSE_BRACE",
        "LESS_THAN",
        "GREATER_THAN",
        "EQUALS",
        "GT_EQ",
        "LT_EQ",
        "NOT_EQ_1",
        "NOT_EQ_2",
        "AT",
        "ARROW",
        "ADD_ASSIGN",
        "SUB_ASSIGN",
        "MULT_ASSIGN",
        "AT_ASSIGN",
        "DIV_ASSIGN",
        "MOD_ASSIGN",
        "AND_ASSIGN",
        "OR_ASSIGN",
        "XOR_ASSIGN",
        "LEFT_SHIFT_ASSIGN",
        "RIGHT_SHIFT_ASSIGN",
        "POWER_ASSIGN",
        "IDIV_ASSIGN",
        "EXPR_ASSIGN",
        "EXCL",
        "NEWLINE",
        "SKIP_",
        "UNKNOWN_CHAR",
        "FSTRING_MIDDLE",
        "FSTRING_START",
        "FSTRING_END",
    ]

    RULE_fandango = 0
    RULE_program = 1
    RULE_statement = 2
    RULE_production = 3
    RULE_alternative = 4
    RULE_concatenation = 5
    RULE_operator = 6
    RULE_kleene = 7
    RULE_plus = 8
    RULE_option = 9
    RULE_repeat = 10
    RULE_symbol = 11
    RULE_char_set = 12
    RULE_constraint = 13
    RULE_implies = 14
    RULE_quantifier = 15
    RULE_formula_disjunction = 16
    RULE_formula_conjunction = 17
    RULE_formula_atom = 18
    RULE_formula_comparison = 19
    RULE_expr = 20
    RULE_selector_length = 21
    RULE_selector = 22
    RULE_selection = 23
    RULE_base_selection = 24
    RULE_rs_pairs = 25
    RULE_rs_pair = 26
    RULE_rs_slices = 27
    RULE_rs_slice = 28
    RULE_python = 29
    RULE_python_tag = 30
    RULE_python_file = 31
    RULE_interactive = 32
    RULE_eval = 33
    RULE_func_type = 34
    RULE_statements = 35
    RULE_stmt = 36
    RULE_statement_newline = 37
    RULE_simple_stmts = 38
    RULE_simple_stmt = 39
    RULE_compound_stmt = 40
    RULE_assignment = 41
    RULE_annotated_rhs = 42
    RULE_augassign = 43
    RULE_return_stmt = 44
    RULE_raise_stmt = 45
    RULE_global_stmt = 46
    RULE_nonlocal_stmt = 47
    RULE_del_stmt = 48
    RULE_yield_stmt = 49
    RULE_assert_stmt = 50
    RULE_import_stmt = 51
    RULE_import_name = 52
    RULE_import_from = 53
    RULE_import_from_targets = 54
    RULE_import_from_as_names = 55
    RULE_import_from_as_name = 56
    RULE_dotted_as_names = 57
    RULE_dotted_as_name = 58
    RULE_dotted_name = 59
    RULE_block = 60
    RULE_decorators = 61
    RULE_class_def = 62
    RULE_class_def_raw = 63
    RULE_function_def = 64
    RULE_function_def_raw = 65
    RULE_params = 66
    RULE_parameters = 67
    RULE_slash_no_default = 68
    RULE_slash_with_default = 69
    RULE_star_etc = 70
    RULE_kwds = 71
    RULE_param_no_default = 72
    RULE_param_no_default_star_annotation = 73
    RULE_param_with_default = 74
    RULE_param_maybe_default = 75
    RULE_param = 76
    RULE_param_star_annotation = 77
    RULE_annotation = 78
    RULE_star_annotation = 79
    RULE_default = 80
    RULE_if_stmt = 81
    RULE_elif_stmt = 82
    RULE_else_block = 83
    RULE_while_stmt = 84
    RULE_for_stmt = 85
    RULE_with_stmt = 86
    RULE_with_item = 87
    RULE_try_stmt = 88
    RULE_except_block = 89
    RULE_except_star_block = 90
    RULE_finally_block = 91
    RULE_match_stmt = 92
    RULE_subject_expr = 93
    RULE_case_block = 94
    RULE_guard = 95
    RULE_patterns = 96
    RULE_pattern = 97
    RULE_as_pattern = 98
    RULE_or_pattern = 99
    RULE_closed_pattern = 100
    RULE_literal_pattern = 101
    RULE_literal_expr = 102
    RULE_complex_number = 103
    RULE_signed_number = 104
    RULE_signed_real_number = 105
    RULE_real_number = 106
    RULE_imaginary_number = 107
    RULE_capture_pattern = 108
    RULE_pattern_capture_target = 109
    RULE_wildcard_pattern = 110
    RULE_value_pattern = 111
    RULE_attr = 112
    RULE_name_or_attr = 113
    RULE_group_pattern = 114
    RULE_sequence_pattern = 115
    RULE_open_sequence_pattern = 116
    RULE_maybe_sequence_pattern = 117
    RULE_maybe_star_pattern = 118
    RULE_star_pattern = 119
    RULE_mapping_pattern = 120
    RULE_items_pattern = 121
    RULE_key_value_pattern = 122
    RULE_double_star_pattern = 123
    RULE_class_pattern = 124
    RULE_positional_patterns = 125
    RULE_keyword_patterns = 126
    RULE_keyword_pattern = 127
    RULE_type_alias = 128
    RULE_type_params = 129
    RULE_type_param_seq = 130
    RULE_type_param = 131
    RULE_type_param_bound = 132
    RULE_expressions = 133
    RULE_expression = 134
    RULE_yield_expr = 135
    RULE_star_expressions = 136
    RULE_star_expression = 137
    RULE_star_named_expressions = 138
    RULE_star_named_expression = 139
    RULE_assignment_expression = 140
    RULE_named_expression = 141
    RULE_disjunction = 142
    RULE_conjunction = 143
    RULE_inversion = 144
    RULE_comparison = 145
    RULE_compare_op_bitwise_or_pair = 146
    RULE_eq_bitwise_or = 147
    RULE_noteq_bitwise_or = 148
    RULE_lte_bitwise_or = 149
    RULE_lt_bitwise_or = 150
    RULE_gte_bitwise_or = 151
    RULE_gt_bitwise_or = 152
    RULE_notin_bitwise_or = 153
    RULE_in_bitwise_or = 154
    RULE_isnot_bitwise_or = 155
    RULE_is_bitwise_or = 156
    RULE_bitwise_or = 157
    RULE_bitwise_xor = 158
    RULE_bitwise_and = 159
    RULE_shift_expr = 160
    RULE_sum = 161
    RULE_term = 162
    RULE_factor = 163
    RULE_power = 164
    RULE_await_primary = 165
    RULE_primary = 166
    RULE_slices = 167
    RULE_slice = 168
    RULE_atom = 169
    RULE_group = 170
    RULE_lambdef = 171
    RULE_lambda_params = 172
    RULE_lambda_parameters = 173
    RULE_lambda_slash_no_default = 174
    RULE_lambda_slash_with_default = 175
    RULE_lambda_star_etc = 176
    RULE_lambda_kwds = 177
    RULE_lambda_param_no_default = 178
    RULE_lambda_param_with_default = 179
    RULE_lambda_param_maybe_default = 180
    RULE_lambda_param = 181
    RULE_fstring_middle = 182
    RULE_fstring_replacement_field = 183
    RULE_fstring_conversion = 184
    RULE_fstring_full_format_spec = 185
    RULE_fstring_format_spec = 186
    RULE_fstring = 187
    RULE_string = 188
    RULE_strings = 189
    RULE_list = 190
    RULE_tuple = 191
    RULE_set = 192
    RULE_dict = 193
    RULE_double_starred_kvpairs = 194
    RULE_double_starred_kvpair = 195
    RULE_kvpair = 196
    RULE_for_if_clauses = 197
    RULE_for_if_clause = 198
    RULE_listcomp = 199
    RULE_setcomp = 200
    RULE_genexp = 201
    RULE_dictcomp = 202
    RULE_arguments = 203
    RULE_args = 204
    RULE_arg = 205
    RULE_kwargs = 206
    RULE_starred_expression = 207
    RULE_kwarg_or_starred = 208
    RULE_kwarg_or_double_starred = 209
    RULE_star_targets = 210
    RULE_star_targets_list_seq = 211
    RULE_star_targets_tuple_seq = 212
    RULE_star_target = 213
    RULE_target_with_star_atom = 214
    RULE_star_atom = 215
    RULE_single_target = 216
    RULE_single_subscript_attribute_target = 217
    RULE_t_primary = 218
    RULE_del_targets = 219
    RULE_del_target = 220
    RULE_del_t_atom = 221
    RULE_type_expressions = 222
    RULE_func_type_comment = 223

    ruleNames = [
        "fandango",
        "program",
        "statement",
        "production",
        "alternative",
        "concatenation",
        "operator",
        "kleene",
        "plus",
        "option",
        "repeat",
        "symbol",
        "char_set",
        "constraint",
        "implies",
        "quantifier",
        "formula_disjunction",
        "formula_conjunction",
        "formula_atom",
        "formula_comparison",
        "expr",
        "selector_length",
        "selector",
        "selection",
        "base_selection",
        "rs_pairs",
        "rs_pair",
        "rs_slices",
        "rs_slice",
        "python",
        "python_tag",
        "python_file",
        "interactive",
        "eval",
        "func_type",
        "statements",
        "stmt",
        "statement_newline",
        "simple_stmts",
        "simple_stmt",
        "compound_stmt",
        "assignment",
        "annotated_rhs",
        "augassign",
        "return_stmt",
        "raise_stmt",
        "global_stmt",
        "nonlocal_stmt",
        "del_stmt",
        "yield_stmt",
        "assert_stmt",
        "import_stmt",
        "import_name",
        "import_from",
        "import_from_targets",
        "import_from_as_names",
        "import_from_as_name",
        "dotted_as_names",
        "dotted_as_name",
        "dotted_name",
        "block",
        "decorators",
        "class_def",
        "class_def_raw",
        "function_def",
        "function_def_raw",
        "params",
        "parameters",
        "slash_no_default",
        "slash_with_default",
        "star_etc",
        "kwds",
        "param_no_default",
        "param_no_default_star_annotation",
        "param_with_default",
        "param_maybe_default",
        "param",
        "param_star_annotation",
        "annotation",
        "star_annotation",
        "default",
        "if_stmt",
        "elif_stmt",
        "else_block",
        "while_stmt",
        "for_stmt",
        "with_stmt",
        "with_item",
        "try_stmt",
        "except_block",
        "except_star_block",
        "finally_block",
        "match_stmt",
        "subject_expr",
        "case_block",
        "guard",
        "patterns",
        "pattern",
        "as_pattern",
        "or_pattern",
        "closed_pattern",
        "literal_pattern",
        "literal_expr",
        "complex_number",
        "signed_number",
        "signed_real_number",
        "real_number",
        "imaginary_number",
        "capture_pattern",
        "pattern_capture_target",
        "wildcard_pattern",
        "value_pattern",
        "attr",
        "name_or_attr",
        "group_pattern",
        "sequence_pattern",
        "open_sequence_pattern",
        "maybe_sequence_pattern",
        "maybe_star_pattern",
        "star_pattern",
        "mapping_pattern",
        "items_pattern",
        "key_value_pattern",
        "double_star_pattern",
        "class_pattern",
        "positional_patterns",
        "keyword_patterns",
        "keyword_pattern",
        "type_alias",
        "type_params",
        "type_param_seq",
        "type_param",
        "type_param_bound",
        "expressions",
        "expression",
        "yield_expr",
        "star_expressions",
        "star_expression",
        "star_named_expressions",
        "star_named_expression",
        "assignment_expression",
        "named_expression",
        "disjunction",
        "conjunction",
        "inversion",
        "comparison",
        "compare_op_bitwise_or_pair",
        "eq_bitwise_or",
        "noteq_bitwise_or",
        "lte_bitwise_or",
        "lt_bitwise_or",
        "gte_bitwise_or",
        "gt_bitwise_or",
        "notin_bitwise_or",
        "in_bitwise_or",
        "isnot_bitwise_or",
        "is_bitwise_or",
        "bitwise_or",
        "bitwise_xor",
        "bitwise_and",
        "shift_expr",
        "sum",
        "term",
        "factor",
        "power",
        "await_primary",
        "primary",
        "slices",
        "slice",
        "atom",
        "group",
        "lambdef",
        "lambda_params",
        "lambda_parameters",
        "lambda_slash_no_default",
        "lambda_slash_with_default",
        "lambda_star_etc",
        "lambda_kwds",
        "lambda_param_no_default",
        "lambda_param_with_default",
        "lambda_param_maybe_default",
        "lambda_param",
        "fstring_middle",
        "fstring_replacement_field",
        "fstring_conversion",
        "fstring_full_format_spec",
        "fstring_format_spec",
        "fstring",
        "string",
        "strings",
        "list",
        "tuple",
        "set",
        "dict",
        "double_starred_kvpairs",
        "double_starred_kvpair",
        "kvpair",
        "for_if_clauses",
        "for_if_clause",
        "listcomp",
        "setcomp",
        "genexp",
        "dictcomp",
        "arguments",
        "args",
        "arg",
        "kwargs",
        "starred_expression",
        "kwarg_or_starred",
        "kwarg_or_double_starred",
        "star_targets",
        "star_targets_list_seq",
        "star_targets_tuple_seq",
        "star_target",
        "target_with_star_atom",
        "star_atom",
        "single_target",
        "single_subscript_attribute_target",
        "t_primary",
        "del_targets",
        "del_target",
        "del_t_atom",
        "type_expressions",
        "func_type_comment",
    ]

    EOF = Token.EOF
    INDENT = 1
    DEDENT = 2
    STRING = 3
    NUMBER = 4
    INTEGER = 5
    PYTHON_START = 6
    PYTHON_END = 7
    NONTERMINAL = 8
    AND = 9
    AS = 10
    ASSERT = 11
    ASYNC = 12
    AWAIT = 13
    BREAK = 14
    CASE = 15
    CLASS = 16
    CONTINUE = 17
    DEF = 18
    DEL = 19
    ELIF = 20
    ELSE = 21
    EXCEPT = 22
    FALSE = 23
    FINALLY = 24
    FOR = 25
    FROM = 26
    GLOBAL = 27
    IF = 28
    IMPORT = 29
    IN = 30
    IS = 31
    LAMBDA = 32
    MATCH = 33
    NONE = 34
    NONLOCAL = 35
    NOT = 36
    OR = 37
    PASS = 38
    RAISE = 39
    RETURN = 40
    TRUE = 41
    TRY = 42
    TYPE = 43
    UNDERSCORE = 44
    WHILE = 45
    WHERE = 46
    WITH = 47
    YIELD = 48
    FORALL = 49
    EXISTS = 50
    FITNESS = 51
    NAME = 52
    STRING_LITERAL = 53
    BYTES_LITERAL = 54
    DECIMAL_INTEGER = 55
    OCT_INTEGER = 56
    HEX_INTEGER = 57
    BIN_INTEGER = 58
    FLOAT_NUMBER = 59
    IMAG_NUMBER = 60
    GRAMMAR_ASSIGN = 61
    QUESTION = 62
    BACKSLASH = 63
    DOT = 64
    DOTDOT = 65
    ELLIPSIS = 66
    STAR = 67
    OPEN_PAREN = 68
    CLOSE_PAREN = 69
    COMMA = 70
    COLON = 71
    SEMI_COLON = 72
    POWER = 73
    ASSIGN = 74
    OPEN_BRACK = 75
    CLOSE_BRACK = 76
    OR_OP = 77
    XOR = 78
    AND_OP = 79
    LEFT_SHIFT = 80
    RIGHT_SHIFT = 81
    ADD = 82
    MINUS = 83
    DIV = 84
    MOD = 85
    IDIV = 86
    NOT_OP = 87
    OPEN_BRACE = 88
    CLOSE_BRACE = 89
    LESS_THAN = 90
    GREATER_THAN = 91
    EQUALS = 92
    GT_EQ = 93
    LT_EQ = 94
    NOT_EQ_1 = 95
    NOT_EQ_2 = 96
    AT = 97
    ARROW = 98
    ADD_ASSIGN = 99
    SUB_ASSIGN = 100
    MULT_ASSIGN = 101
    AT_ASSIGN = 102
    DIV_ASSIGN = 103
    MOD_ASSIGN = 104
    AND_ASSIGN = 105
    OR_ASSIGN = 106
    XOR_ASSIGN = 107
    LEFT_SHIFT_ASSIGN = 108
    RIGHT_SHIFT_ASSIGN = 109
    POWER_ASSIGN = 110
    IDIV_ASSIGN = 111
    EXPR_ASSIGN = 112
    EXCL = 113
    NEWLINE = 114
    SKIP_ = 115
    UNKNOWN_CHAR = 116
    FSTRING_MIDDLE = 117
    FSTRING_START = 118
    FSTRING_END = 119

    def __init__(self, input: TokenStream, output: TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.13.2")
        self._interp = ParserATNSimulator(
            self, self.atn, self.decisionsToDFA, self.sharedContextCache
        )
        self._predicates = None

    class FandangoContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def program(self):
            return self.getTypedRuleContext(FandangoParser.ProgramContext, 0)

        def EOF(self):
            return self.getToken(FandangoParser.EOF, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_fandango

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterFandango"):
                listener.enterFandango(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitFandango"):
                listener.exitFandango(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitFandango"):
                return visitor.visitFandango(self)
            else:
                return visitor.visitChildren(self)

    def fandango(self):

        localctx = FandangoParser.FandangoContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_fandango)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 448
            self.program()
            self.state = 449
            self.match(FandangoParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ProgramContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NEWLINE(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.NEWLINE)
            else:
                return self.getToken(FandangoParser.NEWLINE, i)

        def statement(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(FandangoParser.StatementContext)
            else:
                return self.getTypedRuleContext(FandangoParser.StatementContext, i)

        def getRuleIndex(self):
            return FandangoParser.RULE_program

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterProgram"):
                listener.enterProgram(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitProgram"):
                listener.exitProgram(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitProgram"):
                return visitor.visitProgram(self)
            else:
                return visitor.visitChildren(self)

    def program(self):

        localctx = FandangoParser.ProgramContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_program)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 454
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 0, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 451
                    self.match(FandangoParser.NEWLINE)
                self.state = 456
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 0, self._ctx)

            self.state = 466
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while (((_la) & ~0x3F) == 0 and ((1 << _la) & 8989466384365848) != 0) or (
                (((_la - 66)) & ~0x3F) == 0
                and ((1 << (_la - 66)) & 4785076758055431) != 0
            ):
                self.state = 457
                self.statement()
                self.state = 461
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 1, self._ctx)
                while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                    if _alt == 1:
                        self.state = 458
                        self.match(FandangoParser.NEWLINE)
                    self.state = 463
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input, 1, self._ctx)

                self.state = 468
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class StatementContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def production(self):
            return self.getTypedRuleContext(FandangoParser.ProductionContext, 0)

        def constraint(self):
            return self.getTypedRuleContext(FandangoParser.ConstraintContext, 0)

        def python(self):
            return self.getTypedRuleContext(FandangoParser.PythonContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_statement

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterStatement"):
                listener.enterStatement(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitStatement"):
                listener.exitStatement(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitStatement"):
                return visitor.visitStatement(self)
            else:
                return visitor.visitChildren(self)

    def statement(self):

        localctx = FandangoParser.StatementContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_statement)
        try:
            self.state = 472
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 3, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 469
                self.production()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 470
                self.constraint()
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 471
                self.python()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ProductionContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NONTERMINAL(self):
            return self.getToken(FandangoParser.NONTERMINAL, 0)

        def GRAMMAR_ASSIGN(self):
            return self.getToken(FandangoParser.GRAMMAR_ASSIGN, 0)

        def alternative(self):
            return self.getTypedRuleContext(FandangoParser.AlternativeContext, 0)

        def SEMI_COLON(self):
            return self.getToken(FandangoParser.SEMI_COLON, 0)

        def NEWLINE(self):
            return self.getToken(FandangoParser.NEWLINE, 0)

        def EOF(self):
            return self.getToken(FandangoParser.EOF, 0)

        def EXPR_ASSIGN(self):
            return self.getToken(FandangoParser.EXPR_ASSIGN, 0)

        def expression(self):
            return self.getTypedRuleContext(FandangoParser.ExpressionContext, 0)

        def ASSIGN(self):
            return self.getToken(FandangoParser.ASSIGN, 0)

        def COLON(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.COLON)
            else:
                return self.getToken(FandangoParser.COLON, i)

        def getRuleIndex(self):
            return FandangoParser.RULE_production

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterProduction"):
                listener.enterProduction(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitProduction"):
                listener.exitProduction(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitProduction"):
                return visitor.visitProduction(self)
            else:
                return visitor.visitChildren(self)

    def production(self):

        localctx = FandangoParser.ProductionContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_production)
        self._la = 0  # Token type
        try:
            self.state = 502
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 7, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 474
                self.match(FandangoParser.NONTERMINAL)
                self.state = 475
                self.match(FandangoParser.GRAMMAR_ASSIGN)
                self.state = 476
                self.alternative()
                self.state = 479
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == 112:
                    self.state = 477
                    self.match(FandangoParser.EXPR_ASSIGN)
                    self.state = 478
                    self.expression()

                self.state = 481
                _la = self._input.LA(1)
                if not (_la == -1 or _la == 72 or _la == 114):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 483
                self.match(FandangoParser.NONTERMINAL)
                self.state = 484
                self.match(FandangoParser.GRAMMAR_ASSIGN)
                self.state = 485
                self.alternative()
                self.state = 488
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == 74:
                    self.state = 486
                    self.match(FandangoParser.ASSIGN)
                    self.state = 487
                    self.expression()

                self.state = 490
                _la = self._input.LA(1)
                if not (_la == -1 or _la == 72 or _la == 114):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 492
                self.match(FandangoParser.NONTERMINAL)
                self.state = 493
                self.match(FandangoParser.GRAMMAR_ASSIGN)
                self.state = 494
                self.alternative()
                self.state = 498
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == 71:
                    self.state = 495
                    self.match(FandangoParser.COLON)
                    self.state = 496
                    self.match(FandangoParser.COLON)
                    self.state = 497
                    self.expression()

                self.state = 500
                _la = self._input.LA(1)
                if not (_la == -1 or _la == 72 or _la == 114):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class AlternativeContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def concatenation(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(FandangoParser.ConcatenationContext)
            else:
                return self.getTypedRuleContext(FandangoParser.ConcatenationContext, i)

        def OR_OP(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.OR_OP)
            else:
                return self.getToken(FandangoParser.OR_OP, i)

        def getRuleIndex(self):
            return FandangoParser.RULE_alternative

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterAlternative"):
                listener.enterAlternative(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitAlternative"):
                listener.exitAlternative(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitAlternative"):
                return visitor.visitAlternative(self)
            else:
                return visitor.visitChildren(self)

    def alternative(self):

        localctx = FandangoParser.AlternativeContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_alternative)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 504
            self.concatenation()
            self.state = 509
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la == 77:
                self.state = 505
                self.match(FandangoParser.OR_OP)
                self.state = 506
                self.concatenation()
                self.state = 511
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ConcatenationContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def operator(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(FandangoParser.OperatorContext)
            else:
                return self.getTypedRuleContext(FandangoParser.OperatorContext, i)

        def getRuleIndex(self):
            return FandangoParser.RULE_concatenation

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterConcatenation"):
                listener.enterConcatenation(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitConcatenation"):
                listener.exitConcatenation(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitConcatenation"):
                return visitor.visitConcatenation(self)
            else:
                return visitor.visitChildren(self)

    def concatenation(self):

        localctx = FandangoParser.ConcatenationContext(self, self._ctx, self.state)
        self.enterRule(localctx, 10, self.RULE_concatenation)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 512
            self.operator()
            self.state = 516
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 9, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 513
                    self.operator()
                self.state = 518
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 9, self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class OperatorContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def symbol(self):
            return self.getTypedRuleContext(FandangoParser.SymbolContext, 0)

        def kleene(self):
            return self.getTypedRuleContext(FandangoParser.KleeneContext, 0)

        def plus(self):
            return self.getTypedRuleContext(FandangoParser.PlusContext, 0)

        def option(self):
            return self.getTypedRuleContext(FandangoParser.OptionContext, 0)

        def repeat(self):
            return self.getTypedRuleContext(FandangoParser.RepeatContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_operator

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterOperator"):
                listener.enterOperator(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitOperator"):
                listener.exitOperator(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitOperator"):
                return visitor.visitOperator(self)
            else:
                return visitor.visitChildren(self)

    def operator(self):

        localctx = FandangoParser.OperatorContext(self, self._ctx, self.state)
        self.enterRule(localctx, 12, self.RULE_operator)
        try:
            self.state = 524
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 10, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 519
                self.symbol()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 520
                self.kleene()
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 521
                self.plus()
                pass

            elif la_ == 4:
                self.enterOuterAlt(localctx, 4)
                self.state = 522
                self.option()
                pass

            elif la_ == 5:
                self.enterOuterAlt(localctx, 5)
                self.state = 523
                self.repeat()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class KleeneContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def symbol(self):
            return self.getTypedRuleContext(FandangoParser.SymbolContext, 0)

        def STAR(self):
            return self.getToken(FandangoParser.STAR, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_kleene

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterKleene"):
                listener.enterKleene(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitKleene"):
                listener.exitKleene(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitKleene"):
                return visitor.visitKleene(self)
            else:
                return visitor.visitChildren(self)

    def kleene(self):

        localctx = FandangoParser.KleeneContext(self, self._ctx, self.state)
        self.enterRule(localctx, 14, self.RULE_kleene)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 526
            self.symbol()
            self.state = 527
            self.match(FandangoParser.STAR)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class PlusContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def symbol(self):
            return self.getTypedRuleContext(FandangoParser.SymbolContext, 0)

        def ADD(self):
            return self.getToken(FandangoParser.ADD, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_plus

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterPlus"):
                listener.enterPlus(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitPlus"):
                listener.exitPlus(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitPlus"):
                return visitor.visitPlus(self)
            else:
                return visitor.visitChildren(self)

    def plus(self):

        localctx = FandangoParser.PlusContext(self, self._ctx, self.state)
        self.enterRule(localctx, 16, self.RULE_plus)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 529
            self.symbol()
            self.state = 530
            self.match(FandangoParser.ADD)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class OptionContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def symbol(self):
            return self.getTypedRuleContext(FandangoParser.SymbolContext, 0)

        def QUESTION(self):
            return self.getToken(FandangoParser.QUESTION, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_option

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterOption"):
                listener.enterOption(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitOption"):
                listener.exitOption(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitOption"):
                return visitor.visitOption(self)
            else:
                return visitor.visitChildren(self)

    def option(self):

        localctx = FandangoParser.OptionContext(self, self._ctx, self.state)
        self.enterRule(localctx, 18, self.RULE_option)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 532
            self.symbol()
            self.state = 533
            self.match(FandangoParser.QUESTION)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class RepeatContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def symbol(self):
            return self.getTypedRuleContext(FandangoParser.SymbolContext, 0)

        def OPEN_BRACE(self):
            return self.getToken(FandangoParser.OPEN_BRACE, 0)

        def NUMBER(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.NUMBER)
            else:
                return self.getToken(FandangoParser.NUMBER, i)

        def CLOSE_BRACE(self):
            return self.getToken(FandangoParser.CLOSE_BRACE, 0)

        def COMMA(self):
            return self.getToken(FandangoParser.COMMA, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_repeat

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterRepeat"):
                listener.enterRepeat(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitRepeat"):
                listener.exitRepeat(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitRepeat"):
                return visitor.visitRepeat(self)
            else:
                return visitor.visitChildren(self)

    def repeat(self):

        localctx = FandangoParser.RepeatContext(self, self._ctx, self.state)
        self.enterRule(localctx, 20, self.RULE_repeat)
        self._la = 0  # Token type
        try:
            self.state = 551
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 13, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 535
                self.symbol()
                self.state = 536
                self.match(FandangoParser.OPEN_BRACE)
                self.state = 537
                self.match(FandangoParser.NUMBER)
                self.state = 538
                self.match(FandangoParser.CLOSE_BRACE)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 540
                self.symbol()
                self.state = 541
                self.match(FandangoParser.OPEN_BRACE)
                self.state = 543
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == 4:
                    self.state = 542
                    self.match(FandangoParser.NUMBER)

                self.state = 545
                self.match(FandangoParser.COMMA)
                self.state = 547
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == 4:
                    self.state = 546
                    self.match(FandangoParser.NUMBER)

                self.state = 549
                self.match(FandangoParser.CLOSE_BRACE)
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class SymbolContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NONTERMINAL(self):
            return self.getToken(FandangoParser.NONTERMINAL, 0)

        def STRING(self):
            return self.getToken(FandangoParser.STRING, 0)

        def NUMBER(self):
            return self.getToken(FandangoParser.NUMBER, 0)

        def OPEN_PAREN(self):
            return self.getToken(FandangoParser.OPEN_PAREN, 0)

        def alternative(self):
            return self.getTypedRuleContext(FandangoParser.AlternativeContext, 0)

        def CLOSE_PAREN(self):
            return self.getToken(FandangoParser.CLOSE_PAREN, 0)

        def char_set(self):
            return self.getTypedRuleContext(FandangoParser.Char_setContext, 0)

        def NEWLINE(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.NEWLINE)
            else:
                return self.getToken(FandangoParser.NEWLINE, i)

        def getRuleIndex(self):
            return FandangoParser.RULE_symbol

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterSymbol"):
                listener.enterSymbol(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitSymbol"):
                listener.exitSymbol(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitSymbol"):
                return visitor.visitSymbol(self)
            else:
                return visitor.visitChildren(self)

    def symbol(self):

        localctx = FandangoParser.SymbolContext(self, self._ctx, self.state)
        self.enterRule(localctx, 22, self.RULE_symbol)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 556
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la == 114:
                self.state = 553
                self.match(FandangoParser.NEWLINE)
                self.state = 558
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 567
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [8]:
                self.state = 559
                self.match(FandangoParser.NONTERMINAL)
                pass
            elif token in [3]:
                self.state = 560
                self.match(FandangoParser.STRING)
                pass
            elif token in [4]:
                self.state = 561
                self.match(FandangoParser.NUMBER)
                pass
            elif token in [68]:
                self.state = 562
                self.match(FandangoParser.OPEN_PAREN)
                self.state = 563
                self.alternative()
                self.state = 564
                self.match(FandangoParser.CLOSE_PAREN)
                pass
            elif token in [75]:
                self.state = 566
                self.char_set()
                pass
            else:
                raise NoViableAltException(self)

            self.state = 572
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 16, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 569
                    self.match(FandangoParser.NEWLINE)
                self.state = 574
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 16, self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Char_setContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def OPEN_BRACK(self):
            return self.getToken(FandangoParser.OPEN_BRACK, 0)

        def STRING(self):
            return self.getToken(FandangoParser.STRING, 0)

        def CLOSE_BRACK(self):
            return self.getToken(FandangoParser.CLOSE_BRACK, 0)

        def XOR(self):
            return self.getToken(FandangoParser.XOR, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_char_set

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterChar_set"):
                listener.enterChar_set(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitChar_set"):
                listener.exitChar_set(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitChar_set"):
                return visitor.visitChar_set(self)
            else:
                return visitor.visitChildren(self)

    def char_set(self):

        localctx = FandangoParser.Char_setContext(self, self._ctx, self.state)
        self.enterRule(localctx, 24, self.RULE_char_set)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 575
            self.match(FandangoParser.OPEN_BRACK)
            self.state = 577
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == 78:
                self.state = 576
                self.match(FandangoParser.XOR)

            self.state = 579
            self.match(FandangoParser.STRING)
            self.state = 580
            self.match(FandangoParser.CLOSE_BRACK)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ConstraintContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def WHERE(self):
            return self.getToken(FandangoParser.WHERE, 0)

        def implies(self):
            return self.getTypedRuleContext(FandangoParser.ImpliesContext, 0)

        def SEMI_COLON(self):
            return self.getToken(FandangoParser.SEMI_COLON, 0)

        def NEWLINE(self):
            return self.getToken(FandangoParser.NEWLINE, 0)

        def EOF(self):
            return self.getToken(FandangoParser.EOF, 0)

        def FITNESS(self):
            return self.getToken(FandangoParser.FITNESS, 0)

        def expr(self):
            return self.getTypedRuleContext(FandangoParser.ExprContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_constraint

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterConstraint"):
                listener.enterConstraint(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitConstraint"):
                listener.exitConstraint(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitConstraint"):
                return visitor.visitConstraint(self)
            else:
                return visitor.visitChildren(self)

    def constraint(self):

        localctx = FandangoParser.ConstraintContext(self, self._ctx, self.state)
        self.enterRule(localctx, 26, self.RULE_constraint)
        self._la = 0  # Token type
        try:
            self.state = 598
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 18, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 582
                self.match(FandangoParser.WHERE)
                self.state = 583
                self.implies()
                self.state = 584
                _la = self._input.LA(1)
                if not (_la == -1 or _la == 72 or _la == 114):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 586
                self.match(FandangoParser.WHERE)
                self.state = 587
                self.match(FandangoParser.FITNESS)
                self.state = 588
                self.expr()
                self.state = 589
                _la = self._input.LA(1)
                if not (_la == -1 or _la == 72 or _la == 114):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 591
                self.implies()
                self.state = 592
                self.match(FandangoParser.SEMI_COLON)
                pass

            elif la_ == 4:
                self.enterOuterAlt(localctx, 4)
                self.state = 594
                self.match(FandangoParser.FITNESS)
                self.state = 595
                self.expr()
                self.state = 596
                self.match(FandangoParser.SEMI_COLON)
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ImpliesContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def quantifier(self):
            return self.getTypedRuleContext(FandangoParser.QuantifierContext, 0)

        def ARROW(self):
            return self.getToken(FandangoParser.ARROW, 0)

        def implies(self):
            return self.getTypedRuleContext(FandangoParser.ImpliesContext, 0)

        def NEWLINE(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.NEWLINE)
            else:
                return self.getToken(FandangoParser.NEWLINE, i)

        def getRuleIndex(self):
            return FandangoParser.RULE_implies

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterImplies"):
                listener.enterImplies(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitImplies"):
                listener.exitImplies(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitImplies"):
                return visitor.visitImplies(self)
            else:
                return visitor.visitChildren(self)

    def implies(self):

        localctx = FandangoParser.ImpliesContext(self, self._ctx, self.state)
        self.enterRule(localctx, 28, self.RULE_implies)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 603
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 19, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 600
                    self.match(FandangoParser.NEWLINE)
                self.state = 605
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 19, self._ctx)

            self.state = 611
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 20, self._ctx)
            if la_ == 1:
                self.state = 606
                self.quantifier()
                self.state = 607
                self.match(FandangoParser.ARROW)
                self.state = 608
                self.implies()
                pass

            elif la_ == 2:
                self.state = 610
                self.quantifier()
                pass

            self.state = 616
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 21, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 613
                    self.match(FandangoParser.NEWLINE)
                self.state = 618
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 21, self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class QuantifierContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def FORALL(self):
            return self.getToken(FandangoParser.FORALL, 0)

        def NONTERMINAL(self):
            return self.getToken(FandangoParser.NONTERMINAL, 0)

        def IN(self):
            return self.getToken(FandangoParser.IN, 0)

        def selector(self):
            return self.getTypedRuleContext(FandangoParser.SelectorContext, 0)

        def COLON(self):
            return self.getToken(FandangoParser.COLON, 0)

        def quantifier(self):
            return self.getTypedRuleContext(FandangoParser.QuantifierContext, 0)

        def EXISTS(self):
            return self.getToken(FandangoParser.EXISTS, 0)

        def formula_disjunction(self):
            return self.getTypedRuleContext(
                FandangoParser.Formula_disjunctionContext, 0
            )

        def NEWLINE(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.NEWLINE)
            else:
                return self.getToken(FandangoParser.NEWLINE, i)

        def getRuleIndex(self):
            return FandangoParser.RULE_quantifier

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterQuantifier"):
                listener.enterQuantifier(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitQuantifier"):
                listener.exitQuantifier(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitQuantifier"):
                return visitor.visitQuantifier(self)
            else:
                return visitor.visitChildren(self)

    def quantifier(self):

        localctx = FandangoParser.QuantifierContext(self, self._ctx, self.state)
        self.enterRule(localctx, 30, self.RULE_quantifier)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 622
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 22, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 619
                    self.match(FandangoParser.NEWLINE)
                self.state = 624
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 22, self._ctx)

            self.state = 640
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [49]:
                self.state = 625
                self.match(FandangoParser.FORALL)
                self.state = 626
                self.match(FandangoParser.NONTERMINAL)
                self.state = 627
                self.match(FandangoParser.IN)
                self.state = 628
                self.selector(0)
                self.state = 629
                self.match(FandangoParser.COLON)
                self.state = 630
                self.quantifier()
                pass
            elif token in [50]:
                self.state = 632
                self.match(FandangoParser.EXISTS)
                self.state = 633
                self.match(FandangoParser.NONTERMINAL)
                self.state = 634
                self.match(FandangoParser.IN)
                self.state = 635
                self.selector(0)
                self.state = 636
                self.match(FandangoParser.COLON)
                self.state = 637
                self.quantifier()
                pass
            elif token in [
                3,
                4,
                8,
                13,
                23,
                34,
                36,
                41,
                52,
                66,
                68,
                75,
                77,
                82,
                83,
                87,
                88,
                114,
                118,
            ]:
                self.state = 639
                self.formula_disjunction()
                pass
            else:
                raise NoViableAltException(self)

            self.state = 645
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 24, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 642
                    self.match(FandangoParser.NEWLINE)
                self.state = 647
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 24, self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Formula_disjunctionContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def formula_conjunction(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(
                    FandangoParser.Formula_conjunctionContext
                )
            else:
                return self.getTypedRuleContext(
                    FandangoParser.Formula_conjunctionContext, i
                )

        def NEWLINE(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.NEWLINE)
            else:
                return self.getToken(FandangoParser.NEWLINE, i)

        def OR(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.OR)
            else:
                return self.getToken(FandangoParser.OR, i)

        def getRuleIndex(self):
            return FandangoParser.RULE_formula_disjunction

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterFormula_disjunction"):
                listener.enterFormula_disjunction(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitFormula_disjunction"):
                listener.exitFormula_disjunction(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitFormula_disjunction"):
                return visitor.visitFormula_disjunction(self)
            else:
                return visitor.visitChildren(self)

    def formula_disjunction(self):

        localctx = FandangoParser.Formula_disjunctionContext(
            self, self._ctx, self.state
        )
        self.enterRule(localctx, 32, self.RULE_formula_disjunction)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 651
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 25, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 648
                    self.match(FandangoParser.NEWLINE)
                self.state = 653
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 25, self._ctx)

            self.state = 654
            self.formula_conjunction()
            self.state = 659
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la == 37:
                self.state = 655
                self.match(FandangoParser.OR)
                self.state = 656
                self.formula_conjunction()
                self.state = 661
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 665
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 27, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 662
                    self.match(FandangoParser.NEWLINE)
                self.state = 667
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 27, self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Formula_conjunctionContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def formula_atom(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(FandangoParser.Formula_atomContext)
            else:
                return self.getTypedRuleContext(FandangoParser.Formula_atomContext, i)

        def NEWLINE(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.NEWLINE)
            else:
                return self.getToken(FandangoParser.NEWLINE, i)

        def AND(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.AND)
            else:
                return self.getToken(FandangoParser.AND, i)

        def getRuleIndex(self):
            return FandangoParser.RULE_formula_conjunction

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterFormula_conjunction"):
                listener.enterFormula_conjunction(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitFormula_conjunction"):
                listener.exitFormula_conjunction(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitFormula_conjunction"):
                return visitor.visitFormula_conjunction(self)
            else:
                return visitor.visitChildren(self)

    def formula_conjunction(self):

        localctx = FandangoParser.Formula_conjunctionContext(
            self, self._ctx, self.state
        )
        self.enterRule(localctx, 34, self.RULE_formula_conjunction)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 671
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 28, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 668
                    self.match(FandangoParser.NEWLINE)
                self.state = 673
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 28, self._ctx)

            self.state = 674
            self.formula_atom()
            self.state = 679
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la == 9:
                self.state = 675
                self.match(FandangoParser.AND)
                self.state = 676
                self.formula_atom()
                self.state = 681
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 685
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 30, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 682
                    self.match(FandangoParser.NEWLINE)
                self.state = 687
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 30, self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Formula_atomContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def formula_comparison(self):
            return self.getTypedRuleContext(FandangoParser.Formula_comparisonContext, 0)

        def OPEN_PAREN(self):
            return self.getToken(FandangoParser.OPEN_PAREN, 0)

        def implies(self):
            return self.getTypedRuleContext(FandangoParser.ImpliesContext, 0)

        def CLOSE_PAREN(self):
            return self.getToken(FandangoParser.CLOSE_PAREN, 0)

        def expr(self):
            return self.getTypedRuleContext(FandangoParser.ExprContext, 0)

        def NEWLINE(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.NEWLINE)
            else:
                return self.getToken(FandangoParser.NEWLINE, i)

        def getRuleIndex(self):
            return FandangoParser.RULE_formula_atom

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterFormula_atom"):
                listener.enterFormula_atom(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitFormula_atom"):
                listener.exitFormula_atom(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitFormula_atom"):
                return visitor.visitFormula_atom(self)
            else:
                return visitor.visitChildren(self)

    def formula_atom(self):

        localctx = FandangoParser.Formula_atomContext(self, self._ctx, self.state)
        self.enterRule(localctx, 36, self.RULE_formula_atom)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 691
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la == 114:
                self.state = 688
                self.match(FandangoParser.NEWLINE)
                self.state = 693
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 700
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 32, self._ctx)
            if la_ == 1:
                self.state = 694
                self.formula_comparison()
                pass

            elif la_ == 2:
                self.state = 695
                self.match(FandangoParser.OPEN_PAREN)
                self.state = 696
                self.implies()
                self.state = 697
                self.match(FandangoParser.CLOSE_PAREN)
                pass

            elif la_ == 3:
                self.state = 699
                self.expr()
                pass

            self.state = 705
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 33, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 702
                    self.match(FandangoParser.NEWLINE)
                self.state = 707
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 33, self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Formula_comparisonContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def expr(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(FandangoParser.ExprContext)
            else:
                return self.getTypedRuleContext(FandangoParser.ExprContext, i)

        def LESS_THAN(self):
            return self.getToken(FandangoParser.LESS_THAN, 0)

        def GREATER_THAN(self):
            return self.getToken(FandangoParser.GREATER_THAN, 0)

        def EQUALS(self):
            return self.getToken(FandangoParser.EQUALS, 0)

        def GT_EQ(self):
            return self.getToken(FandangoParser.GT_EQ, 0)

        def LT_EQ(self):
            return self.getToken(FandangoParser.LT_EQ, 0)

        def NOT_EQ_1(self):
            return self.getToken(FandangoParser.NOT_EQ_1, 0)

        def NOT_EQ_2(self):
            return self.getToken(FandangoParser.NOT_EQ_2, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_formula_comparison

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterFormula_comparison"):
                listener.enterFormula_comparison(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitFormula_comparison"):
                listener.exitFormula_comparison(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitFormula_comparison"):
                return visitor.visitFormula_comparison(self)
            else:
                return visitor.visitChildren(self)

    def formula_comparison(self):

        localctx = FandangoParser.Formula_comparisonContext(self, self._ctx, self.state)
        self.enterRule(localctx, 38, self.RULE_formula_comparison)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 708
            self.expr()
            self.state = 709
            _la = self._input.LA(1)
            if not (((((_la - 90)) & ~0x3F) == 0 and ((1 << (_la - 90)) & 127) != 0)):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
            self.state = 710
            self.expr()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ExprContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def selector_length(self):
            return self.getTypedRuleContext(FandangoParser.Selector_lengthContext, 0)

        def inversion(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(FandangoParser.InversionContext)
            else:
                return self.getTypedRuleContext(FandangoParser.InversionContext, i)

        def IF(self):
            return self.getToken(FandangoParser.IF, 0)

        def ELSE(self):
            return self.getToken(FandangoParser.ELSE, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_expr

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterExpr"):
                listener.enterExpr(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitExpr"):
                listener.exitExpr(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitExpr"):
                return visitor.visitExpr(self)
            else:
                return visitor.visitChildren(self)

    def expr(self):

        localctx = FandangoParser.ExprContext(self, self._ctx, self.state)
        self.enterRule(localctx, 40, self.RULE_expr)
        try:
            self.state = 720
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 34, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 712
                self.selector_length()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 713
                self.inversion()
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 714
                self.inversion()
                self.state = 715
                self.match(FandangoParser.IF)
                self.state = 716
                self.inversion()
                self.state = 717
                self.match(FandangoParser.ELSE)
                self.state = 718
                self.inversion()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Selector_lengthContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def OR_OP(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.OR_OP)
            else:
                return self.getToken(FandangoParser.OR_OP, i)

        def selector(self):
            return self.getTypedRuleContext(FandangoParser.SelectorContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_selector_length

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterSelector_length"):
                listener.enterSelector_length(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitSelector_length"):
                listener.exitSelector_length(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitSelector_length"):
                return visitor.visitSelector_length(self)
            else:
                return visitor.visitChildren(self)

    def selector_length(self):

        localctx = FandangoParser.Selector_lengthContext(self, self._ctx, self.state)
        self.enterRule(localctx, 42, self.RULE_selector_length)
        try:
            self.state = 727
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [77]:
                self.enterOuterAlt(localctx, 1)
                self.state = 722
                self.match(FandangoParser.OR_OP)
                self.state = 723
                self.selector(0)
                self.state = 724
                self.match(FandangoParser.OR_OP)
                pass
            elif token in [8, 68]:
                self.enterOuterAlt(localctx, 2)
                self.state = 726
                self.selector(0)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class SelectorContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def selection(self):
            return self.getTypedRuleContext(FandangoParser.SelectionContext, 0)

        def selector(self):
            return self.getTypedRuleContext(FandangoParser.SelectorContext, 0)

        def DOT(self):
            return self.getToken(FandangoParser.DOT, 0)

        def DOTDOT(self):
            return self.getToken(FandangoParser.DOTDOT, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_selector

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterSelector"):
                listener.enterSelector(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitSelector"):
                listener.exitSelector(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitSelector"):
                return visitor.visitSelector(self)
            else:
                return visitor.visitChildren(self)

    def selector(self, _p: int = 0):
        _parentctx = self._ctx
        _parentState = self.state
        localctx = FandangoParser.SelectorContext(self, self._ctx, _parentState)
        _prevctx = localctx
        _startState = 44
        self.enterRecursionRule(localctx, 44, self.RULE_selector, _p)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 730
            self.selection()
            self._ctx.stop = self._input.LT(-1)
            self.state = 740
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 37, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    if self._parseListeners is not None:
                        self.triggerExitRuleEvent()
                    _prevctx = localctx
                    self.state = 738
                    self._errHandler.sync(self)
                    la_ = self._interp.adaptivePredict(self._input, 36, self._ctx)
                    if la_ == 1:
                        localctx = FandangoParser.SelectorContext(
                            self, _parentctx, _parentState
                        )
                        self.pushNewRecursionContext(
                            localctx, _startState, self.RULE_selector
                        )
                        self.state = 732
                        if not self.precpred(self._ctx, 2):
                            from antlr4.error.Errors import FailedPredicateException

                            raise FailedPredicateException(
                                self, "self.precpred(self._ctx, 2)"
                            )
                        self.state = 733
                        self.match(FandangoParser.DOT)
                        self.state = 734
                        self.selection()
                        pass

                    elif la_ == 2:
                        localctx = FandangoParser.SelectorContext(
                            self, _parentctx, _parentState
                        )
                        self.pushNewRecursionContext(
                            localctx, _startState, self.RULE_selector
                        )
                        self.state = 735
                        if not self.precpred(self._ctx, 1):
                            from antlr4.error.Errors import FailedPredicateException

                            raise FailedPredicateException(
                                self, "self.precpred(self._ctx, 1)"
                            )
                        self.state = 736
                        self.match(FandangoParser.DOTDOT)
                        self.state = 737
                        self.selection()
                        pass

                self.state = 742
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 37, self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.unrollRecursionContexts(_parentctx)
        return localctx

    class SelectionContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def base_selection(self):
            return self.getTypedRuleContext(FandangoParser.Base_selectionContext, 0)

        def OPEN_BRACK(self):
            return self.getToken(FandangoParser.OPEN_BRACK, 0)

        def rs_slices(self):
            return self.getTypedRuleContext(FandangoParser.Rs_slicesContext, 0)

        def CLOSE_BRACK(self):
            return self.getToken(FandangoParser.CLOSE_BRACK, 0)

        def OPEN_BRACE(self):
            return self.getToken(FandangoParser.OPEN_BRACE, 0)

        def rs_pairs(self):
            return self.getTypedRuleContext(FandangoParser.Rs_pairsContext, 0)

        def CLOSE_BRACE(self):
            return self.getToken(FandangoParser.CLOSE_BRACE, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_selection

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterSelection"):
                listener.enterSelection(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitSelection"):
                listener.exitSelection(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitSelection"):
                return visitor.visitSelection(self)
            else:
                return visitor.visitChildren(self)

    def selection(self):

        localctx = FandangoParser.SelectionContext(self, self._ctx, self.state)
        self.enterRule(localctx, 46, self.RULE_selection)
        try:
            self.state = 754
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 38, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 743
                self.base_selection()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 744
                self.base_selection()
                self.state = 745
                self.match(FandangoParser.OPEN_BRACK)
                self.state = 746
                self.rs_slices()
                self.state = 747
                self.match(FandangoParser.CLOSE_BRACK)
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 749
                self.base_selection()
                self.state = 750
                self.match(FandangoParser.OPEN_BRACE)
                self.state = 751
                self.rs_pairs()
                self.state = 752
                self.match(FandangoParser.CLOSE_BRACE)
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Base_selectionContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NONTERMINAL(self):
            return self.getToken(FandangoParser.NONTERMINAL, 0)

        def OPEN_PAREN(self):
            return self.getToken(FandangoParser.OPEN_PAREN, 0)

        def selector(self):
            return self.getTypedRuleContext(FandangoParser.SelectorContext, 0)

        def CLOSE_PAREN(self):
            return self.getToken(FandangoParser.CLOSE_PAREN, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_base_selection

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterBase_selection"):
                listener.enterBase_selection(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitBase_selection"):
                listener.exitBase_selection(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitBase_selection"):
                return visitor.visitBase_selection(self)
            else:
                return visitor.visitChildren(self)

    def base_selection(self):

        localctx = FandangoParser.Base_selectionContext(self, self._ctx, self.state)
        self.enterRule(localctx, 48, self.RULE_base_selection)
        try:
            self.state = 761
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [8]:
                self.enterOuterAlt(localctx, 1)
                self.state = 756
                self.match(FandangoParser.NONTERMINAL)
                pass
            elif token in [68]:
                self.enterOuterAlt(localctx, 2)
                self.state = 757
                self.match(FandangoParser.OPEN_PAREN)
                self.state = 758
                self.selector(0)
                self.state = 759
                self.match(FandangoParser.CLOSE_PAREN)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Rs_pairsContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def rs_pair(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(FandangoParser.Rs_pairContext)
            else:
                return self.getTypedRuleContext(FandangoParser.Rs_pairContext, i)

        def COMMA(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.COMMA)
            else:
                return self.getToken(FandangoParser.COMMA, i)

        def getRuleIndex(self):
            return FandangoParser.RULE_rs_pairs

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterRs_pairs"):
                listener.enterRs_pairs(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitRs_pairs"):
                listener.exitRs_pairs(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitRs_pairs"):
                return visitor.visitRs_pairs(self)
            else:
                return visitor.visitChildren(self)

    def rs_pairs(self):

        localctx = FandangoParser.Rs_pairsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 50, self.RULE_rs_pairs)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 763
            self.rs_pair()
            self.state = 768
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 40, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 764
                    self.match(FandangoParser.COMMA)
                    self.state = 765
                    self.rs_pair()
                self.state = 770
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 40, self._ctx)

            self.state = 772
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == 70:
                self.state = 771
                self.match(FandangoParser.COMMA)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Rs_pairContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def STAR(self):
            return self.getToken(FandangoParser.STAR, 0)

        def NONTERMINAL(self):
            return self.getToken(FandangoParser.NONTERMINAL, 0)

        def COLON(self):
            return self.getToken(FandangoParser.COLON, 0)

        def rs_slice(self):
            return self.getTypedRuleContext(FandangoParser.Rs_sliceContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_rs_pair

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterRs_pair"):
                listener.enterRs_pair(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitRs_pair"):
                listener.exitRs_pair(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitRs_pair"):
                return visitor.visitRs_pair(self)
            else:
                return visitor.visitChildren(self)

    def rs_pair(self):

        localctx = FandangoParser.Rs_pairContext(self, self._ctx, self.state)
        self.enterRule(localctx, 52, self.RULE_rs_pair)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 774
            self.match(FandangoParser.STAR)
            self.state = 775
            self.match(FandangoParser.NONTERMINAL)
            self.state = 778
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == 71:
                self.state = 776
                self.match(FandangoParser.COLON)
                self.state = 777
                self.rs_slice()

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Rs_slicesContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def rs_slice(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(FandangoParser.Rs_sliceContext)
            else:
                return self.getTypedRuleContext(FandangoParser.Rs_sliceContext, i)

        def COMMA(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.COMMA)
            else:
                return self.getToken(FandangoParser.COMMA, i)

        def getRuleIndex(self):
            return FandangoParser.RULE_rs_slices

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterRs_slices"):
                listener.enterRs_slices(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitRs_slices"):
                listener.exitRs_slices(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitRs_slices"):
                return visitor.visitRs_slices(self)
            else:
                return visitor.visitChildren(self)

    def rs_slices(self):

        localctx = FandangoParser.Rs_slicesContext(self, self._ctx, self.state)
        self.enterRule(localctx, 54, self.RULE_rs_slices)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 780
            self.rs_slice()
            self.state = 785
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 43, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 781
                    self.match(FandangoParser.COMMA)
                    self.state = 782
                    self.rs_slice()
                self.state = 787
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 43, self._ctx)

            self.state = 789
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == 70:
                self.state = 788
                self.match(FandangoParser.COMMA)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Rs_sliceContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NUMBER(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.NUMBER)
            else:
                return self.getToken(FandangoParser.NUMBER, i)

        def COLON(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.COLON)
            else:
                return self.getToken(FandangoParser.COLON, i)

        def getRuleIndex(self):
            return FandangoParser.RULE_rs_slice

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterRs_slice"):
                listener.enterRs_slice(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitRs_slice"):
                listener.exitRs_slice(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitRs_slice"):
                return visitor.visitRs_slice(self)
            else:
                return visitor.visitChildren(self)

    def rs_slice(self):

        localctx = FandangoParser.Rs_sliceContext(self, self._ctx, self.state)
        self.enterRule(localctx, 56, self.RULE_rs_slice)
        self._la = 0  # Token type
        try:
            self.state = 810
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 50, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 791
                self.match(FandangoParser.NUMBER)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 793
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == 4:
                    self.state = 792
                    self.match(FandangoParser.NUMBER)

                self.state = 795
                self.match(FandangoParser.COLON)
                self.state = 797
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == 4:
                    self.state = 796
                    self.match(FandangoParser.NUMBER)

                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 800
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == 4:
                    self.state = 799
                    self.match(FandangoParser.NUMBER)

                self.state = 802
                self.match(FandangoParser.COLON)
                self.state = 804
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == 4:
                    self.state = 803
                    self.match(FandangoParser.NUMBER)

                self.state = 806
                self.match(FandangoParser.COLON)
                self.state = 808
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == 4:
                    self.state = 807
                    self.match(FandangoParser.NUMBER)

                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class PythonContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def compound_stmt(self):
            return self.getTypedRuleContext(FandangoParser.Compound_stmtContext, 0)

        def simple_stmt(self):
            return self.getTypedRuleContext(FandangoParser.Simple_stmtContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_python

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterPython"):
                listener.enterPython(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitPython"):
                listener.exitPython(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitPython"):
                return visitor.visitPython(self)
            else:
                return visitor.visitChildren(self)

    def python(self):

        localctx = FandangoParser.PythonContext(self, self._ctx, self.state)
        self.enterRule(localctx, 58, self.RULE_python)
        try:
            self.state = 814
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [12, 16, 18, 25, 28, 33, 42, 45, 47, 97]:
                self.enterOuterAlt(localctx, 1)
                self.state = 812
                self.compound_stmt()
                pass
            elif token in [
                3,
                4,
                8,
                11,
                13,
                14,
                17,
                19,
                23,
                26,
                27,
                29,
                32,
                34,
                35,
                36,
                38,
                39,
                40,
                41,
                43,
                48,
                52,
                66,
                67,
                68,
                75,
                77,
                82,
                83,
                87,
                88,
                118,
            ]:
                self.enterOuterAlt(localctx, 2)
                self.state = 813
                self.simple_stmt()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Python_tagContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def stmt(self):
            return self.getTypedRuleContext(FandangoParser.StmtContext, 0)

        def NEWLINE(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.NEWLINE)
            else:
                return self.getToken(FandangoParser.NEWLINE, i)

        def getRuleIndex(self):
            return FandangoParser.RULE_python_tag

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterPython_tag"):
                listener.enterPython_tag(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitPython_tag"):
                listener.exitPython_tag(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitPython_tag"):
                return visitor.visitPython_tag(self)
            else:
                return visitor.visitChildren(self)

    def python_tag(self):

        localctx = FandangoParser.Python_tagContext(self, self._ctx, self.state)
        self.enterRule(localctx, 60, self.RULE_python_tag)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 819
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la == 114:
                self.state = 816
                self.match(FandangoParser.NEWLINE)
                self.state = 821
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 822
            self.stmt()
            self.state = 826
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la == 114:
                self.state = 823
                self.match(FandangoParser.NEWLINE)
                self.state = 828
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Python_fileContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def statements(self):
            return self.getTypedRuleContext(FandangoParser.StatementsContext, 0)

        def EOF(self):
            return self.getToken(FandangoParser.EOF, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_python_file

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterPython_file"):
                listener.enterPython_file(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitPython_file"):
                listener.exitPython_file(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitPython_file"):
                return visitor.visitPython_file(self)
            else:
                return visitor.visitChildren(self)

    def python_file(self):

        localctx = FandangoParser.Python_fileContext(self, self._ctx, self.state)
        self.enterRule(localctx, 62, self.RULE_python_file)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 830
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if (((_la) & ~0x3F) == 0 and ((1 << _la) & 4978447966239000) != 0) or (
                (((_la - 66)) & ~0x3F) == 0
                and ((1 << (_la - 66)) & 4785076758055431) != 0
            ):
                self.state = 829
                self.statements()

            self.state = 833
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 55, self._ctx)
            if la_ == 1:
                self.state = 832
                self.match(FandangoParser.EOF)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class InteractiveContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def statement_newline(self):
            return self.getTypedRuleContext(FandangoParser.Statement_newlineContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_interactive

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterInteractive"):
                listener.enterInteractive(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitInteractive"):
                listener.exitInteractive(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitInteractive"):
                return visitor.visitInteractive(self)
            else:
                return visitor.visitChildren(self)

    def interactive(self):

        localctx = FandangoParser.InteractiveContext(self, self._ctx, self.state)
        self.enterRule(localctx, 64, self.RULE_interactive)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 835
            self.statement_newline()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class EvalContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def expressions(self):
            return self.getTypedRuleContext(FandangoParser.ExpressionsContext, 0)

        def NEWLINE(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.NEWLINE)
            else:
                return self.getToken(FandangoParser.NEWLINE, i)

        def EOF(self):
            return self.getToken(FandangoParser.EOF, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_eval

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterEval"):
                listener.enterEval(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitEval"):
                listener.exitEval(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitEval"):
                return visitor.visitEval(self)
            else:
                return visitor.visitChildren(self)

    def eval_(self):

        localctx = FandangoParser.EvalContext(self, self._ctx, self.state)
        self.enterRule(localctx, 66, self.RULE_eval)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 837
            self.expressions()
            self.state = 841
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la == 114:
                self.state = 838
                self.match(FandangoParser.NEWLINE)
                self.state = 843
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 845
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 57, self._ctx)
            if la_ == 1:
                self.state = 844
                self.match(FandangoParser.EOF)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Func_typeContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def OPEN_PAREN(self):
            return self.getToken(FandangoParser.OPEN_PAREN, 0)

        def CLOSE_PAREN(self):
            return self.getToken(FandangoParser.CLOSE_PAREN, 0)

        def ARROW(self):
            return self.getToken(FandangoParser.ARROW, 0)

        def expression(self):
            return self.getTypedRuleContext(FandangoParser.ExpressionContext, 0)

        def type_expressions(self):
            return self.getTypedRuleContext(FandangoParser.Type_expressionsContext, 0)

        def NEWLINE(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.NEWLINE)
            else:
                return self.getToken(FandangoParser.NEWLINE, i)

        def EOF(self):
            return self.getToken(FandangoParser.EOF, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_func_type

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterFunc_type"):
                listener.enterFunc_type(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitFunc_type"):
                listener.exitFunc_type(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitFunc_type"):
                return visitor.visitFunc_type(self)
            else:
                return visitor.visitChildren(self)

    def func_type(self):

        localctx = FandangoParser.Func_typeContext(self, self._ctx, self.state)
        self.enterRule(localctx, 68, self.RULE_func_type)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 847
            self.match(FandangoParser.OPEN_PAREN)
            self.state = 849
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if (((_la) & ~0x3F) == 0 and ((1 << _la) & 4505888853336344) != 0) or (
                (((_la - 66)) & ~0x3F) == 0
                and ((1 << (_la - 66)) & 4503599633861255) != 0
            ):
                self.state = 848
                self.type_expressions()

            self.state = 851
            self.match(FandangoParser.CLOSE_PAREN)
            self.state = 852
            self.match(FandangoParser.ARROW)
            self.state = 853
            self.expression()
            self.state = 857
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la == 114:
                self.state = 854
                self.match(FandangoParser.NEWLINE)
                self.state = 859
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 861
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 60, self._ctx)
            if la_ == 1:
                self.state = 860
                self.match(FandangoParser.EOF)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class StatementsContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def stmt(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(FandangoParser.StmtContext)
            else:
                return self.getTypedRuleContext(FandangoParser.StmtContext, i)

        def NEWLINE(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.NEWLINE)
            else:
                return self.getToken(FandangoParser.NEWLINE, i)

        def getRuleIndex(self):
            return FandangoParser.RULE_statements

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterStatements"):
                listener.enterStatements(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitStatements"):
                listener.exitStatements(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitStatements"):
                return visitor.visitStatements(self)
            else:
                return visitor.visitChildren(self)

    def statements(self):

        localctx = FandangoParser.StatementsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 70, self.RULE_statements)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 865
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 865
                self._errHandler.sync(self)
                token = self._input.LA(1)
                if token in [
                    3,
                    4,
                    8,
                    11,
                    12,
                    13,
                    14,
                    16,
                    17,
                    18,
                    19,
                    23,
                    25,
                    26,
                    27,
                    28,
                    29,
                    32,
                    33,
                    34,
                    35,
                    36,
                    38,
                    39,
                    40,
                    41,
                    42,
                    43,
                    45,
                    47,
                    48,
                    52,
                    66,
                    67,
                    68,
                    75,
                    77,
                    82,
                    83,
                    87,
                    88,
                    97,
                    118,
                ]:
                    self.state = 863
                    self.stmt()
                    pass
                elif token in [114]:
                    self.state = 864
                    self.match(FandangoParser.NEWLINE)
                    pass
                else:
                    raise NoViableAltException(self)

                self.state = 867
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not (
                    (((_la) & ~0x3F) == 0 and ((1 << _la) & 4978447966239000) != 0)
                    or (
                        (((_la - 66)) & ~0x3F) == 0
                        and ((1 << (_la - 66)) & 4785076758055431) != 0
                    )
                ):
                    break

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class StmtContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def compound_stmt(self):
            return self.getTypedRuleContext(FandangoParser.Compound_stmtContext, 0)

        def simple_stmts(self):
            return self.getTypedRuleContext(FandangoParser.Simple_stmtsContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_stmt

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterStmt"):
                listener.enterStmt(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitStmt"):
                listener.exitStmt(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitStmt"):
                return visitor.visitStmt(self)
            else:
                return visitor.visitChildren(self)

    def stmt(self):

        localctx = FandangoParser.StmtContext(self, self._ctx, self.state)
        self.enterRule(localctx, 72, self.RULE_stmt)
        try:
            self.state = 871
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [12, 16, 18, 25, 28, 33, 42, 45, 47, 97]:
                self.enterOuterAlt(localctx, 1)
                self.state = 869
                self.compound_stmt()
                pass
            elif token in [
                3,
                4,
                8,
                11,
                13,
                14,
                17,
                19,
                23,
                26,
                27,
                29,
                32,
                34,
                35,
                36,
                38,
                39,
                40,
                41,
                43,
                48,
                52,
                66,
                67,
                68,
                75,
                77,
                82,
                83,
                87,
                88,
                118,
            ]:
                self.enterOuterAlt(localctx, 2)
                self.state = 870
                self.simple_stmts()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Statement_newlineContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def compound_stmt(self):
            return self.getTypedRuleContext(FandangoParser.Compound_stmtContext, 0)

        def NEWLINE(self):
            return self.getToken(FandangoParser.NEWLINE, 0)

        def simple_stmts(self):
            return self.getTypedRuleContext(FandangoParser.Simple_stmtsContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_statement_newline

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterStatement_newline"):
                listener.enterStatement_newline(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitStatement_newline"):
                listener.exitStatement_newline(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitStatement_newline"):
                return visitor.visitStatement_newline(self)
            else:
                return visitor.visitChildren(self)

    def statement_newline(self):

        localctx = FandangoParser.Statement_newlineContext(self, self._ctx, self.state)
        self.enterRule(localctx, 74, self.RULE_statement_newline)
        try:
            self.state = 878
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [12, 16, 18, 25, 28, 33, 42, 45, 47, 97]:
                self.enterOuterAlt(localctx, 1)
                self.state = 873
                self.compound_stmt()
                self.state = 874
                self.match(FandangoParser.NEWLINE)
                pass
            elif token in [
                3,
                4,
                8,
                11,
                13,
                14,
                17,
                19,
                23,
                26,
                27,
                29,
                32,
                34,
                35,
                36,
                38,
                39,
                40,
                41,
                43,
                48,
                52,
                66,
                67,
                68,
                75,
                77,
                82,
                83,
                87,
                88,
                118,
            ]:
                self.enterOuterAlt(localctx, 2)
                self.state = 876
                self.simple_stmts()
                pass
            elif token in [114]:
                self.enterOuterAlt(localctx, 3)
                self.state = 877
                self.match(FandangoParser.NEWLINE)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Simple_stmtsContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def simple_stmt(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(FandangoParser.Simple_stmtContext)
            else:
                return self.getTypedRuleContext(FandangoParser.Simple_stmtContext, i)

        def EOF(self):
            return self.getToken(FandangoParser.EOF, 0)

        def SEMI_COLON(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.SEMI_COLON)
            else:
                return self.getToken(FandangoParser.SEMI_COLON, i)

        def NEWLINE(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.NEWLINE)
            else:
                return self.getToken(FandangoParser.NEWLINE, i)

        def getRuleIndex(self):
            return FandangoParser.RULE_simple_stmts

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterSimple_stmts"):
                listener.enterSimple_stmts(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitSimple_stmts"):
                listener.exitSimple_stmts(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitSimple_stmts"):
                return visitor.visitSimple_stmts(self)
            else:
                return visitor.visitChildren(self)

    def simple_stmts(self):

        localctx = FandangoParser.Simple_stmtsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 76, self.RULE_simple_stmts)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 880
            self.simple_stmt()
            self.state = 885
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 65, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 881
                    self.match(FandangoParser.SEMI_COLON)
                    self.state = 882
                    self.simple_stmt()
                self.state = 887
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 65, self._ctx)

            self.state = 889
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == 72:
                self.state = 888
                self.match(FandangoParser.SEMI_COLON)

            self.state = 903
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 69, self._ctx)
            if la_ == 1:
                self.state = 892
                self._errHandler.sync(self)
                _alt = 1
                while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                    if _alt == 1:
                        self.state = 891
                        self.match(FandangoParser.NEWLINE)

                    else:
                        raise NoViableAltException(self)
                    self.state = 894
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input, 67, self._ctx)

                pass

            elif la_ == 2:
                self.state = 899
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la == 114:
                    self.state = 896
                    self.match(FandangoParser.NEWLINE)
                    self.state = 901
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                self.state = 902
                self.match(FandangoParser.EOF)
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Simple_stmtContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def assignment(self):
            return self.getTypedRuleContext(FandangoParser.AssignmentContext, 0)

        def type_alias(self):
            return self.getTypedRuleContext(FandangoParser.Type_aliasContext, 0)

        def star_expressions(self):
            return self.getTypedRuleContext(FandangoParser.Star_expressionsContext, 0)

        def return_stmt(self):
            return self.getTypedRuleContext(FandangoParser.Return_stmtContext, 0)

        def import_stmt(self):
            return self.getTypedRuleContext(FandangoParser.Import_stmtContext, 0)

        def raise_stmt(self):
            return self.getTypedRuleContext(FandangoParser.Raise_stmtContext, 0)

        def PASS(self):
            return self.getToken(FandangoParser.PASS, 0)

        def del_stmt(self):
            return self.getTypedRuleContext(FandangoParser.Del_stmtContext, 0)

        def yield_stmt(self):
            return self.getTypedRuleContext(FandangoParser.Yield_stmtContext, 0)

        def assert_stmt(self):
            return self.getTypedRuleContext(FandangoParser.Assert_stmtContext, 0)

        def BREAK(self):
            return self.getToken(FandangoParser.BREAK, 0)

        def CONTINUE(self):
            return self.getToken(FandangoParser.CONTINUE, 0)

        def global_stmt(self):
            return self.getTypedRuleContext(FandangoParser.Global_stmtContext, 0)

        def nonlocal_stmt(self):
            return self.getTypedRuleContext(FandangoParser.Nonlocal_stmtContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_simple_stmt

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterSimple_stmt"):
                listener.enterSimple_stmt(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitSimple_stmt"):
                listener.exitSimple_stmt(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitSimple_stmt"):
                return visitor.visitSimple_stmt(self)
            else:
                return visitor.visitChildren(self)

    def simple_stmt(self):

        localctx = FandangoParser.Simple_stmtContext(self, self._ctx, self.state)
        self.enterRule(localctx, 78, self.RULE_simple_stmt)
        try:
            self.state = 919
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 70, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 905
                self.assignment()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 906
                self.type_alias()
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 907
                self.star_expressions()
                pass

            elif la_ == 4:
                self.enterOuterAlt(localctx, 4)
                self.state = 908
                self.return_stmt()
                pass

            elif la_ == 5:
                self.enterOuterAlt(localctx, 5)
                self.state = 909
                self.import_stmt()
                pass

            elif la_ == 6:
                self.enterOuterAlt(localctx, 6)
                self.state = 910
                self.raise_stmt()
                pass

            elif la_ == 7:
                self.enterOuterAlt(localctx, 7)
                self.state = 911
                self.match(FandangoParser.PASS)
                pass

            elif la_ == 8:
                self.enterOuterAlt(localctx, 8)
                self.state = 912
                self.del_stmt()
                pass

            elif la_ == 9:
                self.enterOuterAlt(localctx, 9)
                self.state = 913
                self.yield_stmt()
                pass

            elif la_ == 10:
                self.enterOuterAlt(localctx, 10)
                self.state = 914
                self.assert_stmt()
                pass

            elif la_ == 11:
                self.enterOuterAlt(localctx, 11)
                self.state = 915
                self.match(FandangoParser.BREAK)
                pass

            elif la_ == 12:
                self.enterOuterAlt(localctx, 12)
                self.state = 916
                self.match(FandangoParser.CONTINUE)
                pass

            elif la_ == 13:
                self.enterOuterAlt(localctx, 13)
                self.state = 917
                self.global_stmt()
                pass

            elif la_ == 14:
                self.enterOuterAlt(localctx, 14)
                self.state = 918
                self.nonlocal_stmt()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Compound_stmtContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def function_def(self):
            return self.getTypedRuleContext(FandangoParser.Function_defContext, 0)

        def if_stmt(self):
            return self.getTypedRuleContext(FandangoParser.If_stmtContext, 0)

        def class_def(self):
            return self.getTypedRuleContext(FandangoParser.Class_defContext, 0)

        def with_stmt(self):
            return self.getTypedRuleContext(FandangoParser.With_stmtContext, 0)

        def for_stmt(self):
            return self.getTypedRuleContext(FandangoParser.For_stmtContext, 0)

        def try_stmt(self):
            return self.getTypedRuleContext(FandangoParser.Try_stmtContext, 0)

        def while_stmt(self):
            return self.getTypedRuleContext(FandangoParser.While_stmtContext, 0)

        def match_stmt(self):
            return self.getTypedRuleContext(FandangoParser.Match_stmtContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_compound_stmt

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterCompound_stmt"):
                listener.enterCompound_stmt(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitCompound_stmt"):
                listener.exitCompound_stmt(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitCompound_stmt"):
                return visitor.visitCompound_stmt(self)
            else:
                return visitor.visitChildren(self)

    def compound_stmt(self):

        localctx = FandangoParser.Compound_stmtContext(self, self._ctx, self.state)
        self.enterRule(localctx, 80, self.RULE_compound_stmt)
        try:
            self.state = 929
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 71, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 921
                self.function_def()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 922
                self.if_stmt()
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 923
                self.class_def()
                pass

            elif la_ == 4:
                self.enterOuterAlt(localctx, 4)
                self.state = 924
                self.with_stmt()
                pass

            elif la_ == 5:
                self.enterOuterAlt(localctx, 5)
                self.state = 925
                self.for_stmt()
                pass

            elif la_ == 6:
                self.enterOuterAlt(localctx, 6)
                self.state = 926
                self.try_stmt()
                pass

            elif la_ == 7:
                self.enterOuterAlt(localctx, 7)
                self.state = 927
                self.while_stmt()
                pass

            elif la_ == 8:
                self.enterOuterAlt(localctx, 8)
                self.state = 928
                self.match_stmt()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class AssignmentContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NAME(self):
            return self.getToken(FandangoParser.NAME, 0)

        def COLON(self):
            return self.getToken(FandangoParser.COLON, 0)

        def expression(self):
            return self.getTypedRuleContext(FandangoParser.ExpressionContext, 0)

        def ASSIGN(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.ASSIGN)
            else:
                return self.getToken(FandangoParser.ASSIGN, i)

        def annotated_rhs(self):
            return self.getTypedRuleContext(FandangoParser.Annotated_rhsContext, 0)

        def OPEN_PAREN(self):
            return self.getToken(FandangoParser.OPEN_PAREN, 0)

        def single_target(self):
            return self.getTypedRuleContext(FandangoParser.Single_targetContext, 0)

        def CLOSE_PAREN(self):
            return self.getToken(FandangoParser.CLOSE_PAREN, 0)

        def single_subscript_attribute_target(self):
            return self.getTypedRuleContext(
                FandangoParser.Single_subscript_attribute_targetContext, 0
            )

        def yield_expr(self):
            return self.getTypedRuleContext(FandangoParser.Yield_exprContext, 0)

        def star_expressions(self):
            return self.getTypedRuleContext(FandangoParser.Star_expressionsContext, 0)

        def star_targets(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(FandangoParser.Star_targetsContext)
            else:
                return self.getTypedRuleContext(FandangoParser.Star_targetsContext, i)

        def augassign(self):
            return self.getTypedRuleContext(FandangoParser.AugassignContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_assignment

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterAssignment"):
                listener.enterAssignment(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitAssignment"):
                listener.exitAssignment(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitAssignment"):
                return visitor.visitAssignment(self)
            else:
                return visitor.visitChildren(self)

    def assignment(self):

        localctx = FandangoParser.AssignmentContext(self, self._ctx, self.state)
        self.enterRule(localctx, 82, self.RULE_assignment)
        self._la = 0  # Token type
        try:
            self.state = 968
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 78, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 931
                self.match(FandangoParser.NAME)
                self.state = 932
                self.match(FandangoParser.COLON)
                self.state = 933
                self.expression()
                self.state = 936
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == 74:
                    self.state = 934
                    self.match(FandangoParser.ASSIGN)
                    self.state = 935
                    self.annotated_rhs()

                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 943
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input, 73, self._ctx)
                if la_ == 1:
                    self.state = 938
                    self.match(FandangoParser.OPEN_PAREN)
                    self.state = 939
                    self.single_target()
                    self.state = 940
                    self.match(FandangoParser.CLOSE_PAREN)
                    pass

                elif la_ == 2:
                    self.state = 942
                    self.single_subscript_attribute_target()
                    pass

                self.state = 945
                self.match(FandangoParser.COLON)
                self.state = 946
                self.expression()
                self.state = 949
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == 74:
                    self.state = 947
                    self.match(FandangoParser.ASSIGN)
                    self.state = 948
                    self.annotated_rhs()

                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 954
                self._errHandler.sync(self)
                _alt = 1
                while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                    if _alt == 1:
                        self.state = 951
                        self.star_targets()
                        self.state = 952
                        self.match(FandangoParser.ASSIGN)

                    else:
                        raise NoViableAltException(self)
                    self.state = 956
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input, 75, self._ctx)

                self.state = 960
                self._errHandler.sync(self)
                token = self._input.LA(1)
                if token in [48]:
                    self.state = 958
                    self.yield_expr()
                    pass
                elif token in [
                    3,
                    4,
                    8,
                    13,
                    23,
                    32,
                    34,
                    36,
                    41,
                    52,
                    66,
                    67,
                    68,
                    75,
                    77,
                    82,
                    83,
                    87,
                    88,
                    118,
                ]:
                    self.state = 959
                    self.star_expressions()
                    pass
                else:
                    raise NoViableAltException(self)

                pass

            elif la_ == 4:
                self.enterOuterAlt(localctx, 4)
                self.state = 962
                self.single_target()
                self.state = 963
                self.augassign()
                self.state = 966
                self._errHandler.sync(self)
                token = self._input.LA(1)
                if token in [48]:
                    self.state = 964
                    self.yield_expr()
                    pass
                elif token in [
                    3,
                    4,
                    8,
                    13,
                    23,
                    32,
                    34,
                    36,
                    41,
                    52,
                    66,
                    67,
                    68,
                    75,
                    77,
                    82,
                    83,
                    87,
                    88,
                    118,
                ]:
                    self.state = 965
                    self.star_expressions()
                    pass
                else:
                    raise NoViableAltException(self)

                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Annotated_rhsContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def yield_expr(self):
            return self.getTypedRuleContext(FandangoParser.Yield_exprContext, 0)

        def star_expressions(self):
            return self.getTypedRuleContext(FandangoParser.Star_expressionsContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_annotated_rhs

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterAnnotated_rhs"):
                listener.enterAnnotated_rhs(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitAnnotated_rhs"):
                listener.exitAnnotated_rhs(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitAnnotated_rhs"):
                return visitor.visitAnnotated_rhs(self)
            else:
                return visitor.visitChildren(self)

    def annotated_rhs(self):

        localctx = FandangoParser.Annotated_rhsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 84, self.RULE_annotated_rhs)
        try:
            self.state = 972
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [48]:
                self.enterOuterAlt(localctx, 1)
                self.state = 970
                self.yield_expr()
                pass
            elif token in [
                3,
                4,
                8,
                13,
                23,
                32,
                34,
                36,
                41,
                52,
                66,
                67,
                68,
                75,
                77,
                82,
                83,
                87,
                88,
                118,
            ]:
                self.enterOuterAlt(localctx, 2)
                self.state = 971
                self.star_expressions()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class AugassignContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ADD_ASSIGN(self):
            return self.getToken(FandangoParser.ADD_ASSIGN, 0)

        def SUB_ASSIGN(self):
            return self.getToken(FandangoParser.SUB_ASSIGN, 0)

        def MULT_ASSIGN(self):
            return self.getToken(FandangoParser.MULT_ASSIGN, 0)

        def AT_ASSIGN(self):
            return self.getToken(FandangoParser.AT_ASSIGN, 0)

        def DIV_ASSIGN(self):
            return self.getToken(FandangoParser.DIV_ASSIGN, 0)

        def MOD_ASSIGN(self):
            return self.getToken(FandangoParser.MOD_ASSIGN, 0)

        def AND_ASSIGN(self):
            return self.getToken(FandangoParser.AND_ASSIGN, 0)

        def OR_ASSIGN(self):
            return self.getToken(FandangoParser.OR_ASSIGN, 0)

        def XOR_ASSIGN(self):
            return self.getToken(FandangoParser.XOR_ASSIGN, 0)

        def LEFT_SHIFT_ASSIGN(self):
            return self.getToken(FandangoParser.LEFT_SHIFT_ASSIGN, 0)

        def RIGHT_SHIFT_ASSIGN(self):
            return self.getToken(FandangoParser.RIGHT_SHIFT_ASSIGN, 0)

        def POWER_ASSIGN(self):
            return self.getToken(FandangoParser.POWER_ASSIGN, 0)

        def IDIV_ASSIGN(self):
            return self.getToken(FandangoParser.IDIV_ASSIGN, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_augassign

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterAugassign"):
                listener.enterAugassign(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitAugassign"):
                listener.exitAugassign(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitAugassign"):
                return visitor.visitAugassign(self)
            else:
                return visitor.visitChildren(self)

    def augassign(self):

        localctx = FandangoParser.AugassignContext(self, self._ctx, self.state)
        self.enterRule(localctx, 86, self.RULE_augassign)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 974
            _la = self._input.LA(1)
            if not (((((_la - 99)) & ~0x3F) == 0 and ((1 << (_la - 99)) & 8191) != 0)):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Return_stmtContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def RETURN(self):
            return self.getToken(FandangoParser.RETURN, 0)

        def star_expressions(self):
            return self.getTypedRuleContext(FandangoParser.Star_expressionsContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_return_stmt

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterReturn_stmt"):
                listener.enterReturn_stmt(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitReturn_stmt"):
                listener.exitReturn_stmt(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitReturn_stmt"):
                return visitor.visitReturn_stmt(self)
            else:
                return visitor.visitChildren(self)

    def return_stmt(self):

        localctx = FandangoParser.Return_stmtContext(self, self._ctx, self.state)
        self.enterRule(localctx, 88, self.RULE_return_stmt)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 976
            self.match(FandangoParser.RETURN)
            self.state = 978
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 80, self._ctx)
            if la_ == 1:
                self.state = 977
                self.star_expressions()

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Raise_stmtContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def RAISE(self):
            return self.getToken(FandangoParser.RAISE, 0)

        def expression(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(FandangoParser.ExpressionContext)
            else:
                return self.getTypedRuleContext(FandangoParser.ExpressionContext, i)

        def FROM(self):
            return self.getToken(FandangoParser.FROM, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_raise_stmt

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterRaise_stmt"):
                listener.enterRaise_stmt(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitRaise_stmt"):
                listener.exitRaise_stmt(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitRaise_stmt"):
                return visitor.visitRaise_stmt(self)
            else:
                return visitor.visitChildren(self)

    def raise_stmt(self):

        localctx = FandangoParser.Raise_stmtContext(self, self._ctx, self.state)
        self.enterRule(localctx, 90, self.RULE_raise_stmt)
        try:
            self.state = 987
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 82, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 980
                self.match(FandangoParser.RAISE)
                self.state = 981
                self.expression()
                self.state = 984
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input, 81, self._ctx)
                if la_ == 1:
                    self.state = 982
                    self.match(FandangoParser.FROM)
                    self.state = 983
                    self.expression()

                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 986
                self.match(FandangoParser.RAISE)
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Global_stmtContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def GLOBAL(self):
            return self.getToken(FandangoParser.GLOBAL, 0)

        def NAME(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.NAME)
            else:
                return self.getToken(FandangoParser.NAME, i)

        def COMMA(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.COMMA)
            else:
                return self.getToken(FandangoParser.COMMA, i)

        def getRuleIndex(self):
            return FandangoParser.RULE_global_stmt

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterGlobal_stmt"):
                listener.enterGlobal_stmt(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitGlobal_stmt"):
                listener.exitGlobal_stmt(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitGlobal_stmt"):
                return visitor.visitGlobal_stmt(self)
            else:
                return visitor.visitChildren(self)

    def global_stmt(self):

        localctx = FandangoParser.Global_stmtContext(self, self._ctx, self.state)
        self.enterRule(localctx, 92, self.RULE_global_stmt)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 989
            self.match(FandangoParser.GLOBAL)
            self.state = 990
            self.match(FandangoParser.NAME)
            self.state = 995
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la == 70:
                self.state = 991
                self.match(FandangoParser.COMMA)
                self.state = 992
                self.match(FandangoParser.NAME)
                self.state = 997
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Nonlocal_stmtContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NONLOCAL(self):
            return self.getToken(FandangoParser.NONLOCAL, 0)

        def NAME(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.NAME)
            else:
                return self.getToken(FandangoParser.NAME, i)

        def COMMA(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.COMMA)
            else:
                return self.getToken(FandangoParser.COMMA, i)

        def getRuleIndex(self):
            return FandangoParser.RULE_nonlocal_stmt

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterNonlocal_stmt"):
                listener.enterNonlocal_stmt(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitNonlocal_stmt"):
                listener.exitNonlocal_stmt(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitNonlocal_stmt"):
                return visitor.visitNonlocal_stmt(self)
            else:
                return visitor.visitChildren(self)

    def nonlocal_stmt(self):

        localctx = FandangoParser.Nonlocal_stmtContext(self, self._ctx, self.state)
        self.enterRule(localctx, 94, self.RULE_nonlocal_stmt)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 998
            self.match(FandangoParser.NONLOCAL)
            self.state = 999
            self.match(FandangoParser.NAME)
            self.state = 1004
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la == 70:
                self.state = 1000
                self.match(FandangoParser.COMMA)
                self.state = 1001
                self.match(FandangoParser.NAME)
                self.state = 1006
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Del_stmtContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def DEL(self):
            return self.getToken(FandangoParser.DEL, 0)

        def del_targets(self):
            return self.getTypedRuleContext(FandangoParser.Del_targetsContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_del_stmt

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterDel_stmt"):
                listener.enterDel_stmt(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitDel_stmt"):
                listener.exitDel_stmt(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitDel_stmt"):
                return visitor.visitDel_stmt(self)
            else:
                return visitor.visitChildren(self)

    def del_stmt(self):

        localctx = FandangoParser.Del_stmtContext(self, self._ctx, self.state)
        self.enterRule(localctx, 96, self.RULE_del_stmt)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1007
            self.match(FandangoParser.DEL)
            self.state = 1008
            self.del_targets()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Yield_stmtContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def yield_expr(self):
            return self.getTypedRuleContext(FandangoParser.Yield_exprContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_yield_stmt

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterYield_stmt"):
                listener.enterYield_stmt(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitYield_stmt"):
                listener.exitYield_stmt(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitYield_stmt"):
                return visitor.visitYield_stmt(self)
            else:
                return visitor.visitChildren(self)

    def yield_stmt(self):

        localctx = FandangoParser.Yield_stmtContext(self, self._ctx, self.state)
        self.enterRule(localctx, 98, self.RULE_yield_stmt)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1010
            self.yield_expr()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Assert_stmtContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ASSERT(self):
            return self.getToken(FandangoParser.ASSERT, 0)

        def expression(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(FandangoParser.ExpressionContext)
            else:
                return self.getTypedRuleContext(FandangoParser.ExpressionContext, i)

        def COMMA(self):
            return self.getToken(FandangoParser.COMMA, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_assert_stmt

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterAssert_stmt"):
                listener.enterAssert_stmt(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitAssert_stmt"):
                listener.exitAssert_stmt(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitAssert_stmt"):
                return visitor.visitAssert_stmt(self)
            else:
                return visitor.visitChildren(self)

    def assert_stmt(self):

        localctx = FandangoParser.Assert_stmtContext(self, self._ctx, self.state)
        self.enterRule(localctx, 100, self.RULE_assert_stmt)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1012
            self.match(FandangoParser.ASSERT)
            self.state = 1013
            self.expression()
            self.state = 1016
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == 70:
                self.state = 1014
                self.match(FandangoParser.COMMA)
                self.state = 1015
                self.expression()

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Import_stmtContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def import_name(self):
            return self.getTypedRuleContext(FandangoParser.Import_nameContext, 0)

        def import_from(self):
            return self.getTypedRuleContext(FandangoParser.Import_fromContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_import_stmt

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterImport_stmt"):
                listener.enterImport_stmt(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitImport_stmt"):
                listener.exitImport_stmt(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitImport_stmt"):
                return visitor.visitImport_stmt(self)
            else:
                return visitor.visitChildren(self)

    def import_stmt(self):

        localctx = FandangoParser.Import_stmtContext(self, self._ctx, self.state)
        self.enterRule(localctx, 102, self.RULE_import_stmt)
        try:
            self.state = 1020
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [29]:
                self.enterOuterAlt(localctx, 1)
                self.state = 1018
                self.import_name()
                pass
            elif token in [26]:
                self.enterOuterAlt(localctx, 2)
                self.state = 1019
                self.import_from()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Import_nameContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def IMPORT(self):
            return self.getToken(FandangoParser.IMPORT, 0)

        def dotted_as_names(self):
            return self.getTypedRuleContext(FandangoParser.Dotted_as_namesContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_import_name

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterImport_name"):
                listener.enterImport_name(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitImport_name"):
                listener.exitImport_name(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitImport_name"):
                return visitor.visitImport_name(self)
            else:
                return visitor.visitChildren(self)

    def import_name(self):

        localctx = FandangoParser.Import_nameContext(self, self._ctx, self.state)
        self.enterRule(localctx, 104, self.RULE_import_name)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1022
            self.match(FandangoParser.IMPORT)
            self.state = 1023
            self.dotted_as_names()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Import_fromContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def FROM(self):
            return self.getToken(FandangoParser.FROM, 0)

        def dotted_name(self):
            return self.getTypedRuleContext(FandangoParser.Dotted_nameContext, 0)

        def IMPORT(self):
            return self.getToken(FandangoParser.IMPORT, 0)

        def import_from_targets(self):
            return self.getTypedRuleContext(
                FandangoParser.Import_from_targetsContext, 0
            )

        def DOT(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.DOT)
            else:
                return self.getToken(FandangoParser.DOT, i)

        def ELLIPSIS(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.ELLIPSIS)
            else:
                return self.getToken(FandangoParser.ELLIPSIS, i)

        def getRuleIndex(self):
            return FandangoParser.RULE_import_from

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterImport_from"):
                listener.enterImport_from(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitImport_from"):
                listener.exitImport_from(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitImport_from"):
                return visitor.visitImport_from(self)
            else:
                return visitor.visitChildren(self)

    def import_from(self):

        localctx = FandangoParser.Import_fromContext(self, self._ctx, self.state)
        self.enterRule(localctx, 106, self.RULE_import_from)
        self._la = 0  # Token type
        try:
            self.state = 1044
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 89, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 1025
                self.match(FandangoParser.FROM)
                self.state = 1029
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la == 64 or _la == 66:
                    self.state = 1026
                    _la = self._input.LA(1)
                    if not (_la == 64 or _la == 66):
                        self._errHandler.recoverInline(self)
                    else:
                        self._errHandler.reportMatch(self)
                        self.consume()
                    self.state = 1031
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                self.state = 1032
                self.dotted_name(0)
                self.state = 1033
                self.match(FandangoParser.IMPORT)
                self.state = 1034
                self.import_from_targets()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 1036
                self.match(FandangoParser.FROM)
                self.state = 1038
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while True:
                    self.state = 1037
                    _la = self._input.LA(1)
                    if not (_la == 64 or _la == 66):
                        self._errHandler.recoverInline(self)
                    else:
                        self._errHandler.reportMatch(self)
                        self.consume()
                    self.state = 1040
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)
                    if not (_la == 64 or _la == 66):
                        break

                self.state = 1042
                self.match(FandangoParser.IMPORT)
                self.state = 1043
                self.import_from_targets()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Import_from_targetsContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def OPEN_PAREN(self):
            return self.getToken(FandangoParser.OPEN_PAREN, 0)

        def import_from_as_names(self):
            return self.getTypedRuleContext(
                FandangoParser.Import_from_as_namesContext, 0
            )

        def CLOSE_PAREN(self):
            return self.getToken(FandangoParser.CLOSE_PAREN, 0)

        def COMMA(self):
            return self.getToken(FandangoParser.COMMA, 0)

        def STAR(self):
            return self.getToken(FandangoParser.STAR, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_import_from_targets

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterImport_from_targets"):
                listener.enterImport_from_targets(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitImport_from_targets"):
                listener.exitImport_from_targets(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitImport_from_targets"):
                return visitor.visitImport_from_targets(self)
            else:
                return visitor.visitChildren(self)

    def import_from_targets(self):

        localctx = FandangoParser.Import_from_targetsContext(
            self, self._ctx, self.state
        )
        self.enterRule(localctx, 108, self.RULE_import_from_targets)
        self._la = 0  # Token type
        try:
            self.state = 1055
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [68]:
                self.enterOuterAlt(localctx, 1)
                self.state = 1046
                self.match(FandangoParser.OPEN_PAREN)
                self.state = 1047
                self.import_from_as_names()
                self.state = 1049
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == 70:
                    self.state = 1048
                    self.match(FandangoParser.COMMA)

                self.state = 1051
                self.match(FandangoParser.CLOSE_PAREN)
                pass
            elif token in [52]:
                self.enterOuterAlt(localctx, 2)
                self.state = 1053
                self.import_from_as_names()
                pass
            elif token in [67]:
                self.enterOuterAlt(localctx, 3)
                self.state = 1054
                self.match(FandangoParser.STAR)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Import_from_as_namesContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def import_from_as_name(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(
                    FandangoParser.Import_from_as_nameContext
                )
            else:
                return self.getTypedRuleContext(
                    FandangoParser.Import_from_as_nameContext, i
                )

        def COMMA(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.COMMA)
            else:
                return self.getToken(FandangoParser.COMMA, i)

        def getRuleIndex(self):
            return FandangoParser.RULE_import_from_as_names

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterImport_from_as_names"):
                listener.enterImport_from_as_names(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitImport_from_as_names"):
                listener.exitImport_from_as_names(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitImport_from_as_names"):
                return visitor.visitImport_from_as_names(self)
            else:
                return visitor.visitChildren(self)

    def import_from_as_names(self):

        localctx = FandangoParser.Import_from_as_namesContext(
            self, self._ctx, self.state
        )
        self.enterRule(localctx, 110, self.RULE_import_from_as_names)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1057
            self.import_from_as_name()
            self.state = 1062
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 92, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 1058
                    self.match(FandangoParser.COMMA)
                    self.state = 1059
                    self.import_from_as_name()
                self.state = 1064
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 92, self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Import_from_as_nameContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NAME(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.NAME)
            else:
                return self.getToken(FandangoParser.NAME, i)

        def AS(self):
            return self.getToken(FandangoParser.AS, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_import_from_as_name

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterImport_from_as_name"):
                listener.enterImport_from_as_name(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitImport_from_as_name"):
                listener.exitImport_from_as_name(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitImport_from_as_name"):
                return visitor.visitImport_from_as_name(self)
            else:
                return visitor.visitChildren(self)

    def import_from_as_name(self):

        localctx = FandangoParser.Import_from_as_nameContext(
            self, self._ctx, self.state
        )
        self.enterRule(localctx, 112, self.RULE_import_from_as_name)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1065
            self.match(FandangoParser.NAME)
            self.state = 1068
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == 10:
                self.state = 1066
                self.match(FandangoParser.AS)
                self.state = 1067
                self.match(FandangoParser.NAME)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Dotted_as_namesContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def dotted_as_name(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(FandangoParser.Dotted_as_nameContext)
            else:
                return self.getTypedRuleContext(FandangoParser.Dotted_as_nameContext, i)

        def COMMA(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.COMMA)
            else:
                return self.getToken(FandangoParser.COMMA, i)

        def getRuleIndex(self):
            return FandangoParser.RULE_dotted_as_names

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterDotted_as_names"):
                listener.enterDotted_as_names(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitDotted_as_names"):
                listener.exitDotted_as_names(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitDotted_as_names"):
                return visitor.visitDotted_as_names(self)
            else:
                return visitor.visitChildren(self)

    def dotted_as_names(self):

        localctx = FandangoParser.Dotted_as_namesContext(self, self._ctx, self.state)
        self.enterRule(localctx, 114, self.RULE_dotted_as_names)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1070
            self.dotted_as_name()
            self.state = 1075
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la == 70:
                self.state = 1071
                self.match(FandangoParser.COMMA)
                self.state = 1072
                self.dotted_as_name()
                self.state = 1077
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Dotted_as_nameContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def dotted_name(self):
            return self.getTypedRuleContext(FandangoParser.Dotted_nameContext, 0)

        def AS(self):
            return self.getToken(FandangoParser.AS, 0)

        def NAME(self):
            return self.getToken(FandangoParser.NAME, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_dotted_as_name

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterDotted_as_name"):
                listener.enterDotted_as_name(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitDotted_as_name"):
                listener.exitDotted_as_name(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitDotted_as_name"):
                return visitor.visitDotted_as_name(self)
            else:
                return visitor.visitChildren(self)

    def dotted_as_name(self):

        localctx = FandangoParser.Dotted_as_nameContext(self, self._ctx, self.state)
        self.enterRule(localctx, 116, self.RULE_dotted_as_name)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1078
            self.dotted_name(0)
            self.state = 1081
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == 10:
                self.state = 1079
                self.match(FandangoParser.AS)
                self.state = 1080
                self.match(FandangoParser.NAME)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Dotted_nameContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NAME(self):
            return self.getToken(FandangoParser.NAME, 0)

        def dotted_name(self):
            return self.getTypedRuleContext(FandangoParser.Dotted_nameContext, 0)

        def DOT(self):
            return self.getToken(FandangoParser.DOT, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_dotted_name

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterDotted_name"):
                listener.enterDotted_name(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitDotted_name"):
                listener.exitDotted_name(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitDotted_name"):
                return visitor.visitDotted_name(self)
            else:
                return visitor.visitChildren(self)

    def dotted_name(self, _p: int = 0):
        _parentctx = self._ctx
        _parentState = self.state
        localctx = FandangoParser.Dotted_nameContext(self, self._ctx, _parentState)
        _prevctx = localctx
        _startState = 118
        self.enterRecursionRule(localctx, 118, self.RULE_dotted_name, _p)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1084
            self.match(FandangoParser.NAME)
            self._ctx.stop = self._input.LT(-1)
            self.state = 1091
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 96, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    if self._parseListeners is not None:
                        self.triggerExitRuleEvent()
                    _prevctx = localctx
                    localctx = FandangoParser.Dotted_nameContext(
                        self, _parentctx, _parentState
                    )
                    self.pushNewRecursionContext(
                        localctx, _startState, self.RULE_dotted_name
                    )
                    self.state = 1086
                    if not self.precpred(self._ctx, 2):
                        from antlr4.error.Errors import FailedPredicateException

                        raise FailedPredicateException(
                            self, "self.precpred(self._ctx, 2)"
                        )
                    self.state = 1087
                    self.match(FandangoParser.DOT)
                    self.state = 1088
                    self.match(FandangoParser.NAME)
                self.state = 1093
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 96, self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.unrollRecursionContexts(_parentctx)
        return localctx

    class BlockContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NEWLINE(self):
            return self.getToken(FandangoParser.NEWLINE, 0)

        def INDENT(self):
            return self.getToken(FandangoParser.INDENT, 0)

        def statements(self):
            return self.getTypedRuleContext(FandangoParser.StatementsContext, 0)

        def DEDENT(self):
            return self.getToken(FandangoParser.DEDENT, 0)

        def simple_stmts(self):
            return self.getTypedRuleContext(FandangoParser.Simple_stmtsContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_block

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterBlock"):
                listener.enterBlock(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitBlock"):
                listener.exitBlock(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitBlock"):
                return visitor.visitBlock(self)
            else:
                return visitor.visitChildren(self)

    def block(self):

        localctx = FandangoParser.BlockContext(self, self._ctx, self.state)
        self.enterRule(localctx, 120, self.RULE_block)
        try:
            self.state = 1100
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [114]:
                self.enterOuterAlt(localctx, 1)
                self.state = 1094
                self.match(FandangoParser.NEWLINE)
                self.state = 1095
                self.match(FandangoParser.INDENT)
                self.state = 1096
                self.statements()
                self.state = 1097
                self.match(FandangoParser.DEDENT)
                pass
            elif token in [
                3,
                4,
                8,
                11,
                13,
                14,
                17,
                19,
                23,
                26,
                27,
                29,
                32,
                34,
                35,
                36,
                38,
                39,
                40,
                41,
                43,
                48,
                52,
                66,
                67,
                68,
                75,
                77,
                82,
                83,
                87,
                88,
                118,
            ]:
                self.enterOuterAlt(localctx, 2)
                self.state = 1099
                self.simple_stmts()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class DecoratorsContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def AT(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.AT)
            else:
                return self.getToken(FandangoParser.AT, i)

        def named_expression(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(FandangoParser.Named_expressionContext)
            else:
                return self.getTypedRuleContext(
                    FandangoParser.Named_expressionContext, i
                )

        def NEWLINE(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.NEWLINE)
            else:
                return self.getToken(FandangoParser.NEWLINE, i)

        def getRuleIndex(self):
            return FandangoParser.RULE_decorators

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterDecorators"):
                listener.enterDecorators(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitDecorators"):
                listener.exitDecorators(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitDecorators"):
                return visitor.visitDecorators(self)
            else:
                return visitor.visitChildren(self)

    def decorators(self):

        localctx = FandangoParser.DecoratorsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 122, self.RULE_decorators)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1106
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 1102
                self.match(FandangoParser.AT)
                self.state = 1103
                self.named_expression()
                self.state = 1104
                self.match(FandangoParser.NEWLINE)
                self.state = 1108
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not (_la == 97):
                    break

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Class_defContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def class_def_raw(self):
            return self.getTypedRuleContext(FandangoParser.Class_def_rawContext, 0)

        def decorators(self):
            return self.getTypedRuleContext(FandangoParser.DecoratorsContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_class_def

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterClass_def"):
                listener.enterClass_def(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitClass_def"):
                listener.exitClass_def(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitClass_def"):
                return visitor.visitClass_def(self)
            else:
                return visitor.visitChildren(self)

    def class_def(self):

        localctx = FandangoParser.Class_defContext(self, self._ctx, self.state)
        self.enterRule(localctx, 124, self.RULE_class_def)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1111
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == 97:
                self.state = 1110
                self.decorators()

            self.state = 1113
            self.class_def_raw()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Class_def_rawContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def CLASS(self):
            return self.getToken(FandangoParser.CLASS, 0)

        def NAME(self):
            return self.getToken(FandangoParser.NAME, 0)

        def COLON(self):
            return self.getToken(FandangoParser.COLON, 0)

        def block(self):
            return self.getTypedRuleContext(FandangoParser.BlockContext, 0)

        def type_params(self):
            return self.getTypedRuleContext(FandangoParser.Type_paramsContext, 0)

        def OPEN_PAREN(self):
            return self.getToken(FandangoParser.OPEN_PAREN, 0)

        def CLOSE_PAREN(self):
            return self.getToken(FandangoParser.CLOSE_PAREN, 0)

        def arguments(self):
            return self.getTypedRuleContext(FandangoParser.ArgumentsContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_class_def_raw

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterClass_def_raw"):
                listener.enterClass_def_raw(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitClass_def_raw"):
                listener.exitClass_def_raw(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitClass_def_raw"):
                return visitor.visitClass_def_raw(self)
            else:
                return visitor.visitChildren(self)

    def class_def_raw(self):

        localctx = FandangoParser.Class_def_rawContext(self, self._ctx, self.state)
        self.enterRule(localctx, 126, self.RULE_class_def_raw)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1115
            self.match(FandangoParser.CLASS)
            self.state = 1116
            self.match(FandangoParser.NAME)
            self.state = 1118
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == 75:
                self.state = 1117
                self.type_params()

            self.state = 1125
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == 68:
                self.state = 1120
                self.match(FandangoParser.OPEN_PAREN)
                self.state = 1122
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if (((_la) & ~0x3F) == 0 and ((1 << _la) & 4505888853336344) != 0) or (
                    (((_la - 66)) & ~0x3F) == 0
                    and ((1 << (_la - 66)) & 4503599633861255) != 0
                ):
                    self.state = 1121
                    self.arguments()

                self.state = 1124
                self.match(FandangoParser.CLOSE_PAREN)

            self.state = 1127
            self.match(FandangoParser.COLON)
            self.state = 1128
            self.block()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Function_defContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def function_def_raw(self):
            return self.getTypedRuleContext(FandangoParser.Function_def_rawContext, 0)

        def decorators(self):
            return self.getTypedRuleContext(FandangoParser.DecoratorsContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_function_def

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterFunction_def"):
                listener.enterFunction_def(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitFunction_def"):
                listener.exitFunction_def(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitFunction_def"):
                return visitor.visitFunction_def(self)
            else:
                return visitor.visitChildren(self)

    def function_def(self):

        localctx = FandangoParser.Function_defContext(self, self._ctx, self.state)
        self.enterRule(localctx, 128, self.RULE_function_def)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1131
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == 97:
                self.state = 1130
                self.decorators()

            self.state = 1133
            self.function_def_raw()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Function_def_rawContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def DEF(self):
            return self.getToken(FandangoParser.DEF, 0)

        def NAME(self):
            return self.getToken(FandangoParser.NAME, 0)

        def OPEN_PAREN(self):
            return self.getToken(FandangoParser.OPEN_PAREN, 0)

        def CLOSE_PAREN(self):
            return self.getToken(FandangoParser.CLOSE_PAREN, 0)

        def COLON(self):
            return self.getToken(FandangoParser.COLON, 0)

        def block(self):
            return self.getTypedRuleContext(FandangoParser.BlockContext, 0)

        def ASYNC(self):
            return self.getToken(FandangoParser.ASYNC, 0)

        def type_params(self):
            return self.getTypedRuleContext(FandangoParser.Type_paramsContext, 0)

        def params(self):
            return self.getTypedRuleContext(FandangoParser.ParamsContext, 0)

        def ARROW(self):
            return self.getToken(FandangoParser.ARROW, 0)

        def expression(self):
            return self.getTypedRuleContext(FandangoParser.ExpressionContext, 0)

        def func_type_comment(self):
            return self.getTypedRuleContext(FandangoParser.Func_type_commentContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_function_def_raw

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterFunction_def_raw"):
                listener.enterFunction_def_raw(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitFunction_def_raw"):
                listener.exitFunction_def_raw(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitFunction_def_raw"):
                return visitor.visitFunction_def_raw(self)
            else:
                return visitor.visitChildren(self)

    def function_def_raw(self):

        localctx = FandangoParser.Function_def_rawContext(self, self._ctx, self.state)
        self.enterRule(localctx, 130, self.RULE_function_def_raw)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1136
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == 12:
                self.state = 1135
                self.match(FandangoParser.ASYNC)

            self.state = 1138
            self.match(FandangoParser.DEF)
            self.state = 1139
            self.match(FandangoParser.NAME)
            self.state = 1141
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == 75:
                self.state = 1140
                self.type_params()

            self.state = 1143
            self.match(FandangoParser.OPEN_PAREN)
            self.state = 1145
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if (((_la - 52)) & ~0x3F) == 0 and ((1 << (_la - 52)) & 2129921) != 0:
                self.state = 1144
                self.params()

            self.state = 1147
            self.match(FandangoParser.CLOSE_PAREN)
            self.state = 1150
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == 98:
                self.state = 1148
                self.match(FandangoParser.ARROW)
                self.state = 1149
                self.expression()

            self.state = 1152
            self.match(FandangoParser.COLON)
            self.state = 1154
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 108, self._ctx)
            if la_ == 1:
                self.state = 1153
                self.func_type_comment()

            self.state = 1156
            self.block()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ParamsContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def parameters(self):
            return self.getTypedRuleContext(FandangoParser.ParametersContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_params

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterParams"):
                listener.enterParams(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitParams"):
                listener.exitParams(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitParams"):
                return visitor.visitParams(self)
            else:
                return visitor.visitChildren(self)

    def params(self):

        localctx = FandangoParser.ParamsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 132, self.RULE_params)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1158
            self.parameters()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ParametersContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def slash_no_default(self):
            return self.getTypedRuleContext(FandangoParser.Slash_no_defaultContext, 0)

        def param_no_default(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(FandangoParser.Param_no_defaultContext)
            else:
                return self.getTypedRuleContext(
                    FandangoParser.Param_no_defaultContext, i
                )

        def param_with_default(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(
                    FandangoParser.Param_with_defaultContext
                )
            else:
                return self.getTypedRuleContext(
                    FandangoParser.Param_with_defaultContext, i
                )

        def star_etc(self):
            return self.getTypedRuleContext(FandangoParser.Star_etcContext, 0)

        def slash_with_default(self):
            return self.getTypedRuleContext(FandangoParser.Slash_with_defaultContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_parameters

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterParameters"):
                listener.enterParameters(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitParameters"):
                listener.exitParameters(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitParameters"):
                return visitor.visitParameters(self)
            else:
                return visitor.visitChildren(self)

    def parameters(self):

        localctx = FandangoParser.ParametersContext(self, self._ctx, self.state)
        self.enterRule(localctx, 134, self.RULE_parameters)
        self._la = 0  # Token type
        try:
            self.state = 1209
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 119, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 1160
                self.slash_no_default()
                self.state = 1164
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 109, self._ctx)
                while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                    if _alt == 1:
                        self.state = 1161
                        self.param_no_default()
                    self.state = 1166
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input, 109, self._ctx)

                self.state = 1170
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la == 52:
                    self.state = 1167
                    self.param_with_default()
                    self.state = 1172
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                self.state = 1174
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == 67 or _la == 73:
                    self.state = 1173
                    self.star_etc()

                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 1176
                self.slash_with_default()
                self.state = 1180
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la == 52:
                    self.state = 1177
                    self.param_with_default()
                    self.state = 1182
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                self.state = 1184
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == 67 or _la == 73:
                    self.state = 1183
                    self.star_etc()

                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 1187
                self._errHandler.sync(self)
                _alt = 1
                while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                    if _alt == 1:
                        self.state = 1186
                        self.param_no_default()

                    else:
                        raise NoViableAltException(self)
                    self.state = 1189
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input, 114, self._ctx)

                self.state = 1194
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la == 52:
                    self.state = 1191
                    self.param_with_default()
                    self.state = 1196
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                self.state = 1198
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == 67 or _la == 73:
                    self.state = 1197
                    self.star_etc()

                pass

            elif la_ == 4:
                self.enterOuterAlt(localctx, 4)
                self.state = 1201
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while True:
                    self.state = 1200
                    self.param_with_default()
                    self.state = 1203
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)
                    if not (_la == 52):
                        break

                self.state = 1206
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == 67 or _la == 73:
                    self.state = 1205
                    self.star_etc()

                pass

            elif la_ == 5:
                self.enterOuterAlt(localctx, 5)
                self.state = 1208
                self.star_etc()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Slash_no_defaultContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def DIV(self):
            return self.getToken(FandangoParser.DIV, 0)

        def param_no_default(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(FandangoParser.Param_no_defaultContext)
            else:
                return self.getTypedRuleContext(
                    FandangoParser.Param_no_defaultContext, i
                )

        def COMMA(self):
            return self.getToken(FandangoParser.COMMA, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_slash_no_default

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterSlash_no_default"):
                listener.enterSlash_no_default(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitSlash_no_default"):
                listener.exitSlash_no_default(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitSlash_no_default"):
                return visitor.visitSlash_no_default(self)
            else:
                return visitor.visitChildren(self)

    def slash_no_default(self):

        localctx = FandangoParser.Slash_no_defaultContext(self, self._ctx, self.state)
        self.enterRule(localctx, 136, self.RULE_slash_no_default)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1212
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 1211
                self.param_no_default()
                self.state = 1214
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not (_la == 52):
                    break

            self.state = 1216
            self.match(FandangoParser.DIV)
            self.state = 1218
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == 70:
                self.state = 1217
                self.match(FandangoParser.COMMA)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Slash_with_defaultContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def DIV(self):
            return self.getToken(FandangoParser.DIV, 0)

        def param_no_default(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(FandangoParser.Param_no_defaultContext)
            else:
                return self.getTypedRuleContext(
                    FandangoParser.Param_no_defaultContext, i
                )

        def param_with_default(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(
                    FandangoParser.Param_with_defaultContext
                )
            else:
                return self.getTypedRuleContext(
                    FandangoParser.Param_with_defaultContext, i
                )

        def COMMA(self):
            return self.getToken(FandangoParser.COMMA, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_slash_with_default

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterSlash_with_default"):
                listener.enterSlash_with_default(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitSlash_with_default"):
                listener.exitSlash_with_default(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitSlash_with_default"):
                return visitor.visitSlash_with_default(self)
            else:
                return visitor.visitChildren(self)

    def slash_with_default(self):

        localctx = FandangoParser.Slash_with_defaultContext(self, self._ctx, self.state)
        self.enterRule(localctx, 138, self.RULE_slash_with_default)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1223
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 122, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 1220
                    self.param_no_default()
                self.state = 1225
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 122, self._ctx)

            self.state = 1227
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 1226
                self.param_with_default()
                self.state = 1229
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not (_la == 52):
                    break

            self.state = 1231
            self.match(FandangoParser.DIV)
            self.state = 1233
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == 70:
                self.state = 1232
                self.match(FandangoParser.COMMA)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Star_etcContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def STAR(self):
            return self.getToken(FandangoParser.STAR, 0)

        def param_no_default(self):
            return self.getTypedRuleContext(FandangoParser.Param_no_defaultContext, 0)

        def param_maybe_default(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(
                    FandangoParser.Param_maybe_defaultContext
                )
            else:
                return self.getTypedRuleContext(
                    FandangoParser.Param_maybe_defaultContext, i
                )

        def kwds(self):
            return self.getTypedRuleContext(FandangoParser.KwdsContext, 0)

        def param_no_default_star_annotation(self):
            return self.getTypedRuleContext(
                FandangoParser.Param_no_default_star_annotationContext, 0
            )

        def COMMA(self):
            return self.getToken(FandangoParser.COMMA, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_star_etc

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterStar_etc"):
                listener.enterStar_etc(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitStar_etc"):
                listener.exitStar_etc(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitStar_etc"):
                return visitor.visitStar_etc(self)
            else:
                return visitor.visitChildren(self)

    def star_etc(self):

        localctx = FandangoParser.Star_etcContext(self, self._ctx, self.state)
        self.enterRule(localctx, 140, self.RULE_star_etc)
        self._la = 0  # Token type
        try:
            self.state = 1268
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 131, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 1235
                self.match(FandangoParser.STAR)
                self.state = 1236
                self.param_no_default()
                self.state = 1240
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la == 52:
                    self.state = 1237
                    self.param_maybe_default()
                    self.state = 1242
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                self.state = 1244
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == 73:
                    self.state = 1243
                    self.kwds()

                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 1246
                self.match(FandangoParser.STAR)
                self.state = 1247
                self.param_no_default_star_annotation()
                self.state = 1251
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la == 52:
                    self.state = 1248
                    self.param_maybe_default()
                    self.state = 1253
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                self.state = 1255
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == 73:
                    self.state = 1254
                    self.kwds()

                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 1257
                self.match(FandangoParser.STAR)
                self.state = 1258
                self.match(FandangoParser.COMMA)
                self.state = 1260
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while True:
                    self.state = 1259
                    self.param_maybe_default()
                    self.state = 1262
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)
                    if not (_la == 52):
                        break

                self.state = 1265
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == 73:
                    self.state = 1264
                    self.kwds()

                pass

            elif la_ == 4:
                self.enterOuterAlt(localctx, 4)
                self.state = 1267
                self.kwds()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class KwdsContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def POWER(self):
            return self.getToken(FandangoParser.POWER, 0)

        def param_no_default(self):
            return self.getTypedRuleContext(FandangoParser.Param_no_defaultContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_kwds

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterKwds"):
                listener.enterKwds(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitKwds"):
                listener.exitKwds(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitKwds"):
                return visitor.visitKwds(self)
            else:
                return visitor.visitChildren(self)

    def kwds(self):

        localctx = FandangoParser.KwdsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 142, self.RULE_kwds)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1270
            self.match(FandangoParser.POWER)
            self.state = 1271
            self.param_no_default()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Param_no_defaultContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def param(self):
            return self.getTypedRuleContext(FandangoParser.ParamContext, 0)

        def COMMA(self):
            return self.getToken(FandangoParser.COMMA, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_param_no_default

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterParam_no_default"):
                listener.enterParam_no_default(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitParam_no_default"):
                listener.exitParam_no_default(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitParam_no_default"):
                return visitor.visitParam_no_default(self)
            else:
                return visitor.visitChildren(self)

    def param_no_default(self):

        localctx = FandangoParser.Param_no_defaultContext(self, self._ctx, self.state)
        self.enterRule(localctx, 144, self.RULE_param_no_default)
        try:
            self.state = 1277
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 132, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 1273
                self.param()
                self.state = 1274
                self.match(FandangoParser.COMMA)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 1276
                self.param()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Param_no_default_star_annotationContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def param_star_annotation(self):
            return self.getTypedRuleContext(
                FandangoParser.Param_star_annotationContext, 0
            )

        def COMMA(self):
            return self.getToken(FandangoParser.COMMA, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_param_no_default_star_annotation

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterParam_no_default_star_annotation"):
                listener.enterParam_no_default_star_annotation(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitParam_no_default_star_annotation"):
                listener.exitParam_no_default_star_annotation(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitParam_no_default_star_annotation"):
                return visitor.visitParam_no_default_star_annotation(self)
            else:
                return visitor.visitChildren(self)

    def param_no_default_star_annotation(self):

        localctx = FandangoParser.Param_no_default_star_annotationContext(
            self, self._ctx, self.state
        )
        self.enterRule(localctx, 146, self.RULE_param_no_default_star_annotation)
        try:
            self.state = 1283
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 133, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 1279
                self.param_star_annotation()
                self.state = 1280
                self.match(FandangoParser.COMMA)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 1282
                self.param_star_annotation()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Param_with_defaultContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def param(self):
            return self.getTypedRuleContext(FandangoParser.ParamContext, 0)

        def default(self):
            return self.getTypedRuleContext(FandangoParser.DefaultContext, 0)

        def COMMA(self):
            return self.getToken(FandangoParser.COMMA, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_param_with_default

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterParam_with_default"):
                listener.enterParam_with_default(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitParam_with_default"):
                listener.exitParam_with_default(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitParam_with_default"):
                return visitor.visitParam_with_default(self)
            else:
                return visitor.visitChildren(self)

    def param_with_default(self):

        localctx = FandangoParser.Param_with_defaultContext(self, self._ctx, self.state)
        self.enterRule(localctx, 148, self.RULE_param_with_default)
        try:
            self.state = 1292
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 134, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 1285
                self.param()
                self.state = 1286
                self.default()
                self.state = 1287
                self.match(FandangoParser.COMMA)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 1289
                self.param()
                self.state = 1290
                self.default()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Param_maybe_defaultContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def param(self):
            return self.getTypedRuleContext(FandangoParser.ParamContext, 0)

        def COMMA(self):
            return self.getToken(FandangoParser.COMMA, 0)

        def default(self):
            return self.getTypedRuleContext(FandangoParser.DefaultContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_param_maybe_default

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterParam_maybe_default"):
                listener.enterParam_maybe_default(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitParam_maybe_default"):
                listener.exitParam_maybe_default(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitParam_maybe_default"):
                return visitor.visitParam_maybe_default(self)
            else:
                return visitor.visitChildren(self)

    def param_maybe_default(self):

        localctx = FandangoParser.Param_maybe_defaultContext(
            self, self._ctx, self.state
        )
        self.enterRule(localctx, 150, self.RULE_param_maybe_default)
        self._la = 0  # Token type
        try:
            self.state = 1304
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 137, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 1294
                self.param()
                self.state = 1296
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == 74:
                    self.state = 1295
                    self.default()

                self.state = 1298
                self.match(FandangoParser.COMMA)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 1300
                self.param()
                self.state = 1302
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == 74:
                    self.state = 1301
                    self.default()

                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ParamContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NAME(self):
            return self.getToken(FandangoParser.NAME, 0)

        def annotation(self):
            return self.getTypedRuleContext(FandangoParser.AnnotationContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_param

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterParam"):
                listener.enterParam(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitParam"):
                listener.exitParam(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitParam"):
                return visitor.visitParam(self)
            else:
                return visitor.visitChildren(self)

    def param(self):

        localctx = FandangoParser.ParamContext(self, self._ctx, self.state)
        self.enterRule(localctx, 152, self.RULE_param)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1306
            self.match(FandangoParser.NAME)
            self.state = 1308
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == 71:
                self.state = 1307
                self.annotation()

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Param_star_annotationContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NAME(self):
            return self.getToken(FandangoParser.NAME, 0)

        def star_annotation(self):
            return self.getTypedRuleContext(FandangoParser.Star_annotationContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_param_star_annotation

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterParam_star_annotation"):
                listener.enterParam_star_annotation(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitParam_star_annotation"):
                listener.exitParam_star_annotation(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitParam_star_annotation"):
                return visitor.visitParam_star_annotation(self)
            else:
                return visitor.visitChildren(self)

    def param_star_annotation(self):

        localctx = FandangoParser.Param_star_annotationContext(
            self, self._ctx, self.state
        )
        self.enterRule(localctx, 154, self.RULE_param_star_annotation)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1310
            self.match(FandangoParser.NAME)
            self.state = 1311
            self.star_annotation()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class AnnotationContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def COLON(self):
            return self.getToken(FandangoParser.COLON, 0)

        def expression(self):
            return self.getTypedRuleContext(FandangoParser.ExpressionContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_annotation

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterAnnotation"):
                listener.enterAnnotation(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitAnnotation"):
                listener.exitAnnotation(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitAnnotation"):
                return visitor.visitAnnotation(self)
            else:
                return visitor.visitChildren(self)

    def annotation(self):

        localctx = FandangoParser.AnnotationContext(self, self._ctx, self.state)
        self.enterRule(localctx, 156, self.RULE_annotation)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1313
            self.match(FandangoParser.COLON)
            self.state = 1314
            self.expression()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Star_annotationContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def COLON(self):
            return self.getToken(FandangoParser.COLON, 0)

        def star_expression(self):
            return self.getTypedRuleContext(FandangoParser.Star_expressionContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_star_annotation

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterStar_annotation"):
                listener.enterStar_annotation(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitStar_annotation"):
                listener.exitStar_annotation(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitStar_annotation"):
                return visitor.visitStar_annotation(self)
            else:
                return visitor.visitChildren(self)

    def star_annotation(self):

        localctx = FandangoParser.Star_annotationContext(self, self._ctx, self.state)
        self.enterRule(localctx, 158, self.RULE_star_annotation)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1316
            self.match(FandangoParser.COLON)
            self.state = 1317
            self.star_expression()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class DefaultContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ASSIGN(self):
            return self.getToken(FandangoParser.ASSIGN, 0)

        def expression(self):
            return self.getTypedRuleContext(FandangoParser.ExpressionContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_default

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterDefault"):
                listener.enterDefault(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitDefault"):
                listener.exitDefault(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitDefault"):
                return visitor.visitDefault(self)
            else:
                return visitor.visitChildren(self)

    def default(self):

        localctx = FandangoParser.DefaultContext(self, self._ctx, self.state)
        self.enterRule(localctx, 160, self.RULE_default)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1319
            self.match(FandangoParser.ASSIGN)
            self.state = 1320
            self.expression()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class If_stmtContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def IF(self):
            return self.getToken(FandangoParser.IF, 0)

        def named_expression(self):
            return self.getTypedRuleContext(FandangoParser.Named_expressionContext, 0)

        def COLON(self):
            return self.getToken(FandangoParser.COLON, 0)

        def block(self):
            return self.getTypedRuleContext(FandangoParser.BlockContext, 0)

        def elif_stmt(self):
            return self.getTypedRuleContext(FandangoParser.Elif_stmtContext, 0)

        def else_block(self):
            return self.getTypedRuleContext(FandangoParser.Else_blockContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_if_stmt

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterIf_stmt"):
                listener.enterIf_stmt(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitIf_stmt"):
                listener.exitIf_stmt(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitIf_stmt"):
                return visitor.visitIf_stmt(self)
            else:
                return visitor.visitChildren(self)

    def if_stmt(self):

        localctx = FandangoParser.If_stmtContext(self, self._ctx, self.state)
        self.enterRule(localctx, 162, self.RULE_if_stmt)
        self._la = 0  # Token type
        try:
            self.state = 1335
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 140, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 1322
                self.match(FandangoParser.IF)
                self.state = 1323
                self.named_expression()
                self.state = 1324
                self.match(FandangoParser.COLON)
                self.state = 1325
                self.block()
                self.state = 1326
                self.elif_stmt()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 1328
                self.match(FandangoParser.IF)
                self.state = 1329
                self.named_expression()
                self.state = 1330
                self.match(FandangoParser.COLON)
                self.state = 1331
                self.block()
                self.state = 1333
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == 21:
                    self.state = 1332
                    self.else_block()

                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Elif_stmtContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ELIF(self):
            return self.getToken(FandangoParser.ELIF, 0)

        def named_expression(self):
            return self.getTypedRuleContext(FandangoParser.Named_expressionContext, 0)

        def COLON(self):
            return self.getToken(FandangoParser.COLON, 0)

        def block(self):
            return self.getTypedRuleContext(FandangoParser.BlockContext, 0)

        def elif_stmt(self):
            return self.getTypedRuleContext(FandangoParser.Elif_stmtContext, 0)

        def else_block(self):
            return self.getTypedRuleContext(FandangoParser.Else_blockContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_elif_stmt

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterElif_stmt"):
                listener.enterElif_stmt(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitElif_stmt"):
                listener.exitElif_stmt(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitElif_stmt"):
                return visitor.visitElif_stmt(self)
            else:
                return visitor.visitChildren(self)

    def elif_stmt(self):

        localctx = FandangoParser.Elif_stmtContext(self, self._ctx, self.state)
        self.enterRule(localctx, 164, self.RULE_elif_stmt)
        self._la = 0  # Token type
        try:
            self.state = 1350
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 142, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 1337
                self.match(FandangoParser.ELIF)
                self.state = 1338
                self.named_expression()
                self.state = 1339
                self.match(FandangoParser.COLON)
                self.state = 1340
                self.block()
                self.state = 1341
                self.elif_stmt()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 1343
                self.match(FandangoParser.ELIF)
                self.state = 1344
                self.named_expression()
                self.state = 1345
                self.match(FandangoParser.COLON)
                self.state = 1346
                self.block()
                self.state = 1348
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == 21:
                    self.state = 1347
                    self.else_block()

                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Else_blockContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ELSE(self):
            return self.getToken(FandangoParser.ELSE, 0)

        def COLON(self):
            return self.getToken(FandangoParser.COLON, 0)

        def block(self):
            return self.getTypedRuleContext(FandangoParser.BlockContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_else_block

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterElse_block"):
                listener.enterElse_block(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitElse_block"):
                listener.exitElse_block(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitElse_block"):
                return visitor.visitElse_block(self)
            else:
                return visitor.visitChildren(self)

    def else_block(self):

        localctx = FandangoParser.Else_blockContext(self, self._ctx, self.state)
        self.enterRule(localctx, 166, self.RULE_else_block)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1352
            self.match(FandangoParser.ELSE)
            self.state = 1353
            self.match(FandangoParser.COLON)
            self.state = 1354
            self.block()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class While_stmtContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def WHILE(self):
            return self.getToken(FandangoParser.WHILE, 0)

        def named_expression(self):
            return self.getTypedRuleContext(FandangoParser.Named_expressionContext, 0)

        def COLON(self):
            return self.getToken(FandangoParser.COLON, 0)

        def block(self):
            return self.getTypedRuleContext(FandangoParser.BlockContext, 0)

        def else_block(self):
            return self.getTypedRuleContext(FandangoParser.Else_blockContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_while_stmt

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterWhile_stmt"):
                listener.enterWhile_stmt(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitWhile_stmt"):
                listener.exitWhile_stmt(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitWhile_stmt"):
                return visitor.visitWhile_stmt(self)
            else:
                return visitor.visitChildren(self)

    def while_stmt(self):

        localctx = FandangoParser.While_stmtContext(self, self._ctx, self.state)
        self.enterRule(localctx, 168, self.RULE_while_stmt)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1356
            self.match(FandangoParser.WHILE)
            self.state = 1357
            self.named_expression()
            self.state = 1358
            self.match(FandangoParser.COLON)
            self.state = 1359
            self.block()
            self.state = 1361
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == 21:
                self.state = 1360
                self.else_block()

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class For_stmtContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def FOR(self):
            return self.getToken(FandangoParser.FOR, 0)

        def star_targets(self):
            return self.getTypedRuleContext(FandangoParser.Star_targetsContext, 0)

        def IN(self):
            return self.getToken(FandangoParser.IN, 0)

        def star_expressions(self):
            return self.getTypedRuleContext(FandangoParser.Star_expressionsContext, 0)

        def COLON(self):
            return self.getToken(FandangoParser.COLON, 0)

        def block(self):
            return self.getTypedRuleContext(FandangoParser.BlockContext, 0)

        def else_block(self):
            return self.getTypedRuleContext(FandangoParser.Else_blockContext, 0)

        def ASYNC(self):
            return self.getToken(FandangoParser.ASYNC, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_for_stmt

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterFor_stmt"):
                listener.enterFor_stmt(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitFor_stmt"):
                listener.exitFor_stmt(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitFor_stmt"):
                return visitor.visitFor_stmt(self)
            else:
                return visitor.visitChildren(self)

    def for_stmt(self):

        localctx = FandangoParser.For_stmtContext(self, self._ctx, self.state)
        self.enterRule(localctx, 170, self.RULE_for_stmt)
        self._la = 0  # Token type
        try:
            self.state = 1382
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [25]:
                self.enterOuterAlt(localctx, 1)
                self.state = 1363
                self.match(FandangoParser.FOR)
                self.state = 1364
                self.star_targets()
                self.state = 1365
                self.match(FandangoParser.IN)
                self.state = 1366
                self.star_expressions()
                self.state = 1367
                self.match(FandangoParser.COLON)
                self.state = 1368
                self.block()
                self.state = 1370
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == 21:
                    self.state = 1369
                    self.else_block()

                pass
            elif token in [12]:
                self.enterOuterAlt(localctx, 2)
                self.state = 1372
                self.match(FandangoParser.ASYNC)
                self.state = 1373
                self.match(FandangoParser.FOR)
                self.state = 1374
                self.star_targets()
                self.state = 1375
                self.match(FandangoParser.IN)
                self.state = 1376
                self.star_expressions()
                self.state = 1377
                self.match(FandangoParser.COLON)
                self.state = 1378
                self.block()
                self.state = 1380
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == 21:
                    self.state = 1379
                    self.else_block()

                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class With_stmtContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def WITH(self):
            return self.getToken(FandangoParser.WITH, 0)

        def OPEN_PAREN(self):
            return self.getToken(FandangoParser.OPEN_PAREN, 0)

        def with_item(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(FandangoParser.With_itemContext)
            else:
                return self.getTypedRuleContext(FandangoParser.With_itemContext, i)

        def CLOSE_PAREN(self):
            return self.getToken(FandangoParser.CLOSE_PAREN, 0)

        def COLON(self):
            return self.getToken(FandangoParser.COLON, 0)

        def block(self):
            return self.getTypedRuleContext(FandangoParser.BlockContext, 0)

        def COMMA(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.COMMA)
            else:
                return self.getToken(FandangoParser.COMMA, i)

        def ASYNC(self):
            return self.getToken(FandangoParser.ASYNC, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_with_stmt

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterWith_stmt"):
                listener.enterWith_stmt(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitWith_stmt"):
                listener.exitWith_stmt(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitWith_stmt"):
                return visitor.visitWith_stmt(self)
            else:
                return visitor.visitChildren(self)

    def with_stmt(self):

        localctx = FandangoParser.With_stmtContext(self, self._ctx, self.state)
        self.enterRule(localctx, 172, self.RULE_with_stmt)
        self._la = 0  # Token type
        try:
            self.state = 1444
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 153, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 1384
                self.match(FandangoParser.WITH)
                self.state = 1385
                self.match(FandangoParser.OPEN_PAREN)
                self.state = 1386
                self.with_item()
                self.state = 1391
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 147, self._ctx)
                while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                    if _alt == 1:
                        self.state = 1387
                        self.match(FandangoParser.COMMA)
                        self.state = 1388
                        self.with_item()
                    self.state = 1393
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input, 147, self._ctx)

                self.state = 1395
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == 70:
                    self.state = 1394
                    self.match(FandangoParser.COMMA)

                self.state = 1397
                self.match(FandangoParser.CLOSE_PAREN)
                self.state = 1398
                self.match(FandangoParser.COLON)
                self.state = 1399
                self.block()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 1401
                self.match(FandangoParser.WITH)
                self.state = 1402
                self.with_item()
                self.state = 1407
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la == 70:
                    self.state = 1403
                    self.match(FandangoParser.COMMA)
                    self.state = 1404
                    self.with_item()
                    self.state = 1409
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                self.state = 1410
                self.match(FandangoParser.COLON)
                self.state = 1411
                self.block()
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 1413
                self.match(FandangoParser.ASYNC)
                self.state = 1414
                self.match(FandangoParser.WITH)
                self.state = 1415
                self.match(FandangoParser.OPEN_PAREN)
                self.state = 1416
                self.with_item()
                self.state = 1421
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 150, self._ctx)
                while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                    if _alt == 1:
                        self.state = 1417
                        self.match(FandangoParser.COMMA)
                        self.state = 1418
                        self.with_item()
                    self.state = 1423
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input, 150, self._ctx)

                self.state = 1425
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == 70:
                    self.state = 1424
                    self.match(FandangoParser.COMMA)

                self.state = 1427
                self.match(FandangoParser.CLOSE_PAREN)
                self.state = 1428
                self.match(FandangoParser.COLON)
                self.state = 1429
                self.block()
                pass

            elif la_ == 4:
                self.enterOuterAlt(localctx, 4)
                self.state = 1431
                self.match(FandangoParser.ASYNC)
                self.state = 1432
                self.match(FandangoParser.WITH)
                self.state = 1433
                self.with_item()
                self.state = 1438
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la == 70:
                    self.state = 1434
                    self.match(FandangoParser.COMMA)
                    self.state = 1435
                    self.with_item()
                    self.state = 1440
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                self.state = 1441
                self.match(FandangoParser.COLON)
                self.state = 1442
                self.block()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class With_itemContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def expression(self):
            return self.getTypedRuleContext(FandangoParser.ExpressionContext, 0)

        def AS(self):
            return self.getToken(FandangoParser.AS, 0)

        def star_target(self):
            return self.getTypedRuleContext(FandangoParser.Star_targetContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_with_item

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterWith_item"):
                listener.enterWith_item(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitWith_item"):
                listener.exitWith_item(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitWith_item"):
                return visitor.visitWith_item(self)
            else:
                return visitor.visitChildren(self)

    def with_item(self):

        localctx = FandangoParser.With_itemContext(self, self._ctx, self.state)
        self.enterRule(localctx, 174, self.RULE_with_item)
        try:
            self.state = 1451
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 154, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 1446
                self.expression()
                self.state = 1447
                self.match(FandangoParser.AS)
                self.state = 1448
                self.star_target()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 1450
                self.expression()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Try_stmtContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def TRY(self):
            return self.getToken(FandangoParser.TRY, 0)

        def COLON(self):
            return self.getToken(FandangoParser.COLON, 0)

        def block(self):
            return self.getTypedRuleContext(FandangoParser.BlockContext, 0)

        def finally_block(self):
            return self.getTypedRuleContext(FandangoParser.Finally_blockContext, 0)

        def except_block(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(FandangoParser.Except_blockContext)
            else:
                return self.getTypedRuleContext(FandangoParser.Except_blockContext, i)

        def else_block(self):
            return self.getTypedRuleContext(FandangoParser.Else_blockContext, 0)

        def except_star_block(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(
                    FandangoParser.Except_star_blockContext
                )
            else:
                return self.getTypedRuleContext(
                    FandangoParser.Except_star_blockContext, i
                )

        def getRuleIndex(self):
            return FandangoParser.RULE_try_stmt

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterTry_stmt"):
                listener.enterTry_stmt(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitTry_stmt"):
                listener.exitTry_stmt(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitTry_stmt"):
                return visitor.visitTry_stmt(self)
            else:
                return visitor.visitChildren(self)

    def try_stmt(self):

        localctx = FandangoParser.Try_stmtContext(self, self._ctx, self.state)
        self.enterRule(localctx, 176, self.RULE_try_stmt)
        self._la = 0  # Token type
        try:
            self.state = 1486
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 161, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 1453
                self.match(FandangoParser.TRY)
                self.state = 1454
                self.match(FandangoParser.COLON)
                self.state = 1455
                self.block()
                self.state = 1456
                self.finally_block()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 1458
                self.match(FandangoParser.TRY)
                self.state = 1459
                self.match(FandangoParser.COLON)
                self.state = 1460
                self.block()
                self.state = 1462
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while True:
                    self.state = 1461
                    self.except_block()
                    self.state = 1464
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)
                    if not (_la == 22):
                        break

                self.state = 1467
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == 21:
                    self.state = 1466
                    self.else_block()

                self.state = 1470
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == 24:
                    self.state = 1469
                    self.finally_block()

                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 1472
                self.match(FandangoParser.TRY)
                self.state = 1473
                self.match(FandangoParser.COLON)
                self.state = 1474
                self.block()
                self.state = 1476
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while True:
                    self.state = 1475
                    self.except_star_block()
                    self.state = 1478
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)
                    if not (_la == 22):
                        break

                self.state = 1481
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == 21:
                    self.state = 1480
                    self.else_block()

                self.state = 1484
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == 24:
                    self.state = 1483
                    self.finally_block()

                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Except_blockContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def EXCEPT(self):
            return self.getToken(FandangoParser.EXCEPT, 0)

        def expression(self):
            return self.getTypedRuleContext(FandangoParser.ExpressionContext, 0)

        def COLON(self):
            return self.getToken(FandangoParser.COLON, 0)

        def block(self):
            return self.getTypedRuleContext(FandangoParser.BlockContext, 0)

        def AS(self):
            return self.getToken(FandangoParser.AS, 0)

        def NAME(self):
            return self.getToken(FandangoParser.NAME, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_except_block

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterExcept_block"):
                listener.enterExcept_block(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitExcept_block"):
                listener.exitExcept_block(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitExcept_block"):
                return visitor.visitExcept_block(self)
            else:
                return visitor.visitChildren(self)

    def except_block(self):

        localctx = FandangoParser.Except_blockContext(self, self._ctx, self.state)
        self.enterRule(localctx, 178, self.RULE_except_block)
        self._la = 0  # Token type
        try:
            self.state = 1500
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 163, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 1488
                self.match(FandangoParser.EXCEPT)
                self.state = 1489
                self.expression()
                self.state = 1492
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == 10:
                    self.state = 1490
                    self.match(FandangoParser.AS)
                    self.state = 1491
                    self.match(FandangoParser.NAME)

                self.state = 1494
                self.match(FandangoParser.COLON)
                self.state = 1495
                self.block()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 1497
                self.match(FandangoParser.EXCEPT)
                self.state = 1498
                self.match(FandangoParser.COLON)
                self.state = 1499
                self.block()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Except_star_blockContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def EXCEPT(self):
            return self.getToken(FandangoParser.EXCEPT, 0)

        def STAR(self):
            return self.getToken(FandangoParser.STAR, 0)

        def expression(self):
            return self.getTypedRuleContext(FandangoParser.ExpressionContext, 0)

        def COLON(self):
            return self.getToken(FandangoParser.COLON, 0)

        def block(self):
            return self.getTypedRuleContext(FandangoParser.BlockContext, 0)

        def AS(self):
            return self.getToken(FandangoParser.AS, 0)

        def NAME(self):
            return self.getToken(FandangoParser.NAME, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_except_star_block

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterExcept_star_block"):
                listener.enterExcept_star_block(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitExcept_star_block"):
                listener.exitExcept_star_block(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitExcept_star_block"):
                return visitor.visitExcept_star_block(self)
            else:
                return visitor.visitChildren(self)

    def except_star_block(self):

        localctx = FandangoParser.Except_star_blockContext(self, self._ctx, self.state)
        self.enterRule(localctx, 180, self.RULE_except_star_block)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1502
            self.match(FandangoParser.EXCEPT)
            self.state = 1503
            self.match(FandangoParser.STAR)
            self.state = 1504
            self.expression()
            self.state = 1507
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == 10:
                self.state = 1505
                self.match(FandangoParser.AS)
                self.state = 1506
                self.match(FandangoParser.NAME)

            self.state = 1509
            self.match(FandangoParser.COLON)
            self.state = 1510
            self.block()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Finally_blockContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def FINALLY(self):
            return self.getToken(FandangoParser.FINALLY, 0)

        def COLON(self):
            return self.getToken(FandangoParser.COLON, 0)

        def block(self):
            return self.getTypedRuleContext(FandangoParser.BlockContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_finally_block

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterFinally_block"):
                listener.enterFinally_block(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitFinally_block"):
                listener.exitFinally_block(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitFinally_block"):
                return visitor.visitFinally_block(self)
            else:
                return visitor.visitChildren(self)

    def finally_block(self):

        localctx = FandangoParser.Finally_blockContext(self, self._ctx, self.state)
        self.enterRule(localctx, 182, self.RULE_finally_block)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1512
            self.match(FandangoParser.FINALLY)
            self.state = 1513
            self.match(FandangoParser.COLON)
            self.state = 1514
            self.block()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Match_stmtContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def MATCH(self):
            return self.getToken(FandangoParser.MATCH, 0)

        def subject_expr(self):
            return self.getTypedRuleContext(FandangoParser.Subject_exprContext, 0)

        def COLON(self):
            return self.getToken(FandangoParser.COLON, 0)

        def NEWLINE(self):
            return self.getToken(FandangoParser.NEWLINE, 0)

        def INDENT(self):
            return self.getToken(FandangoParser.INDENT, 0)

        def DEDENT(self):
            return self.getToken(FandangoParser.DEDENT, 0)

        def case_block(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(FandangoParser.Case_blockContext)
            else:
                return self.getTypedRuleContext(FandangoParser.Case_blockContext, i)

        def getRuleIndex(self):
            return FandangoParser.RULE_match_stmt

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterMatch_stmt"):
                listener.enterMatch_stmt(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitMatch_stmt"):
                listener.exitMatch_stmt(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitMatch_stmt"):
                return visitor.visitMatch_stmt(self)
            else:
                return visitor.visitChildren(self)

    def match_stmt(self):

        localctx = FandangoParser.Match_stmtContext(self, self._ctx, self.state)
        self.enterRule(localctx, 184, self.RULE_match_stmt)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1516
            self.match(FandangoParser.MATCH)
            self.state = 1517
            self.subject_expr()
            self.state = 1518
            self.match(FandangoParser.COLON)
            self.state = 1519
            self.match(FandangoParser.NEWLINE)
            self.state = 1520
            self.match(FandangoParser.INDENT)
            self.state = 1522
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 1521
                self.case_block()
                self.state = 1524
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not (_la == 15):
                    break

            self.state = 1526
            self.match(FandangoParser.DEDENT)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Subject_exprContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def star_named_expression(self):
            return self.getTypedRuleContext(
                FandangoParser.Star_named_expressionContext, 0
            )

        def COMMA(self):
            return self.getToken(FandangoParser.COMMA, 0)

        def star_named_expressions(self):
            return self.getTypedRuleContext(
                FandangoParser.Star_named_expressionsContext, 0
            )

        def named_expression(self):
            return self.getTypedRuleContext(FandangoParser.Named_expressionContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_subject_expr

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterSubject_expr"):
                listener.enterSubject_expr(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitSubject_expr"):
                listener.exitSubject_expr(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitSubject_expr"):
                return visitor.visitSubject_expr(self)
            else:
                return visitor.visitChildren(self)

    def subject_expr(self):

        localctx = FandangoParser.Subject_exprContext(self, self._ctx, self.state)
        self.enterRule(localctx, 186, self.RULE_subject_expr)
        self._la = 0  # Token type
        try:
            self.state = 1534
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 167, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 1528
                self.star_named_expression()
                self.state = 1529
                self.match(FandangoParser.COMMA)
                self.state = 1531
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if (((_la) & ~0x3F) == 0 and ((1 << _la) & 4505888853336344) != 0) or (
                    (((_la - 66)) & ~0x3F) == 0
                    and ((1 << (_la - 66)) & 4503599633861127) != 0
                ):
                    self.state = 1530
                    self.star_named_expressions()

                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 1533
                self.named_expression()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Case_blockContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def CASE(self):
            return self.getToken(FandangoParser.CASE, 0)

        def patterns(self):
            return self.getTypedRuleContext(FandangoParser.PatternsContext, 0)

        def COLON(self):
            return self.getToken(FandangoParser.COLON, 0)

        def block(self):
            return self.getTypedRuleContext(FandangoParser.BlockContext, 0)

        def guard(self):
            return self.getTypedRuleContext(FandangoParser.GuardContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_case_block

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterCase_block"):
                listener.enterCase_block(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitCase_block"):
                listener.exitCase_block(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitCase_block"):
                return visitor.visitCase_block(self)
            else:
                return visitor.visitChildren(self)

    def case_block(self):

        localctx = FandangoParser.Case_blockContext(self, self._ctx, self.state)
        self.enterRule(localctx, 188, self.RULE_case_block)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1536
            self.match(FandangoParser.CASE)
            self.state = 1537
            self.patterns()
            self.state = 1539
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == 28:
                self.state = 1538
                self.guard()

            self.state = 1541
            self.match(FandangoParser.COLON)
            self.state = 1542
            self.block()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class GuardContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def IF(self):
            return self.getToken(FandangoParser.IF, 0)

        def named_expression(self):
            return self.getTypedRuleContext(FandangoParser.Named_expressionContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_guard

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterGuard"):
                listener.enterGuard(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitGuard"):
                listener.exitGuard(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitGuard"):
                return visitor.visitGuard(self)
            else:
                return visitor.visitChildren(self)

    def guard(self):

        localctx = FandangoParser.GuardContext(self, self._ctx, self.state)
        self.enterRule(localctx, 190, self.RULE_guard)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1544
            self.match(FandangoParser.IF)
            self.state = 1545
            self.named_expression()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class PatternsContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def open_sequence_pattern(self):
            return self.getTypedRuleContext(
                FandangoParser.Open_sequence_patternContext, 0
            )

        def pattern(self):
            return self.getTypedRuleContext(FandangoParser.PatternContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_patterns

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterPatterns"):
                listener.enterPatterns(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitPatterns"):
                listener.exitPatterns(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitPatterns"):
                return visitor.visitPatterns(self)
            else:
                return visitor.visitChildren(self)

    def patterns(self):

        localctx = FandangoParser.PatternsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 192, self.RULE_patterns)
        try:
            self.state = 1549
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 169, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 1547
                self.open_sequence_pattern()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 1548
                self.pattern()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class PatternContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def as_pattern(self):
            return self.getTypedRuleContext(FandangoParser.As_patternContext, 0)

        def or_pattern(self):
            return self.getTypedRuleContext(FandangoParser.Or_patternContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_pattern

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterPattern"):
                listener.enterPattern(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitPattern"):
                listener.exitPattern(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitPattern"):
                return visitor.visitPattern(self)
            else:
                return visitor.visitChildren(self)

    def pattern(self):

        localctx = FandangoParser.PatternContext(self, self._ctx, self.state)
        self.enterRule(localctx, 194, self.RULE_pattern)
        try:
            self.state = 1553
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 170, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 1551
                self.as_pattern()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 1552
                self.or_pattern()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class As_patternContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def or_pattern(self):
            return self.getTypedRuleContext(FandangoParser.Or_patternContext, 0)

        def AS(self):
            return self.getToken(FandangoParser.AS, 0)

        def pattern_capture_target(self):
            return self.getTypedRuleContext(
                FandangoParser.Pattern_capture_targetContext, 0
            )

        def getRuleIndex(self):
            return FandangoParser.RULE_as_pattern

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterAs_pattern"):
                listener.enterAs_pattern(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitAs_pattern"):
                listener.exitAs_pattern(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitAs_pattern"):
                return visitor.visitAs_pattern(self)
            else:
                return visitor.visitChildren(self)

    def as_pattern(self):

        localctx = FandangoParser.As_patternContext(self, self._ctx, self.state)
        self.enterRule(localctx, 196, self.RULE_as_pattern)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1555
            self.or_pattern()
            self.state = 1556
            self.match(FandangoParser.AS)
            self.state = 1557
            self.pattern_capture_target()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Or_patternContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def closed_pattern(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(FandangoParser.Closed_patternContext)
            else:
                return self.getTypedRuleContext(FandangoParser.Closed_patternContext, i)

        def OR_OP(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.OR_OP)
            else:
                return self.getToken(FandangoParser.OR_OP, i)

        def getRuleIndex(self):
            return FandangoParser.RULE_or_pattern

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterOr_pattern"):
                listener.enterOr_pattern(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitOr_pattern"):
                listener.exitOr_pattern(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitOr_pattern"):
                return visitor.visitOr_pattern(self)
            else:
                return visitor.visitChildren(self)

    def or_pattern(self):

        localctx = FandangoParser.Or_patternContext(self, self._ctx, self.state)
        self.enterRule(localctx, 198, self.RULE_or_pattern)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1559
            self.closed_pattern()
            self.state = 1564
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la == 77:
                self.state = 1560
                self.match(FandangoParser.OR_OP)
                self.state = 1561
                self.closed_pattern()
                self.state = 1566
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Closed_patternContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def literal_pattern(self):
            return self.getTypedRuleContext(FandangoParser.Literal_patternContext, 0)

        def capture_pattern(self):
            return self.getTypedRuleContext(FandangoParser.Capture_patternContext, 0)

        def wildcard_pattern(self):
            return self.getTypedRuleContext(FandangoParser.Wildcard_patternContext, 0)

        def value_pattern(self):
            return self.getTypedRuleContext(FandangoParser.Value_patternContext, 0)

        def group_pattern(self):
            return self.getTypedRuleContext(FandangoParser.Group_patternContext, 0)

        def sequence_pattern(self):
            return self.getTypedRuleContext(FandangoParser.Sequence_patternContext, 0)

        def mapping_pattern(self):
            return self.getTypedRuleContext(FandangoParser.Mapping_patternContext, 0)

        def class_pattern(self):
            return self.getTypedRuleContext(FandangoParser.Class_patternContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_closed_pattern

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterClosed_pattern"):
                listener.enterClosed_pattern(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitClosed_pattern"):
                listener.exitClosed_pattern(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitClosed_pattern"):
                return visitor.visitClosed_pattern(self)
            else:
                return visitor.visitChildren(self)

    def closed_pattern(self):

        localctx = FandangoParser.Closed_patternContext(self, self._ctx, self.state)
        self.enterRule(localctx, 200, self.RULE_closed_pattern)
        try:
            self.state = 1575
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 172, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 1567
                self.literal_pattern()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 1568
                self.capture_pattern()
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 1569
                self.wildcard_pattern()
                pass

            elif la_ == 4:
                self.enterOuterAlt(localctx, 4)
                self.state = 1570
                self.value_pattern()
                pass

            elif la_ == 5:
                self.enterOuterAlt(localctx, 5)
                self.state = 1571
                self.group_pattern()
                pass

            elif la_ == 6:
                self.enterOuterAlt(localctx, 6)
                self.state = 1572
                self.sequence_pattern()
                pass

            elif la_ == 7:
                self.enterOuterAlt(localctx, 7)
                self.state = 1573
                self.mapping_pattern()
                pass

            elif la_ == 8:
                self.enterOuterAlt(localctx, 8)
                self.state = 1574
                self.class_pattern()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Literal_patternContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def signed_number(self):
            return self.getTypedRuleContext(FandangoParser.Signed_numberContext, 0)

        def complex_number(self):
            return self.getTypedRuleContext(FandangoParser.Complex_numberContext, 0)

        def strings(self):
            return self.getTypedRuleContext(FandangoParser.StringsContext, 0)

        def NONE(self):
            return self.getToken(FandangoParser.NONE, 0)

        def TRUE(self):
            return self.getToken(FandangoParser.TRUE, 0)

        def FALSE(self):
            return self.getToken(FandangoParser.FALSE, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_literal_pattern

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterLiteral_pattern"):
                listener.enterLiteral_pattern(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitLiteral_pattern"):
                listener.exitLiteral_pattern(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitLiteral_pattern"):
                return visitor.visitLiteral_pattern(self)
            else:
                return visitor.visitChildren(self)

    def literal_pattern(self):

        localctx = FandangoParser.Literal_patternContext(self, self._ctx, self.state)
        self.enterRule(localctx, 202, self.RULE_literal_pattern)
        try:
            self.state = 1583
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 173, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 1577
                self.signed_number()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 1578
                self.complex_number()
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 1579
                self.strings()
                pass

            elif la_ == 4:
                self.enterOuterAlt(localctx, 4)
                self.state = 1580
                self.match(FandangoParser.NONE)
                pass

            elif la_ == 5:
                self.enterOuterAlt(localctx, 5)
                self.state = 1581
                self.match(FandangoParser.TRUE)
                pass

            elif la_ == 6:
                self.enterOuterAlt(localctx, 6)
                self.state = 1582
                self.match(FandangoParser.FALSE)
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Literal_exprContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def signed_number(self):
            return self.getTypedRuleContext(FandangoParser.Signed_numberContext, 0)

        def complex_number(self):
            return self.getTypedRuleContext(FandangoParser.Complex_numberContext, 0)

        def strings(self):
            return self.getTypedRuleContext(FandangoParser.StringsContext, 0)

        def NONE(self):
            return self.getToken(FandangoParser.NONE, 0)

        def TRUE(self):
            return self.getToken(FandangoParser.TRUE, 0)

        def FALSE(self):
            return self.getToken(FandangoParser.FALSE, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_literal_expr

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterLiteral_expr"):
                listener.enterLiteral_expr(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitLiteral_expr"):
                listener.exitLiteral_expr(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitLiteral_expr"):
                return visitor.visitLiteral_expr(self)
            else:
                return visitor.visitChildren(self)

    def literal_expr(self):

        localctx = FandangoParser.Literal_exprContext(self, self._ctx, self.state)
        self.enterRule(localctx, 204, self.RULE_literal_expr)
        try:
            self.state = 1591
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 174, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 1585
                self.signed_number()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 1586
                self.complex_number()
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 1587
                self.strings()
                pass

            elif la_ == 4:
                self.enterOuterAlt(localctx, 4)
                self.state = 1588
                self.match(FandangoParser.NONE)
                pass

            elif la_ == 5:
                self.enterOuterAlt(localctx, 5)
                self.state = 1589
                self.match(FandangoParser.TRUE)
                pass

            elif la_ == 6:
                self.enterOuterAlt(localctx, 6)
                self.state = 1590
                self.match(FandangoParser.FALSE)
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Complex_numberContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def signed_real_number(self):
            return self.getTypedRuleContext(FandangoParser.Signed_real_numberContext, 0)

        def ADD(self):
            return self.getToken(FandangoParser.ADD, 0)

        def imaginary_number(self):
            return self.getTypedRuleContext(FandangoParser.Imaginary_numberContext, 0)

        def MINUS(self):
            return self.getToken(FandangoParser.MINUS, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_complex_number

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterComplex_number"):
                listener.enterComplex_number(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitComplex_number"):
                listener.exitComplex_number(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitComplex_number"):
                return visitor.visitComplex_number(self)
            else:
                return visitor.visitChildren(self)

    def complex_number(self):

        localctx = FandangoParser.Complex_numberContext(self, self._ctx, self.state)
        self.enterRule(localctx, 206, self.RULE_complex_number)
        try:
            self.state = 1601
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 175, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 1593
                self.signed_real_number()
                self.state = 1594
                self.match(FandangoParser.ADD)
                self.state = 1595
                self.imaginary_number()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 1597
                self.signed_real_number()
                self.state = 1598
                self.match(FandangoParser.MINUS)
                self.state = 1599
                self.imaginary_number()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Signed_numberContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NUMBER(self):
            return self.getToken(FandangoParser.NUMBER, 0)

        def MINUS(self):
            return self.getToken(FandangoParser.MINUS, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_signed_number

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterSigned_number"):
                listener.enterSigned_number(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitSigned_number"):
                listener.exitSigned_number(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitSigned_number"):
                return visitor.visitSigned_number(self)
            else:
                return visitor.visitChildren(self)

    def signed_number(self):

        localctx = FandangoParser.Signed_numberContext(self, self._ctx, self.state)
        self.enterRule(localctx, 208, self.RULE_signed_number)
        try:
            self.state = 1606
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [4]:
                self.enterOuterAlt(localctx, 1)
                self.state = 1603
                self.match(FandangoParser.NUMBER)
                pass
            elif token in [83]:
                self.enterOuterAlt(localctx, 2)
                self.state = 1604
                self.match(FandangoParser.MINUS)
                self.state = 1605
                self.match(FandangoParser.NUMBER)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Signed_real_numberContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def real_number(self):
            return self.getTypedRuleContext(FandangoParser.Real_numberContext, 0)

        def MINUS(self):
            return self.getToken(FandangoParser.MINUS, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_signed_real_number

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterSigned_real_number"):
                listener.enterSigned_real_number(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitSigned_real_number"):
                listener.exitSigned_real_number(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitSigned_real_number"):
                return visitor.visitSigned_real_number(self)
            else:
                return visitor.visitChildren(self)

    def signed_real_number(self):

        localctx = FandangoParser.Signed_real_numberContext(self, self._ctx, self.state)
        self.enterRule(localctx, 210, self.RULE_signed_real_number)
        try:
            self.state = 1611
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [4]:
                self.enterOuterAlt(localctx, 1)
                self.state = 1608
                self.real_number()
                pass
            elif token in [83]:
                self.enterOuterAlt(localctx, 2)
                self.state = 1609
                self.match(FandangoParser.MINUS)
                self.state = 1610
                self.real_number()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Real_numberContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NUMBER(self):
            return self.getToken(FandangoParser.NUMBER, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_real_number

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterReal_number"):
                listener.enterReal_number(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitReal_number"):
                listener.exitReal_number(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitReal_number"):
                return visitor.visitReal_number(self)
            else:
                return visitor.visitChildren(self)

    def real_number(self):

        localctx = FandangoParser.Real_numberContext(self, self._ctx, self.state)
        self.enterRule(localctx, 212, self.RULE_real_number)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1613
            self.match(FandangoParser.NUMBER)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Imaginary_numberContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NUMBER(self):
            return self.getToken(FandangoParser.NUMBER, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_imaginary_number

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterImaginary_number"):
                listener.enterImaginary_number(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitImaginary_number"):
                listener.exitImaginary_number(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitImaginary_number"):
                return visitor.visitImaginary_number(self)
            else:
                return visitor.visitChildren(self)

    def imaginary_number(self):

        localctx = FandangoParser.Imaginary_numberContext(self, self._ctx, self.state)
        self.enterRule(localctx, 214, self.RULE_imaginary_number)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1615
            self.match(FandangoParser.NUMBER)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Capture_patternContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def pattern_capture_target(self):
            return self.getTypedRuleContext(
                FandangoParser.Pattern_capture_targetContext, 0
            )

        def getRuleIndex(self):
            return FandangoParser.RULE_capture_pattern

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterCapture_pattern"):
                listener.enterCapture_pattern(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitCapture_pattern"):
                listener.exitCapture_pattern(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitCapture_pattern"):
                return visitor.visitCapture_pattern(self)
            else:
                return visitor.visitChildren(self)

    def capture_pattern(self):

        localctx = FandangoParser.Capture_patternContext(self, self._ctx, self.state)
        self.enterRule(localctx, 216, self.RULE_capture_pattern)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1617
            self.pattern_capture_target()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Pattern_capture_targetContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NAME(self):
            return self.getToken(FandangoParser.NAME, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_pattern_capture_target

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterPattern_capture_target"):
                listener.enterPattern_capture_target(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitPattern_capture_target"):
                listener.exitPattern_capture_target(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitPattern_capture_target"):
                return visitor.visitPattern_capture_target(self)
            else:
                return visitor.visitChildren(self)

    def pattern_capture_target(self):

        localctx = FandangoParser.Pattern_capture_targetContext(
            self, self._ctx, self.state
        )
        self.enterRule(localctx, 218, self.RULE_pattern_capture_target)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1619
            self.match(FandangoParser.NAME)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Wildcard_patternContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def UNDERSCORE(self):
            return self.getToken(FandangoParser.UNDERSCORE, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_wildcard_pattern

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterWildcard_pattern"):
                listener.enterWildcard_pattern(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitWildcard_pattern"):
                listener.exitWildcard_pattern(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitWildcard_pattern"):
                return visitor.visitWildcard_pattern(self)
            else:
                return visitor.visitChildren(self)

    def wildcard_pattern(self):

        localctx = FandangoParser.Wildcard_patternContext(self, self._ctx, self.state)
        self.enterRule(localctx, 220, self.RULE_wildcard_pattern)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1621
            self.match(FandangoParser.UNDERSCORE)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Value_patternContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def attr(self):
            return self.getTypedRuleContext(FandangoParser.AttrContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_value_pattern

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterValue_pattern"):
                listener.enterValue_pattern(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitValue_pattern"):
                listener.exitValue_pattern(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitValue_pattern"):
                return visitor.visitValue_pattern(self)
            else:
                return visitor.visitChildren(self)

    def value_pattern(self):

        localctx = FandangoParser.Value_patternContext(self, self._ctx, self.state)
        self.enterRule(localctx, 222, self.RULE_value_pattern)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1623
            self.attr()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class AttrContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def name_or_attr(self):
            return self.getTypedRuleContext(FandangoParser.Name_or_attrContext, 0)

        def DOT(self):
            return self.getToken(FandangoParser.DOT, 0)

        def NAME(self):
            return self.getToken(FandangoParser.NAME, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_attr

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterAttr"):
                listener.enterAttr(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitAttr"):
                listener.exitAttr(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitAttr"):
                return visitor.visitAttr(self)
            else:
                return visitor.visitChildren(self)

    def attr(self):

        localctx = FandangoParser.AttrContext(self, self._ctx, self.state)
        self.enterRule(localctx, 224, self.RULE_attr)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1625
            self.name_or_attr(0)
            self.state = 1626
            self.match(FandangoParser.DOT)
            self.state = 1627
            self.match(FandangoParser.NAME)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Name_or_attrContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NAME(self):
            return self.getToken(FandangoParser.NAME, 0)

        def name_or_attr(self):
            return self.getTypedRuleContext(FandangoParser.Name_or_attrContext, 0)

        def DOT(self):
            return self.getToken(FandangoParser.DOT, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_name_or_attr

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterName_or_attr"):
                listener.enterName_or_attr(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitName_or_attr"):
                listener.exitName_or_attr(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitName_or_attr"):
                return visitor.visitName_or_attr(self)
            else:
                return visitor.visitChildren(self)

    def name_or_attr(self, _p: int = 0):
        _parentctx = self._ctx
        _parentState = self.state
        localctx = FandangoParser.Name_or_attrContext(self, self._ctx, _parentState)
        _prevctx = localctx
        _startState = 226
        self.enterRecursionRule(localctx, 226, self.RULE_name_or_attr, _p)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1630
            self.match(FandangoParser.NAME)
            self._ctx.stop = self._input.LT(-1)
            self.state = 1637
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 178, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    if self._parseListeners is not None:
                        self.triggerExitRuleEvent()
                    _prevctx = localctx
                    localctx = FandangoParser.Name_or_attrContext(
                        self, _parentctx, _parentState
                    )
                    self.pushNewRecursionContext(
                        localctx, _startState, self.RULE_name_or_attr
                    )
                    self.state = 1632
                    if not self.precpred(self._ctx, 2):
                        from antlr4.error.Errors import FailedPredicateException

                        raise FailedPredicateException(
                            self, "self.precpred(self._ctx, 2)"
                        )
                    self.state = 1633
                    self.match(FandangoParser.DOT)
                    self.state = 1634
                    self.match(FandangoParser.NAME)
                self.state = 1639
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 178, self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.unrollRecursionContexts(_parentctx)
        return localctx

    class Group_patternContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def OPEN_PAREN(self):
            return self.getToken(FandangoParser.OPEN_PAREN, 0)

        def pattern(self):
            return self.getTypedRuleContext(FandangoParser.PatternContext, 0)

        def CLOSE_PAREN(self):
            return self.getToken(FandangoParser.CLOSE_PAREN, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_group_pattern

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterGroup_pattern"):
                listener.enterGroup_pattern(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitGroup_pattern"):
                listener.exitGroup_pattern(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitGroup_pattern"):
                return visitor.visitGroup_pattern(self)
            else:
                return visitor.visitChildren(self)

    def group_pattern(self):

        localctx = FandangoParser.Group_patternContext(self, self._ctx, self.state)
        self.enterRule(localctx, 228, self.RULE_group_pattern)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1640
            self.match(FandangoParser.OPEN_PAREN)
            self.state = 1641
            self.pattern()
            self.state = 1642
            self.match(FandangoParser.CLOSE_PAREN)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Sequence_patternContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def OPEN_BRACK(self):
            return self.getToken(FandangoParser.OPEN_BRACK, 0)

        def CLOSE_BRACK(self):
            return self.getToken(FandangoParser.CLOSE_BRACK, 0)

        def maybe_sequence_pattern(self):
            return self.getTypedRuleContext(
                FandangoParser.Maybe_sequence_patternContext, 0
            )

        def OPEN_PAREN(self):
            return self.getToken(FandangoParser.OPEN_PAREN, 0)

        def CLOSE_PAREN(self):
            return self.getToken(FandangoParser.CLOSE_PAREN, 0)

        def open_sequence_pattern(self):
            return self.getTypedRuleContext(
                FandangoParser.Open_sequence_patternContext, 0
            )

        def getRuleIndex(self):
            return FandangoParser.RULE_sequence_pattern

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterSequence_pattern"):
                listener.enterSequence_pattern(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitSequence_pattern"):
                listener.exitSequence_pattern(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitSequence_pattern"):
                return visitor.visitSequence_pattern(self)
            else:
                return visitor.visitChildren(self)

    def sequence_pattern(self):

        localctx = FandangoParser.Sequence_patternContext(self, self._ctx, self.state)
        self.enterRule(localctx, 230, self.RULE_sequence_pattern)
        self._la = 0  # Token type
        try:
            self.state = 1654
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [75]:
                self.enterOuterAlt(localctx, 1)
                self.state = 1644
                self.match(FandangoParser.OPEN_BRACK)
                self.state = 1646
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if (((_la) & ~0x3F) == 0 and ((1 << _la) & 4523408024928280) != 0) or (
                    (((_la - 67)) & ~0x3F) == 0
                    and ((1 << (_la - 67)) & 2251799815848195) != 0
                ):
                    self.state = 1645
                    self.maybe_sequence_pattern()

                self.state = 1648
                self.match(FandangoParser.CLOSE_BRACK)
                pass
            elif token in [68]:
                self.enterOuterAlt(localctx, 2)
                self.state = 1649
                self.match(FandangoParser.OPEN_PAREN)
                self.state = 1651
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if (((_la) & ~0x3F) == 0 and ((1 << _la) & 4523408024928280) != 0) or (
                    (((_la - 67)) & ~0x3F) == 0
                    and ((1 << (_la - 67)) & 2251799815848195) != 0
                ):
                    self.state = 1650
                    self.open_sequence_pattern()

                self.state = 1653
                self.match(FandangoParser.CLOSE_PAREN)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Open_sequence_patternContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def maybe_star_pattern(self):
            return self.getTypedRuleContext(FandangoParser.Maybe_star_patternContext, 0)

        def COMMA(self):
            return self.getToken(FandangoParser.COMMA, 0)

        def maybe_sequence_pattern(self):
            return self.getTypedRuleContext(
                FandangoParser.Maybe_sequence_patternContext, 0
            )

        def getRuleIndex(self):
            return FandangoParser.RULE_open_sequence_pattern

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterOpen_sequence_pattern"):
                listener.enterOpen_sequence_pattern(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitOpen_sequence_pattern"):
                listener.exitOpen_sequence_pattern(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitOpen_sequence_pattern"):
                return visitor.visitOpen_sequence_pattern(self)
            else:
                return visitor.visitChildren(self)

    def open_sequence_pattern(self):

        localctx = FandangoParser.Open_sequence_patternContext(
            self, self._ctx, self.state
        )
        self.enterRule(localctx, 232, self.RULE_open_sequence_pattern)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1656
            self.maybe_star_pattern()
            self.state = 1657
            self.match(FandangoParser.COMMA)
            self.state = 1659
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if (((_la) & ~0x3F) == 0 and ((1 << _la) & 4523408024928280) != 0) or (
                (((_la - 67)) & ~0x3F) == 0
                and ((1 << (_la - 67)) & 2251799815848195) != 0
            ):
                self.state = 1658
                self.maybe_sequence_pattern()

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Maybe_sequence_patternContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def maybe_star_pattern(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(
                    FandangoParser.Maybe_star_patternContext
                )
            else:
                return self.getTypedRuleContext(
                    FandangoParser.Maybe_star_patternContext, i
                )

        def COMMA(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.COMMA)
            else:
                return self.getToken(FandangoParser.COMMA, i)

        def getRuleIndex(self):
            return FandangoParser.RULE_maybe_sequence_pattern

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterMaybe_sequence_pattern"):
                listener.enterMaybe_sequence_pattern(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitMaybe_sequence_pattern"):
                listener.exitMaybe_sequence_pattern(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitMaybe_sequence_pattern"):
                return visitor.visitMaybe_sequence_pattern(self)
            else:
                return visitor.visitChildren(self)

    def maybe_sequence_pattern(self):

        localctx = FandangoParser.Maybe_sequence_patternContext(
            self, self._ctx, self.state
        )
        self.enterRule(localctx, 234, self.RULE_maybe_sequence_pattern)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1661
            self.maybe_star_pattern()
            self.state = 1666
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 183, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 1662
                    self.match(FandangoParser.COMMA)
                    self.state = 1663
                    self.maybe_star_pattern()
                self.state = 1668
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 183, self._ctx)

            self.state = 1670
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == 70:
                self.state = 1669
                self.match(FandangoParser.COMMA)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Maybe_star_patternContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def star_pattern(self):
            return self.getTypedRuleContext(FandangoParser.Star_patternContext, 0)

        def pattern(self):
            return self.getTypedRuleContext(FandangoParser.PatternContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_maybe_star_pattern

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterMaybe_star_pattern"):
                listener.enterMaybe_star_pattern(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitMaybe_star_pattern"):
                listener.exitMaybe_star_pattern(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitMaybe_star_pattern"):
                return visitor.visitMaybe_star_pattern(self)
            else:
                return visitor.visitChildren(self)

    def maybe_star_pattern(self):

        localctx = FandangoParser.Maybe_star_patternContext(self, self._ctx, self.state)
        self.enterRule(localctx, 236, self.RULE_maybe_star_pattern)
        try:
            self.state = 1674
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [67]:
                self.enterOuterAlt(localctx, 1)
                self.state = 1672
                self.star_pattern()
                pass
            elif token in [3, 4, 23, 34, 41, 44, 52, 68, 75, 83, 88, 118]:
                self.enterOuterAlt(localctx, 2)
                self.state = 1673
                self.pattern()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Star_patternContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def STAR(self):
            return self.getToken(FandangoParser.STAR, 0)

        def pattern_capture_target(self):
            return self.getTypedRuleContext(
                FandangoParser.Pattern_capture_targetContext, 0
            )

        def wildcard_pattern(self):
            return self.getTypedRuleContext(FandangoParser.Wildcard_patternContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_star_pattern

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterStar_pattern"):
                listener.enterStar_pattern(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitStar_pattern"):
                listener.exitStar_pattern(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitStar_pattern"):
                return visitor.visitStar_pattern(self)
            else:
                return visitor.visitChildren(self)

    def star_pattern(self):

        localctx = FandangoParser.Star_patternContext(self, self._ctx, self.state)
        self.enterRule(localctx, 238, self.RULE_star_pattern)
        try:
            self.state = 1680
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 186, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 1676
                self.match(FandangoParser.STAR)
                self.state = 1677
                self.pattern_capture_target()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 1678
                self.match(FandangoParser.STAR)
                self.state = 1679
                self.wildcard_pattern()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Mapping_patternContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def OPEN_BRACE(self):
            return self.getToken(FandangoParser.OPEN_BRACE, 0)

        def CLOSE_BRACE(self):
            return self.getToken(FandangoParser.CLOSE_BRACE, 0)

        def double_star_pattern(self):
            return self.getTypedRuleContext(
                FandangoParser.Double_star_patternContext, 0
            )

        def COMMA(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.COMMA)
            else:
                return self.getToken(FandangoParser.COMMA, i)

        def items_pattern(self):
            return self.getTypedRuleContext(FandangoParser.Items_patternContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_mapping_pattern

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterMapping_pattern"):
                listener.enterMapping_pattern(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitMapping_pattern"):
                listener.exitMapping_pattern(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitMapping_pattern"):
                return visitor.visitMapping_pattern(self)
            else:
                return visitor.visitChildren(self)

    def mapping_pattern(self):

        localctx = FandangoParser.Mapping_patternContext(self, self._ctx, self.state)
        self.enterRule(localctx, 240, self.RULE_mapping_pattern)
        self._la = 0  # Token type
        try:
            self.state = 1707
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 190, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 1682
                self.match(FandangoParser.OPEN_BRACE)
                self.state = 1683
                self.match(FandangoParser.CLOSE_BRACE)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 1684
                self.match(FandangoParser.OPEN_BRACE)
                self.state = 1685
                self.double_star_pattern()
                self.state = 1687
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == 70:
                    self.state = 1686
                    self.match(FandangoParser.COMMA)

                self.state = 1689
                self.match(FandangoParser.CLOSE_BRACE)
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 1691
                self.match(FandangoParser.OPEN_BRACE)
                self.state = 1692
                self.items_pattern()
                self.state = 1693
                self.match(FandangoParser.COMMA)
                self.state = 1694
                self.double_star_pattern()
                self.state = 1696
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == 70:
                    self.state = 1695
                    self.match(FandangoParser.COMMA)

                self.state = 1698
                self.match(FandangoParser.CLOSE_BRACE)
                pass

            elif la_ == 4:
                self.enterOuterAlt(localctx, 4)
                self.state = 1700
                self.match(FandangoParser.OPEN_BRACE)
                self.state = 1701
                self.items_pattern()
                self.state = 1703
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == 70:
                    self.state = 1702
                    self.match(FandangoParser.COMMA)

                self.state = 1705
                self.match(FandangoParser.CLOSE_BRACE)
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Items_patternContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def key_value_pattern(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(
                    FandangoParser.Key_value_patternContext
                )
            else:
                return self.getTypedRuleContext(
                    FandangoParser.Key_value_patternContext, i
                )

        def COMMA(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.COMMA)
            else:
                return self.getToken(FandangoParser.COMMA, i)

        def getRuleIndex(self):
            return FandangoParser.RULE_items_pattern

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterItems_pattern"):
                listener.enterItems_pattern(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitItems_pattern"):
                listener.exitItems_pattern(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitItems_pattern"):
                return visitor.visitItems_pattern(self)
            else:
                return visitor.visitChildren(self)

    def items_pattern(self):

        localctx = FandangoParser.Items_patternContext(self, self._ctx, self.state)
        self.enterRule(localctx, 242, self.RULE_items_pattern)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1709
            self.key_value_pattern()
            self.state = 1714
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 191, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 1710
                    self.match(FandangoParser.COMMA)
                    self.state = 1711
                    self.key_value_pattern()
                self.state = 1716
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 191, self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Key_value_patternContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def COLON(self):
            return self.getToken(FandangoParser.COLON, 0)

        def pattern(self):
            return self.getTypedRuleContext(FandangoParser.PatternContext, 0)

        def literal_expr(self):
            return self.getTypedRuleContext(FandangoParser.Literal_exprContext, 0)

        def attr(self):
            return self.getTypedRuleContext(FandangoParser.AttrContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_key_value_pattern

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterKey_value_pattern"):
                listener.enterKey_value_pattern(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitKey_value_pattern"):
                listener.exitKey_value_pattern(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitKey_value_pattern"):
                return visitor.visitKey_value_pattern(self)
            else:
                return visitor.visitChildren(self)

    def key_value_pattern(self):

        localctx = FandangoParser.Key_value_patternContext(self, self._ctx, self.state)
        self.enterRule(localctx, 244, self.RULE_key_value_pattern)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1719
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [3, 4, 23, 34, 41, 83, 118]:
                self.state = 1717
                self.literal_expr()
                pass
            elif token in [52]:
                self.state = 1718
                self.attr()
                pass
            else:
                raise NoViableAltException(self)

            self.state = 1721
            self.match(FandangoParser.COLON)
            self.state = 1722
            self.pattern()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Double_star_patternContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def POWER(self):
            return self.getToken(FandangoParser.POWER, 0)

        def pattern_capture_target(self):
            return self.getTypedRuleContext(
                FandangoParser.Pattern_capture_targetContext, 0
            )

        def getRuleIndex(self):
            return FandangoParser.RULE_double_star_pattern

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterDouble_star_pattern"):
                listener.enterDouble_star_pattern(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitDouble_star_pattern"):
                listener.exitDouble_star_pattern(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitDouble_star_pattern"):
                return visitor.visitDouble_star_pattern(self)
            else:
                return visitor.visitChildren(self)

    def double_star_pattern(self):

        localctx = FandangoParser.Double_star_patternContext(
            self, self._ctx, self.state
        )
        self.enterRule(localctx, 246, self.RULE_double_star_pattern)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1724
            self.match(FandangoParser.POWER)
            self.state = 1725
            self.pattern_capture_target()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Class_patternContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def name_or_attr(self):
            return self.getTypedRuleContext(FandangoParser.Name_or_attrContext, 0)

        def OPEN_PAREN(self):
            return self.getToken(FandangoParser.OPEN_PAREN, 0)

        def CLOSE_PAREN(self):
            return self.getToken(FandangoParser.CLOSE_PAREN, 0)

        def positional_patterns(self):
            return self.getTypedRuleContext(
                FandangoParser.Positional_patternsContext, 0
            )

        def COMMA(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.COMMA)
            else:
                return self.getToken(FandangoParser.COMMA, i)

        def keyword_patterns(self):
            return self.getTypedRuleContext(FandangoParser.Keyword_patternsContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_class_pattern

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterClass_pattern"):
                listener.enterClass_pattern(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitClass_pattern"):
                listener.exitClass_pattern(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitClass_pattern"):
                return visitor.visitClass_pattern(self)
            else:
                return visitor.visitChildren(self)

    def class_pattern(self):

        localctx = FandangoParser.Class_patternContext(self, self._ctx, self.state)
        self.enterRule(localctx, 248, self.RULE_class_pattern)
        self._la = 0  # Token type
        try:
            self.state = 1757
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 196, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 1727
                self.name_or_attr(0)
                self.state = 1728
                self.match(FandangoParser.OPEN_PAREN)
                self.state = 1729
                self.match(FandangoParser.CLOSE_PAREN)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 1731
                self.name_or_attr(0)
                self.state = 1732
                self.match(FandangoParser.OPEN_PAREN)
                self.state = 1733
                self.positional_patterns()
                self.state = 1735
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == 70:
                    self.state = 1734
                    self.match(FandangoParser.COMMA)

                self.state = 1737
                self.match(FandangoParser.CLOSE_PAREN)
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 1739
                self.name_or_attr(0)
                self.state = 1740
                self.match(FandangoParser.OPEN_PAREN)
                self.state = 1741
                self.keyword_patterns()
                self.state = 1743
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == 70:
                    self.state = 1742
                    self.match(FandangoParser.COMMA)

                self.state = 1745
                self.match(FandangoParser.CLOSE_PAREN)
                pass

            elif la_ == 4:
                self.enterOuterAlt(localctx, 4)
                self.state = 1747
                self.name_or_attr(0)
                self.state = 1748
                self.match(FandangoParser.OPEN_PAREN)
                self.state = 1749
                self.positional_patterns()
                self.state = 1750
                self.match(FandangoParser.COMMA)
                self.state = 1751
                self.keyword_patterns()
                self.state = 1753
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == 70:
                    self.state = 1752
                    self.match(FandangoParser.COMMA)

                self.state = 1755
                self.match(FandangoParser.CLOSE_PAREN)
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Positional_patternsContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def pattern(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(FandangoParser.PatternContext)
            else:
                return self.getTypedRuleContext(FandangoParser.PatternContext, i)

        def COMMA(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.COMMA)
            else:
                return self.getToken(FandangoParser.COMMA, i)

        def getRuleIndex(self):
            return FandangoParser.RULE_positional_patterns

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterPositional_patterns"):
                listener.enterPositional_patterns(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitPositional_patterns"):
                listener.exitPositional_patterns(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitPositional_patterns"):
                return visitor.visitPositional_patterns(self)
            else:
                return visitor.visitChildren(self)

    def positional_patterns(self):

        localctx = FandangoParser.Positional_patternsContext(
            self, self._ctx, self.state
        )
        self.enterRule(localctx, 250, self.RULE_positional_patterns)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1759
            self.pattern()
            self.state = 1764
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 197, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 1760
                    self.match(FandangoParser.COMMA)
                    self.state = 1761
                    self.pattern()
                self.state = 1766
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 197, self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Keyword_patternsContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def keyword_pattern(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(FandangoParser.Keyword_patternContext)
            else:
                return self.getTypedRuleContext(
                    FandangoParser.Keyword_patternContext, i
                )

        def COMMA(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.COMMA)
            else:
                return self.getToken(FandangoParser.COMMA, i)

        def getRuleIndex(self):
            return FandangoParser.RULE_keyword_patterns

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterKeyword_patterns"):
                listener.enterKeyword_patterns(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitKeyword_patterns"):
                listener.exitKeyword_patterns(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitKeyword_patterns"):
                return visitor.visitKeyword_patterns(self)
            else:
                return visitor.visitChildren(self)

    def keyword_patterns(self):

        localctx = FandangoParser.Keyword_patternsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 252, self.RULE_keyword_patterns)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1767
            self.keyword_pattern()
            self.state = 1772
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 198, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 1768
                    self.match(FandangoParser.COMMA)
                    self.state = 1769
                    self.keyword_pattern()
                self.state = 1774
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 198, self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Keyword_patternContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NAME(self):
            return self.getToken(FandangoParser.NAME, 0)

        def ASSIGN(self):
            return self.getToken(FandangoParser.ASSIGN, 0)

        def pattern(self):
            return self.getTypedRuleContext(FandangoParser.PatternContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_keyword_pattern

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterKeyword_pattern"):
                listener.enterKeyword_pattern(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitKeyword_pattern"):
                listener.exitKeyword_pattern(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitKeyword_pattern"):
                return visitor.visitKeyword_pattern(self)
            else:
                return visitor.visitChildren(self)

    def keyword_pattern(self):

        localctx = FandangoParser.Keyword_patternContext(self, self._ctx, self.state)
        self.enterRule(localctx, 254, self.RULE_keyword_pattern)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1775
            self.match(FandangoParser.NAME)
            self.state = 1776
            self.match(FandangoParser.ASSIGN)
            self.state = 1777
            self.pattern()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Type_aliasContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def TYPE(self):
            return self.getToken(FandangoParser.TYPE, 0)

        def NAME(self):
            return self.getToken(FandangoParser.NAME, 0)

        def ASSIGN(self):
            return self.getToken(FandangoParser.ASSIGN, 0)

        def expression(self):
            return self.getTypedRuleContext(FandangoParser.ExpressionContext, 0)

        def type_params(self):
            return self.getTypedRuleContext(FandangoParser.Type_paramsContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_type_alias

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterType_alias"):
                listener.enterType_alias(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitType_alias"):
                listener.exitType_alias(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitType_alias"):
                return visitor.visitType_alias(self)
            else:
                return visitor.visitChildren(self)

    def type_alias(self):

        localctx = FandangoParser.Type_aliasContext(self, self._ctx, self.state)
        self.enterRule(localctx, 256, self.RULE_type_alias)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1779
            self.match(FandangoParser.TYPE)
            self.state = 1780
            self.match(FandangoParser.NAME)
            self.state = 1782
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == 75:
                self.state = 1781
                self.type_params()

            self.state = 1784
            self.match(FandangoParser.ASSIGN)
            self.state = 1785
            self.expression()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Type_paramsContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def OPEN_BRACK(self):
            return self.getToken(FandangoParser.OPEN_BRACK, 0)

        def type_param_seq(self):
            return self.getTypedRuleContext(FandangoParser.Type_param_seqContext, 0)

        def CLOSE_BRACK(self):
            return self.getToken(FandangoParser.CLOSE_BRACK, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_type_params

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterType_params"):
                listener.enterType_params(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitType_params"):
                listener.exitType_params(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitType_params"):
                return visitor.visitType_params(self)
            else:
                return visitor.visitChildren(self)

    def type_params(self):

        localctx = FandangoParser.Type_paramsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 258, self.RULE_type_params)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1787
            self.match(FandangoParser.OPEN_BRACK)
            self.state = 1788
            self.type_param_seq()
            self.state = 1789
            self.match(FandangoParser.CLOSE_BRACK)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Type_param_seqContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def type_param(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(FandangoParser.Type_paramContext)
            else:
                return self.getTypedRuleContext(FandangoParser.Type_paramContext, i)

        def COMMA(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.COMMA)
            else:
                return self.getToken(FandangoParser.COMMA, i)

        def getRuleIndex(self):
            return FandangoParser.RULE_type_param_seq

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterType_param_seq"):
                listener.enterType_param_seq(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitType_param_seq"):
                listener.exitType_param_seq(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitType_param_seq"):
                return visitor.visitType_param_seq(self)
            else:
                return visitor.visitChildren(self)

    def type_param_seq(self):

        localctx = FandangoParser.Type_param_seqContext(self, self._ctx, self.state)
        self.enterRule(localctx, 260, self.RULE_type_param_seq)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1791
            self.type_param()
            self.state = 1796
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 200, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 1792
                    self.match(FandangoParser.COMMA)
                    self.state = 1793
                    self.type_param()
                self.state = 1798
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 200, self._ctx)

            self.state = 1800
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == 70:
                self.state = 1799
                self.match(FandangoParser.COMMA)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Type_paramContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NAME(self):
            return self.getToken(FandangoParser.NAME, 0)

        def type_param_bound(self):
            return self.getTypedRuleContext(FandangoParser.Type_param_boundContext, 0)

        def STAR(self):
            return self.getToken(FandangoParser.STAR, 0)

        def POWER(self):
            return self.getToken(FandangoParser.POWER, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_type_param

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterType_param"):
                listener.enterType_param(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitType_param"):
                listener.exitType_param(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitType_param"):
                return visitor.visitType_param(self)
            else:
                return visitor.visitChildren(self)

    def type_param(self):

        localctx = FandangoParser.Type_paramContext(self, self._ctx, self.state)
        self.enterRule(localctx, 262, self.RULE_type_param)
        self._la = 0  # Token type
        try:
            self.state = 1810
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [52]:
                self.enterOuterAlt(localctx, 1)
                self.state = 1802
                self.match(FandangoParser.NAME)
                self.state = 1804
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == 71:
                    self.state = 1803
                    self.type_param_bound()

                pass
            elif token in [67]:
                self.enterOuterAlt(localctx, 2)
                self.state = 1806
                self.match(FandangoParser.STAR)
                self.state = 1807
                self.match(FandangoParser.NAME)
                pass
            elif token in [73]:
                self.enterOuterAlt(localctx, 3)
                self.state = 1808
                self.match(FandangoParser.POWER)
                self.state = 1809
                self.match(FandangoParser.NAME)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Type_param_boundContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def COLON(self):
            return self.getToken(FandangoParser.COLON, 0)

        def expression(self):
            return self.getTypedRuleContext(FandangoParser.ExpressionContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_type_param_bound

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterType_param_bound"):
                listener.enterType_param_bound(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitType_param_bound"):
                listener.exitType_param_bound(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitType_param_bound"):
                return visitor.visitType_param_bound(self)
            else:
                return visitor.visitChildren(self)

    def type_param_bound(self):

        localctx = FandangoParser.Type_param_boundContext(self, self._ctx, self.state)
        self.enterRule(localctx, 264, self.RULE_type_param_bound)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1812
            self.match(FandangoParser.COLON)
            self.state = 1813
            self.expression()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ExpressionsContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def expression(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(FandangoParser.ExpressionContext)
            else:
                return self.getTypedRuleContext(FandangoParser.ExpressionContext, i)

        def COMMA(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.COMMA)
            else:
                return self.getToken(FandangoParser.COMMA, i)

        def getRuleIndex(self):
            return FandangoParser.RULE_expressions

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterExpressions"):
                listener.enterExpressions(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitExpressions"):
                listener.exitExpressions(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitExpressions"):
                return visitor.visitExpressions(self)
            else:
                return visitor.visitChildren(self)

    def expressions(self):

        localctx = FandangoParser.ExpressionsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 266, self.RULE_expressions)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1815
            self.expression()
            self.state = 1820
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 204, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 1816
                    self.match(FandangoParser.COMMA)
                    self.state = 1817
                    self.expression()
                self.state = 1822
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 204, self._ctx)

            self.state = 1824
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == 70:
                self.state = 1823
                self.match(FandangoParser.COMMA)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ExpressionContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def disjunction(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(FandangoParser.DisjunctionContext)
            else:
                return self.getTypedRuleContext(FandangoParser.DisjunctionContext, i)

        def IF(self):
            return self.getToken(FandangoParser.IF, 0)

        def ELSE(self):
            return self.getToken(FandangoParser.ELSE, 0)

        def expression(self):
            return self.getTypedRuleContext(FandangoParser.ExpressionContext, 0)

        def lambdef(self):
            return self.getTypedRuleContext(FandangoParser.LambdefContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_expression

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterExpression"):
                listener.enterExpression(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitExpression"):
                listener.exitExpression(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitExpression"):
                return visitor.visitExpression(self)
            else:
                return visitor.visitChildren(self)

    def expression(self):

        localctx = FandangoParser.ExpressionContext(self, self._ctx, self.state)
        self.enterRule(localctx, 268, self.RULE_expression)
        try:
            self.state = 1834
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 206, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 1826
                self.disjunction()
                self.state = 1827
                self.match(FandangoParser.IF)
                self.state = 1828
                self.disjunction()
                self.state = 1829
                self.match(FandangoParser.ELSE)
                self.state = 1830
                self.expression()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 1832
                self.disjunction()
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 1833
                self.lambdef()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Yield_exprContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def YIELD(self):
            return self.getToken(FandangoParser.YIELD, 0)

        def FROM(self):
            return self.getToken(FandangoParser.FROM, 0)

        def expression(self):
            return self.getTypedRuleContext(FandangoParser.ExpressionContext, 0)

        def star_expressions(self):
            return self.getTypedRuleContext(FandangoParser.Star_expressionsContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_yield_expr

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterYield_expr"):
                listener.enterYield_expr(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitYield_expr"):
                listener.exitYield_expr(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitYield_expr"):
                return visitor.visitYield_expr(self)
            else:
                return visitor.visitChildren(self)

    def yield_expr(self):

        localctx = FandangoParser.Yield_exprContext(self, self._ctx, self.state)
        self.enterRule(localctx, 270, self.RULE_yield_expr)
        try:
            self.state = 1843
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 208, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 1836
                self.match(FandangoParser.YIELD)
                self.state = 1837
                self.match(FandangoParser.FROM)
                self.state = 1838
                self.expression()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 1839
                self.match(FandangoParser.YIELD)
                self.state = 1841
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input, 207, self._ctx)
                if la_ == 1:
                    self.state = 1840
                    self.star_expressions()

                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Star_expressionsContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def star_expression(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(FandangoParser.Star_expressionContext)
            else:
                return self.getTypedRuleContext(
                    FandangoParser.Star_expressionContext, i
                )

        def COMMA(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.COMMA)
            else:
                return self.getToken(FandangoParser.COMMA, i)

        def getRuleIndex(self):
            return FandangoParser.RULE_star_expressions

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterStar_expressions"):
                listener.enterStar_expressions(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitStar_expressions"):
                listener.exitStar_expressions(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitStar_expressions"):
                return visitor.visitStar_expressions(self)
            else:
                return visitor.visitChildren(self)

    def star_expressions(self):

        localctx = FandangoParser.Star_expressionsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 272, self.RULE_star_expressions)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1845
            self.star_expression()
            self.state = 1850
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 209, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 1846
                    self.match(FandangoParser.COMMA)
                    self.state = 1847
                    self.star_expression()
                self.state = 1852
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 209, self._ctx)

            self.state = 1854
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == 70:
                self.state = 1853
                self.match(FandangoParser.COMMA)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Star_expressionContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def STAR(self):
            return self.getToken(FandangoParser.STAR, 0)

        def bitwise_or(self):
            return self.getTypedRuleContext(FandangoParser.Bitwise_orContext, 0)

        def expression(self):
            return self.getTypedRuleContext(FandangoParser.ExpressionContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_star_expression

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterStar_expression"):
                listener.enterStar_expression(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitStar_expression"):
                listener.exitStar_expression(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitStar_expression"):
                return visitor.visitStar_expression(self)
            else:
                return visitor.visitChildren(self)

    def star_expression(self):

        localctx = FandangoParser.Star_expressionContext(self, self._ctx, self.state)
        self.enterRule(localctx, 274, self.RULE_star_expression)
        try:
            self.state = 1859
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [67]:
                self.enterOuterAlt(localctx, 1)
                self.state = 1856
                self.match(FandangoParser.STAR)
                self.state = 1857
                self.bitwise_or(0)
                pass
            elif token in [
                3,
                4,
                8,
                13,
                23,
                32,
                34,
                36,
                41,
                52,
                66,
                68,
                75,
                77,
                82,
                83,
                87,
                88,
                118,
            ]:
                self.enterOuterAlt(localctx, 2)
                self.state = 1858
                self.expression()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Star_named_expressionsContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def star_named_expression(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(
                    FandangoParser.Star_named_expressionContext
                )
            else:
                return self.getTypedRuleContext(
                    FandangoParser.Star_named_expressionContext, i
                )

        def COMMA(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.COMMA)
            else:
                return self.getToken(FandangoParser.COMMA, i)

        def getRuleIndex(self):
            return FandangoParser.RULE_star_named_expressions

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterStar_named_expressions"):
                listener.enterStar_named_expressions(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitStar_named_expressions"):
                listener.exitStar_named_expressions(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitStar_named_expressions"):
                return visitor.visitStar_named_expressions(self)
            else:
                return visitor.visitChildren(self)

    def star_named_expressions(self):

        localctx = FandangoParser.Star_named_expressionsContext(
            self, self._ctx, self.state
        )
        self.enterRule(localctx, 276, self.RULE_star_named_expressions)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1861
            self.star_named_expression()
            self.state = 1866
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 212, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 1862
                    self.match(FandangoParser.COMMA)
                    self.state = 1863
                    self.star_named_expression()
                self.state = 1868
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 212, self._ctx)

            self.state = 1870
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == 70:
                self.state = 1869
                self.match(FandangoParser.COMMA)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Star_named_expressionContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def STAR(self):
            return self.getToken(FandangoParser.STAR, 0)

        def bitwise_or(self):
            return self.getTypedRuleContext(FandangoParser.Bitwise_orContext, 0)

        def named_expression(self):
            return self.getTypedRuleContext(FandangoParser.Named_expressionContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_star_named_expression

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterStar_named_expression"):
                listener.enterStar_named_expression(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitStar_named_expression"):
                listener.exitStar_named_expression(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitStar_named_expression"):
                return visitor.visitStar_named_expression(self)
            else:
                return visitor.visitChildren(self)

    def star_named_expression(self):

        localctx = FandangoParser.Star_named_expressionContext(
            self, self._ctx, self.state
        )
        self.enterRule(localctx, 278, self.RULE_star_named_expression)
        try:
            self.state = 1875
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [67]:
                self.enterOuterAlt(localctx, 1)
                self.state = 1872
                self.match(FandangoParser.STAR)
                self.state = 1873
                self.bitwise_or(0)
                pass
            elif token in [
                3,
                4,
                8,
                13,
                23,
                32,
                34,
                36,
                41,
                52,
                66,
                68,
                75,
                77,
                82,
                83,
                87,
                88,
                118,
            ]:
                self.enterOuterAlt(localctx, 2)
                self.state = 1874
                self.named_expression()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Assignment_expressionContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NAME(self):
            return self.getToken(FandangoParser.NAME, 0)

        def EXPR_ASSIGN(self):
            return self.getToken(FandangoParser.EXPR_ASSIGN, 0)

        def expression(self):
            return self.getTypedRuleContext(FandangoParser.ExpressionContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_assignment_expression

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterAssignment_expression"):
                listener.enterAssignment_expression(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitAssignment_expression"):
                listener.exitAssignment_expression(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitAssignment_expression"):
                return visitor.visitAssignment_expression(self)
            else:
                return visitor.visitChildren(self)

    def assignment_expression(self):

        localctx = FandangoParser.Assignment_expressionContext(
            self, self._ctx, self.state
        )
        self.enterRule(localctx, 280, self.RULE_assignment_expression)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1877
            self.match(FandangoParser.NAME)
            self.state = 1878
            self.match(FandangoParser.EXPR_ASSIGN)
            self.state = 1879
            self.expression()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Named_expressionContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def assignment_expression(self):
            return self.getTypedRuleContext(
                FandangoParser.Assignment_expressionContext, 0
            )

        def expression(self):
            return self.getTypedRuleContext(FandangoParser.ExpressionContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_named_expression

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterNamed_expression"):
                listener.enterNamed_expression(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitNamed_expression"):
                listener.exitNamed_expression(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitNamed_expression"):
                return visitor.visitNamed_expression(self)
            else:
                return visitor.visitChildren(self)

    def named_expression(self):

        localctx = FandangoParser.Named_expressionContext(self, self._ctx, self.state)
        self.enterRule(localctx, 282, self.RULE_named_expression)
        try:
            self.state = 1883
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 215, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 1881
                self.assignment_expression()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 1882
                self.expression()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class DisjunctionContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def conjunction(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(FandangoParser.ConjunctionContext)
            else:
                return self.getTypedRuleContext(FandangoParser.ConjunctionContext, i)

        def OR(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.OR)
            else:
                return self.getToken(FandangoParser.OR, i)

        def getRuleIndex(self):
            return FandangoParser.RULE_disjunction

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterDisjunction"):
                listener.enterDisjunction(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitDisjunction"):
                listener.exitDisjunction(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitDisjunction"):
                return visitor.visitDisjunction(self)
            else:
                return visitor.visitChildren(self)

    def disjunction(self):

        localctx = FandangoParser.DisjunctionContext(self, self._ctx, self.state)
        self.enterRule(localctx, 284, self.RULE_disjunction)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1885
            self.conjunction()
            self.state = 1890
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la == 37:
                self.state = 1886
                self.match(FandangoParser.OR)
                self.state = 1887
                self.conjunction()
                self.state = 1892
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ConjunctionContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def inversion(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(FandangoParser.InversionContext)
            else:
                return self.getTypedRuleContext(FandangoParser.InversionContext, i)

        def AND(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.AND)
            else:
                return self.getToken(FandangoParser.AND, i)

        def getRuleIndex(self):
            return FandangoParser.RULE_conjunction

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterConjunction"):
                listener.enterConjunction(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitConjunction"):
                listener.exitConjunction(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitConjunction"):
                return visitor.visitConjunction(self)
            else:
                return visitor.visitChildren(self)

    def conjunction(self):

        localctx = FandangoParser.ConjunctionContext(self, self._ctx, self.state)
        self.enterRule(localctx, 286, self.RULE_conjunction)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1893
            self.inversion()
            self.state = 1898
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la == 9:
                self.state = 1894
                self.match(FandangoParser.AND)
                self.state = 1895
                self.inversion()
                self.state = 1900
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class InversionContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NOT(self):
            return self.getToken(FandangoParser.NOT, 0)

        def inversion(self):
            return self.getTypedRuleContext(FandangoParser.InversionContext, 0)

        def comparison(self):
            return self.getTypedRuleContext(FandangoParser.ComparisonContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_inversion

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterInversion"):
                listener.enterInversion(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitInversion"):
                listener.exitInversion(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitInversion"):
                return visitor.visitInversion(self)
            else:
                return visitor.visitChildren(self)

    def inversion(self):

        localctx = FandangoParser.InversionContext(self, self._ctx, self.state)
        self.enterRule(localctx, 288, self.RULE_inversion)
        try:
            self.state = 1904
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [36]:
                self.enterOuterAlt(localctx, 1)
                self.state = 1901
                self.match(FandangoParser.NOT)
                self.state = 1902
                self.inversion()
                pass
            elif token in [
                3,
                4,
                8,
                13,
                23,
                34,
                41,
                52,
                66,
                68,
                75,
                77,
                82,
                83,
                87,
                88,
                118,
            ]:
                self.enterOuterAlt(localctx, 2)
                self.state = 1903
                self.comparison()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ComparisonContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def bitwise_or(self):
            return self.getTypedRuleContext(FandangoParser.Bitwise_orContext, 0)

        def compare_op_bitwise_or_pair(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(
                    FandangoParser.Compare_op_bitwise_or_pairContext
                )
            else:
                return self.getTypedRuleContext(
                    FandangoParser.Compare_op_bitwise_or_pairContext, i
                )

        def getRuleIndex(self):
            return FandangoParser.RULE_comparison

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterComparison"):
                listener.enterComparison(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitComparison"):
                listener.exitComparison(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitComparison"):
                return visitor.visitComparison(self)
            else:
                return visitor.visitChildren(self)

    def comparison(self):

        localctx = FandangoParser.ComparisonContext(self, self._ctx, self.state)
        self.enterRule(localctx, 290, self.RULE_comparison)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1906
            self.bitwise_or(0)
            self.state = 1910
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 219, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 1907
                    self.compare_op_bitwise_or_pair()
                self.state = 1912
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 219, self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Compare_op_bitwise_or_pairContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def eq_bitwise_or(self):
            return self.getTypedRuleContext(FandangoParser.Eq_bitwise_orContext, 0)

        def noteq_bitwise_or(self):
            return self.getTypedRuleContext(FandangoParser.Noteq_bitwise_orContext, 0)

        def lte_bitwise_or(self):
            return self.getTypedRuleContext(FandangoParser.Lte_bitwise_orContext, 0)

        def lt_bitwise_or(self):
            return self.getTypedRuleContext(FandangoParser.Lt_bitwise_orContext, 0)

        def gte_bitwise_or(self):
            return self.getTypedRuleContext(FandangoParser.Gte_bitwise_orContext, 0)

        def gt_bitwise_or(self):
            return self.getTypedRuleContext(FandangoParser.Gt_bitwise_orContext, 0)

        def notin_bitwise_or(self):
            return self.getTypedRuleContext(FandangoParser.Notin_bitwise_orContext, 0)

        def in_bitwise_or(self):
            return self.getTypedRuleContext(FandangoParser.In_bitwise_orContext, 0)

        def isnot_bitwise_or(self):
            return self.getTypedRuleContext(FandangoParser.Isnot_bitwise_orContext, 0)

        def is_bitwise_or(self):
            return self.getTypedRuleContext(FandangoParser.Is_bitwise_orContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_compare_op_bitwise_or_pair

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterCompare_op_bitwise_or_pair"):
                listener.enterCompare_op_bitwise_or_pair(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitCompare_op_bitwise_or_pair"):
                listener.exitCompare_op_bitwise_or_pair(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitCompare_op_bitwise_or_pair"):
                return visitor.visitCompare_op_bitwise_or_pair(self)
            else:
                return visitor.visitChildren(self)

    def compare_op_bitwise_or_pair(self):

        localctx = FandangoParser.Compare_op_bitwise_or_pairContext(
            self, self._ctx, self.state
        )
        self.enterRule(localctx, 292, self.RULE_compare_op_bitwise_or_pair)
        try:
            self.state = 1923
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 220, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 1913
                self.eq_bitwise_or()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 1914
                self.noteq_bitwise_or()
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 1915
                self.lte_bitwise_or()
                pass

            elif la_ == 4:
                self.enterOuterAlt(localctx, 4)
                self.state = 1916
                self.lt_bitwise_or()
                pass

            elif la_ == 5:
                self.enterOuterAlt(localctx, 5)
                self.state = 1917
                self.gte_bitwise_or()
                pass

            elif la_ == 6:
                self.enterOuterAlt(localctx, 6)
                self.state = 1918
                self.gt_bitwise_or()
                pass

            elif la_ == 7:
                self.enterOuterAlt(localctx, 7)
                self.state = 1919
                self.notin_bitwise_or()
                pass

            elif la_ == 8:
                self.enterOuterAlt(localctx, 8)
                self.state = 1920
                self.in_bitwise_or()
                pass

            elif la_ == 9:
                self.enterOuterAlt(localctx, 9)
                self.state = 1921
                self.isnot_bitwise_or()
                pass

            elif la_ == 10:
                self.enterOuterAlt(localctx, 10)
                self.state = 1922
                self.is_bitwise_or()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Eq_bitwise_orContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def EQUALS(self):
            return self.getToken(FandangoParser.EQUALS, 0)

        def bitwise_or(self):
            return self.getTypedRuleContext(FandangoParser.Bitwise_orContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_eq_bitwise_or

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterEq_bitwise_or"):
                listener.enterEq_bitwise_or(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitEq_bitwise_or"):
                listener.exitEq_bitwise_or(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitEq_bitwise_or"):
                return visitor.visitEq_bitwise_or(self)
            else:
                return visitor.visitChildren(self)

    def eq_bitwise_or(self):

        localctx = FandangoParser.Eq_bitwise_orContext(self, self._ctx, self.state)
        self.enterRule(localctx, 294, self.RULE_eq_bitwise_or)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1925
            self.match(FandangoParser.EQUALS)
            self.state = 1926
            self.bitwise_or(0)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Noteq_bitwise_orContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NOT_EQ_2(self):
            return self.getToken(FandangoParser.NOT_EQ_2, 0)

        def bitwise_or(self):
            return self.getTypedRuleContext(FandangoParser.Bitwise_orContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_noteq_bitwise_or

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterNoteq_bitwise_or"):
                listener.enterNoteq_bitwise_or(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitNoteq_bitwise_or"):
                listener.exitNoteq_bitwise_or(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitNoteq_bitwise_or"):
                return visitor.visitNoteq_bitwise_or(self)
            else:
                return visitor.visitChildren(self)

    def noteq_bitwise_or(self):

        localctx = FandangoParser.Noteq_bitwise_orContext(self, self._ctx, self.state)
        self.enterRule(localctx, 296, self.RULE_noteq_bitwise_or)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1928
            self.match(FandangoParser.NOT_EQ_2)
            self.state = 1929
            self.bitwise_or(0)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Lte_bitwise_orContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def LT_EQ(self):
            return self.getToken(FandangoParser.LT_EQ, 0)

        def bitwise_or(self):
            return self.getTypedRuleContext(FandangoParser.Bitwise_orContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_lte_bitwise_or

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterLte_bitwise_or"):
                listener.enterLte_bitwise_or(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitLte_bitwise_or"):
                listener.exitLte_bitwise_or(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitLte_bitwise_or"):
                return visitor.visitLte_bitwise_or(self)
            else:
                return visitor.visitChildren(self)

    def lte_bitwise_or(self):

        localctx = FandangoParser.Lte_bitwise_orContext(self, self._ctx, self.state)
        self.enterRule(localctx, 298, self.RULE_lte_bitwise_or)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1931
            self.match(FandangoParser.LT_EQ)
            self.state = 1932
            self.bitwise_or(0)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Lt_bitwise_orContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def LESS_THAN(self):
            return self.getToken(FandangoParser.LESS_THAN, 0)

        def bitwise_or(self):
            return self.getTypedRuleContext(FandangoParser.Bitwise_orContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_lt_bitwise_or

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterLt_bitwise_or"):
                listener.enterLt_bitwise_or(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitLt_bitwise_or"):
                listener.exitLt_bitwise_or(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitLt_bitwise_or"):
                return visitor.visitLt_bitwise_or(self)
            else:
                return visitor.visitChildren(self)

    def lt_bitwise_or(self):

        localctx = FandangoParser.Lt_bitwise_orContext(self, self._ctx, self.state)
        self.enterRule(localctx, 300, self.RULE_lt_bitwise_or)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1934
            self.match(FandangoParser.LESS_THAN)
            self.state = 1935
            self.bitwise_or(0)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Gte_bitwise_orContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def GT_EQ(self):
            return self.getToken(FandangoParser.GT_EQ, 0)

        def bitwise_or(self):
            return self.getTypedRuleContext(FandangoParser.Bitwise_orContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_gte_bitwise_or

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterGte_bitwise_or"):
                listener.enterGte_bitwise_or(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitGte_bitwise_or"):
                listener.exitGte_bitwise_or(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitGte_bitwise_or"):
                return visitor.visitGte_bitwise_or(self)
            else:
                return visitor.visitChildren(self)

    def gte_bitwise_or(self):

        localctx = FandangoParser.Gte_bitwise_orContext(self, self._ctx, self.state)
        self.enterRule(localctx, 302, self.RULE_gte_bitwise_or)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1937
            self.match(FandangoParser.GT_EQ)
            self.state = 1938
            self.bitwise_or(0)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Gt_bitwise_orContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def GREATER_THAN(self):
            return self.getToken(FandangoParser.GREATER_THAN, 0)

        def bitwise_or(self):
            return self.getTypedRuleContext(FandangoParser.Bitwise_orContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_gt_bitwise_or

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterGt_bitwise_or"):
                listener.enterGt_bitwise_or(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitGt_bitwise_or"):
                listener.exitGt_bitwise_or(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitGt_bitwise_or"):
                return visitor.visitGt_bitwise_or(self)
            else:
                return visitor.visitChildren(self)

    def gt_bitwise_or(self):

        localctx = FandangoParser.Gt_bitwise_orContext(self, self._ctx, self.state)
        self.enterRule(localctx, 304, self.RULE_gt_bitwise_or)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1940
            self.match(FandangoParser.GREATER_THAN)
            self.state = 1941
            self.bitwise_or(0)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Notin_bitwise_orContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NOT(self):
            return self.getToken(FandangoParser.NOT, 0)

        def IN(self):
            return self.getToken(FandangoParser.IN, 0)

        def bitwise_or(self):
            return self.getTypedRuleContext(FandangoParser.Bitwise_orContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_notin_bitwise_or

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterNotin_bitwise_or"):
                listener.enterNotin_bitwise_or(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitNotin_bitwise_or"):
                listener.exitNotin_bitwise_or(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitNotin_bitwise_or"):
                return visitor.visitNotin_bitwise_or(self)
            else:
                return visitor.visitChildren(self)

    def notin_bitwise_or(self):

        localctx = FandangoParser.Notin_bitwise_orContext(self, self._ctx, self.state)
        self.enterRule(localctx, 306, self.RULE_notin_bitwise_or)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1943
            self.match(FandangoParser.NOT)
            self.state = 1944
            self.match(FandangoParser.IN)
            self.state = 1945
            self.bitwise_or(0)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class In_bitwise_orContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def IN(self):
            return self.getToken(FandangoParser.IN, 0)

        def bitwise_or(self):
            return self.getTypedRuleContext(FandangoParser.Bitwise_orContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_in_bitwise_or

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterIn_bitwise_or"):
                listener.enterIn_bitwise_or(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitIn_bitwise_or"):
                listener.exitIn_bitwise_or(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitIn_bitwise_or"):
                return visitor.visitIn_bitwise_or(self)
            else:
                return visitor.visitChildren(self)

    def in_bitwise_or(self):

        localctx = FandangoParser.In_bitwise_orContext(self, self._ctx, self.state)
        self.enterRule(localctx, 308, self.RULE_in_bitwise_or)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1947
            self.match(FandangoParser.IN)
            self.state = 1948
            self.bitwise_or(0)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Isnot_bitwise_orContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def IS(self):
            return self.getToken(FandangoParser.IS, 0)

        def NOT(self):
            return self.getToken(FandangoParser.NOT, 0)

        def bitwise_or(self):
            return self.getTypedRuleContext(FandangoParser.Bitwise_orContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_isnot_bitwise_or

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterIsnot_bitwise_or"):
                listener.enterIsnot_bitwise_or(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitIsnot_bitwise_or"):
                listener.exitIsnot_bitwise_or(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitIsnot_bitwise_or"):
                return visitor.visitIsnot_bitwise_or(self)
            else:
                return visitor.visitChildren(self)

    def isnot_bitwise_or(self):

        localctx = FandangoParser.Isnot_bitwise_orContext(self, self._ctx, self.state)
        self.enterRule(localctx, 310, self.RULE_isnot_bitwise_or)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1950
            self.match(FandangoParser.IS)
            self.state = 1951
            self.match(FandangoParser.NOT)
            self.state = 1952
            self.bitwise_or(0)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Is_bitwise_orContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def IS(self):
            return self.getToken(FandangoParser.IS, 0)

        def bitwise_or(self):
            return self.getTypedRuleContext(FandangoParser.Bitwise_orContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_is_bitwise_or

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterIs_bitwise_or"):
                listener.enterIs_bitwise_or(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitIs_bitwise_or"):
                listener.exitIs_bitwise_or(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitIs_bitwise_or"):
                return visitor.visitIs_bitwise_or(self)
            else:
                return visitor.visitChildren(self)

    def is_bitwise_or(self):

        localctx = FandangoParser.Is_bitwise_orContext(self, self._ctx, self.state)
        self.enterRule(localctx, 312, self.RULE_is_bitwise_or)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1954
            self.match(FandangoParser.IS)
            self.state = 1955
            self.bitwise_or(0)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Bitwise_orContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def bitwise_xor(self):
            return self.getTypedRuleContext(FandangoParser.Bitwise_xorContext, 0)

        def bitwise_or(self):
            return self.getTypedRuleContext(FandangoParser.Bitwise_orContext, 0)

        def OR_OP(self):
            return self.getToken(FandangoParser.OR_OP, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_bitwise_or

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterBitwise_or"):
                listener.enterBitwise_or(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitBitwise_or"):
                listener.exitBitwise_or(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitBitwise_or"):
                return visitor.visitBitwise_or(self)
            else:
                return visitor.visitChildren(self)

    def bitwise_or(self, _p: int = 0):
        _parentctx = self._ctx
        _parentState = self.state
        localctx = FandangoParser.Bitwise_orContext(self, self._ctx, _parentState)
        _prevctx = localctx
        _startState = 314
        self.enterRecursionRule(localctx, 314, self.RULE_bitwise_or, _p)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1958
            self.bitwise_xor(0)
            self._ctx.stop = self._input.LT(-1)
            self.state = 1965
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 221, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    if self._parseListeners is not None:
                        self.triggerExitRuleEvent()
                    _prevctx = localctx
                    localctx = FandangoParser.Bitwise_orContext(
                        self, _parentctx, _parentState
                    )
                    self.pushNewRecursionContext(
                        localctx, _startState, self.RULE_bitwise_or
                    )
                    self.state = 1960
                    if not self.precpred(self._ctx, 2):
                        from antlr4.error.Errors import FailedPredicateException

                        raise FailedPredicateException(
                            self, "self.precpred(self._ctx, 2)"
                        )
                    self.state = 1961
                    self.match(FandangoParser.OR_OP)
                    self.state = 1962
                    self.bitwise_xor(0)
                self.state = 1967
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 221, self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.unrollRecursionContexts(_parentctx)
        return localctx

    class Bitwise_xorContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def bitwise_and(self):
            return self.getTypedRuleContext(FandangoParser.Bitwise_andContext, 0)

        def bitwise_xor(self):
            return self.getTypedRuleContext(FandangoParser.Bitwise_xorContext, 0)

        def XOR(self):
            return self.getToken(FandangoParser.XOR, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_bitwise_xor

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterBitwise_xor"):
                listener.enterBitwise_xor(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitBitwise_xor"):
                listener.exitBitwise_xor(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitBitwise_xor"):
                return visitor.visitBitwise_xor(self)
            else:
                return visitor.visitChildren(self)

    def bitwise_xor(self, _p: int = 0):
        _parentctx = self._ctx
        _parentState = self.state
        localctx = FandangoParser.Bitwise_xorContext(self, self._ctx, _parentState)
        _prevctx = localctx
        _startState = 316
        self.enterRecursionRule(localctx, 316, self.RULE_bitwise_xor, _p)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1969
            self.bitwise_and(0)
            self._ctx.stop = self._input.LT(-1)
            self.state = 1976
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 222, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    if self._parseListeners is not None:
                        self.triggerExitRuleEvent()
                    _prevctx = localctx
                    localctx = FandangoParser.Bitwise_xorContext(
                        self, _parentctx, _parentState
                    )
                    self.pushNewRecursionContext(
                        localctx, _startState, self.RULE_bitwise_xor
                    )
                    self.state = 1971
                    if not self.precpred(self._ctx, 2):
                        from antlr4.error.Errors import FailedPredicateException

                        raise FailedPredicateException(
                            self, "self.precpred(self._ctx, 2)"
                        )
                    self.state = 1972
                    self.match(FandangoParser.XOR)
                    self.state = 1973
                    self.bitwise_and(0)
                self.state = 1978
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 222, self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.unrollRecursionContexts(_parentctx)
        return localctx

    class Bitwise_andContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def shift_expr(self):
            return self.getTypedRuleContext(FandangoParser.Shift_exprContext, 0)

        def bitwise_and(self):
            return self.getTypedRuleContext(FandangoParser.Bitwise_andContext, 0)

        def AND_OP(self):
            return self.getToken(FandangoParser.AND_OP, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_bitwise_and

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterBitwise_and"):
                listener.enterBitwise_and(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitBitwise_and"):
                listener.exitBitwise_and(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitBitwise_and"):
                return visitor.visitBitwise_and(self)
            else:
                return visitor.visitChildren(self)

    def bitwise_and(self, _p: int = 0):
        _parentctx = self._ctx
        _parentState = self.state
        localctx = FandangoParser.Bitwise_andContext(self, self._ctx, _parentState)
        _prevctx = localctx
        _startState = 318
        self.enterRecursionRule(localctx, 318, self.RULE_bitwise_and, _p)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1980
            self.shift_expr(0)
            self._ctx.stop = self._input.LT(-1)
            self.state = 1987
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 223, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    if self._parseListeners is not None:
                        self.triggerExitRuleEvent()
                    _prevctx = localctx
                    localctx = FandangoParser.Bitwise_andContext(
                        self, _parentctx, _parentState
                    )
                    self.pushNewRecursionContext(
                        localctx, _startState, self.RULE_bitwise_and
                    )
                    self.state = 1982
                    if not self.precpred(self._ctx, 2):
                        from antlr4.error.Errors import FailedPredicateException

                        raise FailedPredicateException(
                            self, "self.precpred(self._ctx, 2)"
                        )
                    self.state = 1983
                    self.match(FandangoParser.AND_OP)
                    self.state = 1984
                    self.shift_expr(0)
                self.state = 1989
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 223, self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.unrollRecursionContexts(_parentctx)
        return localctx

    class Shift_exprContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def sum_(self):
            return self.getTypedRuleContext(FandangoParser.SumContext, 0)

        def shift_expr(self):
            return self.getTypedRuleContext(FandangoParser.Shift_exprContext, 0)

        def LEFT_SHIFT(self):
            return self.getToken(FandangoParser.LEFT_SHIFT, 0)

        def RIGHT_SHIFT(self):
            return self.getToken(FandangoParser.RIGHT_SHIFT, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_shift_expr

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterShift_expr"):
                listener.enterShift_expr(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitShift_expr"):
                listener.exitShift_expr(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitShift_expr"):
                return visitor.visitShift_expr(self)
            else:
                return visitor.visitChildren(self)

    def shift_expr(self, _p: int = 0):
        _parentctx = self._ctx
        _parentState = self.state
        localctx = FandangoParser.Shift_exprContext(self, self._ctx, _parentState)
        _prevctx = localctx
        _startState = 320
        self.enterRecursionRule(localctx, 320, self.RULE_shift_expr, _p)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1991
            self.sum_(0)
            self._ctx.stop = self._input.LT(-1)
            self.state = 2001
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 225, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    if self._parseListeners is not None:
                        self.triggerExitRuleEvent()
                    _prevctx = localctx
                    self.state = 1999
                    self._errHandler.sync(self)
                    la_ = self._interp.adaptivePredict(self._input, 224, self._ctx)
                    if la_ == 1:
                        localctx = FandangoParser.Shift_exprContext(
                            self, _parentctx, _parentState
                        )
                        self.pushNewRecursionContext(
                            localctx, _startState, self.RULE_shift_expr
                        )
                        self.state = 1993
                        if not self.precpred(self._ctx, 3):
                            from antlr4.error.Errors import FailedPredicateException

                            raise FailedPredicateException(
                                self, "self.precpred(self._ctx, 3)"
                            )
                        self.state = 1994
                        self.match(FandangoParser.LEFT_SHIFT)
                        self.state = 1995
                        self.sum_(0)
                        pass

                    elif la_ == 2:
                        localctx = FandangoParser.Shift_exprContext(
                            self, _parentctx, _parentState
                        )
                        self.pushNewRecursionContext(
                            localctx, _startState, self.RULE_shift_expr
                        )
                        self.state = 1996
                        if not self.precpred(self._ctx, 2):
                            from antlr4.error.Errors import FailedPredicateException

                            raise FailedPredicateException(
                                self, "self.precpred(self._ctx, 2)"
                            )
                        self.state = 1997
                        self.match(FandangoParser.RIGHT_SHIFT)
                        self.state = 1998
                        self.sum_(0)
                        pass

                self.state = 2003
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 225, self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.unrollRecursionContexts(_parentctx)
        return localctx

    class SumContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def term(self):
            return self.getTypedRuleContext(FandangoParser.TermContext, 0)

        def sum_(self):
            return self.getTypedRuleContext(FandangoParser.SumContext, 0)

        def ADD(self):
            return self.getToken(FandangoParser.ADD, 0)

        def MINUS(self):
            return self.getToken(FandangoParser.MINUS, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_sum

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterSum"):
                listener.enterSum(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitSum"):
                listener.exitSum(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitSum"):
                return visitor.visitSum(self)
            else:
                return visitor.visitChildren(self)

    def sum_(self, _p: int = 0):
        _parentctx = self._ctx
        _parentState = self.state
        localctx = FandangoParser.SumContext(self, self._ctx, _parentState)
        _prevctx = localctx
        _startState = 322
        self.enterRecursionRule(localctx, 322, self.RULE_sum, _p)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 2005
            self.term(0)
            self._ctx.stop = self._input.LT(-1)
            self.state = 2015
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 227, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    if self._parseListeners is not None:
                        self.triggerExitRuleEvent()
                    _prevctx = localctx
                    self.state = 2013
                    self._errHandler.sync(self)
                    la_ = self._interp.adaptivePredict(self._input, 226, self._ctx)
                    if la_ == 1:
                        localctx = FandangoParser.SumContext(
                            self, _parentctx, _parentState
                        )
                        self.pushNewRecursionContext(
                            localctx, _startState, self.RULE_sum
                        )
                        self.state = 2007
                        if not self.precpred(self._ctx, 3):
                            from antlr4.error.Errors import FailedPredicateException

                            raise FailedPredicateException(
                                self, "self.precpred(self._ctx, 3)"
                            )
                        self.state = 2008
                        self.match(FandangoParser.ADD)
                        self.state = 2009
                        self.term(0)
                        pass

                    elif la_ == 2:
                        localctx = FandangoParser.SumContext(
                            self, _parentctx, _parentState
                        )
                        self.pushNewRecursionContext(
                            localctx, _startState, self.RULE_sum
                        )
                        self.state = 2010
                        if not self.precpred(self._ctx, 2):
                            from antlr4.error.Errors import FailedPredicateException

                            raise FailedPredicateException(
                                self, "self.precpred(self._ctx, 2)"
                            )
                        self.state = 2011
                        self.match(FandangoParser.MINUS)
                        self.state = 2012
                        self.term(0)
                        pass

                self.state = 2017
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 227, self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.unrollRecursionContexts(_parentctx)
        return localctx

    class TermContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def factor(self):
            return self.getTypedRuleContext(FandangoParser.FactorContext, 0)

        def term(self):
            return self.getTypedRuleContext(FandangoParser.TermContext, 0)

        def STAR(self):
            return self.getToken(FandangoParser.STAR, 0)

        def DIV(self):
            return self.getToken(FandangoParser.DIV, 0)

        def IDIV(self):
            return self.getToken(FandangoParser.IDIV, 0)

        def MOD(self):
            return self.getToken(FandangoParser.MOD, 0)

        def AT(self):
            return self.getToken(FandangoParser.AT, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_term

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterTerm"):
                listener.enterTerm(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitTerm"):
                listener.exitTerm(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitTerm"):
                return visitor.visitTerm(self)
            else:
                return visitor.visitChildren(self)

    def term(self, _p: int = 0):
        _parentctx = self._ctx
        _parentState = self.state
        localctx = FandangoParser.TermContext(self, self._ctx, _parentState)
        _prevctx = localctx
        _startState = 324
        self.enterRecursionRule(localctx, 324, self.RULE_term, _p)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 2019
            self.factor()
            self._ctx.stop = self._input.LT(-1)
            self.state = 2038
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 229, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    if self._parseListeners is not None:
                        self.triggerExitRuleEvent()
                    _prevctx = localctx
                    self.state = 2036
                    self._errHandler.sync(self)
                    la_ = self._interp.adaptivePredict(self._input, 228, self._ctx)
                    if la_ == 1:
                        localctx = FandangoParser.TermContext(
                            self, _parentctx, _parentState
                        )
                        self.pushNewRecursionContext(
                            localctx, _startState, self.RULE_term
                        )
                        self.state = 2021
                        if not self.precpred(self._ctx, 6):
                            from antlr4.error.Errors import FailedPredicateException

                            raise FailedPredicateException(
                                self, "self.precpred(self._ctx, 6)"
                            )
                        self.state = 2022
                        self.match(FandangoParser.STAR)
                        self.state = 2023
                        self.factor()
                        pass

                    elif la_ == 2:
                        localctx = FandangoParser.TermContext(
                            self, _parentctx, _parentState
                        )
                        self.pushNewRecursionContext(
                            localctx, _startState, self.RULE_term
                        )
                        self.state = 2024
                        if not self.precpred(self._ctx, 5):
                            from antlr4.error.Errors import FailedPredicateException

                            raise FailedPredicateException(
                                self, "self.precpred(self._ctx, 5)"
                            )
                        self.state = 2025
                        self.match(FandangoParser.DIV)
                        self.state = 2026
                        self.factor()
                        pass

                    elif la_ == 3:
                        localctx = FandangoParser.TermContext(
                            self, _parentctx, _parentState
                        )
                        self.pushNewRecursionContext(
                            localctx, _startState, self.RULE_term
                        )
                        self.state = 2027
                        if not self.precpred(self._ctx, 4):
                            from antlr4.error.Errors import FailedPredicateException

                            raise FailedPredicateException(
                                self, "self.precpred(self._ctx, 4)"
                            )
                        self.state = 2028
                        self.match(FandangoParser.IDIV)
                        self.state = 2029
                        self.factor()
                        pass

                    elif la_ == 4:
                        localctx = FandangoParser.TermContext(
                            self, _parentctx, _parentState
                        )
                        self.pushNewRecursionContext(
                            localctx, _startState, self.RULE_term
                        )
                        self.state = 2030
                        if not self.precpred(self._ctx, 3):
                            from antlr4.error.Errors import FailedPredicateException

                            raise FailedPredicateException(
                                self, "self.precpred(self._ctx, 3)"
                            )
                        self.state = 2031
                        self.match(FandangoParser.MOD)
                        self.state = 2032
                        self.factor()
                        pass

                    elif la_ == 5:
                        localctx = FandangoParser.TermContext(
                            self, _parentctx, _parentState
                        )
                        self.pushNewRecursionContext(
                            localctx, _startState, self.RULE_term
                        )
                        self.state = 2033
                        if not self.precpred(self._ctx, 2):
                            from antlr4.error.Errors import FailedPredicateException

                            raise FailedPredicateException(
                                self, "self.precpred(self._ctx, 2)"
                            )
                        self.state = 2034
                        self.match(FandangoParser.AT)
                        self.state = 2035
                        self.factor()
                        pass

                self.state = 2040
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 229, self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.unrollRecursionContexts(_parentctx)
        return localctx

    class FactorContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ADD(self):
            return self.getToken(FandangoParser.ADD, 0)

        def factor(self):
            return self.getTypedRuleContext(FandangoParser.FactorContext, 0)

        def MINUS(self):
            return self.getToken(FandangoParser.MINUS, 0)

        def NOT_OP(self):
            return self.getToken(FandangoParser.NOT_OP, 0)

        def power(self):
            return self.getTypedRuleContext(FandangoParser.PowerContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_factor

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterFactor"):
                listener.enterFactor(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitFactor"):
                listener.exitFactor(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitFactor"):
                return visitor.visitFactor(self)
            else:
                return visitor.visitChildren(self)

    def factor(self):

        localctx = FandangoParser.FactorContext(self, self._ctx, self.state)
        self.enterRule(localctx, 326, self.RULE_factor)
        try:
            self.state = 2048
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [82]:
                self.enterOuterAlt(localctx, 1)
                self.state = 2041
                self.match(FandangoParser.ADD)
                self.state = 2042
                self.factor()
                pass
            elif token in [83]:
                self.enterOuterAlt(localctx, 2)
                self.state = 2043
                self.match(FandangoParser.MINUS)
                self.state = 2044
                self.factor()
                pass
            elif token in [87]:
                self.enterOuterAlt(localctx, 3)
                self.state = 2045
                self.match(FandangoParser.NOT_OP)
                self.state = 2046
                self.factor()
                pass
            elif token in [3, 4, 8, 13, 23, 34, 41, 52, 66, 68, 75, 77, 88, 118]:
                self.enterOuterAlt(localctx, 4)
                self.state = 2047
                self.power()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class PowerContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def await_primary(self):
            return self.getTypedRuleContext(FandangoParser.Await_primaryContext, 0)

        def POWER(self):
            return self.getToken(FandangoParser.POWER, 0)

        def factor(self):
            return self.getTypedRuleContext(FandangoParser.FactorContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_power

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterPower"):
                listener.enterPower(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitPower"):
                listener.exitPower(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitPower"):
                return visitor.visitPower(self)
            else:
                return visitor.visitChildren(self)

    def power(self):

        localctx = FandangoParser.PowerContext(self, self._ctx, self.state)
        self.enterRule(localctx, 328, self.RULE_power)
        try:
            self.state = 2055
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 231, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 2050
                self.await_primary()
                self.state = 2051
                self.match(FandangoParser.POWER)
                self.state = 2052
                self.factor()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 2054
                self.await_primary()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Await_primaryContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def AWAIT(self):
            return self.getToken(FandangoParser.AWAIT, 0)

        def primary(self):
            return self.getTypedRuleContext(FandangoParser.PrimaryContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_await_primary

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterAwait_primary"):
                listener.enterAwait_primary(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitAwait_primary"):
                listener.exitAwait_primary(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitAwait_primary"):
                return visitor.visitAwait_primary(self)
            else:
                return visitor.visitChildren(self)

    def await_primary(self):

        localctx = FandangoParser.Await_primaryContext(self, self._ctx, self.state)
        self.enterRule(localctx, 330, self.RULE_await_primary)
        try:
            self.state = 2060
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [13]:
                self.enterOuterAlt(localctx, 1)
                self.state = 2057
                self.match(FandangoParser.AWAIT)
                self.state = 2058
                self.primary(0)
                pass
            elif token in [3, 4, 8, 23, 34, 41, 52, 66, 68, 75, 77, 88, 118]:
                self.enterOuterAlt(localctx, 2)
                self.state = 2059
                self.primary(0)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class PrimaryContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def atom(self):
            return self.getTypedRuleContext(FandangoParser.AtomContext, 0)

        def primary(self):
            return self.getTypedRuleContext(FandangoParser.PrimaryContext, 0)

        def DOT(self):
            return self.getToken(FandangoParser.DOT, 0)

        def NAME(self):
            return self.getToken(FandangoParser.NAME, 0)

        def genexp(self):
            return self.getTypedRuleContext(FandangoParser.GenexpContext, 0)

        def OPEN_PAREN(self):
            return self.getToken(FandangoParser.OPEN_PAREN, 0)

        def CLOSE_PAREN(self):
            return self.getToken(FandangoParser.CLOSE_PAREN, 0)

        def arguments(self):
            return self.getTypedRuleContext(FandangoParser.ArgumentsContext, 0)

        def OPEN_BRACK(self):
            return self.getToken(FandangoParser.OPEN_BRACK, 0)

        def slices(self):
            return self.getTypedRuleContext(FandangoParser.SlicesContext, 0)

        def CLOSE_BRACK(self):
            return self.getToken(FandangoParser.CLOSE_BRACK, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_primary

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterPrimary"):
                listener.enterPrimary(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitPrimary"):
                listener.exitPrimary(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitPrimary"):
                return visitor.visitPrimary(self)
            else:
                return visitor.visitChildren(self)

    def primary(self, _p: int = 0):
        _parentctx = self._ctx
        _parentState = self.state
        localctx = FandangoParser.PrimaryContext(self, self._ctx, _parentState)
        _prevctx = localctx
        _startState = 332
        self.enterRecursionRule(localctx, 332, self.RULE_primary, _p)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 2063
            self.atom()
            self._ctx.stop = self._input.LT(-1)
            self.state = 2083
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 235, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    if self._parseListeners is not None:
                        self.triggerExitRuleEvent()
                    _prevctx = localctx
                    self.state = 2081
                    self._errHandler.sync(self)
                    la_ = self._interp.adaptivePredict(self._input, 234, self._ctx)
                    if la_ == 1:
                        localctx = FandangoParser.PrimaryContext(
                            self, _parentctx, _parentState
                        )
                        self.pushNewRecursionContext(
                            localctx, _startState, self.RULE_primary
                        )
                        self.state = 2065
                        if not self.precpred(self._ctx, 5):
                            from antlr4.error.Errors import FailedPredicateException

                            raise FailedPredicateException(
                                self, "self.precpred(self._ctx, 5)"
                            )
                        self.state = 2066
                        self.match(FandangoParser.DOT)
                        self.state = 2067
                        self.match(FandangoParser.NAME)
                        pass

                    elif la_ == 2:
                        localctx = FandangoParser.PrimaryContext(
                            self, _parentctx, _parentState
                        )
                        self.pushNewRecursionContext(
                            localctx, _startState, self.RULE_primary
                        )
                        self.state = 2068
                        if not self.precpred(self._ctx, 4):
                            from antlr4.error.Errors import FailedPredicateException

                            raise FailedPredicateException(
                                self, "self.precpred(self._ctx, 4)"
                            )
                        self.state = 2069
                        self.genexp()
                        pass

                    elif la_ == 3:
                        localctx = FandangoParser.PrimaryContext(
                            self, _parentctx, _parentState
                        )
                        self.pushNewRecursionContext(
                            localctx, _startState, self.RULE_primary
                        )
                        self.state = 2070
                        if not self.precpred(self._ctx, 3):
                            from antlr4.error.Errors import FailedPredicateException

                            raise FailedPredicateException(
                                self, "self.precpred(self._ctx, 3)"
                            )
                        self.state = 2071
                        self.match(FandangoParser.OPEN_PAREN)
                        self.state = 2073
                        self._errHandler.sync(self)
                        _la = self._input.LA(1)
                        if (
                            ((_la) & ~0x3F) == 0
                            and ((1 << _la) & 4505888853336344) != 0
                        ) or (
                            (((_la - 66)) & ~0x3F) == 0
                            and ((1 << (_la - 66)) & 4503599633861255) != 0
                        ):
                            self.state = 2072
                            self.arguments()

                        self.state = 2075
                        self.match(FandangoParser.CLOSE_PAREN)
                        pass

                    elif la_ == 4:
                        localctx = FandangoParser.PrimaryContext(
                            self, _parentctx, _parentState
                        )
                        self.pushNewRecursionContext(
                            localctx, _startState, self.RULE_primary
                        )
                        self.state = 2076
                        if not self.precpred(self._ctx, 2):
                            from antlr4.error.Errors import FailedPredicateException

                            raise FailedPredicateException(
                                self, "self.precpred(self._ctx, 2)"
                            )
                        self.state = 2077
                        self.match(FandangoParser.OPEN_BRACK)
                        self.state = 2078
                        self.slices()
                        self.state = 2079
                        self.match(FandangoParser.CLOSE_BRACK)
                        pass

                self.state = 2085
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 235, self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.unrollRecursionContexts(_parentctx)
        return localctx

    class SlicesContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def slice_(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(FandangoParser.SliceContext)
            else:
                return self.getTypedRuleContext(FandangoParser.SliceContext, i)

        def starred_expression(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(
                    FandangoParser.Starred_expressionContext
                )
            else:
                return self.getTypedRuleContext(
                    FandangoParser.Starred_expressionContext, i
                )

        def COMMA(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.COMMA)
            else:
                return self.getToken(FandangoParser.COMMA, i)

        def getRuleIndex(self):
            return FandangoParser.RULE_slices

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterSlices"):
                listener.enterSlices(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitSlices"):
                listener.exitSlices(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitSlices"):
                return visitor.visitSlices(self)
            else:
                return visitor.visitChildren(self)

    def slices(self):

        localctx = FandangoParser.SlicesContext(self, self._ctx, self.state)
        self.enterRule(localctx, 334, self.RULE_slices)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 2088
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [
                3,
                4,
                8,
                13,
                23,
                32,
                34,
                36,
                41,
                52,
                66,
                68,
                71,
                75,
                77,
                82,
                83,
                87,
                88,
                118,
            ]:
                self.state = 2086
                self.slice_()
                pass
            elif token in [67]:
                self.state = 2087
                self.starred_expression()
                pass
            else:
                raise NoViableAltException(self)

            self.state = 2097
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 238, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 2090
                    self.match(FandangoParser.COMMA)
                    self.state = 2093
                    self._errHandler.sync(self)
                    token = self._input.LA(1)
                    if token in [
                        3,
                        4,
                        8,
                        13,
                        23,
                        32,
                        34,
                        36,
                        41,
                        52,
                        66,
                        68,
                        71,
                        75,
                        77,
                        82,
                        83,
                        87,
                        88,
                        118,
                    ]:
                        self.state = 2091
                        self.slice_()
                        pass
                    elif token in [67]:
                        self.state = 2092
                        self.starred_expression()
                        pass
                    else:
                        raise NoViableAltException(self)

                self.state = 2099
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 238, self._ctx)

            self.state = 2101
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == 70:
                self.state = 2100
                self.match(FandangoParser.COMMA)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class SliceContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def COLON(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.COLON)
            else:
                return self.getToken(FandangoParser.COLON, i)

        def expression(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(FandangoParser.ExpressionContext)
            else:
                return self.getTypedRuleContext(FandangoParser.ExpressionContext, i)

        def named_expression(self):
            return self.getTypedRuleContext(FandangoParser.Named_expressionContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_slice

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterSlice"):
                listener.enterSlice(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitSlice"):
                listener.exitSlice(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitSlice"):
                return visitor.visitSlice(self)
            else:
                return visitor.visitChildren(self)

    def slice_(self):

        localctx = FandangoParser.SliceContext(self, self._ctx, self.state)
        self.enterRule(localctx, 336, self.RULE_slice)
        self._la = 0  # Token type
        try:
            self.state = 2117
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 244, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 2104
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if (((_la) & ~0x3F) == 0 and ((1 << _la) & 4505888853336344) != 0) or (
                    (((_la - 66)) & ~0x3F) == 0
                    and ((1 << (_la - 66)) & 4503599633861125) != 0
                ):
                    self.state = 2103
                    self.expression()

                self.state = 2106
                self.match(FandangoParser.COLON)
                self.state = 2108
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if (((_la) & ~0x3F) == 0 and ((1 << _la) & 4505888853336344) != 0) or (
                    (((_la - 66)) & ~0x3F) == 0
                    and ((1 << (_la - 66)) & 4503599633861125) != 0
                ):
                    self.state = 2107
                    self.expression()

                self.state = 2114
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == 71:
                    self.state = 2110
                    self.match(FandangoParser.COLON)
                    self.state = 2112
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)
                    if (
                        ((_la) & ~0x3F) == 0 and ((1 << _la) & 4505888853336344) != 0
                    ) or (
                        (((_la - 66)) & ~0x3F) == 0
                        and ((1 << (_la - 66)) & 4503599633861125) != 0
                    ):
                        self.state = 2111
                        self.expression()

                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 2116
                self.named_expression()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class AtomContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def selector_length(self):
            return self.getTypedRuleContext(FandangoParser.Selector_lengthContext, 0)

        def NAME(self):
            return self.getToken(FandangoParser.NAME, 0)

        def TRUE(self):
            return self.getToken(FandangoParser.TRUE, 0)

        def FALSE(self):
            return self.getToken(FandangoParser.FALSE, 0)

        def NONE(self):
            return self.getToken(FandangoParser.NONE, 0)

        def strings(self):
            return self.getTypedRuleContext(FandangoParser.StringsContext, 0)

        def NUMBER(self):
            return self.getToken(FandangoParser.NUMBER, 0)

        def tuple_(self):
            return self.getTypedRuleContext(FandangoParser.TupleContext, 0)

        def group(self):
            return self.getTypedRuleContext(FandangoParser.GroupContext, 0)

        def genexp(self):
            return self.getTypedRuleContext(FandangoParser.GenexpContext, 0)

        def list_(self):
            return self.getTypedRuleContext(FandangoParser.ListContext, 0)

        def listcomp(self):
            return self.getTypedRuleContext(FandangoParser.ListcompContext, 0)

        def dict_(self):
            return self.getTypedRuleContext(FandangoParser.DictContext, 0)

        def set_(self):
            return self.getTypedRuleContext(FandangoParser.SetContext, 0)

        def dictcomp(self):
            return self.getTypedRuleContext(FandangoParser.DictcompContext, 0)

        def setcomp(self):
            return self.getTypedRuleContext(FandangoParser.SetcompContext, 0)

        def ELLIPSIS(self):
            return self.getToken(FandangoParser.ELLIPSIS, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_atom

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterAtom"):
                listener.enterAtom(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitAtom"):
                listener.exitAtom(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitAtom"):
                return visitor.visitAtom(self)
            else:
                return visitor.visitChildren(self)

    def atom(self):

        localctx = FandangoParser.AtomContext(self, self._ctx, self.state)
        self.enterRule(localctx, 338, self.RULE_atom)
        try:
            self.state = 2142
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 248, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 2119
                self.selector_length()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 2120
                self.match(FandangoParser.NAME)
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 2121
                self.match(FandangoParser.TRUE)
                pass

            elif la_ == 4:
                self.enterOuterAlt(localctx, 4)
                self.state = 2122
                self.match(FandangoParser.FALSE)
                pass

            elif la_ == 5:
                self.enterOuterAlt(localctx, 5)
                self.state = 2123
                self.match(FandangoParser.NONE)
                pass

            elif la_ == 6:
                self.enterOuterAlt(localctx, 6)
                self.state = 2124
                self.strings()
                pass

            elif la_ == 7:
                self.enterOuterAlt(localctx, 7)
                self.state = 2125
                self.match(FandangoParser.NUMBER)
                pass

            elif la_ == 8:
                self.enterOuterAlt(localctx, 8)
                self.state = 2129
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input, 245, self._ctx)
                if la_ == 1:
                    self.state = 2126
                    self.tuple_()
                    pass

                elif la_ == 2:
                    self.state = 2127
                    self.group()
                    pass

                elif la_ == 3:
                    self.state = 2128
                    self.genexp()
                    pass

                pass

            elif la_ == 9:
                self.enterOuterAlt(localctx, 9)
                self.state = 2133
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input, 246, self._ctx)
                if la_ == 1:
                    self.state = 2131
                    self.list_()
                    pass

                elif la_ == 2:
                    self.state = 2132
                    self.listcomp()
                    pass

                pass

            elif la_ == 10:
                self.enterOuterAlt(localctx, 10)
                self.state = 2139
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input, 247, self._ctx)
                if la_ == 1:
                    self.state = 2135
                    self.dict_()
                    pass

                elif la_ == 2:
                    self.state = 2136
                    self.set_()
                    pass

                elif la_ == 3:
                    self.state = 2137
                    self.dictcomp()
                    pass

                elif la_ == 4:
                    self.state = 2138
                    self.setcomp()
                    pass

                pass

            elif la_ == 11:
                self.enterOuterAlt(localctx, 11)
                self.state = 2141
                self.match(FandangoParser.ELLIPSIS)
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class GroupContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def OPEN_PAREN(self):
            return self.getToken(FandangoParser.OPEN_PAREN, 0)

        def CLOSE_PAREN(self):
            return self.getToken(FandangoParser.CLOSE_PAREN, 0)

        def yield_expr(self):
            return self.getTypedRuleContext(FandangoParser.Yield_exprContext, 0)

        def named_expression(self):
            return self.getTypedRuleContext(FandangoParser.Named_expressionContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_group

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterGroup"):
                listener.enterGroup(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitGroup"):
                listener.exitGroup(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitGroup"):
                return visitor.visitGroup(self)
            else:
                return visitor.visitChildren(self)

    def group(self):

        localctx = FandangoParser.GroupContext(self, self._ctx, self.state)
        self.enterRule(localctx, 340, self.RULE_group)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 2144
            self.match(FandangoParser.OPEN_PAREN)
            self.state = 2147
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [48]:
                self.state = 2145
                self.yield_expr()
                pass
            elif token in [
                3,
                4,
                8,
                13,
                23,
                32,
                34,
                36,
                41,
                52,
                66,
                68,
                75,
                77,
                82,
                83,
                87,
                88,
                118,
            ]:
                self.state = 2146
                self.named_expression()
                pass
            else:
                raise NoViableAltException(self)

            self.state = 2149
            self.match(FandangoParser.CLOSE_PAREN)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class LambdefContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def LAMBDA(self):
            return self.getToken(FandangoParser.LAMBDA, 0)

        def COLON(self):
            return self.getToken(FandangoParser.COLON, 0)

        def expression(self):
            return self.getTypedRuleContext(FandangoParser.ExpressionContext, 0)

        def lambda_params(self):
            return self.getTypedRuleContext(FandangoParser.Lambda_paramsContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_lambdef

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterLambdef"):
                listener.enterLambdef(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitLambdef"):
                listener.exitLambdef(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitLambdef"):
                return visitor.visitLambdef(self)
            else:
                return visitor.visitChildren(self)

    def lambdef(self):

        localctx = FandangoParser.LambdefContext(self, self._ctx, self.state)
        self.enterRule(localctx, 342, self.RULE_lambdef)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 2151
            self.match(FandangoParser.LAMBDA)
            self.state = 2153
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if (((_la - 52)) & ~0x3F) == 0 and ((1 << (_la - 52)) & 2129921) != 0:
                self.state = 2152
                self.lambda_params()

            self.state = 2155
            self.match(FandangoParser.COLON)
            self.state = 2156
            self.expression()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Lambda_paramsContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def lambda_parameters(self):
            return self.getTypedRuleContext(FandangoParser.Lambda_parametersContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_lambda_params

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterLambda_params"):
                listener.enterLambda_params(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitLambda_params"):
                listener.exitLambda_params(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitLambda_params"):
                return visitor.visitLambda_params(self)
            else:
                return visitor.visitChildren(self)

    def lambda_params(self):

        localctx = FandangoParser.Lambda_paramsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 344, self.RULE_lambda_params)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 2158
            self.lambda_parameters()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Lambda_parametersContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def lambda_slash_no_default(self):
            return self.getTypedRuleContext(
                FandangoParser.Lambda_slash_no_defaultContext, 0
            )

        def lambda_param_no_default(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(
                    FandangoParser.Lambda_param_no_defaultContext
                )
            else:
                return self.getTypedRuleContext(
                    FandangoParser.Lambda_param_no_defaultContext, i
                )

        def lambda_param_with_default(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(
                    FandangoParser.Lambda_param_with_defaultContext
                )
            else:
                return self.getTypedRuleContext(
                    FandangoParser.Lambda_param_with_defaultContext, i
                )

        def lambda_star_etc(self):
            return self.getTypedRuleContext(FandangoParser.Lambda_star_etcContext, 0)

        def lambda_slash_with_default(self):
            return self.getTypedRuleContext(
                FandangoParser.Lambda_slash_with_defaultContext, 0
            )

        def getRuleIndex(self):
            return FandangoParser.RULE_lambda_parameters

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterLambda_parameters"):
                listener.enterLambda_parameters(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitLambda_parameters"):
                listener.exitLambda_parameters(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitLambda_parameters"):
                return visitor.visitLambda_parameters(self)
            else:
                return visitor.visitChildren(self)

    def lambda_parameters(self):

        localctx = FandangoParser.Lambda_parametersContext(self, self._ctx, self.state)
        self.enterRule(localctx, 346, self.RULE_lambda_parameters)
        self._la = 0  # Token type
        try:
            self.state = 2209
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 261, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 2160
                self.lambda_slash_no_default()
                self.state = 2164
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 251, self._ctx)
                while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                    if _alt == 1:
                        self.state = 2161
                        self.lambda_param_no_default()
                    self.state = 2166
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input, 251, self._ctx)

                self.state = 2170
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la == 52:
                    self.state = 2167
                    self.lambda_param_with_default()
                    self.state = 2172
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                self.state = 2174
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == 67 or _la == 73:
                    self.state = 2173
                    self.lambda_star_etc()

                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 2176
                self.lambda_slash_with_default()
                self.state = 2180
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la == 52:
                    self.state = 2177
                    self.lambda_param_with_default()
                    self.state = 2182
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                self.state = 2184
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == 67 or _la == 73:
                    self.state = 2183
                    self.lambda_star_etc()

                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 2187
                self._errHandler.sync(self)
                _alt = 1
                while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                    if _alt == 1:
                        self.state = 2186
                        self.lambda_param_no_default()

                    else:
                        raise NoViableAltException(self)
                    self.state = 2189
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input, 256, self._ctx)

                self.state = 2194
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la == 52:
                    self.state = 2191
                    self.lambda_param_with_default()
                    self.state = 2196
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                self.state = 2198
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == 67 or _la == 73:
                    self.state = 2197
                    self.lambda_star_etc()

                pass

            elif la_ == 4:
                self.enterOuterAlt(localctx, 4)
                self.state = 2201
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while True:
                    self.state = 2200
                    self.lambda_param_with_default()
                    self.state = 2203
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)
                    if not (_la == 52):
                        break

                self.state = 2206
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == 67 or _la == 73:
                    self.state = 2205
                    self.lambda_star_etc()

                pass

            elif la_ == 5:
                self.enterOuterAlt(localctx, 5)
                self.state = 2208
                self.lambda_star_etc()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Lambda_slash_no_defaultContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def DIV(self):
            return self.getToken(FandangoParser.DIV, 0)

        def lambda_param_no_default(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(
                    FandangoParser.Lambda_param_no_defaultContext
                )
            else:
                return self.getTypedRuleContext(
                    FandangoParser.Lambda_param_no_defaultContext, i
                )

        def COMMA(self):
            return self.getToken(FandangoParser.COMMA, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_lambda_slash_no_default

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterLambda_slash_no_default"):
                listener.enterLambda_slash_no_default(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitLambda_slash_no_default"):
                listener.exitLambda_slash_no_default(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitLambda_slash_no_default"):
                return visitor.visitLambda_slash_no_default(self)
            else:
                return visitor.visitChildren(self)

    def lambda_slash_no_default(self):

        localctx = FandangoParser.Lambda_slash_no_defaultContext(
            self, self._ctx, self.state
        )
        self.enterRule(localctx, 348, self.RULE_lambda_slash_no_default)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 2212
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 2211
                self.lambda_param_no_default()
                self.state = 2214
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not (_la == 52):
                    break

            self.state = 2216
            self.match(FandangoParser.DIV)
            self.state = 2218
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == 70:
                self.state = 2217
                self.match(FandangoParser.COMMA)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Lambda_slash_with_defaultContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def DIV(self):
            return self.getToken(FandangoParser.DIV, 0)

        def lambda_param_no_default(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(
                    FandangoParser.Lambda_param_no_defaultContext
                )
            else:
                return self.getTypedRuleContext(
                    FandangoParser.Lambda_param_no_defaultContext, i
                )

        def lambda_param_with_default(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(
                    FandangoParser.Lambda_param_with_defaultContext
                )
            else:
                return self.getTypedRuleContext(
                    FandangoParser.Lambda_param_with_defaultContext, i
                )

        def COMMA(self):
            return self.getToken(FandangoParser.COMMA, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_lambda_slash_with_default

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterLambda_slash_with_default"):
                listener.enterLambda_slash_with_default(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitLambda_slash_with_default"):
                listener.exitLambda_slash_with_default(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitLambda_slash_with_default"):
                return visitor.visitLambda_slash_with_default(self)
            else:
                return visitor.visitChildren(self)

    def lambda_slash_with_default(self):

        localctx = FandangoParser.Lambda_slash_with_defaultContext(
            self, self._ctx, self.state
        )
        self.enterRule(localctx, 350, self.RULE_lambda_slash_with_default)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 2223
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 264, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 2220
                    self.lambda_param_no_default()
                self.state = 2225
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 264, self._ctx)

            self.state = 2227
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 2226
                self.lambda_param_with_default()
                self.state = 2229
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not (_la == 52):
                    break

            self.state = 2231
            self.match(FandangoParser.DIV)
            self.state = 2233
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == 70:
                self.state = 2232
                self.match(FandangoParser.COMMA)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Lambda_star_etcContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def STAR(self):
            return self.getToken(FandangoParser.STAR, 0)

        def lambda_param_no_default(self):
            return self.getTypedRuleContext(
                FandangoParser.Lambda_param_no_defaultContext, 0
            )

        def lambda_param_maybe_default(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(
                    FandangoParser.Lambda_param_maybe_defaultContext
                )
            else:
                return self.getTypedRuleContext(
                    FandangoParser.Lambda_param_maybe_defaultContext, i
                )

        def lambda_kwds(self):
            return self.getTypedRuleContext(FandangoParser.Lambda_kwdsContext, 0)

        def COMMA(self):
            return self.getToken(FandangoParser.COMMA, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_lambda_star_etc

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterLambda_star_etc"):
                listener.enterLambda_star_etc(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitLambda_star_etc"):
                listener.exitLambda_star_etc(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitLambda_star_etc"):
                return visitor.visitLambda_star_etc(self)
            else:
                return visitor.visitChildren(self)

    def lambda_star_etc(self):

        localctx = FandangoParser.Lambda_star_etcContext(self, self._ctx, self.state)
        self.enterRule(localctx, 352, self.RULE_lambda_star_etc)
        self._la = 0  # Token type
        try:
            self.state = 2257
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 271, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 2235
                self.match(FandangoParser.STAR)
                self.state = 2236
                self.lambda_param_no_default()
                self.state = 2240
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la == 52:
                    self.state = 2237
                    self.lambda_param_maybe_default()
                    self.state = 2242
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                self.state = 2244
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == 73:
                    self.state = 2243
                    self.lambda_kwds()

                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 2246
                self.match(FandangoParser.STAR)
                self.state = 2247
                self.match(FandangoParser.COMMA)
                self.state = 2249
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while True:
                    self.state = 2248
                    self.lambda_param_maybe_default()
                    self.state = 2251
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)
                    if not (_la == 52):
                        break

                self.state = 2254
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == 73:
                    self.state = 2253
                    self.lambda_kwds()

                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 2256
                self.lambda_kwds()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Lambda_kwdsContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def POWER(self):
            return self.getToken(FandangoParser.POWER, 0)

        def lambda_param_no_default(self):
            return self.getTypedRuleContext(
                FandangoParser.Lambda_param_no_defaultContext, 0
            )

        def getRuleIndex(self):
            return FandangoParser.RULE_lambda_kwds

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterLambda_kwds"):
                listener.enterLambda_kwds(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitLambda_kwds"):
                listener.exitLambda_kwds(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitLambda_kwds"):
                return visitor.visitLambda_kwds(self)
            else:
                return visitor.visitChildren(self)

    def lambda_kwds(self):

        localctx = FandangoParser.Lambda_kwdsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 354, self.RULE_lambda_kwds)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 2259
            self.match(FandangoParser.POWER)
            self.state = 2260
            self.lambda_param_no_default()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Lambda_param_no_defaultContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def lambda_param(self):
            return self.getTypedRuleContext(FandangoParser.Lambda_paramContext, 0)

        def COMMA(self):
            return self.getToken(FandangoParser.COMMA, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_lambda_param_no_default

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterLambda_param_no_default"):
                listener.enterLambda_param_no_default(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitLambda_param_no_default"):
                listener.exitLambda_param_no_default(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitLambda_param_no_default"):
                return visitor.visitLambda_param_no_default(self)
            else:
                return visitor.visitChildren(self)

    def lambda_param_no_default(self):

        localctx = FandangoParser.Lambda_param_no_defaultContext(
            self, self._ctx, self.state
        )
        self.enterRule(localctx, 356, self.RULE_lambda_param_no_default)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 2262
            self.lambda_param()
            self.state = 2264
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == 70:
                self.state = 2263
                self.match(FandangoParser.COMMA)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Lambda_param_with_defaultContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def lambda_param(self):
            return self.getTypedRuleContext(FandangoParser.Lambda_paramContext, 0)

        def default(self):
            return self.getTypedRuleContext(FandangoParser.DefaultContext, 0)

        def COMMA(self):
            return self.getToken(FandangoParser.COMMA, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_lambda_param_with_default

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterLambda_param_with_default"):
                listener.enterLambda_param_with_default(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitLambda_param_with_default"):
                listener.exitLambda_param_with_default(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitLambda_param_with_default"):
                return visitor.visitLambda_param_with_default(self)
            else:
                return visitor.visitChildren(self)

    def lambda_param_with_default(self):

        localctx = FandangoParser.Lambda_param_with_defaultContext(
            self, self._ctx, self.state
        )
        self.enterRule(localctx, 358, self.RULE_lambda_param_with_default)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 2266
            self.lambda_param()
            self.state = 2267
            self.default()
            self.state = 2269
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == 70:
                self.state = 2268
                self.match(FandangoParser.COMMA)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Lambda_param_maybe_defaultContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def lambda_param(self):
            return self.getTypedRuleContext(FandangoParser.Lambda_paramContext, 0)

        def default(self):
            return self.getTypedRuleContext(FandangoParser.DefaultContext, 0)

        def COMMA(self):
            return self.getToken(FandangoParser.COMMA, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_lambda_param_maybe_default

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterLambda_param_maybe_default"):
                listener.enterLambda_param_maybe_default(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitLambda_param_maybe_default"):
                listener.exitLambda_param_maybe_default(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitLambda_param_maybe_default"):
                return visitor.visitLambda_param_maybe_default(self)
            else:
                return visitor.visitChildren(self)

    def lambda_param_maybe_default(self):

        localctx = FandangoParser.Lambda_param_maybe_defaultContext(
            self, self._ctx, self.state
        )
        self.enterRule(localctx, 360, self.RULE_lambda_param_maybe_default)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 2271
            self.lambda_param()
            self.state = 2273
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == 74:
                self.state = 2272
                self.default()

            self.state = 2276
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == 70:
                self.state = 2275
                self.match(FandangoParser.COMMA)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Lambda_paramContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NAME(self):
            return self.getToken(FandangoParser.NAME, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_lambda_param

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterLambda_param"):
                listener.enterLambda_param(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitLambda_param"):
                listener.exitLambda_param(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitLambda_param"):
                return visitor.visitLambda_param(self)
            else:
                return visitor.visitChildren(self)

    def lambda_param(self):

        localctx = FandangoParser.Lambda_paramContext(self, self._ctx, self.state)
        self.enterRule(localctx, 362, self.RULE_lambda_param)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 2278
            self.match(FandangoParser.NAME)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Fstring_middleContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def fstring_replacement_field(self):
            return self.getTypedRuleContext(
                FandangoParser.Fstring_replacement_fieldContext, 0
            )

        def FSTRING_MIDDLE(self):
            return self.getToken(FandangoParser.FSTRING_MIDDLE, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_fstring_middle

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterFstring_middle"):
                listener.enterFstring_middle(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitFstring_middle"):
                listener.exitFstring_middle(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitFstring_middle"):
                return visitor.visitFstring_middle(self)
            else:
                return visitor.visitChildren(self)

    def fstring_middle(self):

        localctx = FandangoParser.Fstring_middleContext(self, self._ctx, self.state)
        self.enterRule(localctx, 364, self.RULE_fstring_middle)
        try:
            self.state = 2282
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [88]:
                self.enterOuterAlt(localctx, 1)
                self.state = 2280
                self.fstring_replacement_field()
                pass
            elif token in [117]:
                self.enterOuterAlt(localctx, 2)
                self.state = 2281
                self.match(FandangoParser.FSTRING_MIDDLE)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Fstring_replacement_fieldContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def OPEN_BRACE(self):
            return self.getToken(FandangoParser.OPEN_BRACE, 0)

        def CLOSE_BRACE(self):
            return self.getToken(FandangoParser.CLOSE_BRACE, 0)

        def yield_expr(self):
            return self.getTypedRuleContext(FandangoParser.Yield_exprContext, 0)

        def star_expressions(self):
            return self.getTypedRuleContext(FandangoParser.Star_expressionsContext, 0)

        def ASSIGN(self):
            return self.getToken(FandangoParser.ASSIGN, 0)

        def fstring_conversion(self):
            return self.getTypedRuleContext(FandangoParser.Fstring_conversionContext, 0)

        def fstring_full_format_spec(self):
            return self.getTypedRuleContext(
                FandangoParser.Fstring_full_format_specContext, 0
            )

        def getRuleIndex(self):
            return FandangoParser.RULE_fstring_replacement_field

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterFstring_replacement_field"):
                listener.enterFstring_replacement_field(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitFstring_replacement_field"):
                listener.exitFstring_replacement_field(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitFstring_replacement_field"):
                return visitor.visitFstring_replacement_field(self)
            else:
                return visitor.visitChildren(self)

    def fstring_replacement_field(self):

        localctx = FandangoParser.Fstring_replacement_fieldContext(
            self, self._ctx, self.state
        )
        self.enterRule(localctx, 366, self.RULE_fstring_replacement_field)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 2284
            self.match(FandangoParser.OPEN_BRACE)
            self.state = 2287
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [48]:
                self.state = 2285
                self.yield_expr()
                pass
            elif token in [
                3,
                4,
                8,
                13,
                23,
                32,
                34,
                36,
                41,
                52,
                66,
                67,
                68,
                75,
                77,
                82,
                83,
                87,
                88,
                118,
            ]:
                self.state = 2286
                self.star_expressions()
                pass
            else:
                raise NoViableAltException(self)

            self.state = 2290
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == 74:
                self.state = 2289
                self.match(FandangoParser.ASSIGN)

            self.state = 2293
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 279, self._ctx)
            if la_ == 1:
                self.state = 2292
                self.fstring_conversion()

            self.state = 2296
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == 71:
                self.state = 2295
                self.fstring_full_format_spec()

            self.state = 2298
            self.match(FandangoParser.CLOSE_BRACE)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Fstring_conversionContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def EXCL(self):
            return self.getToken(FandangoParser.EXCL, 0)

        def NAME(self):
            return self.getToken(FandangoParser.NAME, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_fstring_conversion

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterFstring_conversion"):
                listener.enterFstring_conversion(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitFstring_conversion"):
                listener.exitFstring_conversion(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitFstring_conversion"):
                return visitor.visitFstring_conversion(self)
            else:
                return visitor.visitChildren(self)

    def fstring_conversion(self):

        localctx = FandangoParser.Fstring_conversionContext(self, self._ctx, self.state)
        self.enterRule(localctx, 368, self.RULE_fstring_conversion)
        try:
            self.state = 2303
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [71, 89]:
                self.enterOuterAlt(localctx, 1)

                pass
            elif token in [113]:
                self.enterOuterAlt(localctx, 2)
                self.state = 2301
                self.match(FandangoParser.EXCL)
                self.state = 2302
                self.match(FandangoParser.NAME)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Fstring_full_format_specContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def COLON(self):
            return self.getToken(FandangoParser.COLON, 0)

        def fstring_format_spec(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(
                    FandangoParser.Fstring_format_specContext
                )
            else:
                return self.getTypedRuleContext(
                    FandangoParser.Fstring_format_specContext, i
                )

        def getRuleIndex(self):
            return FandangoParser.RULE_fstring_full_format_spec

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterFstring_full_format_spec"):
                listener.enterFstring_full_format_spec(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitFstring_full_format_spec"):
                listener.exitFstring_full_format_spec(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitFstring_full_format_spec"):
                return visitor.visitFstring_full_format_spec(self)
            else:
                return visitor.visitChildren(self)

    def fstring_full_format_spec(self):

        localctx = FandangoParser.Fstring_full_format_specContext(
            self, self._ctx, self.state
        )
        self.enterRule(localctx, 370, self.RULE_fstring_full_format_spec)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 2305
            self.match(FandangoParser.COLON)
            self.state = 2309
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la == 88 or _la == 117:
                self.state = 2306
                self.fstring_format_spec()
                self.state = 2311
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Fstring_format_specContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def FSTRING_MIDDLE(self):
            return self.getToken(FandangoParser.FSTRING_MIDDLE, 0)

        def fstring_replacement_field(self):
            return self.getTypedRuleContext(
                FandangoParser.Fstring_replacement_fieldContext, 0
            )

        def getRuleIndex(self):
            return FandangoParser.RULE_fstring_format_spec

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterFstring_format_spec"):
                listener.enterFstring_format_spec(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitFstring_format_spec"):
                listener.exitFstring_format_spec(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitFstring_format_spec"):
                return visitor.visitFstring_format_spec(self)
            else:
                return visitor.visitChildren(self)

    def fstring_format_spec(self):

        localctx = FandangoParser.Fstring_format_specContext(
            self, self._ctx, self.state
        )
        self.enterRule(localctx, 372, self.RULE_fstring_format_spec)
        try:
            self.state = 2314
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [117]:
                self.enterOuterAlt(localctx, 1)
                self.state = 2312
                self.match(FandangoParser.FSTRING_MIDDLE)
                pass
            elif token in [88]:
                self.enterOuterAlt(localctx, 2)
                self.state = 2313
                self.fstring_replacement_field()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class FstringContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def FSTRING_START(self):
            return self.getToken(FandangoParser.FSTRING_START, 0)

        def FSTRING_END(self):
            return self.getToken(FandangoParser.FSTRING_END, 0)

        def fstring_middle(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(FandangoParser.Fstring_middleContext)
            else:
                return self.getTypedRuleContext(FandangoParser.Fstring_middleContext, i)

        def getRuleIndex(self):
            return FandangoParser.RULE_fstring

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterFstring"):
                listener.enterFstring(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitFstring"):
                listener.exitFstring(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitFstring"):
                return visitor.visitFstring(self)
            else:
                return visitor.visitChildren(self)

    def fstring(self):

        localctx = FandangoParser.FstringContext(self, self._ctx, self.state)
        self.enterRule(localctx, 374, self.RULE_fstring)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 2316
            self.match(FandangoParser.FSTRING_START)
            self.state = 2320
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la == 88 or _la == 117:
                self.state = 2317
                self.fstring_middle()
                self.state = 2322
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 2323
            self.match(FandangoParser.FSTRING_END)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class StringContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def STRING(self):
            return self.getToken(FandangoParser.STRING, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_string

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterString"):
                listener.enterString(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitString"):
                listener.exitString(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitString"):
                return visitor.visitString(self)
            else:
                return visitor.visitChildren(self)

    def string(self):

        localctx = FandangoParser.StringContext(self, self._ctx, self.state)
        self.enterRule(localctx, 376, self.RULE_string)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 2325
            self.match(FandangoParser.STRING)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class StringsContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def fstring(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(FandangoParser.FstringContext)
            else:
                return self.getTypedRuleContext(FandangoParser.FstringContext, i)

        def string(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(FandangoParser.StringContext)
            else:
                return self.getTypedRuleContext(FandangoParser.StringContext, i)

        def getRuleIndex(self):
            return FandangoParser.RULE_strings

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterStrings"):
                listener.enterStrings(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitStrings"):
                listener.exitStrings(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitStrings"):
                return visitor.visitStrings(self)
            else:
                return visitor.visitChildren(self)

    def strings(self):

        localctx = FandangoParser.StringsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 378, self.RULE_strings)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 2329
            self._errHandler.sync(self)
            _alt = 1
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 2329
                    self._errHandler.sync(self)
                    token = self._input.LA(1)
                    if token in [118]:
                        self.state = 2327
                        self.fstring()
                        pass
                    elif token in [3]:
                        self.state = 2328
                        self.string()
                        pass
                    else:
                        raise NoViableAltException(self)

                else:
                    raise NoViableAltException(self)
                self.state = 2331
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 286, self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ListContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def OPEN_BRACK(self):
            return self.getToken(FandangoParser.OPEN_BRACK, 0)

        def CLOSE_BRACK(self):
            return self.getToken(FandangoParser.CLOSE_BRACK, 0)

        def star_named_expressions(self):
            return self.getTypedRuleContext(
                FandangoParser.Star_named_expressionsContext, 0
            )

        def getRuleIndex(self):
            return FandangoParser.RULE_list

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterList"):
                listener.enterList(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitList"):
                listener.exitList(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitList"):
                return visitor.visitList(self)
            else:
                return visitor.visitChildren(self)

    def list_(self):

        localctx = FandangoParser.ListContext(self, self._ctx, self.state)
        self.enterRule(localctx, 380, self.RULE_list)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 2333
            self.match(FandangoParser.OPEN_BRACK)
            self.state = 2335
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if (((_la) & ~0x3F) == 0 and ((1 << _la) & 4505888853336344) != 0) or (
                (((_la - 66)) & ~0x3F) == 0
                and ((1 << (_la - 66)) & 4503599633861127) != 0
            ):
                self.state = 2334
                self.star_named_expressions()

            self.state = 2337
            self.match(FandangoParser.CLOSE_BRACK)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class TupleContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def OPEN_PAREN(self):
            return self.getToken(FandangoParser.OPEN_PAREN, 0)

        def CLOSE_PAREN(self):
            return self.getToken(FandangoParser.CLOSE_PAREN, 0)

        def star_named_expression(self):
            return self.getTypedRuleContext(
                FandangoParser.Star_named_expressionContext, 0
            )

        def COMMA(self):
            return self.getToken(FandangoParser.COMMA, 0)

        def star_named_expressions(self):
            return self.getTypedRuleContext(
                FandangoParser.Star_named_expressionsContext, 0
            )

        def getRuleIndex(self):
            return FandangoParser.RULE_tuple

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterTuple"):
                listener.enterTuple(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitTuple"):
                listener.exitTuple(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitTuple"):
                return visitor.visitTuple(self)
            else:
                return visitor.visitChildren(self)

    def tuple_(self):

        localctx = FandangoParser.TupleContext(self, self._ctx, self.state)
        self.enterRule(localctx, 382, self.RULE_tuple)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 2339
            self.match(FandangoParser.OPEN_PAREN)
            self.state = 2344
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if (((_la) & ~0x3F) == 0 and ((1 << _la) & 4505888853336344) != 0) or (
                (((_la - 66)) & ~0x3F) == 0
                and ((1 << (_la - 66)) & 4503599633861127) != 0
            ):
                self.state = 2340
                self.star_named_expression()
                self.state = 2341
                self.match(FandangoParser.COMMA)

                self.state = 2342
                self.star_named_expressions()

            self.state = 2346
            self.match(FandangoParser.CLOSE_PAREN)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class SetContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def OPEN_BRACE(self):
            return self.getToken(FandangoParser.OPEN_BRACE, 0)

        def star_named_expressions(self):
            return self.getTypedRuleContext(
                FandangoParser.Star_named_expressionsContext, 0
            )

        def CLOSE_BRACE(self):
            return self.getToken(FandangoParser.CLOSE_BRACE, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_set

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterSet"):
                listener.enterSet(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitSet"):
                listener.exitSet(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitSet"):
                return visitor.visitSet(self)
            else:
                return visitor.visitChildren(self)

    def set_(self):

        localctx = FandangoParser.SetContext(self, self._ctx, self.state)
        self.enterRule(localctx, 384, self.RULE_set)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 2348
            self.match(FandangoParser.OPEN_BRACE)
            self.state = 2349
            self.star_named_expressions()
            self.state = 2350
            self.match(FandangoParser.CLOSE_BRACE)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class DictContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def OPEN_BRACE(self):
            return self.getToken(FandangoParser.OPEN_BRACE, 0)

        def CLOSE_BRACE(self):
            return self.getToken(FandangoParser.CLOSE_BRACE, 0)

        def double_starred_kvpairs(self):
            return self.getTypedRuleContext(
                FandangoParser.Double_starred_kvpairsContext, 0
            )

        def getRuleIndex(self):
            return FandangoParser.RULE_dict

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterDict"):
                listener.enterDict(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitDict"):
                listener.exitDict(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitDict"):
                return visitor.visitDict(self)
            else:
                return visitor.visitChildren(self)

    def dict_(self):

        localctx = FandangoParser.DictContext(self, self._ctx, self.state)
        self.enterRule(localctx, 386, self.RULE_dict)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 2352
            self.match(FandangoParser.OPEN_BRACE)
            self.state = 2354
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if (((_la) & ~0x3F) == 0 and ((1 << _la) & 4505888853336344) != 0) or (
                (((_la - 66)) & ~0x3F) == 0
                and ((1 << (_la - 66)) & 4503599633861253) != 0
            ):
                self.state = 2353
                self.double_starred_kvpairs()

            self.state = 2356
            self.match(FandangoParser.CLOSE_BRACE)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Double_starred_kvpairsContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def double_starred_kvpair(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(
                    FandangoParser.Double_starred_kvpairContext
                )
            else:
                return self.getTypedRuleContext(
                    FandangoParser.Double_starred_kvpairContext, i
                )

        def COMMA(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.COMMA)
            else:
                return self.getToken(FandangoParser.COMMA, i)

        def getRuleIndex(self):
            return FandangoParser.RULE_double_starred_kvpairs

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterDouble_starred_kvpairs"):
                listener.enterDouble_starred_kvpairs(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitDouble_starred_kvpairs"):
                listener.exitDouble_starred_kvpairs(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitDouble_starred_kvpairs"):
                return visitor.visitDouble_starred_kvpairs(self)
            else:
                return visitor.visitChildren(self)

    def double_starred_kvpairs(self):

        localctx = FandangoParser.Double_starred_kvpairsContext(
            self, self._ctx, self.state
        )
        self.enterRule(localctx, 388, self.RULE_double_starred_kvpairs)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 2358
            self.double_starred_kvpair()
            self.state = 2363
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 290, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 2359
                    self.match(FandangoParser.COMMA)
                    self.state = 2360
                    self.double_starred_kvpair()
                self.state = 2365
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 290, self._ctx)

            self.state = 2367
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == 70:
                self.state = 2366
                self.match(FandangoParser.COMMA)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Double_starred_kvpairContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def POWER(self):
            return self.getToken(FandangoParser.POWER, 0)

        def bitwise_or(self):
            return self.getTypedRuleContext(FandangoParser.Bitwise_orContext, 0)

        def kvpair(self):
            return self.getTypedRuleContext(FandangoParser.KvpairContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_double_starred_kvpair

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterDouble_starred_kvpair"):
                listener.enterDouble_starred_kvpair(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitDouble_starred_kvpair"):
                listener.exitDouble_starred_kvpair(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitDouble_starred_kvpair"):
                return visitor.visitDouble_starred_kvpair(self)
            else:
                return visitor.visitChildren(self)

    def double_starred_kvpair(self):

        localctx = FandangoParser.Double_starred_kvpairContext(
            self, self._ctx, self.state
        )
        self.enterRule(localctx, 390, self.RULE_double_starred_kvpair)
        try:
            self.state = 2372
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [73]:
                self.enterOuterAlt(localctx, 1)
                self.state = 2369
                self.match(FandangoParser.POWER)
                self.state = 2370
                self.bitwise_or(0)
                pass
            elif token in [
                3,
                4,
                8,
                13,
                23,
                32,
                34,
                36,
                41,
                52,
                66,
                68,
                75,
                77,
                82,
                83,
                87,
                88,
                118,
            ]:
                self.enterOuterAlt(localctx, 2)
                self.state = 2371
                self.kvpair()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class KvpairContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def expression(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(FandangoParser.ExpressionContext)
            else:
                return self.getTypedRuleContext(FandangoParser.ExpressionContext, i)

        def COLON(self):
            return self.getToken(FandangoParser.COLON, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_kvpair

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterKvpair"):
                listener.enterKvpair(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitKvpair"):
                listener.exitKvpair(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitKvpair"):
                return visitor.visitKvpair(self)
            else:
                return visitor.visitChildren(self)

    def kvpair(self):

        localctx = FandangoParser.KvpairContext(self, self._ctx, self.state)
        self.enterRule(localctx, 392, self.RULE_kvpair)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 2374
            self.expression()
            self.state = 2375
            self.match(FandangoParser.COLON)
            self.state = 2376
            self.expression()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class For_if_clausesContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def for_if_clause(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(FandangoParser.For_if_clauseContext)
            else:
                return self.getTypedRuleContext(FandangoParser.For_if_clauseContext, i)

        def getRuleIndex(self):
            return FandangoParser.RULE_for_if_clauses

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterFor_if_clauses"):
                listener.enterFor_if_clauses(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitFor_if_clauses"):
                listener.exitFor_if_clauses(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitFor_if_clauses"):
                return visitor.visitFor_if_clauses(self)
            else:
                return visitor.visitChildren(self)

    def for_if_clauses(self):

        localctx = FandangoParser.For_if_clausesContext(self, self._ctx, self.state)
        self.enterRule(localctx, 394, self.RULE_for_if_clauses)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 2379
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 2378
                self.for_if_clause()
                self.state = 2381
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not (_la == 12 or _la == 25):
                    break

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class For_if_clauseContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def FOR(self):
            return self.getToken(FandangoParser.FOR, 0)

        def star_targets(self):
            return self.getTypedRuleContext(FandangoParser.Star_targetsContext, 0)

        def IN(self):
            return self.getToken(FandangoParser.IN, 0)

        def disjunction(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(FandangoParser.DisjunctionContext)
            else:
                return self.getTypedRuleContext(FandangoParser.DisjunctionContext, i)

        def ASYNC(self):
            return self.getToken(FandangoParser.ASYNC, 0)

        def IF(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.IF)
            else:
                return self.getToken(FandangoParser.IF, i)

        def getRuleIndex(self):
            return FandangoParser.RULE_for_if_clause

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterFor_if_clause"):
                listener.enterFor_if_clause(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitFor_if_clause"):
                listener.exitFor_if_clause(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitFor_if_clause"):
                return visitor.visitFor_if_clause(self)
            else:
                return visitor.visitChildren(self)

    def for_if_clause(self):

        localctx = FandangoParser.For_if_clauseContext(self, self._ctx, self.state)
        self.enterRule(localctx, 396, self.RULE_for_if_clause)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 2384
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == 12:
                self.state = 2383
                self.match(FandangoParser.ASYNC)

            self.state = 2386
            self.match(FandangoParser.FOR)
            self.state = 2387
            self.star_targets()
            self.state = 2388
            self.match(FandangoParser.IN)
            self.state = 2389
            self.disjunction()
            self.state = 2394
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la == 28:
                self.state = 2390
                self.match(FandangoParser.IF)
                self.state = 2391
                self.disjunction()
                self.state = 2396
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ListcompContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def OPEN_BRACK(self):
            return self.getToken(FandangoParser.OPEN_BRACK, 0)

        def named_expression(self):
            return self.getTypedRuleContext(FandangoParser.Named_expressionContext, 0)

        def for_if_clauses(self):
            return self.getTypedRuleContext(FandangoParser.For_if_clausesContext, 0)

        def CLOSE_BRACK(self):
            return self.getToken(FandangoParser.CLOSE_BRACK, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_listcomp

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterListcomp"):
                listener.enterListcomp(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitListcomp"):
                listener.exitListcomp(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitListcomp"):
                return visitor.visitListcomp(self)
            else:
                return visitor.visitChildren(self)

    def listcomp(self):

        localctx = FandangoParser.ListcompContext(self, self._ctx, self.state)
        self.enterRule(localctx, 398, self.RULE_listcomp)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 2397
            self.match(FandangoParser.OPEN_BRACK)
            self.state = 2398
            self.named_expression()
            self.state = 2399
            self.for_if_clauses()
            self.state = 2400
            self.match(FandangoParser.CLOSE_BRACK)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class SetcompContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def OPEN_BRACE(self):
            return self.getToken(FandangoParser.OPEN_BRACE, 0)

        def named_expression(self):
            return self.getTypedRuleContext(FandangoParser.Named_expressionContext, 0)

        def for_if_clauses(self):
            return self.getTypedRuleContext(FandangoParser.For_if_clausesContext, 0)

        def CLOSE_BRACE(self):
            return self.getToken(FandangoParser.CLOSE_BRACE, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_setcomp

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterSetcomp"):
                listener.enterSetcomp(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitSetcomp"):
                listener.exitSetcomp(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitSetcomp"):
                return visitor.visitSetcomp(self)
            else:
                return visitor.visitChildren(self)

    def setcomp(self):

        localctx = FandangoParser.SetcompContext(self, self._ctx, self.state)
        self.enterRule(localctx, 400, self.RULE_setcomp)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 2402
            self.match(FandangoParser.OPEN_BRACE)
            self.state = 2403
            self.named_expression()
            self.state = 2404
            self.for_if_clauses()
            self.state = 2405
            self.match(FandangoParser.CLOSE_BRACE)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class GenexpContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def OPEN_PAREN(self):
            return self.getToken(FandangoParser.OPEN_PAREN, 0)

        def for_if_clauses(self):
            return self.getTypedRuleContext(FandangoParser.For_if_clausesContext, 0)

        def CLOSE_PAREN(self):
            return self.getToken(FandangoParser.CLOSE_PAREN, 0)

        def assignment_expression(self):
            return self.getTypedRuleContext(
                FandangoParser.Assignment_expressionContext, 0
            )

        def expression(self):
            return self.getTypedRuleContext(FandangoParser.ExpressionContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_genexp

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterGenexp"):
                listener.enterGenexp(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitGenexp"):
                listener.exitGenexp(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitGenexp"):
                return visitor.visitGenexp(self)
            else:
                return visitor.visitChildren(self)

    def genexp(self):

        localctx = FandangoParser.GenexpContext(self, self._ctx, self.state)
        self.enterRule(localctx, 402, self.RULE_genexp)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 2407
            self.match(FandangoParser.OPEN_PAREN)
            self.state = 2410
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 296, self._ctx)
            if la_ == 1:
                self.state = 2408
                self.assignment_expression()
                pass

            elif la_ == 2:
                self.state = 2409
                self.expression()
                pass

            self.state = 2412
            self.for_if_clauses()
            self.state = 2413
            self.match(FandangoParser.CLOSE_PAREN)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class DictcompContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def OPEN_BRACE(self):
            return self.getToken(FandangoParser.OPEN_BRACE, 0)

        def kvpair(self):
            return self.getTypedRuleContext(FandangoParser.KvpairContext, 0)

        def for_if_clauses(self):
            return self.getTypedRuleContext(FandangoParser.For_if_clausesContext, 0)

        def CLOSE_BRACE(self):
            return self.getToken(FandangoParser.CLOSE_BRACE, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_dictcomp

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterDictcomp"):
                listener.enterDictcomp(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitDictcomp"):
                listener.exitDictcomp(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitDictcomp"):
                return visitor.visitDictcomp(self)
            else:
                return visitor.visitChildren(self)

    def dictcomp(self):

        localctx = FandangoParser.DictcompContext(self, self._ctx, self.state)
        self.enterRule(localctx, 404, self.RULE_dictcomp)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 2415
            self.match(FandangoParser.OPEN_BRACE)
            self.state = 2416
            self.kvpair()
            self.state = 2417
            self.for_if_clauses()
            self.state = 2418
            self.match(FandangoParser.CLOSE_BRACE)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ArgumentsContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def args(self):
            return self.getTypedRuleContext(FandangoParser.ArgsContext, 0)

        def COMMA(self):
            return self.getToken(FandangoParser.COMMA, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_arguments

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterArguments"):
                listener.enterArguments(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitArguments"):
                listener.exitArguments(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitArguments"):
                return visitor.visitArguments(self)
            else:
                return visitor.visitChildren(self)

    def arguments(self):

        localctx = FandangoParser.ArgumentsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 406, self.RULE_arguments)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 2420
            self.args()
            self.state = 2422
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == 70:
                self.state = 2421
                self.match(FandangoParser.COMMA)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ArgsContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def arg(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(FandangoParser.ArgContext)
            else:
                return self.getTypedRuleContext(FandangoParser.ArgContext, i)

        def COMMA(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.COMMA)
            else:
                return self.getToken(FandangoParser.COMMA, i)

        def kwargs(self):
            return self.getTypedRuleContext(FandangoParser.KwargsContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_args

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterArgs"):
                listener.enterArgs(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitArgs"):
                listener.exitArgs(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitArgs"):
                return visitor.visitArgs(self)
            else:
                return visitor.visitChildren(self)

    def args(self):

        localctx = FandangoParser.ArgsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 408, self.RULE_args)
        try:
            self.state = 2437
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 300, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 2424
                self.arg()
                self.state = 2429
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 298, self._ctx)
                while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                    if _alt == 1:
                        self.state = 2425
                        self.match(FandangoParser.COMMA)
                        self.state = 2426
                        self.arg()
                    self.state = 2431
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input, 298, self._ctx)

                self.state = 2434
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input, 299, self._ctx)
                if la_ == 1:
                    self.state = 2432
                    self.match(FandangoParser.COMMA)
                    self.state = 2433
                    self.kwargs()

                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 2436
                self.kwargs()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ArgContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def starred_expression(self):
            return self.getTypedRuleContext(FandangoParser.Starred_expressionContext, 0)

        def assignment_expression(self):
            return self.getTypedRuleContext(
                FandangoParser.Assignment_expressionContext, 0
            )

        def expression(self):
            return self.getTypedRuleContext(FandangoParser.ExpressionContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_arg

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterArg"):
                listener.enterArg(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitArg"):
                listener.exitArg(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitArg"):
                return visitor.visitArg(self)
            else:
                return visitor.visitChildren(self)

    def arg(self):

        localctx = FandangoParser.ArgContext(self, self._ctx, self.state)
        self.enterRule(localctx, 410, self.RULE_arg)
        try:
            self.state = 2442
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 301, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 2439
                self.starred_expression()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 2440
                self.assignment_expression()
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 2441
                self.expression()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class KwargsContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def kwarg_or_starred(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(FandangoParser.Kwarg_or_starredContext)
            else:
                return self.getTypedRuleContext(
                    FandangoParser.Kwarg_or_starredContext, i
                )

        def COMMA(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.COMMA)
            else:
                return self.getToken(FandangoParser.COMMA, i)

        def kwarg_or_double_starred(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(
                    FandangoParser.Kwarg_or_double_starredContext
                )
            else:
                return self.getTypedRuleContext(
                    FandangoParser.Kwarg_or_double_starredContext, i
                )

        def getRuleIndex(self):
            return FandangoParser.RULE_kwargs

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterKwargs"):
                listener.enterKwargs(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitKwargs"):
                listener.exitKwargs(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitKwargs"):
                return visitor.visitKwargs(self)
            else:
                return visitor.visitChildren(self)

    def kwargs(self):

        localctx = FandangoParser.KwargsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 412, self.RULE_kwargs)
        try:
            self.state = 2477
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 306, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 2444
                self.kwarg_or_starred()
                self.state = 2449
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 302, self._ctx)
                while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                    if _alt == 1:
                        self.state = 2445
                        self.match(FandangoParser.COMMA)
                        self.state = 2446
                        self.kwarg_or_starred()
                    self.state = 2451
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input, 302, self._ctx)

                self.state = 2452
                self.match(FandangoParser.COMMA)
                self.state = 2453
                self.kwarg_or_double_starred()
                self.state = 2458
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 303, self._ctx)
                while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                    if _alt == 1:
                        self.state = 2454
                        self.match(FandangoParser.COMMA)
                        self.state = 2455
                        self.kwarg_or_double_starred()
                    self.state = 2460
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input, 303, self._ctx)

                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 2461
                self.kwarg_or_starred()
                self.state = 2466
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 304, self._ctx)
                while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                    if _alt == 1:
                        self.state = 2462
                        self.match(FandangoParser.COMMA)
                        self.state = 2463
                        self.kwarg_or_starred()
                    self.state = 2468
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input, 304, self._ctx)

                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 2469
                self.kwarg_or_double_starred()
                self.state = 2474
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 305, self._ctx)
                while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                    if _alt == 1:
                        self.state = 2470
                        self.match(FandangoParser.COMMA)
                        self.state = 2471
                        self.kwarg_or_double_starred()
                    self.state = 2476
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input, 305, self._ctx)

                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Starred_expressionContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def STAR(self):
            return self.getToken(FandangoParser.STAR, 0)

        def expression(self):
            return self.getTypedRuleContext(FandangoParser.ExpressionContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_starred_expression

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterStarred_expression"):
                listener.enterStarred_expression(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitStarred_expression"):
                listener.exitStarred_expression(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitStarred_expression"):
                return visitor.visitStarred_expression(self)
            else:
                return visitor.visitChildren(self)

    def starred_expression(self):

        localctx = FandangoParser.Starred_expressionContext(self, self._ctx, self.state)
        self.enterRule(localctx, 414, self.RULE_starred_expression)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 2479
            self.match(FandangoParser.STAR)
            self.state = 2480
            self.expression()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Kwarg_or_starredContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NAME(self):
            return self.getToken(FandangoParser.NAME, 0)

        def ASSIGN(self):
            return self.getToken(FandangoParser.ASSIGN, 0)

        def expression(self):
            return self.getTypedRuleContext(FandangoParser.ExpressionContext, 0)

        def starred_expression(self):
            return self.getTypedRuleContext(FandangoParser.Starred_expressionContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_kwarg_or_starred

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterKwarg_or_starred"):
                listener.enterKwarg_or_starred(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitKwarg_or_starred"):
                listener.exitKwarg_or_starred(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitKwarg_or_starred"):
                return visitor.visitKwarg_or_starred(self)
            else:
                return visitor.visitChildren(self)

    def kwarg_or_starred(self):

        localctx = FandangoParser.Kwarg_or_starredContext(self, self._ctx, self.state)
        self.enterRule(localctx, 416, self.RULE_kwarg_or_starred)
        try:
            self.state = 2486
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [52]:
                self.enterOuterAlt(localctx, 1)
                self.state = 2482
                self.match(FandangoParser.NAME)
                self.state = 2483
                self.match(FandangoParser.ASSIGN)
                self.state = 2484
                self.expression()
                pass
            elif token in [67]:
                self.enterOuterAlt(localctx, 2)
                self.state = 2485
                self.starred_expression()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Kwarg_or_double_starredContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NAME(self):
            return self.getToken(FandangoParser.NAME, 0)

        def ASSIGN(self):
            return self.getToken(FandangoParser.ASSIGN, 0)

        def expression(self):
            return self.getTypedRuleContext(FandangoParser.ExpressionContext, 0)

        def POWER(self):
            return self.getToken(FandangoParser.POWER, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_kwarg_or_double_starred

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterKwarg_or_double_starred"):
                listener.enterKwarg_or_double_starred(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitKwarg_or_double_starred"):
                listener.exitKwarg_or_double_starred(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitKwarg_or_double_starred"):
                return visitor.visitKwarg_or_double_starred(self)
            else:
                return visitor.visitChildren(self)

    def kwarg_or_double_starred(self):

        localctx = FandangoParser.Kwarg_or_double_starredContext(
            self, self._ctx, self.state
        )
        self.enterRule(localctx, 418, self.RULE_kwarg_or_double_starred)
        try:
            self.state = 2493
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [52]:
                self.enterOuterAlt(localctx, 1)
                self.state = 2488
                self.match(FandangoParser.NAME)
                self.state = 2489
                self.match(FandangoParser.ASSIGN)
                self.state = 2490
                self.expression()
                pass
            elif token in [73]:
                self.enterOuterAlt(localctx, 2)
                self.state = 2491
                self.match(FandangoParser.POWER)
                self.state = 2492
                self.expression()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Star_targetsContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def star_target(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(FandangoParser.Star_targetContext)
            else:
                return self.getTypedRuleContext(FandangoParser.Star_targetContext, i)

        def COMMA(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.COMMA)
            else:
                return self.getToken(FandangoParser.COMMA, i)

        def getRuleIndex(self):
            return FandangoParser.RULE_star_targets

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterStar_targets"):
                listener.enterStar_targets(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitStar_targets"):
                listener.exitStar_targets(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitStar_targets"):
                return visitor.visitStar_targets(self)
            else:
                return visitor.visitChildren(self)

    def star_targets(self):

        localctx = FandangoParser.Star_targetsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 420, self.RULE_star_targets)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 2495
            self.star_target()
            self.state = 2500
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 309, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 2496
                    self.match(FandangoParser.COMMA)
                    self.state = 2497
                    self.star_target()
                self.state = 2502
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 309, self._ctx)

            self.state = 2504
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == 70:
                self.state = 2503
                self.match(FandangoParser.COMMA)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Star_targets_list_seqContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def star_target(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(FandangoParser.Star_targetContext)
            else:
                return self.getTypedRuleContext(FandangoParser.Star_targetContext, i)

        def COMMA(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.COMMA)
            else:
                return self.getToken(FandangoParser.COMMA, i)

        def getRuleIndex(self):
            return FandangoParser.RULE_star_targets_list_seq

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterStar_targets_list_seq"):
                listener.enterStar_targets_list_seq(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitStar_targets_list_seq"):
                listener.exitStar_targets_list_seq(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitStar_targets_list_seq"):
                return visitor.visitStar_targets_list_seq(self)
            else:
                return visitor.visitChildren(self)

    def star_targets_list_seq(self):

        localctx = FandangoParser.Star_targets_list_seqContext(
            self, self._ctx, self.state
        )
        self.enterRule(localctx, 422, self.RULE_star_targets_list_seq)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 2506
            self.star_target()
            self.state = 2511
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 311, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 2507
                    self.match(FandangoParser.COMMA)
                    self.state = 2508
                    self.star_target()
                self.state = 2513
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 311, self._ctx)

            self.state = 2515
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == 70:
                self.state = 2514
                self.match(FandangoParser.COMMA)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Star_targets_tuple_seqContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def star_target(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(FandangoParser.Star_targetContext)
            else:
                return self.getTypedRuleContext(FandangoParser.Star_targetContext, i)

        def COMMA(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.COMMA)
            else:
                return self.getToken(FandangoParser.COMMA, i)

        def getRuleIndex(self):
            return FandangoParser.RULE_star_targets_tuple_seq

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterStar_targets_tuple_seq"):
                listener.enterStar_targets_tuple_seq(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitStar_targets_tuple_seq"):
                listener.exitStar_targets_tuple_seq(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitStar_targets_tuple_seq"):
                return visitor.visitStar_targets_tuple_seq(self)
            else:
                return visitor.visitChildren(self)

    def star_targets_tuple_seq(self):

        localctx = FandangoParser.Star_targets_tuple_seqContext(
            self, self._ctx, self.state
        )
        self.enterRule(localctx, 424, self.RULE_star_targets_tuple_seq)
        self._la = 0  # Token type
        try:
            self.state = 2530
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 315, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 2517
                self.star_target()
                self.state = 2520
                self._errHandler.sync(self)
                _alt = 1
                while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                    if _alt == 1:
                        self.state = 2518
                        self.match(FandangoParser.COMMA)
                        self.state = 2519
                        self.star_target()

                    else:
                        raise NoViableAltException(self)
                    self.state = 2522
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input, 313, self._ctx)

                self.state = 2525
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == 70:
                    self.state = 2524
                    self.match(FandangoParser.COMMA)

                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 2527
                self.star_target()
                self.state = 2528
                self.match(FandangoParser.COMMA)
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Star_targetContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def STAR(self):
            return self.getToken(FandangoParser.STAR, 0)

        def star_target(self):
            return self.getTypedRuleContext(FandangoParser.Star_targetContext, 0)

        def target_with_star_atom(self):
            return self.getTypedRuleContext(
                FandangoParser.Target_with_star_atomContext, 0
            )

        def getRuleIndex(self):
            return FandangoParser.RULE_star_target

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterStar_target"):
                listener.enterStar_target(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitStar_target"):
                listener.exitStar_target(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitStar_target"):
                return visitor.visitStar_target(self)
            else:
                return visitor.visitChildren(self)

    def star_target(self):

        localctx = FandangoParser.Star_targetContext(self, self._ctx, self.state)
        self.enterRule(localctx, 426, self.RULE_star_target)
        try:
            self.state = 2535
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [67]:
                self.enterOuterAlt(localctx, 1)
                self.state = 2532
                self.match(FandangoParser.STAR)
                self.state = 2533
                self.star_target()
                pass
            elif token in [3, 4, 8, 23, 34, 41, 52, 66, 68, 75, 77, 88, 118]:
                self.enterOuterAlt(localctx, 2)
                self.state = 2534
                self.target_with_star_atom()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Target_with_star_atomContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def t_primary(self):
            return self.getTypedRuleContext(FandangoParser.T_primaryContext, 0)

        def DOT(self):
            return self.getToken(FandangoParser.DOT, 0)

        def NAME(self):
            return self.getToken(FandangoParser.NAME, 0)

        def OPEN_BRACK(self):
            return self.getToken(FandangoParser.OPEN_BRACK, 0)

        def slices(self):
            return self.getTypedRuleContext(FandangoParser.SlicesContext, 0)

        def CLOSE_BRACK(self):
            return self.getToken(FandangoParser.CLOSE_BRACK, 0)

        def star_atom(self):
            return self.getTypedRuleContext(FandangoParser.Star_atomContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_target_with_star_atom

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterTarget_with_star_atom"):
                listener.enterTarget_with_star_atom(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitTarget_with_star_atom"):
                listener.exitTarget_with_star_atom(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitTarget_with_star_atom"):
                return visitor.visitTarget_with_star_atom(self)
            else:
                return visitor.visitChildren(self)

    def target_with_star_atom(self):

        localctx = FandangoParser.Target_with_star_atomContext(
            self, self._ctx, self.state
        )
        self.enterRule(localctx, 428, self.RULE_target_with_star_atom)
        try:
            self.state = 2547
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 317, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 2537
                self.t_primary(0)
                self.state = 2538
                self.match(FandangoParser.DOT)
                self.state = 2539
                self.match(FandangoParser.NAME)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 2541
                self.t_primary(0)
                self.state = 2542
                self.match(FandangoParser.OPEN_BRACK)
                self.state = 2543
                self.slices()
                self.state = 2544
                self.match(FandangoParser.CLOSE_BRACK)
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 2546
                self.star_atom()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Star_atomContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NAME(self):
            return self.getToken(FandangoParser.NAME, 0)

        def OPEN_PAREN(self):
            return self.getToken(FandangoParser.OPEN_PAREN, 0)

        def target_with_star_atom(self):
            return self.getTypedRuleContext(
                FandangoParser.Target_with_star_atomContext, 0
            )

        def CLOSE_PAREN(self):
            return self.getToken(FandangoParser.CLOSE_PAREN, 0)

        def star_targets_tuple_seq(self):
            return self.getTypedRuleContext(
                FandangoParser.Star_targets_tuple_seqContext, 0
            )

        def OPEN_BRACK(self):
            return self.getToken(FandangoParser.OPEN_BRACK, 0)

        def CLOSE_BRACK(self):
            return self.getToken(FandangoParser.CLOSE_BRACK, 0)

        def star_targets_list_seq(self):
            return self.getTypedRuleContext(
                FandangoParser.Star_targets_list_seqContext, 0
            )

        def getRuleIndex(self):
            return FandangoParser.RULE_star_atom

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterStar_atom"):
                listener.enterStar_atom(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitStar_atom"):
                listener.exitStar_atom(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitStar_atom"):
                return visitor.visitStar_atom(self)
            else:
                return visitor.visitChildren(self)

    def star_atom(self):

        localctx = FandangoParser.Star_atomContext(self, self._ctx, self.state)
        self.enterRule(localctx, 430, self.RULE_star_atom)
        self._la = 0  # Token type
        try:
            self.state = 2564
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 320, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 2549
                self.match(FandangoParser.NAME)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 2550
                self.match(FandangoParser.OPEN_PAREN)
                self.state = 2551
                self.target_with_star_atom()
                self.state = 2552
                self.match(FandangoParser.CLOSE_PAREN)
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 2554
                self.match(FandangoParser.OPEN_PAREN)
                self.state = 2556
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if (((_la) & ~0x3F) == 0 and ((1 << _la) & 4505815838884120) != 0) or (
                    (((_la - 66)) & ~0x3F) == 0
                    and ((1 << (_la - 66)) & 4503599631567367) != 0
                ):
                    self.state = 2555
                    self.star_targets_tuple_seq()

                self.state = 2558
                self.match(FandangoParser.CLOSE_PAREN)
                pass

            elif la_ == 4:
                self.enterOuterAlt(localctx, 4)
                self.state = 2559
                self.match(FandangoParser.OPEN_BRACK)
                self.state = 2561
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if (((_la) & ~0x3F) == 0 and ((1 << _la) & 4505815838884120) != 0) or (
                    (((_la - 66)) & ~0x3F) == 0
                    and ((1 << (_la - 66)) & 4503599631567367) != 0
                ):
                    self.state = 2560
                    self.star_targets_list_seq()

                self.state = 2563
                self.match(FandangoParser.CLOSE_BRACK)
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Single_targetContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def single_subscript_attribute_target(self):
            return self.getTypedRuleContext(
                FandangoParser.Single_subscript_attribute_targetContext, 0
            )

        def NAME(self):
            return self.getToken(FandangoParser.NAME, 0)

        def OPEN_PAREN(self):
            return self.getToken(FandangoParser.OPEN_PAREN, 0)

        def single_target(self):
            return self.getTypedRuleContext(FandangoParser.Single_targetContext, 0)

        def CLOSE_PAREN(self):
            return self.getToken(FandangoParser.CLOSE_PAREN, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_single_target

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterSingle_target"):
                listener.enterSingle_target(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitSingle_target"):
                listener.exitSingle_target(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitSingle_target"):
                return visitor.visitSingle_target(self)
            else:
                return visitor.visitChildren(self)

    def single_target(self):

        localctx = FandangoParser.Single_targetContext(self, self._ctx, self.state)
        self.enterRule(localctx, 432, self.RULE_single_target)
        try:
            self.state = 2572
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 321, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 2566
                self.single_subscript_attribute_target()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 2567
                self.match(FandangoParser.NAME)
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 2568
                self.match(FandangoParser.OPEN_PAREN)
                self.state = 2569
                self.single_target()
                self.state = 2570
                self.match(FandangoParser.CLOSE_PAREN)
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Single_subscript_attribute_targetContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def t_primary(self):
            return self.getTypedRuleContext(FandangoParser.T_primaryContext, 0)

        def DOT(self):
            return self.getToken(FandangoParser.DOT, 0)

        def NAME(self):
            return self.getToken(FandangoParser.NAME, 0)

        def OPEN_BRACK(self):
            return self.getToken(FandangoParser.OPEN_BRACK, 0)

        def slices(self):
            return self.getTypedRuleContext(FandangoParser.SlicesContext, 0)

        def CLOSE_BRACK(self):
            return self.getToken(FandangoParser.CLOSE_BRACK, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_single_subscript_attribute_target

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterSingle_subscript_attribute_target"):
                listener.enterSingle_subscript_attribute_target(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitSingle_subscript_attribute_target"):
                listener.exitSingle_subscript_attribute_target(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitSingle_subscript_attribute_target"):
                return visitor.visitSingle_subscript_attribute_target(self)
            else:
                return visitor.visitChildren(self)

    def single_subscript_attribute_target(self):

        localctx = FandangoParser.Single_subscript_attribute_targetContext(
            self, self._ctx, self.state
        )
        self.enterRule(localctx, 434, self.RULE_single_subscript_attribute_target)
        try:
            self.state = 2583
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 322, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 2574
                self.t_primary(0)
                self.state = 2575
                self.match(FandangoParser.DOT)
                self.state = 2576
                self.match(FandangoParser.NAME)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 2578
                self.t_primary(0)
                self.state = 2579
                self.match(FandangoParser.OPEN_BRACK)
                self.state = 2580
                self.slices()
                self.state = 2581
                self.match(FandangoParser.CLOSE_BRACK)
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class T_primaryContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def atom(self):
            return self.getTypedRuleContext(FandangoParser.AtomContext, 0)

        def t_primary(self):
            return self.getTypedRuleContext(FandangoParser.T_primaryContext, 0)

        def DOT(self):
            return self.getToken(FandangoParser.DOT, 0)

        def NAME(self):
            return self.getToken(FandangoParser.NAME, 0)

        def OPEN_BRACK(self):
            return self.getToken(FandangoParser.OPEN_BRACK, 0)

        def slices(self):
            return self.getTypedRuleContext(FandangoParser.SlicesContext, 0)

        def CLOSE_BRACK(self):
            return self.getToken(FandangoParser.CLOSE_BRACK, 0)

        def genexp(self):
            return self.getTypedRuleContext(FandangoParser.GenexpContext, 0)

        def OPEN_PAREN(self):
            return self.getToken(FandangoParser.OPEN_PAREN, 0)

        def CLOSE_PAREN(self):
            return self.getToken(FandangoParser.CLOSE_PAREN, 0)

        def arguments(self):
            return self.getTypedRuleContext(FandangoParser.ArgumentsContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_t_primary

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterT_primary"):
                listener.enterT_primary(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitT_primary"):
                listener.exitT_primary(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitT_primary"):
                return visitor.visitT_primary(self)
            else:
                return visitor.visitChildren(self)

    def t_primary(self, _p: int = 0):
        _parentctx = self._ctx
        _parentState = self.state
        localctx = FandangoParser.T_primaryContext(self, self._ctx, _parentState)
        _prevctx = localctx
        _startState = 436
        self.enterRecursionRule(localctx, 436, self.RULE_t_primary, _p)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 2586
            self.atom()
            self._ctx.stop = self._input.LT(-1)
            self.state = 2606
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 325, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    if self._parseListeners is not None:
                        self.triggerExitRuleEvent()
                    _prevctx = localctx
                    self.state = 2604
                    self._errHandler.sync(self)
                    la_ = self._interp.adaptivePredict(self._input, 324, self._ctx)
                    if la_ == 1:
                        localctx = FandangoParser.T_primaryContext(
                            self, _parentctx, _parentState
                        )
                        self.pushNewRecursionContext(
                            localctx, _startState, self.RULE_t_primary
                        )
                        self.state = 2588
                        if not self.precpred(self._ctx, 5):
                            from antlr4.error.Errors import FailedPredicateException

                            raise FailedPredicateException(
                                self, "self.precpred(self._ctx, 5)"
                            )
                        self.state = 2589
                        self.match(FandangoParser.DOT)
                        self.state = 2590
                        self.match(FandangoParser.NAME)
                        pass

                    elif la_ == 2:
                        localctx = FandangoParser.T_primaryContext(
                            self, _parentctx, _parentState
                        )
                        self.pushNewRecursionContext(
                            localctx, _startState, self.RULE_t_primary
                        )
                        self.state = 2591
                        if not self.precpred(self._ctx, 4):
                            from antlr4.error.Errors import FailedPredicateException

                            raise FailedPredicateException(
                                self, "self.precpred(self._ctx, 4)"
                            )
                        self.state = 2592
                        self.match(FandangoParser.OPEN_BRACK)
                        self.state = 2593
                        self.slices()
                        self.state = 2594
                        self.match(FandangoParser.CLOSE_BRACK)
                        pass

                    elif la_ == 3:
                        localctx = FandangoParser.T_primaryContext(
                            self, _parentctx, _parentState
                        )
                        self.pushNewRecursionContext(
                            localctx, _startState, self.RULE_t_primary
                        )
                        self.state = 2596
                        if not self.precpred(self._ctx, 3):
                            from antlr4.error.Errors import FailedPredicateException

                            raise FailedPredicateException(
                                self, "self.precpred(self._ctx, 3)"
                            )
                        self.state = 2597
                        self.genexp()
                        pass

                    elif la_ == 4:
                        localctx = FandangoParser.T_primaryContext(
                            self, _parentctx, _parentState
                        )
                        self.pushNewRecursionContext(
                            localctx, _startState, self.RULE_t_primary
                        )
                        self.state = 2598
                        if not self.precpred(self._ctx, 2):
                            from antlr4.error.Errors import FailedPredicateException

                            raise FailedPredicateException(
                                self, "self.precpred(self._ctx, 2)"
                            )
                        self.state = 2599
                        self.match(FandangoParser.OPEN_PAREN)
                        self.state = 2601
                        self._errHandler.sync(self)
                        _la = self._input.LA(1)
                        if (
                            ((_la) & ~0x3F) == 0
                            and ((1 << _la) & 4505888853336344) != 0
                        ) or (
                            (((_la - 66)) & ~0x3F) == 0
                            and ((1 << (_la - 66)) & 4503599633861255) != 0
                        ):
                            self.state = 2600
                            self.arguments()

                        self.state = 2603
                        self.match(FandangoParser.CLOSE_PAREN)
                        pass

                self.state = 2608
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 325, self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.unrollRecursionContexts(_parentctx)
        return localctx

    class Del_targetsContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def del_target(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(FandangoParser.Del_targetContext)
            else:
                return self.getTypedRuleContext(FandangoParser.Del_targetContext, i)

        def COMMA(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.COMMA)
            else:
                return self.getToken(FandangoParser.COMMA, i)

        def getRuleIndex(self):
            return FandangoParser.RULE_del_targets

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterDel_targets"):
                listener.enterDel_targets(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitDel_targets"):
                listener.exitDel_targets(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitDel_targets"):
                return visitor.visitDel_targets(self)
            else:
                return visitor.visitChildren(self)

    def del_targets(self):

        localctx = FandangoParser.Del_targetsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 438, self.RULE_del_targets)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 2609
            self.del_target()
            self.state = 2614
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 326, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 2610
                    self.match(FandangoParser.COMMA)
                    self.state = 2611
                    self.del_target()
                self.state = 2616
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 326, self._ctx)

            self.state = 2618
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == 70:
                self.state = 2617
                self.match(FandangoParser.COMMA)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Del_targetContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def t_primary(self):
            return self.getTypedRuleContext(FandangoParser.T_primaryContext, 0)

        def DOT(self):
            return self.getToken(FandangoParser.DOT, 0)

        def NAME(self):
            return self.getToken(FandangoParser.NAME, 0)

        def OPEN_BRACK(self):
            return self.getToken(FandangoParser.OPEN_BRACK, 0)

        def slices(self):
            return self.getTypedRuleContext(FandangoParser.SlicesContext, 0)

        def CLOSE_BRACK(self):
            return self.getToken(FandangoParser.CLOSE_BRACK, 0)

        def del_t_atom(self):
            return self.getTypedRuleContext(FandangoParser.Del_t_atomContext, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_del_target

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterDel_target"):
                listener.enterDel_target(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitDel_target"):
                listener.exitDel_target(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitDel_target"):
                return visitor.visitDel_target(self)
            else:
                return visitor.visitChildren(self)

    def del_target(self):

        localctx = FandangoParser.Del_targetContext(self, self._ctx, self.state)
        self.enterRule(localctx, 440, self.RULE_del_target)
        try:
            self.state = 2630
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 328, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 2620
                self.t_primary(0)
                self.state = 2621
                self.match(FandangoParser.DOT)
                self.state = 2622
                self.match(FandangoParser.NAME)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 2624
                self.t_primary(0)
                self.state = 2625
                self.match(FandangoParser.OPEN_BRACK)
                self.state = 2626
                self.slices()
                self.state = 2627
                self.match(FandangoParser.CLOSE_BRACK)
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 2629
                self.del_t_atom()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Del_t_atomContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NAME(self):
            return self.getToken(FandangoParser.NAME, 0)

        def OPEN_PAREN(self):
            return self.getToken(FandangoParser.OPEN_PAREN, 0)

        def CLOSE_PAREN(self):
            return self.getToken(FandangoParser.CLOSE_PAREN, 0)

        def del_targets(self):
            return self.getTypedRuleContext(FandangoParser.Del_targetsContext, 0)

        def OPEN_BRACK(self):
            return self.getToken(FandangoParser.OPEN_BRACK, 0)

        def CLOSE_BRACK(self):
            return self.getToken(FandangoParser.CLOSE_BRACK, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_del_t_atom

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterDel_t_atom"):
                listener.enterDel_t_atom(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitDel_t_atom"):
                listener.exitDel_t_atom(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitDel_t_atom"):
                return visitor.visitDel_t_atom(self)
            else:
                return visitor.visitChildren(self)

    def del_t_atom(self):

        localctx = FandangoParser.Del_t_atomContext(self, self._ctx, self.state)
        self.enterRule(localctx, 442, self.RULE_del_t_atom)
        self._la = 0  # Token type
        try:
            self.state = 2643
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [52]:
                self.enterOuterAlt(localctx, 1)
                self.state = 2632
                self.match(FandangoParser.NAME)
                pass
            elif token in [68]:
                self.enterOuterAlt(localctx, 2)
                self.state = 2633
                self.match(FandangoParser.OPEN_PAREN)
                self.state = 2635
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if (((_la) & ~0x3F) == 0 and ((1 << _la) & 4505815838884120) != 0) or (
                    (((_la - 66)) & ~0x3F) == 0
                    and ((1 << (_la - 66)) & 4503599631567365) != 0
                ):
                    self.state = 2634
                    self.del_targets()

                self.state = 2637
                self.match(FandangoParser.CLOSE_PAREN)
                pass
            elif token in [75]:
                self.enterOuterAlt(localctx, 3)
                self.state = 2638
                self.match(FandangoParser.OPEN_BRACK)
                self.state = 2640
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if (((_la) & ~0x3F) == 0 and ((1 << _la) & 4505815838884120) != 0) or (
                    (((_la - 66)) & ~0x3F) == 0
                    and ((1 << (_la - 66)) & 4503599631567365) != 0
                ):
                    self.state = 2639
                    self.del_targets()

                self.state = 2642
                self.match(FandangoParser.CLOSE_BRACK)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Type_expressionsContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def expression(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(FandangoParser.ExpressionContext)
            else:
                return self.getTypedRuleContext(FandangoParser.ExpressionContext, i)

        def COMMA(self, i: int = None):
            if i is None:
                return self.getTokens(FandangoParser.COMMA)
            else:
                return self.getToken(FandangoParser.COMMA, i)

        def STAR(self):
            return self.getToken(FandangoParser.STAR, 0)

        def POWER(self):
            return self.getToken(FandangoParser.POWER, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_type_expressions

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterType_expressions"):
                listener.enterType_expressions(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitType_expressions"):
                listener.exitType_expressions(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitType_expressions"):
                return visitor.visitType_expressions(self)
            else:
                return visitor.visitChildren(self)

    def type_expressions(self):

        localctx = FandangoParser.Type_expressionsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 444, self.RULE_type_expressions)
        self._la = 0  # Token type
        try:
            self.state = 2702
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 336, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 2645
                self.expression()
                self.state = 2650
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 332, self._ctx)
                while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                    if _alt == 1:
                        self.state = 2646
                        self.match(FandangoParser.COMMA)
                        self.state = 2647
                        self.expression()
                    self.state = 2652
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input, 332, self._ctx)

                self.state = 2653
                self.match(FandangoParser.COMMA)
                self.state = 2654
                self.match(FandangoParser.STAR)
                self.state = 2655
                self.expression()
                self.state = 2656
                self.match(FandangoParser.COMMA)
                self.state = 2657
                self.match(FandangoParser.POWER)
                self.state = 2658
                self.expression()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 2660
                self.expression()
                self.state = 2665
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 333, self._ctx)
                while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                    if _alt == 1:
                        self.state = 2661
                        self.match(FandangoParser.COMMA)
                        self.state = 2662
                        self.expression()
                    self.state = 2667
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input, 333, self._ctx)

                self.state = 2668
                self.match(FandangoParser.COMMA)
                self.state = 2669
                self.match(FandangoParser.STAR)
                self.state = 2670
                self.expression()
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 2672
                self.expression()
                self.state = 2677
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 334, self._ctx)
                while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                    if _alt == 1:
                        self.state = 2673
                        self.match(FandangoParser.COMMA)
                        self.state = 2674
                        self.expression()
                    self.state = 2679
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input, 334, self._ctx)

                self.state = 2680
                self.match(FandangoParser.COMMA)
                self.state = 2681
                self.match(FandangoParser.POWER)
                self.state = 2682
                self.expression()
                pass

            elif la_ == 4:
                self.enterOuterAlt(localctx, 4)
                self.state = 2684
                self.match(FandangoParser.STAR)
                self.state = 2685
                self.expression()
                self.state = 2686
                self.match(FandangoParser.COMMA)
                self.state = 2687
                self.match(FandangoParser.POWER)
                self.state = 2688
                self.expression()
                pass

            elif la_ == 5:
                self.enterOuterAlt(localctx, 5)
                self.state = 2690
                self.match(FandangoParser.STAR)
                self.state = 2691
                self.expression()
                pass

            elif la_ == 6:
                self.enterOuterAlt(localctx, 6)
                self.state = 2692
                self.match(FandangoParser.POWER)
                self.state = 2693
                self.expression()
                pass

            elif la_ == 7:
                self.enterOuterAlt(localctx, 7)
                self.state = 2694
                self.expression()
                self.state = 2699
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la == 70:
                    self.state = 2695
                    self.match(FandangoParser.COMMA)
                    self.state = 2696
                    self.expression()
                    self.state = 2701
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Func_type_commentContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NEWLINE(self):
            return self.getToken(FandangoParser.NEWLINE, 0)

        def getRuleIndex(self):
            return FandangoParser.RULE_func_type_comment

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterFunc_type_comment"):
                listener.enterFunc_type_comment(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitFunc_type_comment"):
                listener.exitFunc_type_comment(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitFunc_type_comment"):
                return visitor.visitFunc_type_comment(self)
            else:
                return visitor.visitChildren(self)

    def func_type_comment(self):

        localctx = FandangoParser.Func_type_commentContext(self, self._ctx, self.state)
        self.enterRule(localctx, 446, self.RULE_func_type_comment)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 2704
            self.match(FandangoParser.NEWLINE)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    def sempred(self, localctx: RuleContext, ruleIndex: int, predIndex: int):
        if self._predicates == None:
            self._predicates = dict()
        self._predicates[22] = self.selector_sempred
        self._predicates[59] = self.dotted_name_sempred
        self._predicates[113] = self.name_or_attr_sempred
        self._predicates[157] = self.bitwise_or_sempred
        self._predicates[158] = self.bitwise_xor_sempred
        self._predicates[159] = self.bitwise_and_sempred
        self._predicates[160] = self.shift_expr_sempred
        self._predicates[161] = self.sum_sempred
        self._predicates[162] = self.term_sempred
        self._predicates[166] = self.primary_sempred
        self._predicates[218] = self.t_primary_sempred
        pred = self._predicates.get(ruleIndex, None)
        if pred is None:
            raise Exception("No predicate with index:" + str(ruleIndex))
        else:
            return pred(localctx, predIndex)

    def selector_sempred(self, localctx: SelectorContext, predIndex: int):
        if predIndex == 0:
            return self.precpred(self._ctx, 2)

        if predIndex == 1:
            return self.precpred(self._ctx, 1)

    def dotted_name_sempred(self, localctx: Dotted_nameContext, predIndex: int):
        if predIndex == 2:
            return self.precpred(self._ctx, 2)

    def name_or_attr_sempred(self, localctx: Name_or_attrContext, predIndex: int):
        if predIndex == 3:
            return self.precpred(self._ctx, 2)

    def bitwise_or_sempred(self, localctx: Bitwise_orContext, predIndex: int):
        if predIndex == 4:
            return self.precpred(self._ctx, 2)

    def bitwise_xor_sempred(self, localctx: Bitwise_xorContext, predIndex: int):
        if predIndex == 5:
            return self.precpred(self._ctx, 2)

    def bitwise_and_sempred(self, localctx: Bitwise_andContext, predIndex: int):
        if predIndex == 6:
            return self.precpred(self._ctx, 2)

    def shift_expr_sempred(self, localctx: Shift_exprContext, predIndex: int):
        if predIndex == 7:
            return self.precpred(self._ctx, 3)

        if predIndex == 8:
            return self.precpred(self._ctx, 2)

    def sum_sempred(self, localctx: SumContext, predIndex: int):
        if predIndex == 9:
            return self.precpred(self._ctx, 3)

        if predIndex == 10:
            return self.precpred(self._ctx, 2)

    def term_sempred(self, localctx: TermContext, predIndex: int):
        if predIndex == 11:
            return self.precpred(self._ctx, 6)

        if predIndex == 12:
            return self.precpred(self._ctx, 5)

        if predIndex == 13:
            return self.precpred(self._ctx, 4)

        if predIndex == 14:
            return self.precpred(self._ctx, 3)

        if predIndex == 15:
            return self.precpred(self._ctx, 2)

    def primary_sempred(self, localctx: PrimaryContext, predIndex: int):
        if predIndex == 16:
            return self.precpred(self._ctx, 5)

        if predIndex == 17:
            return self.precpred(self._ctx, 4)

        if predIndex == 18:
            return self.precpred(self._ctx, 3)

        if predIndex == 19:
            return self.precpred(self._ctx, 2)

    def t_primary_sempred(self, localctx: T_primaryContext, predIndex: int):
        if predIndex == 20:
            return self.precpred(self._ctx, 5)

        if predIndex == 21:
            return self.precpred(self._ctx, 4)

        if predIndex == 22:
            return self.precpred(self._ctx, 3)

        if predIndex == 23:
            return self.precpred(self._ctx, 2)
